-----------------------------------------------------------------------------------------------
-- Client Lua Script for OptiPlates  
-- Released under CDDL (Common Development and Distribution Lisense 1.0)
-- Wildstar is Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------

require "Window"
require "ChallengesLib"
require "Unit"
require "GameLib"
require "Apollo"
require "PathMission"
require "Quest"
require "Episode"
require "math"
require "string"
require "DialogSys"
require "PublicEvent"
require "PublicEventObjective"
require "CommunicatorLib"
require "Tooltip"
require "GroupLib"
require "PlayerPathLib"
require "GuildLib"
require "GuildTypeLib"
 
local OptiPlates = {} 
-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
local _AddonVersionMajor = 3
local _AddonVersionMinor = 4
local _AddonVersionPatch = 1
local _AddonVersion = string.format("%d.%d",_AddonVersionMajor,_AddonVersionMinor)
if _AddonVersionPatch ~= nil then
	_AddonVersion = string.format("%d.%d%d",_AddonVersionMajor,_AddonVersionMinor,_AddonVersionPatch)
end
local _AddonInfoObject = nil
local karPathSprite =
{
	[PlayerPathLib.PlayerPathType_Soldier] 		= "CRB_TargetFrameRewardPanelSprites:sprTargetFrame_PathSol",
	[PlayerPathLib.PlayerPathType_Settler] 		= "CRB_TargetFrameRewardPanelSprites:sprTargetFrame_PathSet",
	[PlayerPathLib.PlayerPathType_Scientist] 	= "CRB_TargetFrameRewardPanelSprites:sprTargetFrame_PathSci",
	[PlayerPathLib.PlayerPathType_Explorer] 	= "CRB_TargetFrameRewardPanelSprites:sprTargetFrame_PathExp",
}
local karRewardIcons =
{
	[Unit.CodeEnumRewardInfoType.Quest] 			= { strName = "Quest", strSingle = "CRB_TargetFrameRewardPanelSprites:sprTargetFrame_ActiveQuest", 	strMulti = "CRB_TargetFrameRewardPanelSprites:sprTargetFrame_ActiveQuestMulti" },
	[Unit.CodeEnumRewardInfoType.Challenge] 		= { strName = "Challenge", strSingle = "CRB_TargetFrameRewardPanelSprites:sprTargetFrame_Challenge", 		strMulti = "CRB_TargetFrameRewardPanelSprites:sprTargetFrame_ChallengeMulti" },
	[Unit.CodeEnumRewardInfoType.Explorer] 		= { strName = "Explorer", strSingle = "CRB_TargetFrameRewardPanelSprites:sprTargetFrame_PathExp", 		strMulti = "CRB_TargetFrameRewardPanelSprites:sprTargetFrame_PathExpMulti" },
	[Unit.CodeEnumRewardInfoType.Scientist] 		= { strName = "Scientist", strSingle = "CRB_TargetFrameRewardPanelSprites:sprTargetFrame_PathSci",			strMulti = "CRB_TargetFrameRewardPanelSprites:sprTargetFrame_PathSciMulti" },
	[Unit.CodeEnumRewardInfoType.Soldier] 		= { strName = "Soldier", strSingle = "CRB_TargetFrameRewardPanelSprites:sprTargetFrame_PathSol", 		strMulti = "CRB_TargetFrameRewardPanelSprites:sprTargetFrame_PathSolMulti" },
	[Unit.CodeEnumRewardInfoType.Settler] 		= { strName = "Settler", strSingle = "CRB_TargetFrameRewardPanelSprites:sprTargetFrame_PathSet", 		strMulti = "CRB_TargetFrameRewardPanelSprites:sprTargetFrame_PathSetMulti" },
	[Unit.CodeEnumRewardInfoType.PublicEvent] 	= { strName = "PublicEvent", strSingle = "CRB_TargetFrameRewardPanelSprites:sprTargetFrame_PublicEvent", 	strMulti = "CRB_TargetFrameRewardPanelSprites:sprTargetFrame_PublicEventMulti" },
	[Unit.CodeEnumRewardInfoType.Rival] 			= { strName = "Rival", strSingle = "ClientSprites:Icon_Windows_UI_CRB_Rival", 							strMulti = "ClientSprites:Icon_Windows_UI_CRB_Rival" },
	[Unit.CodeEnumRewardInfoType.Friend] 			= { strName = "Friend", strSingle = "ClientSprites:Icon_Windows_UI_CRB_Friend", 						strMulti = "ClientSprites:Icon_Windows_UI_CRB_Friend" },
	[Unit.CodeEnumRewardInfoType.ScientistSpell]	= { strName = "ScientistSpell", strSingle = "CRB_TargetFrameRewardPanelSprites:sprTargetFrame_PathSciSpell",	strMulti = "CRB_TargetFrameRewardPanelSprites:sprTargetFrame_PathSciSpell" },
	[Unit.CodeEnumRewardInfoType.TSpell]			= { strName = "TSpell", strSingle = "CRB_TargetFrameRewardPanelSprites:sprTargetFrame_ActiveQuest",		strMulti = "CRB_TargetFrameRewardPanelSprites:sprTargetFrame_ActiveQuestMulti" },
	[Unit.CodeEnumRewardInfoType.Contract]			= { strName = "Contract", strSingle = "CRB_TargetFrameRewardPanelSprites:sprTargetFrame_Contract",		strMulti = "CRB_TargetFrameRewardPanelSprites:sprTargetFrame_ContractMulti" },
}

-- Acceptable window heights per font.
local fontHeight = {
	["CRB_Pixel"] 		= 14,
	["CRB_Interface9"]  = 15,
	["CRB_Interface10"] = 17,
	["CRB_Interface11"] = 18,
	["CRB_Interface12"] = 19,
	["CRB_Interface14"] = 22,
	["Nameplates"]		= 15,
}

local healthTexture = {
	"Flat_Health",
	"OptiGrad_1",
	"OptiGrad_2",
	"OptiGrad_3",
	"OptiGrad_4",
	"WhiteFill",
	"OptiGrad_5",
	"OptiGrad_6",
	"OptiGrad_7",
	"OptiGrad_8",
	"OptiGrad_9",
	"OptiGrad_10",
	"OptiGrad_11",
	"OptiGrad_12",
}

local shieldTexture = {
	"Flat_Shield",
	"OptiGrad_1s",
	"OptiGrad_2s",
	"OptiGrad_3s",
	"OptiGrad_4s",
	"WhiteFill",
	"OptiGrad_5",
	"OptiGrad_6",
	"OptiGrad_7",
	"OptiGrad_8",
	"OptiGrad_9s",
	"OptiGrad_10s",
	"OptiGrad_11s",
	"OptiGrad_12s",
}

local className = {
	["4"] = "[md] ",
	["3"] = "[es] ",
	["7"] = "[ss] ",
	["1"] = "[wr] ",
	["5"] = "[st] ",
	["2"] = "[en] ",
}

local arClassToClassIcon = {
	[GameLib.CodeEnumClass.Warrior] 		= "IconSprites:Icon_Windows_UI_CRB_Warrior",
	[GameLib.CodeEnumClass.Engineer] 		= "IconSprites:Icon_Windows_UI_CRB_Engineer",
	[GameLib.CodeEnumClass.Esper] 			= "IconSprites:Icon_Windows_UI_CRB_Esper",
	[GameLib.CodeEnumClass.Medic] 			= "IconSprites:Icon_Windows_UI_CRB_Medic",
	[GameLib.CodeEnumClass.Stalker] 		= "IconSprites:Icon_Windows_UI_CRB_Stalker",
	[GameLib.CodeEnumClass.Spellslinger] 	= "IconSprites:Icon_Windows_UI_CRB_Spellslinger"
}

local arQuestIcons = {
	QuestNew = "ClientSprites:QuestJewel_Offer_Yellow",
	QuestNewGrey = "ClientSprites:QuestJewel_Offer_Grey",
	QuestNewDaily = "ClientSprites:QuestJewel_Offer_Green",
	QuestReceiving = "ClientSprites:QuestJewel_Incomplete_Grey",
	QuestReward = "ClientSprites:QuestJewel_Complete_Green"

}

local arRoleIcons = {
	["dps"] = "TargetFrameSprites:sprRoleMeleeLight",
	["tank"] = "TargetFrameSprites:sprRoleMeleeHeavy",
	["healer"] = "TargetFrameSprites:sprRoleHealer",	
}

local tDisposition = {
	Friendly = Unit.CodeEnumDisposition.Friendly,
	Hostile = Unit.CodeEnumDisposition.Hostile,
	Neutral = Unit.CodeEnumDisposition.Neutral,
	Unknown = Unit.CodeEnumDisposition.Unknown
}

local tSubdueCCStates = {
	Stun = Unit.CodeEnumCCState.Stun,
	Sleep = Unit.CodeEnumCCState.Sleep,
	Disarm = Unit.CodeEnumCCState.Disarm,
	Fear = Unit.CodeEnumCCState.Fear,
	Knockdown = Unit.CodeEnumCCState.Knockdown,
	Subdue = Unit.CodeEnumCCState.Subdue
}

local tNameplateMatrixColors = {
	ApolloColor.new("50929591"),
	ApolloColor.new("xkcdGreen"),
	ApolloColor.new("xkcdRed")
}

local tNameplateMatrixLabelPixie = {
	fRotation = 45/180*math.pi,
	strText = "",
	strFont = "CRB_Interface10_B",
	cr = "xkcdWhite",
	flagsText = {
		DT_VCENTER = true,
		DT_RIGHT = true
	},
	loc =
	{
		fPoints = { 0, 0, 0, 0 },
		nOffsets = { 13, 13, 39, 39 }
	}
}

local tNameplateMatrixInCombatPixie = {
	fRotation = 45/180*math.pi,
	strSprite = "WhiteFill",
	cr = tNameplateMatrixColors[3],
	loc =
	{
		fPoints = { 0, 0, 0, 0 },
		nOffsets = { 13, 13, 39, 39 }
	}
}

local tNameplateMatrixOutOfCombatPixie = {
	fRotation = 45/180*math.pi,
	strSprite = "WhiteFill",
	cr = tNameplateMatrixColors[2],
	loc =
	{
		fPoints = { 0, 0, 0, 0 },
		nOffsets = { 13, 13, 39, 39 }
	}
}

local tNameplateMatrixCategories = {
	"Self",
	"Party",
	--"Guild",
	"FriendlyPlayer",
	"HostilePlayer",
	"FriendlyNPC",
	"Neutral",
	"HostileNPC",
	"FriendlyTarget",
	"NeutralTarget",
	"HostileTarget"
}

local tNameplateMatrixSettings = {
	"Name",
	"Title",
	"Guild",
	"Bars",
	"HPText",
	"SPText",
	"LevelText",
	"InterruptArmor",
	"Castbar",
	"CastText",
	"Resourcebar",
	"ResourceText",
	"Focusbar",
	"FocusText",
	"Sprintbar",
	"SprintText",
	"Distance",
	"Buffs",
	"Eliteness",
	"Class",
	"Role",
	"Heading",
	"ShieldReboot",
	"Nameplate"
}

-- Specifies which settings are configurable for each unit type
local tNameplateMatrixEnabledSettings = {
	Self = {
		Name = true, Guild = true, Bars = true, Castbar = true, Resourcebar = true, Focusbar = true,
		Sprintbar = true, Distance = true, Buffs = true, Class = true, Role = true, Heading = true,
		ShieldReboot = true, 
		Title = true, HPText = true, SPText = true, LevelText = true, CastText = true,
		ResourceText = true, FocusText = true, SprintText = true, InterruptArmor = true,
		Nameplate = true
	},
	Party = {
		Name = true, Guild = true, Bars = true, Castbar = true, Resourcebar = true, Focusbar = true,
		Sprintbar = true, Distance = true, Buffs = true, Class = true, Role = true,
		Title = true, HPText = true, SPText = true, LevelText = true, CastText = true,
		ResourceText = true, FocusText = true, SprintText = true, InterruptArmor = true,
		Nameplate = true
	},
	--Guild = {
	--	Name = true, Guild = true, Bars = true, Castbar = true, Resourcebar = true, Focusbar = true,
	--	Sprintbar = true, Distance = true, Buffs = true, Class = true,
	--	Title = true, HPText = true, SPText = true, LevelText = true, CastText = true,
	--	ResourceText = true, FocusText = true, SprintText = true, InterruptArmor = true
	--},
	FriendlyPlayer = {
		Name = true, Guild = true, Bars = true, Castbar = true, Resourcebar = true, Focusbar = true,
		Sprintbar = true, Distance = true, Buffs = true, Class = true,
		Title = true, HPText = true, SPText = true, LevelText = true, CastText = true,
		ResourceText = true, FocusText = true, SprintText = true, InterruptArmor = true,
		Nameplate = true
	},	
	HostilePlayer = {
		Name = true, Guild = true, Bars = true, Castbar = true, Resourcebar = true, Focusbar = true,
		Sprintbar = true, Distance = true, Buffs = true, Class = true,
		Title = true, HPText = true, SPText = true, LevelText = true, CastText = true,
		ResourceText = true, FocusText = true, SprintText = true, InterruptArmor = true,
		Nameplate = true
	},		
	FriendlyNPC = {
		Name = true, Guild = true, Bars = true, Castbar = true, Distance = true, Buffs = true,
		Eliteness = true,
		HPText = true, SPText = true, LevelText = true, CastText = true, InterruptArmor = true,
		Nameplate = true
	},
	Neutral = {
		Name = true, Guild = true, Bars = true, Castbar = true, Distance = true, Buffs = true,
		Eliteness = true,
		HPText = true, SPText = true, LevelText = true, CastText = true, InterruptArmor = true,
		Nameplate = true
	},
	HostileNPC = {
		Name = true, Guild = true, Bars = true, Castbar = true, Distance = true, Buffs = true,
		Eliteness = true,
		HPText = true, SPText = true, LevelText = true, CastText = true, InterruptArmor = true,
		Nameplate = true
	},
	FriendlyTarget = {
		Name = true, Guild = true, Bars = true, Castbar = true, Resourcebar = true, Focusbar = true,
		Sprintbar = true, Distance = true, Buffs = true, Class = true, Role = true, Eliteness = true,
		Title = true, HPText = true, SPText = true, LevelText = true, CastText = true,
		ResourceText = true, FocusText = true, SprintText = true, InterruptArmor = true
	},
	NeutralTarget = {
		Name = true, Guild = true, Bars = true, Castbar = true, Resourcebar = true, Focusbar = true,
		Sprintbar = true, Distance = true, Buffs = true, Eliteness = true,
		HPText = true, SPText = true, LevelText = true, CastText = true, InterruptArmor = true
	},	
	HostileTarget = {
		Name = true, Guild = true, Bars = true, Castbar = true, Resourcebar = true, Focusbar = true,
		Sprintbar = true, Distance = true, Buffs = true, Class = true, Eliteness = true,
		Title = true, HPText = true, SPText = true, LevelText = true, CastText = true,
		ResourceText = true, FocusText = true, SprintText = true, InterruptArmor = true
	}
}

local tNameplateMatrixCategoryToLetter = {
	Target = "T",
	Self = "M",
	Party = "PP",
	Guild = "GP",
	FriendlyPlayer = "FP",
	HostilePlayer = "HP",
	FriendlyNPC = "F",
	Neutral = "N",
	HostileNPC = "H",
	FriendlyTarget = "FT",
	NeutralTarget = "NT",
	HostileTarget = "HT"
}

local tNameplateMatrixCategoryToColor = {
	Self = ApolloColor.new("xkcdWhite"),
	Party = ApolloColor.new("ff069af3"),
	FriendlyPlayer = ApolloColor.new("xkcdGreen"),
	HostilePlayer = ApolloColor.new("ffe50000"),
	FriendlyNPC = ApolloColor.new("ff76cd26"),
	Neutral = ApolloColor.new("fffff569"),
	HostileNPC = ApolloColor.new("ffd9544d"),
	FriendlyTarget = ApolloColor.new("ff76cd26"),
	NeutralTarget = ApolloColor.new("fffff569"),
	HostileTarget = ApolloColor.new("ffd9544d")
}

local tNameplateMatrixSettingToLetter = {
	Name = "N",
	Title = "T",
	Guild = "G",
	Bars = "B",
	HPText = "HPT",
	SPText = "SPT",
	LevelText = "LT",
	InterruptArmor = "IA",
	Castbar = "C",
	CastText = "CT",
	Resourcebar = "R",
	ResourceText = "RT",
	Focusbar = "F",
	FocusText = "FT",
	Sprintbar = "S",
	SprintText = "ST",
	Distance = "DI",
	Buffs = "D",
	Eliteness = "E",
	Class = "CN",
	Role = "RO",
	Heading = "H",
	ShieldReboot = "SR",
	Nameplate = "Nameplate"
}


local bDebug = false; -- Am I in debug mode? 

local updateBuffs = false; -- Can update buffs yet?
local updateStatistics = false
local buffUpdateInterval = 0.5; -- Interval to update buffs related function (seconds)
local updateStatisticsInterval = 1000
local statisticsLastUpdated = 0

local nOptionChangeMinInterval = 500 -- minimum delay between OptionsChanged calls (for sliders etc)
local nPrevOptionChangeTicks = 0
local bOptionChangeQueued = false

local nRewardUpdateMinInterval = 1500
local nPrevRewardUpdateTicks = 0

local colorpicker = nil
local colorpickerColorObject = CColor.new(1, 1, 1, 1)
local colorpickerwndControlArray = nil

local subdued = false;

local knTargetRange = 40000 -- the distance^2 that normal nameplates should draw within (max targeting range)


local tLegacyTargetIconPixie = {
	strSprite = "BK3:sprHolo_Accent_Rounded",
	cr = white,
	loc =
	{
		fPoints = { 0, 0, 1, 1 },
		nOffsets = { 0, 0, 0, 0 }
	}
}

local tTargetIconPixie = {
	strSprite = "CRB_DatachronSprites:sprDC_GreenPlayRing",
	cr = white,
	loc =
	{
		fPoints = { 0, 0, 0, 0 },
		nOffsets = { 0, 0, 32, 32 }
	}
}

local tLevelPixie = {
	strText = "60",
	strFont = "CRB_Pixel_O",
	loc =
	{
		fPoints = { 0, 0, 0, 0 },
		nOffsets = { 0, 0, 18, 16 }
	}
}

local tInterruptArmorPixie = {
	strSprite = "CRB_ActionBarFrameSprites:sprResourceBar_InterruptCircleOrange",
	strText = "25",
	strFont = "CRB_Interface9_O",
	flagsText = { DT_CENTER = true, DT_VCENTER = true },
	loc =
	{
		fPoints = { 0, 0, 0, 0 },
		nOffsets = { 0, 0, 24, 24 }
	}		
}

local tInterruptArmorInfPixie = {
	strSprite = "IconSprites:Icon_Windows_UI_CRB_Infinity",
	loc =
	{
		fPoints = { 0.5, 0, 0.5, 0 },
		nOffsets = { 0, 0, 24, 24 }
	}		
}

local tHPTextPixie = {
	strText = "30k",
	crText = ApolloColor.new("xkcdAzure"),
	strFont = "CRB_Pixel_O",
	flagsText = { DT_RIGHT = true },
	loc =
	{
		fPoints = { 0, 0, 0, 0 },
		nOffsets = { 0, 0, 40, 15 }
	}
}

local tSPTextPixie = {
	strText = "30k",
	crText = ApolloColor.new("gray"),
	strFont = "CRB_Pixel_O",
	flagsText = { DT_RIGHT = true },
	loc =
	{
		fPoints = { 0, 0, 0, 0 },
		nOffsets = { 0, 0, 40, 15 }
	}
}

local tCleanseIconPixie = {
	strSprite = "CRB_DatachronSprites:sprDC_GreenPlayRing",
	cr = white,
	loc =
	{
		fPoints = { 0, 0, 0, 0 },
		nOffsets = { 0, 0, 32, 32 }
	}
}

local tDistancePixie = {
	strText = "25",
	flagsText = { DT_CENTER = true, DT_VCENTER = true, DT_SINGLELINE = true },
	loc =
	{
		fPoints = { 0, 0, 0, 0 },
		nOffsets = { 0, 0, 24, 24 }
	}		
}

local tHeadingPixie = {
	strSprite = "CRB_MinimapSprites:sprMM_QuestArrow",
	loc =
	{
		fPoints = { 0, 0, 0, 0 },
		nOffsets = { 0, 0, 30, 30 }
	}		
}

local tHeadingTextPixie = {
	strText = "",
	strFont = "CRB_Interface9_O",
	flagsText = { DT_CENTER = true, DT_SINGLELINE = true },	
	loc =
	{
		fPoints = { 0, 0, 0, 0 },
		nOffsets = { 0, 0, 30, 30 }
	}		
}

local tHeadingSegmentText = {
	"S",
	"SE",
	"SE",
	"E",
	"E",
	"NE",
	"NE",
	"N",
	"N",
	"NW",
	"NW",
	"W",
	"W",
	"SW",
	"SW",
	"S"
}
--[[
in tUserSettings, the following table keys are implemented:
default 			- specifies the default value for the setting
defaultifexists 	- specifies the variable to copy the default value from if the variable exists.
					  overrides the "default" setting in that case
percharacter		- setting is not automatically saved or loaded in profiles, has custom loading code in
					  OnSave and OnRestore
setDefaultOnEmptyString - normally default setting is applied only if the setting value is nil, having this
						  key set also sets the value to default if the string is empty ("")
deprecated			- the setting is loaded, but is not saved. allows using variables from a previous version
					  to preserve settings in conjunction with defaultifexists


--]]
local tUserSettings = {
	-- deprecated variables
	setShowIcons = { default = true },
	setShowTitles = { default = false, deprecated = true },
	setShowHealth = { default = true, deprecated = true },
	setShowHealthPercent = { default = false, deprecated = true },
	setShowHealthDeficit = { default = false, deprecated = true },
	setShowShield = { default = true, deprecated = true },
	setShowShieldPercent = { default = false, deprecated = true },
	setShowShieldDeficit = { default = false, deprecated = true },
	setShowFocus = { default = true, deprecated = true },
	setShowFocusPercent = { default = false, deprecated = true },
	setShowLevel = { default = true, deprecated = true },
	setShowArmor = { default = true, deprecated = true },
	setT_M = { default = true, deprecated = true },
	setT_N = { default = true, deprecated = true },
	setT_G = { default = false, deprecated = true },
	setT_B = { default = true, deprecated = true },
	setT_C = { default = true, deprecated = true },
	setT_D = { default = false, deprecated = true },
	setT_DI = { default = false, deprecated = true },
	setT_R = { default = false, deprecated = true },
	setT_F = { default = false, deprecated = true },
	setT_CN = { default = false, deprecated = true },
	setFT_M = { default = false, deprecated = true },
	setStyleLevelIconPositionToNameLeft = { default = true, deprecated = true },	
	setHideNormalTargetIndicator = { default = false, deprecated = true },	
	setSAOSize = { default = 100, deprecated = true },
	setFlatWidth = { default = 80, deprecated = true },
	setFlatHealthHeight = { default = 11, deprecated = true },
	setFlatShieldHeight = { default = 4, deprecated = true },
	setStyleCastbarAboveText = {default = false, deprecated = true }, --
	setStyleCastbarTextInside = {default = false, deprecated = true }, --
	setStyleCastbarTextInsideAutoFit = { default = true, deprecated = true },
	setStyleCastbarShowTimeRemaining = { default = false, deprecated = true }, --
	setStyleCastbarShowTimeRemainingHostile = { default = false, deprecated = true }, --
	setStyleCastbarNoText = { default = false, deprecated = true }, --	
	setStyleClassIconPositionTop = { default = true, deprecated = true },
	setColorCastBar = { default = "feb308", deprecated = true }, -- deprecated in version 3.1
	setIconClass = { default = true, deprecated = true },
	
	
	setClassIconStyle = { default = "Icon" },
	setRoleIconAlwaysShowNameplate = { default = false },
	setHostilePvpOcclusionEnabled = { default = true },
	setHostilePvpOcclusionTimeout = { default = 2000 },
	setHPTextStyle = { default = "Normal" },
	setSPTextStyle = { default = "Normal" },
	setLevelTextPosition = { default = "NameLeft" },
	setClassIconPosition = { default = "NameLeft" },
	setFocusTextStyle = { default = "Normal" },
	setCastbarTextStyle = { default = "Above" },
	setCastbarTimeStyle = { default = "Never" },
	setOneVersionIntegrationEnabled = { default = true },
	setColorPickerIntegrationEnabled = { default = true },
	setQuickselectWindowScaling = { default = 100, percharacter = true },
	setQuickselectWindowNoButtons = { default = false, percharacter = true },
	setColorFriendlyCastBar = { default = "feb308", defaultifexists = "setColorCastBar" },
	setColorHostileCastBar = { default = "feb308", defaultifexists = "setColorCastBar" },
	setShowObjectivesAlways = { default = false },
	setShowFocusTargetIndicator = { default = false, defaultifexists = "setFT_M" },
	
	
	setHideUnimportantNPCNameplates = { default = true },
	setShowInteractableAlways = { default = true },
	setStyleHeadingIndicatorTextFont = { default = { name = "Nameplates", face = "Nameplates", size = 9 } },
	setColorHeadingIndicatorText = { default = "069af3" },
	setColorHeadingIndicatorArrow = { default = "ffffff" },
	setColorLevelText = { default = "ffffff" },
	
	setHarvestablesDatacubeEnabled = { default = false },
	setShowQuestIconNew = { default = false },
	setShowQuestIconNewDaily = { default = false },
	setShowQuestIconReward = { default = false },
	setShowQuestIconReceiving = { default = false },
	setShowQuestIconNewGrey = { default = false },
	
	setStyleQuestIconSize = { default = 30 },
	setStyleQuestIconIgnoreMaximumDistance = { default = true },
	setStyleQuestIconUseDefaultScaling = { default = true },
	setStyleQuestIconAlwaysShowDistance = { default = true },
	setStyleQuestIconMinimumDistance = { default = 0 },
	setStyleQuestIconMinimumDistanceSquared = { default = 0 },
	
	--General Settings
	setNeverShow = { default = false },
	setNeverShowCity = { default = false },
	setEnableOcclusion = { default = false },

	setDrawDistance = { default = 100 },
	setDrawDistanceSquared = { default = 10000 }, -- distance comparisons are done squared to avoid calculating sqrt
	setHealthCutoff = { default = 30 },

	setIconQuest = { default = true },
	setIconMission = { default = true },
	setIconFriend = { default = true },
	setIconEvent = { default = true },
	setIconChallenge = { default = true },
	setIconRival = { default = true },
	setIconTooltips = { default = true },
	
	setShowThreatAlert = { default = false },
	setShowThreatLossAlert = { default = false },
	
	setShowCleanseAlert = { default = false },
	setHideFullFriendlyBars = { default = false },
	setHideFullFriendlyBarsHealthThreshold = { default = 95 },
	setHideFullFriendlyBarsShieldThreshold =  { default = 95 },
	setHideFullFriendlyBarsOneThreshold = { default = false },

	setHideFullFriendlyBarsMine = { default = false },
	setHideFullFriendlyBarsMineHealthThreshold = { default = 95 },
	setHideFullFriendlyBarsMineShieldThreshold =  { default = 95 },
	setHideFullFriendlyBarsMineOneThreshold = { default = false },	
	
	setM_Nameplate = { default = true },
	setM_Nameplate_C = { default = true },
	setPP_Nameplate = { default = true },
	setPP_Nameplate_C = { default = true },
	setFP_Nameplate = { default = true },
	setFP_Nameplate_C = { default = true },
	setHP_Nameplate = { default = true },
	setHP_Nameplate_C = { default = true },
	setF_Nameplate = { default = true },
	setF_Nameplate_C = { default = true },
	setN_Nameplate = { default = true },
	setN_Nameplate_C = { default = true },
	setH_Nameplate = { default = true },
	setH_Nameplate_C = { default = true },
	
	
	-- Targeted Settings
	setFT_N = { default = true, defaultifexists = "setT_N" },
	setFT_T = { default = true, defaultifexists = "setShowTitles" },
	setFT_G = { default = true, defaultifexists = "setT_G" },
	setFT_B = { default = true, defaultifexists = "setT_B" },
	setFT_HPT = { default = true },
	setFT_SPT = { default = true },
	setFT_LT = { default = true, defaultifexists = "setShowLevel" },
	setFT_IA = { default = true, defaultifexists = "setShowArmor" },
	setFT_C = { default = true, defaultifexists = "setT_C" },
	setFT_CT = { default = true },
	setFT_R = { default = false, defaultifexists = "setT_R" },
	setFT_RT = { default = false, defaultifexists = "setT_R" },
	setFT_F = { default = false, defaultifexists = "setT_F" },
	setFT_FT = { default = false, defaultifexists = "setT_F" },
	setFT_S = { default = false },
	setFT_ST = { default = false },
	setFT_DI = { default = false },
	setFT_D = { default = false, defaultifexists = "setT_D" },
	setFT_E = { default = false, defaultifexists = "setT_CN" },
	setFT_CN = { default = false, defaultifexists = "setT_CN" },
	setFT_RO =	{ default = false },
	setFT_N_C = { default = true, defaultifexists = "setT_N" },
	setFT_T_C = { default = false, defaultifexists = "setShowTitles" },
	setFT_G_C = { default = true, defaultifexists = "setT_G" },
	setFT_B_C = { default = true, defaultifexists = "setT_B" },
	setFT_HPT_C = { default = true },
	setFT_SPT_C = { default = true },
	setFT_LT_C = { default = true, defaultifexists = "setShowLevel" },
	setFT_IA_C = { default = true, defaultifexists = "setShowArmor" },
	setFT_C_C = { default = true, defaultifexists = "setT_C" },
	setFT_CT_C = { default = true },
	setFT_R_C = { default = false, defaultifexists = "setT_R" },
	setFT_RT_C = { default = false, defaultifexists = "setT_R" },
	setFT_F_C = { default = false, defaultifexists = "setT_F" },
	setFT_FT_C = { default = false, defaultifexists = "setT_F" },
	setFT_S_C = { default = false },
	setFT_ST_C = { default = false },
	setFT_DI_C = { default = false },
	setFT_D_C = { default = false, defaultifexists = "setT_D" },
	setFT_E_C = { default = false, defaultifexists = "setT_CN" },
	setFT_CN_C = { default = false, defaultifexists = "setT_CN" },
	setFT_RO_C = { default = false },

	setNT_N = { default = true, defaultifexists = "setT_N" },
	setNT_T = { default = true, defaultifexists = "setShowTitles" },
	setNT_G = { default = true, defaultifexists = "setT_G" },
	setNT_B = { default = true, defaultifexists = "setT_B" },
	setNT_HPT = { default = true },
	setNT_SPT = { default = true },
	setNT_LT = { default = true, defaultifexists = "setShowLevel" },
	setNT_IA = { default = true, defaultifexists = "setShowArmor" },
	setNT_C = { default = true, defaultifexists = "setT_C" },
	setNT_CT = { default = true },
	setNT_R = { default = false, defaultifexists = "setT_R" },
	setNT_RT = { default = false, defaultifexists = "setT_R" },
	setNT_F = { default = false, defaultifexists = "setT_F" },
	setNT_FT = { default = false, defaultifexists = "setT_F" },
	setNT_S = { default = false },
	setNT_ST = { default = false },
	setNT_DI = { default = false },
	setNT_D = { default = false, defaultifexists = "setT_D" },
	setNT_E = { default = true, defaultifexists = "setT_CN" },
	setNT_CN = { default = false, defaultifexists = "setT_CN" },
	setNT_RO =	{ default = false },
	setNT_N_C = { default = true, defaultifexists = "setT_N" },
	setNT_T_C = { default = false, defaultifexists = "setShowTitles" },
	setNT_G_C = { default = true, defaultifexists = "setT_G" },
	setNT_B_C = { default = true, defaultifexists = "setT_B" },
	setNT_HPT_C = { default = true },
	setNT_SPT_C = { default = true },
	setNT_LT_C = { default = true, defaultifexists = "setShowLevel" },
	setNT_IA_C = { default = true, defaultifexists = "setShowArmor" },
	setNT_C_C = { default = true, defaultifexists = "setT_C" },
	setNT_CT_C = { default = true },
	setNT_R_C = { default = false, defaultifexists = "setT_R" },
	setNT_RT_C = { default = false, defaultifexists = "setT_R" },
	setNT_F_C = { default = false, defaultifexists = "setT_F" },
	setNT_FT_C = { default = false, defaultifexists = "setT_F" },
	setNT_S_C = { default = false },
	setNT_ST_C = { default = false },
	setNT_DI_C = { default = false },
	setNT_D_C = { default = false, defaultifexists = "setT_D" },
	setNT_E_C = { default = true, defaultifexists = "setT_CN" },
	setNT_CN_C = { default = false, defaultifexists = "setT_CN" },
	setNT_RO_C = { default = false },	

	setHT_N = { default = true, defaultifexists = "setT_N" },
	setHT_T = { default = true, defaultifexists = "setShowTitles" },
	setHT_G = { default = true, defaultifexists = "setT_G" },
	setHT_B = { default = true, defaultifexists = "setT_B" },
	setHT_HPT = { default = true },
	setHT_SPT = { default = true },
	setHT_LT = { default = true, defaultifexists = "setShowLevel" },
	setHT_IA = { default = true, defaultifexists = "setShowArmor" },
	setHT_C = { default = true, defaultifexists = "setT_C" },
	setHT_CT = { default = true },
	setHT_R = { default = false, defaultifexists = "setT_R" },
	setHT_RT = { default = false, defaultifexists = "setT_R" },
	setHT_F = { default = false, defaultifexists = "setT_F" },
	setHT_FT = { default = false, defaultifexists = "setT_F" },
	setHT_S = { default = false },
	setHT_ST = { default = false },
	setHT_DI = { default = false },
	setHT_D = { default = false, defaultifexists = "setT_D" },
	setHT_E = { default = true, defaultifexists = "setT_CN" },
	setHT_CN = { default = false, defaultifexists = "setT_CN" },
	setHT_RO =	{ default = false },
	setHT_N_C = { default = true, defaultifexists = "setT_N" },
	setHT_T_C = { default = false, defaultifexists = "setShowTitles" },
	setHT_G_C = { default = true, defaultifexists = "setT_G" },
	setHT_B_C = { default = true, defaultifexists = "setT_B" },
	setHT_HPT_C = { default = true },
	setHT_SPT_C = { default = true },
	setHT_LT_C = { default = true, defaultifexists = "setShowLevel" },
	setHT_IA_C = { default = true, defaultifexists = "setShowArmor" },
	setHT_C_C = { default = true, defaultifexists = "setT_C" },
	setHT_CT_C = { default = true },
	setHT_R_C = { default = false, defaultifexists = "setT_R" },
	setHT_RT_C = { default = false, defaultifexists = "setT_R" },
	setHT_F_C = { default = false, defaultifexists = "setT_F" },
	setHT_FT_C = { default = false, defaultifexists = "setT_F" },
	setHT_S_C = { default = false },
	setHT_ST_C = { default = false },
	setHT_DI_C = { default = false },
	setHT_D_C = { default = false, defaultifexists = "setT_D" },
	setHT_E_C = { default = true, defaultifexists = "setT_CN" },
	setHT_CN_C = { default = false, defaultifexists = "setT_CN" },
	setHT_RO_C = { default = false },	
	
	setNoTargetSettingsOnSelf = { default = false },
	setFriendlyTargetIndicatorIcon = { default = "CRB_DatachronSprites:sprDC_GreenPlayRing", setDefaultOnEmptyString = true },
	setPartyTargetIndicatorIcon = { default = "CRB_DatachronSprites:sprDC_BluePlayRing", setDefaultOnEmptyString = true },
	setNeutralTargetIndicatorIcon = { default = "CRB_DatachronSprites:sprDC_GreenPlayRing", setDefaultOnEmptyString = true },
	setHostileTargetIndicatorIcon = { default = "CRB_DatachronSprites:sprDC_RedPlayRing", setDefaultOnEmptyString = true },
	
	setFriendlyAltTargetIndicatorIcon = { default = "Crafting_CoordSprites:sprCoord_AdditiveTargetGreenSmall", setDefaultOnEmptyString = true },
	setNeutralAltTargetIndicatorIcon = { default = "Crafting_CoordSprites:sprCoord_AdditiveTargetGreenSmall", setDefaultOnEmptyString = true },
	setHostileAltTargetIndicatorIcon = { default = "Crafting_CoordSprites:sprCoord_AdditiveTargetRedSmall", setDefaultOnEmptyString = true },	
	setShowLegacyTargetIndicator = { default = false },
	setShowNormalTargetIndicator = { default = true, defaultifexists = "setHideNormalTargetIndicator" },
	
	setShowTargetClusterIcons = { default = true },
	setFriendlyClusterIndicatorIcon = { default = "CRB_NameplateSprites:sprNp_TargetSecondaryFriendly", setDefaultOnEmptyString = true },
	setNeutralClusterIndicatorIcon = { default = "CRB_NameplateSprites:sprNp_TargetSecondaryNeutral", setDefaultOnEmptyString = true },
	setHostileClusterIndicatorIcon = { default = "CRB_NameplateSprites:sprNp_TargetSecondaryHostile", setDefaultOnEmptyString = true  },
	
	-- My Nameplate Settings
	setM_N = { default = true }, 
	setM_N_C = { default = true },
	setM_T = { default = true },
	setM_T_C = { default = true },
	setM_HPT = { default = true },
	setM_HPT_C = { default = true },
	setM_SPT = { default = true },
	setM_SPT_C = { default = true },	
	setM_LT = { default = true },
	setM_LT_C = { default = true },	
	setM_IA = { default = true },
	setM_IA_C = { default = true },		
	setM_CT = { default = true },
	setM_CT_C = { default = true },	
	setM_RT = { default = false },
	setM_RT_C = { default = true },	
	setM_FT = { default = false },
	setM_FT_C = { default = false },
	setM_ST = { default = false },
	setM_ST_C = { default = true },
	
	setM_G = { default = true }, 
	setM_G_C = { default = true },
	setM_B = { default = true }, 
	setM_B_C = { default = true },
	setM_C = { default = true }, 
	setM_C_C = { default = true },
	setM_D = { default = false }, 
	setM_D_C = { default = false },
	setM_R = { default = false }, 
	setM_R_C = { default = true },
	setM_F = { default = false }, 
	setM_F_C = { default = false },
	setM_S = { default = false },
	setM_S_C = { default = true },
	setM_H = { default = false },
	setM_H_C = { default = false },
	setM_CN = { default = false },
	setM_CN_C = { default = false },	
	setM_DI = { default = false },
	setM_DI_C = { default = false },
	setM_RO = { default = false },
	setM_RO_C = { default = false },	
	
	setM_SR = { default = false },
	setM_SR_C = { default = false },	
	setHideFullFriendlyBarsMine = { default = false },	
	
	-- Neutral Nameplate Settings
	setN_N = { default = true }, 
	setN_N_C = { default = true }, 
	setN_G = { default = true }, 
	setN_G_C = { default = true }, 
	setN_B = { default = false }, 
	setN_B_C = { default = true },
	setN_HPT = { default = true }, 
	setN_HPT_C = { default = true }, 
	setN_SPT = { default = true }, 
	setN_SPT_C = { default = true }, 
	setN_C = { default = false }, 
	setN_C_C = { default = true }, 
	setN_D = { default = false }, 
	setN_D_C = { default = false }, 
	setN_DI = { default = false }, 
	setN_DI_C = { default = false }, 
	setN_CN = { default = false },
	setN_CN_C = { default = false },
	setN_CT = { default = false },
	setN_CT_C = { default = true },
	setN_E = { default = true },
	setN_E_C = { default = true },
	setN_LT = { default = true },
	setN_LT_C = { default = true },	
	setN_IA = { default = true },
	setN_IA_C = { default = true },	
	
	-- Friendly Nameplate Settings
	setF_N = { default = true }, 
	setF_N_C = { default = true }, 
	setF_G = { default = true }, 
	setF_G_C = { default = true }, 
	setF_B = { default = false }, 
	setF_B_C = { default = true }, 
	setF_C = { default = false }, 
	setF_C_C = { default = false }, 
	setF_D = { default = false }, 
	setF_D_C = { default = false }, 
	setF_DI = { default = false }, 
	setF_DI_C = { default = false }, 
	setF_CT = { default = false },
	setF_CT_C = { default = false },	
	setF_E = { default = false },
	setF_E_C = { default = false },	
	setF_LT = { default = false },
	setF_LT_C = { default = false },		
	setF_IA = { default = false },
	setF_IA_C = { default = false },	
	setF_HPT = { default = true }, 
	setF_HPT_C = { default = true }, 
	setF_SPT = { default = true }, 
	setF_SPT_C = { default = true }, 	
	setF_Important = { default = true },
	setF_Vendor = { default = true },

	-- Friendly PLAYER Nameplate Settings
	setFP_N = { default = true }, 
	setFP_N_C = { default = true },  
	setFP_G = { default = true }, 
	setFP_G_C = { default = true }, 
	setFP_B = { default = false }, 
	setFP_B_C = { default = true }, 
	setFP_C = { default = false }, 
	setFP_C_C = { default = true }, 
	setFP_D = { default = false }, 
	setFP_D_C = { default = false }, 
	setFP_CN = { default = false }, 
	setFP_CN_C = { default = false }, 
	setFP_DI = { default = false }, 
	setFP_DI_C = { default = false }, 
	setFP_R = { default = false }, 
	setFP_R_C = { default = false }, 
	setFP_F = { default = false }, 
	setFP_F_C = { default = false }, 
	setFP_CT = { default = false },
	setFP_CT_C = { default = false },	
	setFP_FT = { default = false },
	setFP_FT_C = { default = false },	
	setFP_RT = { default = false },
	setFP_RT_C = { default = false },	
	setFP_LT = { default = false },
	setFP_LT_C = { default = false },	
	setFP_S = { default = false },
	setFP_S_C = { default = false },		
	setFP_ST = { default = false },
	setFP_ST_C = { default = false },	
	setFP_T = { default = false },
	setFP_T_C = { default = false },	
	setFP_IA = { default = false },
	setFP_IA_C = { default = false },	
	setFP_HPT = { default = true }, 
	setFP_HPT_C = { default = true }, 
	setFP_SPT = { default = true }, 
	setFP_SPT_C = { default = true }, 
	
	-- Party PLAYER Nameplate Settings
	setPP_N = { default = true }, 
	setPP_N_C = { default = true },  
	setPP_G = { default = true }, 
	setPP_G_C = { default = true }, 
	setPP_B = { default = true }, 
	setPP_B_C = { default = true }, 
	setPP_C = { default = false }, 
	setPP_C_C = { default = true }, 
	setPP_D = { default = false }, 
	setPP_D_C = { default = false }, 
	setPP_CN = { default = false }, 
	setPP_CN_C = { default = false },
	setPP_DI = { default = false }, 
	setPP_DI_C = { default = false }, 
	setPP_R = { default = false }, 
	setPP_R_C = { default = false }, 
	setPP_F = { default = false }, 
	setPP_F_C = { default = false }, 
	setPP_RO = { default = false }, 
	setPP_RO_C = { default = false }, 
	setPP_CT = { default = false }, 
	setPP_CT_C = { default = false }, 
	setPP_FT = { default = false }, 
	setPP_FT_C = { default = false }, 
	setPP_LT = { default = false }, 
	setPP_LT_C = { default = false }, 
	setPP_RT = { default = false }, 
	setPP_RT_C = { default = false }, 	
	setPP_S = { default = false }, 
	setPP_S_C = { default = false }, 
	setPP_ST = { default = false }, 
	setPP_ST_C = { default = false }, 
	setPP_T = { default = false }, 
	setPP_T_C = { default = false }, 	
	setPP_IA = { default = false },
	setPP_IA_C = { default = false },		
	setPP_HPT = { default = true }, 
	setPP_HPT_C = { default = true }, 
	setPP_SPT = { default = true }, 
	setPP_SPT_C = { default = true }, 
	
	-- Guild PLAYER Nameplate Settings
	setGP_N = { default = true }, 
	setGP_N_C = { default = true },  
	setGP_G = { default = true }, 
	setGP_G_C = { default = true }, 
	setGP_B = { default = false }, 
	setGP_B_C = { default = false }, 
	setGP_C = { default = false }, 
	setGP_C_C = { default = false }, 
	setGP_D = { default = false }, 
	setGP_D_C = { default = false }, 
	setGP_CN = { default = false }, 
	setGP_CN_C = { default = false },
	setGP_DI = { default = false }, 
	setGP_DI_C = { default = false }, 
	setGP_R = { default = false }, 
	setGP_R_C = { default = false }, 
	setGP_F = { default = false }, 
	setGP_F_C = { default = false }, 
	setGP_CT = { default = false },
	setGP_CT_C = { default = false },	
	setGP_FT = { default = false },
	setGP_FT_C = { default = false },	
	setGP_RT = { default = false },
	setGP_RT_C = { default = false },	
	setGP_LT = { default = false },
	setGP_LT_C = { default = false },	
	setGP_S = { default = false },
	setGP_S_C = { default = false },		
	setGP_ST = { default = false },
	setGP_ST_C = { default = false },	
	setGP_T = { default = false },
	setGP_T_C = { default = false },		
	setGP_IA = { default = false },
	setGP_IA_C = { default = false },	

	setGP_Override = { default = false },
		
	-- Hostile Nameplate Settings
	setH_N = { default = true }, 
	setH_N_C = { default = true },
	setH_G = { default = true }, 
	setH_G_C = { default = true },
	setH_B = { default = false }, 
	setH_B_C = { default = true },
	setH_C = { default = true }, 
	setH_C_C = { default = true },
	setH_D = { default = false }, 
	setH_D_C = { default = false },
	setH_DI = { default = false }, 
	setH_DI_C = { default = false }, 
	setH_CN = { default = false },
	setH_CN_C = { default = false },
	setH_CT = { default = false },
	setH_CT_C = { default = true },
	setH_E = { default = true },
	setH_E_C = { default = true },
	setH_LT = { default = true },
	setH_LT_C = { default = true },	
	setH_IA = { default = true },
	setH_IA_C = { default = true },	
	setH_HPT = { default = true }, 
	setH_HPT_C = { default = true }, 
	setH_SPT = { default = true }, 
	setH_SPT_C = { default = true }, 
	
	-- Hostile PLAYER Nameplate Settings
	setHP_N = { default = true }, 
	setHP_N_C = { default = true },
	setHP_G = { default = true }, 
	setHP_G_C = { default = true },
	setHP_B = { default = true }, 
	setHP_B_C = { default = true },
	setHP_C = { default = true }, 
	setHP_C_C = { default = true },
	setHP_D = { default = false }, 
	setHP_D_C = { default = false },
	setHP_CN = { default = false }, 
	setHP_CN_C = { default = false },
	setHP_DI = { default = false }, 
	setHP_DI_C = { default = false }, 	
	setHP_R = { default = false }, 
	setHP_R_C = { default = false }, 	
	setHP_F = { default = false }, 
	setHP_F_C = { default = false },
	setHP_S = { default = false },
	setHP_S_C = { default = false },	
	setHP_CT = { default = false },
	setHP_CT_C = { default = true },	
	setHP_FT = { default = false },
	setHP_FT_C = { default = false },	
	setHP_RT = { default = false },
	setHP_RT_C = { default = false },	
	setHP_LT = { default = true },
	setHP_LT_C = { default = true },	
	setHP_T = { default = false },
	setHP_T_C = { default = false },		
	setHP_ST = { default = false },
	setHP_ST_C = { default = false },	
	setHP_IA = { default = false },
	setHP_IA_C = { default = true },	
	setHP_HPT = { default = true }, 
	setHP_HPT_C = { default = true }, 
	setHP_SPT = { default = true }, 
	setHP_SPT_C = { default = true }, 	
	-- Pet Settings
	setShowMyPets = { default = true },
	setShowFriendlyPets = { default = false },
	setShowHostilePets = { default = false },
	
	setShowClassColors = { default = false },
	setShowAllyClassColors = { default = false },
	
	-- Style Page
	------------------------------
	setStyle = { default = "sao" },
	setFlatHealthTexture = { default = 11 },
	setFlatShieldTexture = { default = 11 },
	setStyleWidth = { default = 80 },
	setHealthbarHeight = { default = 11 },
	setShieldbarHeight = { default = 4 },	
	setStyleFlatNumberInside = { default = false },
	
	setNameplateVerticalOffset = { default = 0 },
	setNameplateVerticalOffsetMine = { default = 0 },
	setAutoAdjustNameplatesDown = { default = false },
	setOffscreenNameplatesTopOffset = { default = 0 },
	setOffscreenNameplatesMinHeight = { default = 180 },
	setOffscreenNameplatesAdjustAnchorOffset = { default = 0 },

	setStyleCastbarFont = { default = { name = "Nameplates", face = "Nameplates", size = 9 } }, --
	setStyleFocusFont = { default = { name = "CRB_Pixel_O", size = 8, face = "Segoe UI" } }, --
	setStyleClassResourceFont = { default = { name = "CRB_Pixel_O", size = 8, face = "Segoe UI" } }, --
	setStyleUnitNameFont = { default = { name = "Nameplates", face = "Nameplates", size = 9 } }, --
	setStyleUnitGuildFont = { default = { name = "Nameplates", face = "Nameplates", size = 9 } }, --
	setStyleDistanceFont = { default = { name = "CRB_Pixel_O", size = 8, face = "Segoe UI" } }, --

	setStyleCastbarHeight = { default = 5 },--
	setStyleCastbarWidth = { default = 0 },--
	setStyleCastbarPaddingTop = { default = 0 },--
	setStyleCastbarShowMoO = { default = true }, --
	setStyleCastbarShowSubdue = { default = true }, --
	setStyleCastbarTextAutoFit = { default = true, defaultifexists = "setStyleCastbarTextInsideAutoFit" },
	setStyleCastbarAboveResourcebar = { default = false }, --
	setStyleCastbarTextTimeRemaining = { default = "Never" },
	setStyleCastbarTextPosition = { default = "Above" },
	setStyleVerticalAnchor = { default = "Name" },
	
	setStyleFocusbarHeight = { default = 12 }, --
	setStyleResourcebarHeight = { default = 12 }, --
	setStyleCombineResourceBars = { default = false },
	setStyleHideBarsWhenBlinded = { default = false }, --
	
	setStyleShowCleanseIcon = { default = false },
	setStyleCleanseIconSize = { default = 30 },
	setStyleCleanseOutlineShow = { default = true },
	setStyleCleanseOutlineToHealthBar = { default = false },
	setStyleCleanseIconPosition =  { default = "Left" },
	
	setStyleTargetedIconSize = { default = 32 },
	setStyleClassIconSize = { default = 20 },
	setStyleRoleIconSize = { default = 20 },
	setStyleClassIconShowWarrior = { default = true },
	setStyleClassIconShowEngineer = { default = true },
	setStyleClassIconShowEsper = { default = true },
	setStyleClassIconShowMedic = { default = true },
	setStyleClassIconShowStalker = { default = true },
	setStyleClassIconShowSpellslinger = { default = true },
	
	setStyleDashBarHeight = { default = 13 },
	
	setStyleRoleIconShowTank = { default = true },
	setStyleRoleIconShowHealer = { default = true },
	setStyleRoleIconShowDPS = { default = true },
	
	setStyleHeadingArrowSize = { default = 25 },
	setStyleHeadingShowText = { default = false },
	setStyleHeadingArrowBelowTargetMarker = { default = true },
	
	setDefaultScale = { default = 100 },	
	setDistanceScalingScaleFactor = { default = 50 }, -- Minimum scale when doing distance scaling, reached at self.setDistanceScalingMaxRange
	setTargetDefaultScale = { default = 100 },
	setAltTargetDefaultScale = { default = 100 },
	setScalingUpdateMinChange = { default = 5 }, -- Minimum difference between new and old scale before scale gets updated
	setDistanceScalingMinRange = { default = 60 },
	setDistanceScalingMaxRange = { default = 100 },
	setDistanceScalingEnabled = { default = false },	
	setDistanceScalingNormalTargetScale = { default = true },
	setDistanceScalingNormalObjectiveScale = { default = false },
	
	setHealthDeficitScalingEnabled = { default = false },
	setHealthDeficitScalingUseShield = { default = true },
	setHealthDeficitScalingUseHealth = { default = true },
	setHealthDeficitScalingMinCutoff = { default = 25 },
	setHealthDeficitScalingMaxCutoff = { default = 75 },
	setHealthDeficitScalingMaxChange = { default = 1.5 },
	setHealthDeficitScalingPlayerInCombat = { default = false },	
	setHealthDeficitScalingOnlyFriendlyPlayers = { default = false },		
	
	setDefaultOpacity = { default = 100 },
	setBarBGOpacity = { default = 100 },
	setSpeechBubbleOpacity = { default = 30 },
	setTargetOpacityEmphasis = { default = 0 },
	
	setCombatOpacityEnabled = { default = false },
	setCombatOpacityPlayerInCombat = { default = false },
	setCombatOpacityDefaultHostileCasting = { default = false },
	setCombatOpacityHostile = { default = 75 },
	setCombatOpacityFriendly = { default = 75 },

	setDistanceOpacityFactor = { default = 60 },
	setDistanceOpacityMaxRange = { default = 50 },
	setDistanceOpacityEnabled = { default = false },	
	setDistanceOpacityNormalTargetOpacity = { default = true },
	setDistanceOpacityNormalObjectiveOpacity = { default = true },
	
	-- Harvestables
	setHarvestablesEnabled = { default = false },
	setHarvestablesMiningEnabled = { default = false },
	setHarvestablesTreeEnabled = { default = false },
	setHarvestablesRelicEnabled = { default = false },
	setHarvestablesPlantEnabled = { default = false },
	setHarvestablesDistanceEnabled = { default = false },
	setHarvestablesIconEnabled = { default = false },
	setHarvestablesTierEnabled = { default = false },
	setHarvestablesDistanceScalingDisabled = { default = false },
	setHarvestablesNoDistanceHiding = { default = false },
	
	setQuickselectWindowShow = { default = false, percharacter = true },
	setQuickselectWindowLocked = { default = false, percharacter = true },
	setQuickselectWindowHideInCombat = { default = false, percharacter = true },
	setQuickselectWindowPosition = { default = { 10,10,285,40 }, percharacter = true },
	
	-- Buff settings (default buff profile settings in OptiPlates:CreateBuffProfile())
	setBuffSelectedProfiles = { default = {
							M = "Default",
							F = "Default",
							FP = "Default",
							H = "Default",
							HP = "Default",
							PP = "Default",
							N = "Default"
							}
	},
	
	-- Advanced settings
	setNameplateDefaultStratum =  { default = "InWorldHudStratum" },
	setNameplateUpdateInterval = { default = 4 }, -- once every x frames
	setMaximumWindowCacheCount = { default = 300 },
	
	setNameplateBringSelfToFront = { default = false },
	setNameplateBringFocusToFront = { default = false },
	setNameplateBringTankToFront = { default = false },
	setNameplateBringHealerToFront = { default = false },
	setNameplateBringTargetToFront = { default = false },
	setNameplateBringDamagedToFront = { default = false },
	setNameplateBringDamagedToFrontShieldThreshold = { default = 30 },
	setNameplateBringDamagedToFrontHealthThreshold = { default = 75 },
	
	setStyleCleanseIconSprite = { default = "CRB_BreakoutSprites:spr_BreakoutStun_ClockGreen", setDefaultOnEmptyString = true },
	setStyleThreatIconSprite = { default = "CRB_MinimapSprites:sprMM_TargetObjective", setDefaultOnEmptyString = true },	
	
	setMouseTargetingEnabled = { default = true },
	setMouseTargetingButtonLeft = { default = true },
	setMouseTargetingButtonRight = { default = false },
	setMouseTargetingButtonMiddle = { default = false },
	setMouseTargetingModifiersCtrl = { default = false },
	setMouseTargetingModifiersAlt = { default = false },
	setMouseTargetingModifiersShift = { default = false },
	setMouseTargetingClickthrough = { default = false },
	
	setMouseOverEnabled = { default = false },
	setMouseOverModifiersCtrl = { default = false },
	setMouseOverModifiersAlt = { default = false },
	setMouseOverModifiersShift = { default = false },
	
	setMouseOverSettings_F_N = { default = false },
	setMouseOverSettings_F_G = { default = false },
	setMouseOverSettings_F_DI = { default = false },
	setMouseOverSettings_F_D = { default = false },
	setMouseOverSettings_F_B = { default = false },
	setMouseOverSettings_F_F = { default = false },
	setMouseOverSettings_H_N = { default = false },
	setMouseOverSettings_H_G = { default = false },
	setMouseOverSettings_H_DI = { default = false },
	setMouseOverSettings_H_D = { default = false },
	setMouseOverSettings_H_B = { default = false },
	setMouseOverSettings_H_F = { default = false },	
	-- Colors
	setOpacity = { default = "FF" },

	setColorBackgroundBar = { default = "8e8f8d", setDefaultOnEmptyString = true },
	setColorBackdropBar = { default = "585959", setDefaultOnEmptyString = true },

	setColorShieldBar = { default = "75e4ec", setDefaultOnEmptyString = true },
	setColorAbsorbBar = { default = "ff9b00", setDefaultOnEmptyString = true },
	setColorVulnerableBar = { default = "e50000", setDefaultOnEmptyString = true },
	setColorVulnerableTimerBar = { default = "e583c3", setDefaultOnEmptyString = true },

	setColorNeutralName = { default = "fff569", setDefaultOnEmptyString = true },
	setColorNeutralBar = { default = "f3d829", setDefaultOnEmptyString = true },
	setColorNeutralBarCutoff = { default = "c0fb2d", setDefaultOnEmptyString = true },

	setColorFriendlyName = { default = "76cd26", setDefaultOnEmptyString = true },
	setColorFriendlyBar = { default = "15b01a", setDefaultOnEmptyString = true },
	setColorFriendlyBarCutoff = { default = "ff9408", setDefaultOnEmptyString = true },
	
	setColorFriendlyPlayerName = { default = "00add6", setDefaultOnEmptyString = true },
	setColorFriendlyPlayerNameUnflagged = { default = "00add6", setDefaultOnEmptyString = true },
	setColorFriendlyPlayerBar = { default = "15b01a", setDefaultOnEmptyString = true },
	setColorFriendlyPlayerBarCutoff = { default = "f97306", setDefaultOnEmptyString = true },
	
	setColorPartyName = { default = "069af3", setDefaultOnEmptyString = true },
	setColorGuildName = { default = "15b01a", setDefaultOnEmptyString = true },

	setColorHostileName = { default = "d9544d", setDefaultOnEmptyString = true },
	setColorHostileBar = { default = "e50000", setDefaultOnEmptyString = true },
	setColorHostileBarCutoff = { default = "fe02a2", setDefaultOnEmptyString = true },
	
	setColorHostilePlayerName = { default = "e50000", setDefaultOnEmptyString = true },
	setColorHostilePlayerNameUnflagged = { default = "fff569", setDefaultOnEmptyString = true },
	setColorHostilePlayerBar = { default = "e50000", setDefaultOnEmptyString = true },
	setColorHostilePlayerBarCutoff = { default = "fe02a2", setDefaultOnEmptyString = true },
	
	--Class colors
	setColorWarriorBar = { default = "f54f4f", setDefaultOnEmptyString = true },
	setColorStalkerBar = { default = "d23ef4", setDefaultOnEmptyString = true },
	setColorEngineerBar = { default = "efab48", setDefaultOnEmptyString = true },
	setColorEsperBar = { default = "1591db", setDefaultOnEmptyString = true },
	setColorSpellslingerBar = { default = "98c723", setDefaultOnEmptyString = true },
	setColorMedicBar = { default = "ffe757", setDefaultOnEmptyString = true },
	
	setColorThreatOutline = { default = "feb308", setDefaultOnEmptyString = true },
	setColorCleanseOutline = { default = "c20078", setDefaultOnEmptyString = true },
	setColorHarvestables = { default = "fff569", setDefaultOnEmptyString = true },
	setColorFriendlyShieldBarCutoff = { default = "d5b60a", setDefaultOnEmptyString = true },
	setColorFocusBar = { default = "f97306", setDefaultOnEmptyString = true },
	setColorFocusBarText = { default = "ffffff", setDefaultOnEmptyString = true },
	setColorResourceBar = { default = "f97306", setDefaultOnEmptyString = true },
	setColorResourceBarText = { default = "ffffff", setDefaultOnEmptyString = true },
	setColorDashBar = { default = "0343df", setDefaultOnEmptyString = true },
	setColorDashBarText = { default = "ffffff", setDefaultOnEmptyString = true },	
	setColorSprintBar = { default = "75bbfd", setDefaultOnEmptyString = true },
	setColorSprintBarText = { default = "ffffff", setDefaultOnEmptyString = true },		
}

 
-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function OptiPlates:Init()
    Apollo.RegisterAddon(self, true)
end
 

-----------------------------------------------------------------------------------------------
-- OptiPlates OnLoad
-----------------------------------------------------------------------------------------------
function OptiPlates:OnLoad()
	self.playerUnit = nil
	self.playerPosition = nil
	self.focusUnit = nil
    -- Register handlers for events, slash commands and timer, etc.
	Apollo.RegisterEventHandler("UnitGibbed",					"OnUnitGibbed", self)

	Apollo.RegisterSlashCommand("OptiPlates", 					"OnOptiPlatesOn", self)
	Apollo.RegisterSlashCommand("Optiplates", 					"OnOptiPlatesOn", self)
	Apollo.RegisterSlashCommand("optiplates", 					"OnOptiPlatesOn", self)
	
    Apollo.RegisterEventHandler("UnitCreated", 					"OnUnitCreated", self)
	Apollo.RegisterEventHandler("UnitDestroyed", 				"OnUnitDestroyed", self)
	Apollo.RegisterEventHandler("UnitTextBubbleCreate", 		"OnUnitTextBubbleToggled", self)
	Apollo.RegisterEventHandler("UnitTextBubblesDestroyed", 	"OnUnitTextBubbleToggled", self)
	Apollo.RegisterEventHandler("TargetUnitChanged", 			"OnTargetUnitChanged", self)
	Apollo.RegisterEventHandler("AlternateTargetUnitChanged", 	"OnAlternateTargetUnitChanged", self)
	Apollo.RegisterEventHandler("UnitEnteredCombat", 			"OnEnteredCombat", self)
	Apollo.RegisterEventHandler("VarChange_FrameCount", 		"OnFrameVarChange", self)
	Apollo.RegisterEventHandler("NextFrame", 					"OnFrame", self)
	Apollo.RegisterEventHandler("UnitNameChanged", 				"OnUnitNameChanged", self)
	Apollo.RegisterEventHandler("KeyBindingKeyChanged", 		"OnKeyBindingUpdated", self)
	Apollo.RegisterEventHandler("UnitPvpFlagsChanged", 			"OnUnitPvpFlagsChanged", self)
	Apollo.RegisterEventHandler("UnitTitleChanged", 			"OnUnitTitleChanged", self)
	Apollo.RegisterEventHandler("PlayerTitleChange", 			"OnPlayerTitleChanged", self)
	Apollo.RegisterEventHandler("UnitGuildNameplateChanged", 	"OnUnitGuildNameplateChanged",self)
	Apollo.RegisterEventHandler("ApplyCCState", 				"OnApplyCCState", self)
	Apollo.RegisterEventHandler("RemoveCCState", 				"OnRemoveCCState", self)
	Apollo.RegisterEventHandler("UnitLevelChanged", 			"OnUnitLevelChanged", self)
	Apollo.RegisterEventHandler("Mount", 						"OnMount", self)
	
	Apollo.RegisterEventHandler("UnitMemberOfGuildChange", 		"OnUnitMemberOfGuildChange", self)
	Apollo.RegisterEventHandler("GuildChange", 					"OnGuildChange", self)  -- notification that a guild was added / removed.
	Apollo.RegisterEventHandler("ChangeWorld", 					"OptionsChanged", self)
	
	Apollo.RegisterEventHandler("CharacterCreated", 			"OptionsChanged", self)
	Apollo.RegisterEventHandler("ShieldsOverloaded", 			"OnShieldsOverloaded", self)
	
	Apollo.RegisterEventHandler("Group_Updated", 				"OnGroup_Updated", self)
	Apollo.RegisterEventHandler("Group_Join", 					"OnGroup_Join", self)
	Apollo.RegisterEventHandler("Group_Left", 					"OnGroup_Left", self)
	Apollo.RegisterEventHandler("Group_Add",					"OnGroup_Add", self)
	Apollo.RegisterEventHandler("Group_Remove",					"OnGroup_Remove", self) -- Kicked, or someone else leaves (yourself leaving is Group_Leave)
	Apollo.RegisterEventHandler("Group_MemberFlagsChanged",		"OnGroup_MemberFlagsChanged", self)
	Apollo.RegisterEventHandler("Group_FlagsChanged",			"OnGroup_FlagsChanged", self)	
	Apollo.RegisterEventHandler("UnitActivationTypeChanged", 	"OnUnitActivationTypeChanged", self)
	-- These events update the reward icons/quest framing
	--Apollo.RegisterEventHandler(, 			"OnQuestStateChanged", self)
	--Apollo.RegisterEventHandler(, 		"OnQuestObjectiveUpdated", self)
	--Apollo.RegisterEventHandler(, 			"OnContractStateChanged", self)
	--Apollo.RegisterEventHandler(, 	"OnContractObjectiveUpdated", self)

	--Apollo.RegisterEventHandler(, 			"OnPublicEventStart", self)
	--Apollo.RegisterEventHandler(, 	"OnPublicEventObjectiveUpdate", self)
	--Apollo.RegisterEventHandler(, 				"OnPublicEventEnd", self)
	--Apollo.RegisterEventHandler(,				"OnChallengeUpdate", self)
	--Apollo.RegisterEventHandler(, 			"OnChallengeUpdate", self)
	--Apollo.RegisterEventHandler(, 			"OnChallengeUpdate", self)
	--Apollo.RegisterEventHandler(, 			"OnChallengeUpdate", self)
	--Apollo.RegisterEventHandler(,			"OnChallengeActivate", self)
	--Apollo.RegisterEventHandler( 			"OnChallengeUpdate", self)
	--Apollo.RegisterEventHandler(, 		"OnChallengeUpdate", self)
	--Apollo.RegisterEventHandler(, 	"OnPlayerPathMissionChange", self)
	--Apollo.RegisterEventHandler(, 		"OnPlayerPathMissionChange", self)
	--Apollo.RegisterEventHandler(, 	"OnPlayerPathMissionChange", self)
	--Apollo.RegisterEventHandler(, 	"OnPlayerPathMissionChange", self)
	--Apollo.RegisterEventHandler(, 	"OnPlayerPathMissionChange", self)
	
	local tRewardUpdateEvents = {
		"QuestStateChanged", "QuestObjectiveUpdated", "ContractStateChanged","ContractObjectiveUpdated",
		"PublicEventStart", "PublicEventObjectiveUpdate","PublicEventEnd", "PublicEventUnitUpdate",
		"ChallengeAbandon", "ChallengeLeftArea", "ChallengeFailTime", "ChallengeFailArea", "ChallengeActivate",
		"ChallengeCompleted", "ChallengeFailGeneric", "ChallengeUpdated",
		"PlayerPathMissionUnlocked", "PlayerPathMissionUpdate", "PlayerPathMissionComplete", "PlayerPathMissionDeactivate",
		"PlayerPathMissionActivate", "PlayerPathRefresh", 
		"FriendshipAdd", "FriendshipPostRemove", "FriendshipUpdate"
	}
	for idx, event in pairs(tRewardUpdateEvents) do
		Apollo.RegisterEventHandler(event, "OnRewardInfoUpdated", self)
	end
	Apollo.RegisterEventHandler("RefreshHealthShieldBar", 		"OnRefreshHealthShieldBar", self)
	Apollo.RegisterEventHandler("MouseOverUnitChanged", 		"OnMouseOverUnitChanged", self)
	Apollo.RegisterEventHandler("BuffAdded", 					"OnBuffAdded", self)
	Apollo.RegisterEventHandler("BuffRemoved", 					"OnBuffRemoved", self)
	--Apollo.RegisterEventHandler("TargetThreatListUpdated",		"OnTargetThreatListUpdated", self)
	--Apollo.RegisterEventHandler("DamageOrHealingDone", 			"OnDamageOrHealingDone", self)
	
	Apollo.RegisterTimerHandler("InitialParseTimer", "OptionsChanged", self)
	Apollo.CreateTimer("InitialParseTimer", 1.0, false); Apollo.StopTimer("InitialParseTimer");
	
	-- timer to update buffs at a lesser interval
    Apollo.RegisterTimerHandler("BuffUpdater", "UpdateBuffs", self)
	Apollo.CreateTimer("BuffUpdater", buffUpdateInterval, true) 
	
	-- Load custom sprite sheet
	Apollo.LoadSprites("OptiPlates_Sprites.xml")
	
    -- load our forms
	self.xmlSettings = XmlDoc.CreateFromFile("OptiPlatesSettings.xml")
	self.xmlDoc = XmlDoc.CreateFromFile("OptiPlates.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
	self.wndMain = Apollo.LoadForm(self.xmlSettings, "SettingsWindow", nil, self)
	self.wndQuickChange = Apollo.LoadForm(self.xmlSettings, "QuickselectWindow", nil, self)

	self.wndMain:Show(false)
	self.wndQuickChange:Show(false)
	
	self:CreateFontList()
	self.optionsFrameUpdateList = { 1,2,3,4,5,6,7,8 }
	self.optionsStratumList = Apollo.GetStrata()
	self.optionsHPTextStyles = { "Normal", "Percentage", "NormalDeficit", "PercentageDeficit" }
	self.optionsHPTextStyleLabels = { Normal = "Normal", Percentage = "Percentage", NormalDeficit = "Normal - Deficit", PercentageDeficit = "Percentage - Deficit" }
	self.optionsCastbarTextStyles = { "Above", "Below", "Inside" }
	self.optionsCastbarTextStyleLabels = { Above = "Above castbar", Below = "Below castbar", Inside = "Inside castbar" }
	self.optionsCastbarTimeStyles = { "Always", "Hostile", "Never" }
	self.optionsCastbarTimeStyleLabels = { Always = "Always", Hostile = "Hostile units only", Never = "Never" }	
	self.optionsFocusTextStyles = { "Normal", "Percentage"}
	self.optionsFocusTextStyleLabels = { Normal = "Normal", Percentage = "Percentage" }	
	self.optionsLevelPositions = { "Bars", "NameLeft"}
	self.optionsLevelPositionLabels = { Bars = "Next to bars", NameLeft = "Next to name" }	
	self.optionsClassIconPositions = { "Top", "NameLeft"}
	self.optionsClassIconPositionLabels = { Top = "Above name", NameLeft = "Next to name" }
	self.optionsClassIconStyles = { "Icon", "Text", "IconAndText" }
	self.optionsClassIconStyleLabels = { Icon = "Icon", Text = "Text", IconAndText = "Icon and text" }
	
	self.bInitialLoadAllClear 				= false -- delays drawing until everything's come in
	self.refreshHealthShieldBar				= true -- refresh bars after RefreshHealthShieldBar event has fired
	self.arDisplayedNameplates 				= {} -- array of units with a nameplate window object (visible or hidden) 
	self.arUnit2Nameplate 					= {} -- array of unit ids to tNameplate structs
	self.arMount2Unit						= {}
	-- the following is an example of bad optimization (causes messy code). 
	-- array index 1 contains nameplate windows with at least 2 buff/debuff windows
	-- array index 2 contains nameplate windows with less than 2 buff/debuff windows 
	-- idea being to preferrably cache nameplates with buff windows already loaded
	self.arFreeWindows 						= {} -- Cached nameplate windows
	self.arFreeWindows[1] 					= {} -- cached nameplate windows with at least 2 buff windows
	self.arFreeWindows[2] 					= {} -- cached nameplate windows with less than 2 buff windows
	
	self.nFreeWindowsPoolLimit				= 100 -- How many unused nameplate windows are cached. Higher means more memory usage, very high values (>500) are not recommended.
	self.nFreeWindows						= 0 -- How many cached nameplates are available
	self.arColorObjects						= {} -- Cached color objects, populated in self:UpdateColorObjects
	self.arDamagedUnits						= {} -- damaged units since previous frame, updated in OnFrame(), used for bringing damaged nameplates to front
	self.unitLastTarget						= nil -- Current target, updated in self:TargetUnitChanged so previous target is available
	self.squaredDistanceLookupTable			= {} -- cache squared distance to real distance to avoid math.sqrt in OnFrame
	-- populate squared distance lookup table
	for i = 0, knTargetRange do
		self.squaredDistanceLookupTable[i] = math.floor(math.sqrt(i))
	end
	
	-- populate circle unit vectors (used with calculating camera heading)
	self.circleUnitVectorLookupTable = {}
	for i = 0, 360, 2 do
		local rads = i/180*math.pi
		self.circleUnitVectorLookupTable[i] = Vector3.New(math.sin(rads),0,math.cos(rads))
	end
	
	self.setCurrentProfileScope				= GameLib.CodeEnumAddonSaveLevel.Character -- initialize with default profile scope
	self.setCurrentProfileName				= "Default profile"	 -- initialize with default profile name
	self.tInitialNewUnits					= { } -- units that get an onunitcreated event before optionschanged() is first fired
	self.xmlLoaded 							= false
	-- the next is a stupid workaround because GetBuffTooltip() is only available with BuffWindows, and
	-- their buffid can't be changed from code. 32 extra windows, hooray.
	self.wndBuffTooltips					= Apollo.LoadForm(self.xmlDoc, "StupidWorkaroundBuffWindow", nil, self)
	self.wndDebuffTooltips					= Apollo.LoadForm(self.xmlDoc, "StupidWorkaroundDebuffWindow", nil, self)
	self.wndBuffTooltipRefs					= { }
	self.wndDebuffTooltipRefs				= { }
	self.arBuffCasters						= { }
	for j = 1, 16 do
		self.wndBuffTooltipRefs[j] = self.wndBuffTooltips:FindChild(string.format("BufftooltipWindow%d",j))
		self.wndDebuffTooltipRefs[j] = self.wndDebuffTooltips:FindChild(string.format("BufftooltipWindow%d",j))
	end
	-- end of stupid workaround
	self.bStyleSAO = false -- updated in OptionsChanged
	self.bStyleFlat = false -- updated in OptionsChanged
	self.nStyleWidth = 0 -- The width of the current style
	self.nStyleHalfWidth = 0 -- Half width of the current style (to prevent *0.5)
	self.frameCounter = 0 -- Will be initialized to self.setNameplateUpdateInterval whenever it is < 1 and gets decremented by 1 each frame
	self.clickTargetUnit = nil
	self.mouseOverUnit = nil
	self.arDamagedUnits = {}	
	self.arFriendlyUnitNameplates = {}
	
	self.bBlinded = false -- Has the current player blinded status
	self.tGroupUnits = {} -- Current group members in the form of groupId = tNameplate, updated in OnFrame as necessary
	self.bGroupInfoDirty = true -- Updates group info next frame if true
	self.bBuffProfilesDirty = true -- Recalculates everything buff-related next frame if true. Expensive, only used when changing buff profile settings.
	self.bRewardInfoDirty = true
	
	self:UpdateSettingsArrays()
	self:DefaultBuffSettings() -- Creates default buff settings etc
	self:DefaultSettings() -- Populates default settings and makes sure every tUserSettings variable has a value
	self:DefaultColors() -- Apply default colors
	self:UpdateColorObjects() -- Creates cached color objects
	self:ProcessDeprecatedSettings() -- Processes any deprecated settings	
end

function OptiPlates:OnDocLoaded()
	self.xmlLoaded = true
end

function OptiPlates:UpdateBuffs()
	updateBuffs = true;
end

function OptiPlates:CreateFontList()
	local gameFonts = Apollo.GetGameFonts()
	local hash = {}
	local tmpFonts = {}
	for _,v in ipairs(gameFonts) do
	   if (not hash[v.name]) then
		   tmpFonts[#tmpFonts+1] = v
		   hash[v.name] = true
	   end
	end	
	tmpFonts[#tmpFonts+1] = { name = "Nameplates", face = "Nameplates", size = 9, index = #tmpFonts }	
	table.sort(tmpFonts, 
		function(a, b)
			if a.size < b.size then
				return true
			elseif a.size == b.size then
				return a.name < b.name
			else 
				return false 
			end
		end
	)
	self.gameFonts = tmpFonts
end

-- Save User Settings
function OptiPlates:OnSave(eType)
	if eType == GameLib.CodeEnumAddonSaveLevel.Character then 
		local profiles = self.setProfiles[GameLib.CodeEnumAddonSaveLevel.Character]
		local saveArray = {}
		saveArray["currentProfileScope"] = self.setCurrentProfileScope
		saveArray["currentProfileName"] = self.setCurrentProfileName
		saveArray["quickselectShow"] = self.setQuickselectWindowShow
		saveArray["quickselectHideInCombat"] = self.setQuickselectWindowHideInCombat
		saveArray["quickselectPosition"] = self.setQuickselectWindowPosition
		saveArray["quickselectLocked"] = self.setQuickselectWindowLocked
		saveArray["quickselectNoButtons"] = self.setQuickselectWindowNoButtons
		saveArray["quickselectScaling"] = self.setQuickselectWindowScaling
		saveArray["profiles"] = profiles
		return saveArray
	elseif eType == GameLib.CodeEnumAddonSaveLevel.General then 
		local profiles = self.setProfiles[GameLib.CodeEnumAddonSaveLevel.General]
		local saveArray = {}
		saveArray["profiles"] = profiles
		for profileName,profile in pairs(self.setProfiles["Buffprofiles"]) do
			profile.tIncludeBuffs = {}
			profile.tIncludeDebuffs = {}
		end
		saveArray["buffprofiles"] = self.setProfiles["Buffprofiles"]
		return saveArray
	else
		return self.setProfiles[eType]
	end
end


-- Restore Saved User Settings
function OptiPlates:OnRestore(eType, t)
	if self.setProfiles == nil then self.setProfiles = {} end
	-- hopefully character-level save data is the last to be restored
	if eType == GameLib.CodeEnumAddonSaveLevel.Character then 
		self.setCurrentProfileScope = t["currentProfileScope"]
		self.setCurrentProfileName = t["currentProfileName"]
		self.setProfiles[eType] = t["profiles"]
		self:CheckSettingsSanity()
		self:LoadProfile(self.setCurrentProfileScope, self.setCurrentProfileName)
		self.setQuickselectWindowShow = (t["quickselectShow"] or tUserSettings.setQuickselectWindowShow.default)
		self.setQuickselectWindowHideInCombat = (t["quickselectHideInCombat"] or tUserSettings.setQuickselectWindowHideInCombat.default)
		self.setQuickselectWindowPosition = (t["quickselectPosition"] or tUserSettings.setQuickselectWindowPosition.default)
		self.setQuickselectWindowLocked	= (t["quickselectLocked"] or tUserSettings.setQuickselectWindowLocked.default)
		self.setQuickselectWindowScaling	= (t["quickselectScaling"] or tUserSettings.setQuickselectWindowScaling.default)
		self.setQuickselectWindowNoButtons	= (t["quickselectNoButtons"] or tUserSettings.setQuickselectWindowNoButtons.default)
	elseif eType == GameLib.CodeEnumAddonSaveLevel.General then 
		self.setProfiles[eType] = t["profiles"]
		self.setProfiles["Buffprofiles"] = t["buffprofiles"]
	else
		self.setProfiles[eType] = t
	end	
end


-- Open Settings Window
function OptiPlates:OnOptiPlatesOn()
	if #self.wndMain:FindChild("SettingsFrame"):GetChildren() == 0 then
		self:HelperChangeOptionMenu("SettingsGeneral")
	end
	if self.wndMain:FindChild("SettingsAbout") ~= nil then
		updateStatistics = true
	else
		updateStatistics = false
	end
	if self.wndMain:FindChild("SettingsNameplate") ~= nil then
		self.wndMain:FindChild("SettingsFrame"):DestroyChildren()
		self:HelperChangeOptionMenu("SettingsNameplate")
	end
	self.wndMain:Show(true)
end

function OptiPlates:UpdateSettingsArrays()
	local wndSettingsAbout = self.wndMain:FindChild("SettingsAbout")
	if wndSettingsAbout ~= nil then
		self.optionsUI_AboutSettings = { 
			Version = wndSettingsAbout:FindChild("Label_AboutVersionString"),
			MemoryUsage = wndSettingsAbout:FindChild("Label_MemoryUsage"),
			LastSecondTime = wndSettingsAbout:FindChild("Label_LastSecondTime"),
			CachedNameplates = wndSettingsAbout:FindChild("Label_CachedNameplates"),
			
			TotalUnits = wndSettingsAbout:FindChild("Label_TotalUnits"),
			TotalNameplates = wndSettingsAbout:FindChild("Label_TotalNameplates"),
			ShownNameplates = wndSettingsAbout:FindChild("Label_ShownNameplates"),
			ShownPlayerNameplates = wndSettingsAbout:FindChild("Label_ShownPlayerNameplates"),
			BuffWindows = wndSettingsAbout:FindChild("Label_BuffWindows"),
			VisibleBuffWindows = wndSettingsAbout:FindChild("Label_VisibleBuffWindows")
		}
	end
	-- Checkboxes can fire OptionCheckboxChanged as long as they are listed here,
	-- the self[var] variable will then be set to true/false if the checkbox is checked/unchecked
	-- OptionsChanged will also be fired
	self.optionsUI_CheckboxSettings = {
		-- Target nameplate settings
		{wnd = self.wndMain:FindChild("Button_ShowFocusTargetIcon"), var = "setShowFocusTargetIndicator"},
		
		-- Guild nameplate settings
		{wnd = self.wndMain:FindChild("Button_setGP_Override"), var = "setGP_Override"},

		-- Friendly NPC nameplate settings
		{wnd = self.wndMain:FindChild("Button_setF_Important"), var = "setF_Important"},
		{wnd = self.wndMain:FindChild("Button_setF_Vendor"), var = "setF_Vendor"},

		-- My nameplate settings
		{wnd = self.wndMain:FindChild("Button_HideMyFullBars"), var = "setHideFullFriendlyBarsMine", togglewnd = self.wndMain:FindChild("Hidemyfullbars Threshold settings"), parentheighton = 115, parentheightoff = 30},
		{wnd = self.wndMain:FindChild("Button_HideMyFullBarsOneThreshold"), var = "setHideFullFriendlyBarsMineOneThreshold"},
		{wnd = self.wndMain:FindChild("Button_HideFullFriendlyBars"), var = "setHideFullFriendlyBars", togglewnd = self.wndMain:FindChild("Hidefullbars Threshold settings"), parentheighton = 115, parentheightoff = 30},
		{wnd = self.wndMain:FindChild("Button_HideFullFriendlyBarsOneThreshold"), var = "setHideFullFriendlyBarsOneThreshold"},
	
		-- Pet settings
		{wnd = self.wndMain:FindChild("Button_setShowMyPets"), var = "setShowMyPets"},
		{wnd = self.wndMain:FindChild("Button_setShowFriendlyPets"), var = "setShowFriendlyPets"},
		{wnd = self.wndMain:FindChild("Button_setShowHostilePets"), var = "setShowHostilePets"},
		
		{wnd = self.wndMain:FindChild("Button_NeverShow"), var = "setNeverShow"},
		{wnd = self.wndMain:FindChild("Button_NeverShowCity"), var = "setNeverShowCity"},
		{wnd = self.wndMain:FindChild("Button_HideUnimportantFriendlyNameplates"), var = "setHideUnimportantNPCNameplates"},
		{wnd = self.wndMain:FindChild("Button_ShowInteractableAlways"), var = "setShowInteractableAlways"},
		{wnd = self.wndMain:FindChild("Button_ShowObjectivesAlways"), var = "setShowObjectivesAlways"},
		
		{wnd = self.wndMain:FindChild("Button_EnableColorPickerIntegration"), var = "setColorPickerIntegrationEnabled"},
		{wnd = self.wndMain:FindChild("Button_EnableOneVersionIntegration"), var = "setOneVersionIntegrationEnabled"},
		
		
		{wnd = self.wndMain:FindChild("Button_ShowCleanseAlert"), var = "setShowCleanseAlert"},
		
		{wnd = self.wndMain:FindChild("Button_IconQuest"), var = "setIconQuest"},
		{wnd = self.wndMain:FindChild("Button_IconMission"), var = "setIconMission"},
		{wnd = self.wndMain:FindChild("Button_IconFriend"), var = "setIconFriend"},
		{wnd = self.wndMain:FindChild("Button_IconEvent"), var = "setIconEvent"},
		{wnd = self.wndMain:FindChild("Button_IconChallenge"), var = "setIconChallenge"},
		{wnd = self.wndMain:FindChild("Button_IconRival"), var = "setIconRival"},
		{wnd = self.wndMain:FindChild("Button_IconTooltips"), var = "setIconTooltips"},
		
		{wnd = self.wndMain:FindChild("Button_ShowQuestIconNew"), var = "setShowQuestIconNew"},
		{wnd = self.wndMain:FindChild("Button_ShowQuestIconDaily"), var = "setShowQuestIconNewDaily"},
		{wnd = self.wndMain:FindChild("Button_ShowQuestIconGrey"), var = "setShowQuestIconNewGrey"},
		{wnd = self.wndMain:FindChild("Button_ShowQuestIconReward"), var = "setShowQuestIconReward"},
		{wnd = self.wndMain:FindChild("Button_ShowQuestIconReceiving"), var = "setShowQuestIconReceiving"},
		{wnd = self.wndMain:FindChild("Button_ShowQuestIconIgnoreDistance"), var = "setStyleQuestIconIgnoreMaximumDistance"},
		{wnd = self.wndMain:FindChild("Button_ShowQuestIconDefaultScaling"), var = "setStyleQuestIconUseDefaultScaling"},
		{wnd = self.wndMain:FindChild("Button_ShowQuestIconDistanceAlways"), var = "setStyleQuestIconAlwaysShowDistance"},
		
		{wnd = self.wndMain:FindChild("Button_ShowClassColors"), var = "setShowClassColors"},
		{wnd = self.wndMain:FindChild("Button_ShowAllyClassColors"), var = "setShowAllyClassColors"},
		
		
		{wnd = self.wndMain:FindChild("Button_HarvestablesEnabled"), var = "setHarvestablesEnabled"},
		{wnd = self.wndMain:FindChild("Button_HarvestablesDatacubeEnabled"), var = "setHarvestablesDatacubeEnabled"},
		{wnd = self.wndMain:FindChild("Button_HarvestablesMiningEnabled"), var = "setHarvestablesMiningEnabled"},
		{wnd = self.wndMain:FindChild("Button_HarvestablesTreeEnabled"), var = "setHarvestablesTreeEnabled"},
		{wnd = self.wndMain:FindChild("Button_HarvestablesRelicEnabled"), var = "setHarvestablesRelicEnabled"},
		{wnd = self.wndMain:FindChild("Button_HarvestablesPlantEnabled"), var = "setHarvestablesPlantEnabled"},
		{wnd = self.wndMain:FindChild("Button_HarvestablesDistanceEnabled"), var = "setHarvestablesDistanceEnabled"},
		{wnd = self.wndMain:FindChild("Button_HarvestablesIconEnabled"), var = "setHarvestablesIconEnabled"},
		{wnd = self.wndMain:FindChild("Button_HarvestablesTierEnabled"), var = "setHarvestablesTierEnabled"},
		{wnd = self.wndMain:FindChild("Button_HarvestablesDistanceScalingDisabled"), var = "setHarvestablesDistanceScalingDisabled"},
		{wnd = self.wndMain:FindChild("Button_HarvestablesNoDistanceHiding"), var = "setHarvestablesNoDistanceHiding"},
		
		{wnd = self.wndMain:FindChild("Button_NoTargetSettingsOnSelf"), var = "setNoTargetSettingsOnSelf"},
		
		{wnd = self.wndMain:FindChild("Button_StyleFlatNumberInside"), var = "setStyleFlatNumberInside"},
		{wnd = self.wndMain:FindChild("Button_HidePlatesWhenBlinded"), var = "setStyleHideBarsWhenBlinded"},
		{wnd = self.wndMain:FindChild("Button_CastbarShowMoO"), var = "setStyleCastbarShowMoO"},
		{wnd = self.wndMain:FindChild("Button_CastbarShowSubdue"), var = "setStyleCastbarShowSubdue"},
		{wnd = self.wndMain:FindChild("Button_CastbarTextAutoFit"), var = "setStyleCastbarTextAutoFit"},
		{wnd = self.wndMain:FindChild("Button_CastbarAboveResourceBar"), var = "setStyleCastbarAboveResourcebar", bneedsreload = true },
		{wnd = self.wndMain:FindChild("Button_CleanseIconShow"), var = "setStyleShowCleanseIcon" },
		{wnd = self.wndMain:FindChild("Button_CleanseOutlineShow"), var = "setStyleCleanseOutlineShow" },
		{wnd = self.wndMain:FindChild("Button_CleanseColorHealthBar"), var = "setStyleCleanseOutlineToHealthBar" },
		
		{wnd = self.wndMain:FindChild("Button_ShowClusterIcons"), var = "setShowTargetClusterIcons" },
		{wnd = self.wndMain:FindChild("Button_ShowLegacyTargetIcon"), var = "setShowLegacyTargetIndicator" },
		{wnd = self.wndMain:FindChild("Button_ShowNormalTargetIcon"), var = "setShowNormalTargetIndicator" },
		
		{wnd = self.wndMain:FindChild("Button_ResourcebarCombineFocus"), var = "setStyleCombineResourceBars", bneedsreload = true },
		
		{wnd = self.wndMain:FindChild("Button_ClassIconShowWarrior"), var = "setStyleClassIconShowWarrior" },
		{wnd = self.wndMain:FindChild("Button_ClassIconShowEngineer"), var = "setStyleClassIconShowEngineer" },
		{wnd = self.wndMain:FindChild("Button_ClassIconShowEsper"), var = "setStyleClassIconShowEsper" },
		{wnd = self.wndMain:FindChild("Button_ClassIconShowMedic"), var = "setStyleClassIconShowMedic" },
		{wnd = self.wndMain:FindChild("Button_ClassIconShowStalker"), var = "setStyleClassIconShowStalker" },
		{wnd = self.wndMain:FindChild("Button_ClassIconShowSpellslinger"), var = "setStyleClassIconShowSpellslinger" },
		{wnd = self.wndMain:FindChild("Button_RoleIconShowTank"), var = "setStyleRoleIconShowTank" },
		{wnd = self.wndMain:FindChild("Button_RoleIconShowHealer"), var = "setStyleRoleIconShowHealer" },
		{wnd = self.wndMain:FindChild("Button_RoleIconShowDPS"), var = "setStyleRoleIconShowDPS" },
		{wnd = self.wndMain:FindChild("Button_RoleIconShowNameplateAlways"), var = "setRoleIconAlwaysShowNameplate" },
		
		{wnd = self.wndMain:FindChild("Button_ShowHeadingText"), var = "setStyleHeadingShowText" },
		{wnd = self.wndMain:FindChild("Button_ShowHeadingBelowTargetMarker"), var = "setStyleHeadingArrowBelowTargetMarker" },
		
		{wnd = self.wndMain:FindChild("Button_DistanceOpacity"), var = "setDistanceOpacityEnabled", togglewnd = self.wndMain:FindChild("Distance Opacity Settings"), parentheighton = 170, parentheightoff = 30 },
		{wnd = self.wndMain:FindChild("Button_TargetNormalOpacity"), var = "setDistanceOpacityNormalTargetOpacity"},
		{wnd = self.wndMain:FindChild("Button_ObjectiveNormalOpacity"), var = "setDistanceOpacityNormalObjectiveOpacity"},
		{wnd = self.wndMain:FindChild("Button_PvpNameplateOcclusionEnable"), var = "setHostilePvpOcclusionEnabled", togglewnd = self.wndMain:FindChild("PvpNameplateOcclusionConfigWindow"), parentheighton = 30, parentheightoff = 30 },
		
		{wnd = self.wndMain:FindChild("Button_ScaleDistance"), var = "setDistanceScalingEnabled"},
		{wnd = self.wndMain:FindChild("Button_TargetNormalScale"), var = "setDistanceScalingNormalTargetScale"},
		{wnd = self.wndMain:FindChild("Button_ObjectiveNormalScale"), var = "setDistanceScalingNormalObjectiveScale"},
		{wnd = self.wndMain:FindChild("Button_ScaleHealth"), var = "setHealthDeficitScalingEnabled"},
		{wnd = self.wndMain:FindChild("Button_HealthIncludeHealth"), var = "setHealthDeficitScalingUseHealth"},
		{wnd = self.wndMain:FindChild("Button_HealthIncludeShield"), var = "setHealthDeficitScalingUseShield"},
		{wnd = self.wndMain:FindChild("Button_HealthOnlyIfMeInCombat"), var = "setHealthDeficitScalingPlayerInCombat"},
		{wnd = self.wndMain:FindChild("Button_HealthScalingOnlyFriendlyPlayers"), var = "setHealthDeficitScalingOnlyFriendlyPlayers"},
		
		{wnd = self.wndMain:FindChild("Button_NameplateSelfOnTop"), var = "setNameplateBringSelfToFront"},
		{wnd = self.wndMain:FindChild("Button_NameplateTargetOnTop"), var = "setNameplateBringTargetToFront"},
		{wnd = self.wndMain:FindChild("Button_NameplateFocusOnTop"), var = "setNameplateBringFocusToFront"},
		{wnd = self.wndMain:FindChild("Button_NameplateTankOnTop"), var = "setNameplateBringTankToFront"},
		{wnd = self.wndMain:FindChild("Button_NameplateHealerOnTop"), var = "setNameplateBringHealerToFront"},
		{wnd = self.wndMain:FindChild("Button_NameplateDamagedOnTop"), var = "setNameplateBringDamagedToFront"},
		
		{wnd = self.wndMain:FindChild("Button_AdjustOffscreenNameplates"), var = "setAutoAdjustNameplatesDown", togglewnd = self.wndMain:FindChild("Window_AdjustOffscreenNameplates"), parentheighton = 115, parentheightoff = 35 },
		
		{wnd = self.wndMain:FindChild("Button_CombatOpacity"), var = "setCombatOpacityEnabled", togglewnd = self.wndMain:FindChild("Combat Opacity Settings"), parentheighton = 170, parentheightoff = 30},
		{wnd = self.wndMain:FindChild("Button_OpacityOnlyIfMeInCombat"), var = "setCombatOpacityPlayerInCombat"},
		{wnd = self.wndMain:FindChild("Button_OpacityDefaultHostileCasting"), var = "setCombatOpacityDefaultHostileCasting"},
		{wnd = self.wndMain:FindChild("Button_QuickchangeHideInCombat"), var = "setQuickselectWindowHideInCombat"},
		{wnd = self.wndMain:FindChild("Button_QuickchangeShow"), var = "setQuickselectWindowShow"},
		{wnd = self.wndMain:FindChild("Button_QuickchangeNoButtons"), var = "setQuickselectWindowNoButtons"},
		
		
		{wnd = self.wndMain:FindChild("Button_MouseTargetingEnabled"), var = "setMouseTargetingEnabled"},
		{wnd = self.wndMain:FindChild("Button_MouseTargetingButtonLeft"), var = "setMouseTargetingButtonLeft"},
		{wnd = self.wndMain:FindChild("Button_MouseTargetingButtonMiddle"), var = "setMouseTargetingButtonMiddle"},
		{wnd = self.wndMain:FindChild("Button_MouseTargetingButtonRight"), var = "setMouseTargetingButtonRight"},
		{wnd = self.wndMain:FindChild("Button_MouseTargetingModifierCtrl"), var = "setMouseTargetingModifiersCtrl"},
		{wnd = self.wndMain:FindChild("Button_MouseTargetingModifierAlt"), var = "setMouseTargetingModifiersAlt"},
		{wnd = self.wndMain:FindChild("Button_MouseTargetingModifierShift"), var = "setMouseTargetingModifiersShift"},
		{wnd = self.wndMain:FindChild("Button_MouseTargetingClickThrough"), var = "setMouseTargetingClickthrough"},
		
		{wnd = self.wndMain:FindChild("Button_MouseOverEnabled"), var = "setMouseOverEnabled"},
		{wnd = self.wndMain:FindChild("Button_MouseOverModifierCtrl"), var = "setMouseOverModifiersCtrl"},
		{wnd = self.wndMain:FindChild("Button_MouseOverModifierAlt"), var = "setMouseOverModifiersAlt"},
		{wnd = self.wndMain:FindChild("Button_MouseOverModifierShift"), var = "setMouseOverModifiersShift"},
		{wnd = self.wndMain:FindChild("Button_MouseOverOptions_F_N"), var = "setMouseOverSettings_F_N"},
		{wnd = self.wndMain:FindChild("Button_MouseOverOptions_F_DI"), var = "setMouseOverSettings_F_DI"},
		{wnd = self.wndMain:FindChild("Button_MouseOverOptions_F_F"), var = "setMouseOverSettings_F_F"},
		{wnd = self.wndMain:FindChild("Button_MouseOverOptions_F_D"), var = "setMouseOverSettings_F_D"},
		{wnd = self.wndMain:FindChild("Button_MouseOverOptions_F_B"), var = "setMouseOverSettings_F_B"},
		{wnd = self.wndMain:FindChild("Button_MouseOverOptions_F_G"), var = "setMouseOverSettings_F_G"},
		{wnd = self.wndMain:FindChild("Button_MouseOverOptions_H_N"), var = "setMouseOverSettings_H_N"},
		{wnd = self.wndMain:FindChild("Button_MouseOverOptions_H_DI"), var = "setMouseOverSettings_H_DI"},
		{wnd = self.wndMain:FindChild("Button_MouseOverOptions_H_F"), var = "setMouseOverSettings_H_F"},
		{wnd = self.wndMain:FindChild("Button_MouseOverOptions_H_D"), var = "setMouseOverSettings_H_D"},
		{wnd = self.wndMain:FindChild("Button_MouseOverOptions_H_B"), var = "setMouseOverSettings_H_B"},
		{wnd = self.wndMain:FindChild("Button_MouseOverOptions_H_G"), var = "setMouseOverSettings_H_G"},		
		
	}

	-- Editboxes can fire OptionEditboxChanged as long as they are listed here,
	-- the self[var] variable will then be set to the text value of wnd.
	-- if colorlabel is not nil, colorlabel's text color also gets set to the value of var
	-- OptionsChanged will also be fired
	self.optionsUI_EditboxSettings = {
	
	
		{wnd = self.wndMain:FindChild("Edit_BackgroundBar"), var = "setColorBackgroundBar", colorlabel = self.wndMain:FindChild("Label_BackgroundBar")},
		{wnd = self.wndMain:FindChild("Edit_BackdropBar"), var = "setColorBackdropBar", colorlabel = self.wndMain:FindChild("Label_BackdropBar")},
		
		{wnd = self.wndMain:FindChild("Edit_ShieldBar"), var = "setColorShieldBar", colorlabel = self.wndMain:FindChild("Label_ShieldBar")},
		{wnd = self.wndMain:FindChild("Edit_AbsorbBar"), var = "setColorAbsorbBar", colorlabel = self.wndMain:FindChild("Label_AbsorbBar")},
		{wnd = self.wndMain:FindChild("Edit_VulnerableBar"), var = "setColorVulnerableBar", colorlabel = self.wndMain:FindChild("Label_VulnerableBar")},
		{wnd = self.wndMain:FindChild("Edit_VulnerableTimerBar"), var = "setColorVulnerableTimerBar", colorlabel = self.wndMain:FindChild("Label_VulnerableTimerBar")},
		{wnd = self.wndMain:FindChild("Edit_FriendlyCastBar"), var = "setColorFriendlyCastBar", colorlabel = self.wndMain:FindChild("Label_FriendlyCastBar")},
		{wnd = self.wndMain:FindChild("Edit_HostileCastBar"), var = "setColorHostileCastBar", colorlabel = self.wndMain:FindChild("Label_HostileCastBar")},

		{wnd = self.wndMain:FindChild("Edit_NeutralName"), var = "setColorNeutralName", colorlabel = self.wndMain:FindChild("Label_NeutralName")},
		{wnd = self.wndMain:FindChild("Edit_NeutralBar"), var = "setColorNeutralBar", colorlabel = self.wndMain:FindChild("Label_NeutralBar")},
		{wnd = self.wndMain:FindChild("Edit_NeutralBarCutoff"), var = "setColorNeutralBarCutoff", colorlabel = self.wndMain:FindChild("Label_NeutralBarCutoff")},
	
		{wnd = self.wndMain:FindChild("Edit_FriendlyName"), var = "setColorFriendlyName", colorlabel = self.wndMain:FindChild("Label_FriendlyName")},
		{wnd = self.wndMain:FindChild("Edit_FriendlyBar"), var = "setColorFriendlyBar", colorlabel = self.wndMain:FindChild("Label_FriendlyBar")},
		{wnd = self.wndMain:FindChild("Edit_FriendlyBarCutoff"), var = "setColorFriendlyBarCutoff", colorlabel = self.wndMain:FindChild("Label_FriendlyBarCutoff")},	
		{wnd = self.wndMain:FindChild("Edit_FriendlyPlayerName"), var = "setColorFriendlyPlayerName", colorlabel = self.wndMain:FindChild("Label_FriendlyPlayerName")},
		{wnd = self.wndMain:FindChild("Edit_FriendlyPlayerNameUnflagged"), var = "setColorFriendlyPlayerNameUnflagged", colorlabel = self.wndMain:FindChild("Label_FriendlyPlayerNameUnflagged")},
		{wnd = self.wndMain:FindChild("Edit_FriendlyPlayerBar"), var = "setColorFriendlyPlayerBar", colorlabel = self.wndMain:FindChild("Label_FriendlyPlayerBar")},
		{wnd = self.wndMain:FindChild("Edit_FriendlyPlayerBarCutoff"), var = "setColorFriendlyPlayerBarCutoff", colorlabel = self.wndMain:FindChild("Label_FriendlyPlayerBarCutoff")},	

		{wnd = self.wndMain:FindChild("Edit_HostileName"), var = "setColorHostileName", colorlabel = self.wndMain:FindChild("Label_HostileName")},
		{wnd = self.wndMain:FindChild("Edit_HostileBar"), var = "setColorHostileBar", colorlabel = self.wndMain:FindChild("Label_HostileBar")},
		{wnd = self.wndMain:FindChild("Edit_HostileBarCutoff"), var = "setColorHostileBarCutoff", colorlabel = self.wndMain:FindChild("Label_HostileBarCutoff")},	
		{wnd = self.wndMain:FindChild("Edit_HostilePlayerName"), var = "setColorHostilePlayerName", colorlabel = self.wndMain:FindChild("Label_HostilePlayerName")},
		{wnd = self.wndMain:FindChild("Edit_HostilePlayerNameUnflagged"), var = "setColorHostilePlayerNameUnflagged", colorlabel = self.wndMain:FindChild("Label_HostilePlayerNameUnflagged")},
		{wnd = self.wndMain:FindChild("Edit_HostilePlayerBar"), var = "setColorHostilePlayerBar", colorlabel = self.wndMain:FindChild("Label_HostilePlayerBar")},
		{wnd = self.wndMain:FindChild("Edit_HostilePlayerBarCutoff"), var = "setColorHostilePlayerBarCutoff", colorlabel = self.wndMain:FindChild("Label_HostilePlayerBarCutoff")},	
		
		{wnd = self.wndMain:FindChild("Edit_PartyName"), var = "setColorPartyName", colorlabel = self.wndMain:FindChild("Label_PartyName")},		
		{wnd = self.wndMain:FindChild("Edit_GuildName"), var = "setColorGuildName", colorlabel = self.wndMain:FindChild("Label_GuildName")},		

		{wnd = self.wndMain:FindChild("Edit_WarriorBar"), var = "setColorWarriorBar", colorlabel = self.wndMain:FindChild("Label_WarriorBar")},
		{wnd = self.wndMain:FindChild("Edit_StalkerBar"), var = "setColorStalkerBar", colorlabel = self.wndMain:FindChild("Label_StalkerBar")},
		{wnd = self.wndMain:FindChild("Edit_EngineerBar"), var = "setColorEngineerBar", colorlabel = self.wndMain:FindChild("Label_EngineerBar")},
		{wnd = self.wndMain:FindChild("Edit_EsperBar"), var = "setColorEsperBar", colorlabel = self.wndMain:FindChild("Label_EsperBar")},
		{wnd = self.wndMain:FindChild("Edit_SpellslingerBar"), var = "setColorSpellslingerBar", colorlabel = self.wndMain:FindChild("Label_SpellslingerBar")},
		{wnd = self.wndMain:FindChild("Edit_MedicBar"), var = "setColorMedicBar", colorlabel = self.wndMain:FindChild("Label_MedicBar")},
		
		{wnd = self.wndMain:FindChild("Edit_ColorThreatOutline"), var = "setColorThreatOutline", colorlabel = self.wndMain:FindChild("Label_ColorThreatOutline")},
		{wnd = self.wndMain:FindChild("Edit_ColorCleanseOutline"), var = "setColorCleanseOutline", colorlabel = self.wndMain:FindChild("Label_ColorCleanseOutline")},
		{wnd = self.wndMain:FindChild("Edit_ColorHarvestables"), var = "setColorHarvestables", colorlabel = self.wndMain:FindChild("Label_ColorHarvestables")},
		{wnd = self.wndMain:FindChild("Edit_ColorFriendlyShieldBarCutoff"), var = "setColorFriendlyShieldBarCutoff", colorlabel = self.wndMain:FindChild("Label_ColorFriendlyShieldBarCutoff")},
		{wnd = self.wndMain:FindChild("Edit_ColorFocusBar"), var = "setColorFocusBar", colorlabel = self.wndMain:FindChild("Label_ColorFocusBar")},
		{wnd = self.wndMain:FindChild("Edit_ColorFocusBarText"), var = "setColorFocusBarText", colorlabel = self.wndMain:FindChild("Label_ColorFocusBarText")},
		{wnd = self.wndMain:FindChild("Edit_ColorResourceBar"), var = "setColorResourceBar", colorlabel = self.wndMain:FindChild("Label_ColorResourceBar")},
		{wnd = self.wndMain:FindChild("Edit_ColorResourceBarText"), var = "setColorResourceBarText", colorlabel = self.wndMain:FindChild("Label_ColorResourceBarText")},
		{wnd = self.wndMain:FindChild("Edit_ColorDashBar"), var = "setColorDashBar", colorlabel = self.wndMain:FindChild("Label_ColorDashBar")},
		{wnd = self.wndMain:FindChild("Edit_ColorDashBarText"), var = "setColorDashBarText", colorlabel = self.wndMain:FindChild("Label_ColorDashBarText")},
		{wnd = self.wndMain:FindChild("Edit_ColorSprintBar"), var = "setColorSprintBar", colorlabel = self.wndMain:FindChild("Label_ColorSprintBar")},
		{wnd = self.wndMain:FindChild("Edit_ColorSprintBarText"), var = "setColorSprintBarText", colorlabel = self.wndMain:FindChild("Label_ColorSprintBarText")},
		{wnd = self.wndMain:FindChild("Edit_ColorLevelText"), var = "setColorLevelText", colorlabel = self.wndMain:FindChild("Label_ColorLevelText")},
		
		{wnd = self.wndMain:FindChild("Edit_ColorHeadingIndicatorArrow"), var = "setColorHeadingIndicatorArrow", colorlabel = self.wndMain:FindChild("Label_ColorHeadingIndicatorArrow")},
		{wnd = self.wndMain:FindChild("Edit_ColorHeadingIndicatorText"), var = "setColorHeadingIndicatorText", colorlabel = self.wndMain:FindChild("Label_ColorHeadingIndicatorText")},
		
		{wnd = self.wndMain:FindChild("EditBox_ThreatIconSprite"), var = "setStyleThreatIconSprite"},
		{wnd = self.wndMain:FindChild("EditBox_CleanseIconSprite"), var = "setStyleCleanseIconSprite"},
		
		{wnd = self.wndMain:FindChild("EditBox_FriendlyTargetSprite"), var = "setFriendlyTargetIndicatorIcon"},
		{wnd = self.wndMain:FindChild("EditBox_PartyTargetSprite"), var = "setPartyTargetIndicatorIcon"},
		{wnd = self.wndMain:FindChild("EditBox_NeutralTargetSprite"), var = "setNeutralTargetIndicatorIcon"},
		{wnd = self.wndMain:FindChild("EditBox_HostileTargetSprite"), var = "setHostileTargetIndicatorIcon"},
		
		{wnd = self.wndMain:FindChild("EditBox_FriendlyAltTargetSprite"), var = "setFriendlyAltTargetIndicatorIcon"},
		{wnd = self.wndMain:FindChild("EditBox_NeutralAltTargetSprite"), var = "setNeutralAltTargetIndicatorIcon"},
		{wnd = self.wndMain:FindChild("EditBox_HostileAltTargetSprite"), var = "setHostileAltTargetIndicatorIcon"},
		
		{wnd = self.wndMain:FindChild("EditBox_FriendlyClusterSprite"), var = "setFriendlyClusterIndicatorIcon"},
		{wnd = self.wndMain:FindChild("EditBox_NeutralClusterSprite"), var = "setNeutralClusterIndicatorIcon"},
		{wnd = self.wndMain:FindChild("EditBox_HostileClusterSprite"), var = "setHostileClusterIndicatorIcon"},
		
		{wnd = self.wndMain:FindChild("EditBox_PvpNameplateOcclusionTime"), var = "setHostilePvpOcclusionTimeout", vartype = "number" }
	}	
	-- Sliders can fire OptionSliderChanged as long as they are listed here,
	-- the self[var] variable will then be set to the value in wnd
	-- if label is defined, label's text will get set to the same value
	-- if stringformat is defined, label's text will get formatted according to the format
	-- if divideby is defined, the value of var will be divided by divideby before getting shown in label. divideby DOES NOT alter the value set in self[var]
	-- OptionsChanged will also be fired	
	self.optionsUI_SliderSettings = {
		{wnd = self.wndMain:FindChild("Slider_StyleWidth"), var = "setStyleWidth", label = self.wndMain:FindChild("Label_StyleWidthDisplay"), stringformat = "%d" },
		{wnd = self.wndMain:FindChild("Slider_ShieldHeight"), var = "setShieldbarHeight", label = self.wndMain:FindChild("Label_ShieldHeightDisplay"), stringformat = "%d" },
		{wnd = self.wndMain:FindChild("Slider_HealthHeight"), var = "setHealthbarHeight", label = self.wndMain:FindChild("Label_HealthHeightDisplay"), stringformat = "%d" },
		{wnd = self.wndMain:FindChild("Slider_FlatShieldTexture"), var = "setFlatShieldTexture", label = self.wndMain:FindChild("Label_FlatShieldTextureDisplay"), stringformat = "%d" },
		{wnd = self.wndMain:FindChild("Slider_FlatHealthTexture"), var = "setFlatHealthTexture", label = self.wndMain:FindChild("Label_FlatHealthTextureDisplay"), stringformat = "%d" },
		
		{wnd = self.wndMain:FindChild("Slider_DrawDistance"), var = "setDrawDistance", label = self.wndMain:FindChild("Label_DrawDistanceDisplay"), stringformat = "%dm" },
		{wnd = self.wndMain:FindChild("Slider_HealthCutoff"), var = "setHealthCutoff", label = self.wndMain:FindChild("Label_HealthCutoffDisplay"), stringformat = "%d%%" },
		
		{wnd = self.wndMain:FindChild("Slider_HideFullFriendlyBarsShieldThreshold"), var = "setHideFullFriendlyBarsShieldThreshold", label = self.wndMain:FindChild("Label_HideFullFriendlyBarsShieldThresholdDisplay"), stringformat = "%d%%" },
		{wnd = self.wndMain:FindChild("Slider_HideFullFriendlyBarsHealthThreshold"), var = "setHideFullFriendlyBarsHealthThreshold", label = self.wndMain:FindChild("Label_HideFullFriendlyBarsHealthThresholdDisplay"), stringformat = "%d%%" },
		
		{wnd = self.wndMain:FindChild("Slider_HideMyFullBarsShieldThreshold"), var = "setHideFullFriendlyBarsMineShieldThreshold", label = self.wndMain:FindChild("Label_HideMyFullBarsShieldThresholdDisplay"), stringformat = "%d%%" },
		{wnd = self.wndMain:FindChild("Slider_HideMyFullBarsHealthThreshold"), var = "setHideFullFriendlyBarsMineHealthThreshold", label = self.wndMain:FindChild("Label_HideMyFullBarsHealthThresholdDisplay"), stringformat = "%d%%" },
	
		{wnd = self.wndMain:FindChild("Slider_NameplateVerticalOffset"), var = "setNameplateVerticalOffset", label = self.wndMain:FindChild("Label_NameplateVerticalOffsetLabel"), stringformat = "%dpx"},
		{wnd = self.wndMain:FindChild("Slider_MyNameplateVerticalOffset"), var = "setNameplateVerticalOffsetMine", label = self.wndMain:FindChild("Label_MyNameplateVerticalOffsetLabel"), stringformat = "%dpx"},
		{wnd = self.wndMain:FindChild("Slider_AdjustOffscreenNameplatesTop"), var = "setOffscreenNameplatesTopOffset", label = self.wndMain:FindChild("Label_AdjustOffscreenNameplatesTopLabel"), stringformat = "%dpx"},
		{wnd = self.wndMain:FindChild("Slider_AdjustOffscreenNameplatesMinHeight"), var = "setOffscreenNameplatesMinHeight", label = self.wndMain:FindChild("Label_AdjustOffscreenNameplatesMinHeightLabel"), stringformat = "%dpx"},
		{wnd = self.wndMain:FindChild("Slider_AdjustOffscreenNameplatesAnchorDownOffset"), var = "setOffscreenNameplatesAdjustAnchorOffset", label = self.wndMain:FindChild("Label_AdjustOffscreenNameplatesAnchorDownOffsetLabel"), stringformat = "%dpx"},
		
		{wnd = self.wndMain:FindChild("Slider_CastbarHeight"), var = "setStyleCastbarHeight", label = self.wndMain:FindChild("Label_CastbarHeightDisplay"), stringformat = "%dpx"},
		{wnd = self.wndMain:FindChild("Slider_CastbarExtraHeight"), var = "setStyleCastbarPaddingTop", label = self.wndMain:FindChild("Label_CastbarExtraHeightDisplay"), stringformat = "%dpx"},
		{wnd = self.wndMain:FindChild("Slider_CastbarWidth"), var = "setStyleCastbarWidth", label = self.wndMain:FindChild("Label_CastbarWidthDisplay"), stringformat = "%dpx"},
		{wnd = self.wndMain:FindChild("Slider_ResourcebarHeight"), var = "setStyleResourcebarHeight", label = self.wndMain:FindChild("Label_ResourcebarHeightDisplay"), stringformat = "%dpx"},
		{wnd = self.wndMain:FindChild("Slider_FocusbarHeight"), var = "setStyleFocusbarHeight", label = self.wndMain:FindChild("Label_FocusbarHeightDisplay"), stringformat = "%dpx"},
		{wnd = self.wndMain:FindChild("Slider_CleanseIconSize"), var = "setStyleCleanseIconSize", label = self.wndMain:FindChild("Label_CleanseIconSizeDisplay"), stringformat = "%dpx"},
		
		{wnd = self.wndMain:FindChild("Slider_ClassIconSize"), var = "setStyleClassIconSize", label = self.wndMain:FindChild("Label_ClassIconSizeDisplay"), stringformat = "%dpx"},
		{wnd = self.wndMain:FindChild("Slider_RoleIconSize"), var = "setStyleRoleIconSize", label = self.wndMain:FindChild("Label_RoleIconSizeDisplay"), stringformat = "%dpx"},
		{wnd = self.wndMain:FindChild("Slider_TargetIconSize"), var = "setStyleTargetedIconSize", label = self.wndMain:FindChild("Label_TargetIconSizeDisplay"), stringformat = "%dpx"},
		{wnd = self.wndMain:FindChild("Slider_HeadingArrowSize"), var = "setStyleHeadingArrowSize", label = self.wndMain:FindChild("Label_HeadingArrowSizeDisplay"), stringformat = "%dpx"},
		{wnd = self.wndMain:FindChild("Slider_QuestIconSize"), var = "setStyleQuestIconSize", label = self.wndMain:FindChild("Label_QuestIconSizeDisplay"), stringformat = "%dpx"},
		{wnd = self.wndMain:FindChild("Slider_QuestIconMinimumDistance"), var = "setStyleQuestIconMinimumDistance", label = self.wndMain:FindChild("Label_QuestIconMinimumDistanceDisplay"), stringformat = "%dm"},
		
		{wnd = self.wndMain:FindChild("Slider_DefaultScale"), var = "setDefaultScale", label = self.wndMain:FindChild("Label_DefaultScaleDisplay"), stringformat = "%.2f", divideby = 100 },
		{wnd = self.wndMain:FindChild("Slider_ScaleTarget"), var = "setTargetDefaultScale", label = self.wndMain:FindChild("Label_ScaleTargetDisplay"), stringformat = "%.2f", divideby = 100 },
		{wnd = self.wndMain:FindChild("Slider_ScaleAltTarget"), var = "setAltTargetDefaultScale", label = self.wndMain:FindChild("Label_ScaleAltTargetDisplay"), stringformat = "%.2f", divideby = 100 },
		{wnd = self.wndMain:FindChild("Slider_ScaleMinChange"), var = "setScalingUpdateMinChange", label = self.wndMain:FindChild("Label_ScaleMinChangeDisplay"), stringformat = "%.2f", divideby = 100 },
		{wnd = self.wndMain:FindChild("Slider_MinimumDistance"), var = "setDistanceScalingMinRange", label = self.wndMain:FindChild("Label_MinimumDistanceDisplay"), stringformat = "%dm" },
		{wnd = self.wndMain:FindChild("Slider_MaximumDistance"), var = "setDistanceScalingMaxRange", label = self.wndMain:FindChild("Label_MaximumDistanceDisplay"), stringformat = "%dm" },
		{wnd = self.wndMain:FindChild("Slider_DistanceScaleFactor"), var = "setDistanceScalingScaleFactor", label = self.wndMain:FindChild("Label_DistanceScaleFactorDisplay"), stringformat = "%.2f", divideby = 100 },
		{wnd = self.wndMain:FindChild("Slider_MinimumHealth"), var = "setHealthDeficitScalingMinCutoff", label = self.wndMain:FindChild("Label_MinimumHealthDisplay"), stringformat = "%d%%" },
		{wnd = self.wndMain:FindChild("Slider_MaximumHealth"), var = "setHealthDeficitScalingMaxCutoff", label = self.wndMain:FindChild("Label_MaximumHealthDisplay"), stringformat = "%d%%" },
		{wnd = self.wndMain:FindChild("Slider_HealthScaleFactor"), var = "setHealthDeficitScalingMaxChange", label = self.wndMain:FindChild("Label_HealthScaleFactorDisplay"), stringformat = "%.2f", divideby = 100 },
		
		{wnd = self.wndMain:FindChild("Slider_ShieldThresholdZ"), var = "setNameplateBringDamagedToFrontShieldThreshold", label = self.wndMain:FindChild("Label_ShieldThresholdZDisplay"), stringformat = "%d%%" },
		{wnd = self.wndMain:FindChild("Slider_HealthThresholdZ"), var = "setNameplateBringDamagedToFrontHealthThreshold", label = self.wndMain:FindChild("Label_HealthThresholdZDisplay"), stringformat = "%d%%" },
		
		{wnd = self.wndMain:FindChild("Slider_DefaultOpacity"), var = "setDefaultOpacity", label = self.wndMain:FindChild("Label_DefaultOpacityDisplay"), stringformat = "%.2f", divideby = 100 },
		{wnd = self.wndMain:FindChild("Slider_BarBGOpacity"), var = "setBarBGOpacity", label = self.wndMain:FindChild("Label_BarBGOpacityDisplay"), stringformat = "%.2f", divideby = 100 },
		{wnd = self.wndMain:FindChild("Slider_SpeechBubbleOpacity"), var = "setSpeechBubbleOpacity", label = self.wndMain:FindChild("Label_SpeechBubbleOpacityDisplay"), stringformat = "%.2f", divideby = 100 },
		{wnd = self.wndMain:FindChild("Slider_FriendlyCombatOpacity"), var = "setCombatOpacityFriendly", label = self.wndMain:FindChild("Label_FriendlyCombatOpacityDisplay"), stringformat = "%.2f", divideby = 100 },
		{wnd = self.wndMain:FindChild("Slider_HostileCombatOpacity"), var = "setCombatOpacityHostile", label = self.wndMain:FindChild("Label_HostileCombatOpacityDisplay"), stringformat = "%.2f", divideby = 100 },
		{wnd = self.wndMain:FindChild("Slider_TargetOpacityEmphasis"), var = "setTargetOpacityEmphasis", label = self.wndMain:FindChild("Label_TargetOpacityEmphasisDisplay"), stringformat = "%.2f", divideby = 100 },

		{wnd = self.wndMain:FindChild("Slider_OpacityMaximumDistance"), var = "setDistanceOpacityMaxRange", label = self.wndMain:FindChild("Label_OpacityMaximumDistanceDisplay"), stringformat = "%dm" },
		{wnd = self.wndMain:FindChild("Slider_DistanceOpacityChange"), var = "setDistanceOpacityFactor", label = self.wndMain:FindChild("Label_DistanceOpacityChangeDisplay"), stringformat = "%.2f", divideby = 100 },
		
		{wnd = self.wndMain:FindChild("Slider_QuickchangeScaling"), var = "setQuickselectWindowScaling", label = self.wndMain:FindChild("Label_QuickchangeScalingDisplay"), stringformat = "%.2f", divideby = 100 }
	}
	

	self.optionsUI_DropdownListSettings = {
		{ grid = self.wndMain:FindChild("Grid_ChangeDefaultStratumList"),wnd = self.wndMain:FindChild("Window_ChangeDefaultStratumContainer"), btn = self.wndMain:FindChild("Button_ChangeDefaultStratum"), array = "optionsStratumList", var = "setNameplateDefaultStratum" },
		{ grid = self.wndMain:FindChild("Grid_UpdateFrameIntervalList"),wnd = self.wndMain:FindChild("Window_UpdateFrameIntervalContainer"), btn = self.wndMain:FindChild("Button_UpdateFrameInterval"), array = "optionsFrameUpdateList", var = "setNameplateUpdateInterval" },
		{ grid = self.wndMain:FindChild("Grid_HPTextStyleList"),wnd = self.wndMain:FindChild("Window_HPTextStyleListContainer"), btn = self.wndMain:FindChild("Button_HPTextStyleToggle"), array = "optionsHPTextStyles", labelarray = "optionsHPTextStyleLabels", var = "setHPTextStyle" },
		{ grid = self.wndMain:FindChild("Grid_SPTextStyleList"),wnd = self.wndMain:FindChild("Window_SPTextStyleListContainer"), btn = self.wndMain:FindChild("Button_SPTextStyleToggle"), array = "optionsHPTextStyles", labelarray = "optionsHPTextStyleLabels", var = "setSPTextStyle" },
		{ grid = self.wndMain:FindChild("Grid_LevelTextPositionList"),wnd = self.wndMain:FindChild("Window_LevelTextPositionListContainer"), btn = self.wndMain:FindChild("Button_LevelTextPositionToggle"), array = "optionsLevelPositions", labelarray = "optionsLevelPositionLabels", var = "setLevelTextPosition" },
		{ grid = self.wndMain:FindChild("Grid_ClassIconPositionList"),wnd = self.wndMain:FindChild("Window_ClassIconPositionListContainer"), btn = self.wndMain:FindChild("Button_ClassIconPositionToggle"), array = "optionsClassIconPositions", labelarray = "optionsClassIconPositionLabels", var = "setClassIconPosition" },
		{ grid = self.wndMain:FindChild("Grid_CastbarTextStyleList"),wnd = self.wndMain:FindChild("Window_CastbarTextStyleListContainer"), btn = self.wndMain:FindChild("Button_CastbarTextStyleToggle"), array = "optionsCastbarTextStyles", labelarray = "optionsCastbarTextStyleLabels", var = "setCastbarTextStyle" },
		{ grid = self.wndMain:FindChild("Grid_CastbarTimeStyleList"),wnd = self.wndMain:FindChild("Window_CastbarTimeStyleListContainer"), btn = self.wndMain:FindChild("Button_CastbarTimeStyleToggle"), array = "optionsCastbarTimeStyles", labelarray = "optionsCastbarTimeStyleLabels", var = "setCastbarTimeStyle" },
		{ grid = self.wndMain:FindChild("Grid_FocusbarTextStyleList"),wnd = self.wndMain:FindChild("Window_FocusbarTextStyleListContainer"), btn = self.wndMain:FindChild("Button_FocusbarTextStyleToggle"), array = "optionsFocusTextStyles", labelarray = "optionsFocusTextStyleLabels", var = "setFocusTextStyle" },
		{ grid = self.wndMain:FindChild("Grid_ClassIconStyleList"),wnd = self.wndMain:FindChild("Window_ClassIconStyleListContainer"), btn = self.wndMain:FindChild("Button_ClassIconStyleToggle"), array = "optionsClassIconStyles", labelarray = "optionsClassIconStyleLabels", var = "setClassIconStyle" },
	
	}
	
	-- Nav buttons should be here and fire OptionNavChanged
	-- btn is the nav button, and wnd the window to show
	self.optionsUI_NavSettings = {
		{ btn = self.wndMain:FindChild("Nav_General") , wnd = "SettingsGeneral"},
		{ btn = self.wndMain:FindChild("Nav_Nameplate") , wnd = "SettingsNameplate"},
		{ btn = self.wndMain:FindChild("Nav_Colors") , wnd = "SettingsColors"},
		{ btn = self.wndMain:FindChild("Nav_Style") , wnd = "SettingsStyle"},
		{ btn = self.wndMain:FindChild("Nav_Scale") , wnd = "SettingsScale"},
		{ btn = self.wndMain:FindChild("Nav_Opacity") , wnd = "SettingsOpacity"},
		{ btn = self.wndMain:FindChild("Nav_Profiles") , wnd = "SettingsProfiles"},
		{ btn = self.wndMain:FindChild("Nav_Harvesting") , wnd = "SettingsHarvesting"},
		{ btn = self.wndMain:FindChild("Nav_Buffs") , wnd = "SettingsBuffs"},
		{ btn = self.wndMain:FindChild("Nav_Advanced") , wnd = "SettingsAdvanced"},
		{ btn = self.wndMain:FindChild("Nav_Mouse") , wnd = "SettingsMouse"},
		{ btn = self.wndMain:FindChild("Nav_About") , wnd = "SettingsAbout", toggleStats = true}
	}
end


-- Set Default Settings Options
function OptiPlates:DefaultSettings()
	-- copy all default variables to self.*
	for propKey,prop in pairs(tUserSettings) do self[propKey] = prop.default end
	
	if self.setProfiles == nil then self.setProfiles = {} end
	if self.setProfiles[GameLib.CodeEnumAddonSaveLevel.Character] == nil then
		self.setProfiles[GameLib.CodeEnumAddonSaveLevel.Character] = {}
	end
	if self.setProfiles[GameLib.CodeEnumAddonSaveLevel.Realm] == nil then
		self.setProfiles[GameLib.CodeEnumAddonSaveLevel.Realm] = {}
	end
	if self.setProfiles[GameLib.CodeEnumAddonSaveLevel.Account] == nil then
		self.setProfiles[GameLib.CodeEnumAddonSaveLevel.Account] = {}
	end
	if self.setProfiles[GameLib.CodeEnumAddonSaveLevel.General] == nil then
		self.setProfiles[GameLib.CodeEnumAddonSaveLevel.General] = {}
	end
	
	if self.setProfiles[self.setCurrentProfileScope][self.setCurrentProfileName] == nil then
		self.setProfiles[self.setCurrentProfileScope][self.setCurrentProfileName] = {}
	end
	
	self.scopeTextArray = {}
	self.scopeTextArray[GameLib.CodeEnumAddonSaveLevel.Character] = "Character"
	self.scopeTextArray[GameLib.CodeEnumAddonSaveLevel.Realm] = "Realm"	
	self.scopeTextArray[GameLib.CodeEnumAddonSaveLevel.Account] = "Account"
	self.scopeTextArray[GameLib.CodeEnumAddonSaveLevel.General] = "General"	


end

function OptiPlates:DefaultBuffSettings()
	self.defaultBuffProfileSettings = {
		bIncludeAllBuffs = false,
		bExcludeAllBuffs = true,
		bIncludeAllDebuffs = false,
		bExcludeAllDebuffs = true,
		bIncludeAllDispellableBuffs = false, --Spell.CodeEnumSpellClass.BuffDispellable
		bIncludeAllNonDispellableBuffs = false, --Spell.CodeEnumSpellClass.BuffNonDispellable and Spell.CodeEnumSpellClass.BuffNonDispellableRightClickOk
		bIncludeAllDispellableDeBuffs = false, --Spell.CodeEnumSpellClass.DebuffDispellable
		bIncludeAllNonDispellableDeBuffs = false, --Spell.CodeEnumSpellClass.DebuffNonDispellable
		bBuffOnlyShowMine = false,
		bDebuffOnlyShowMine = false,
		
		sIncludeBuffs = "",
		tIncludeBuffs = {  },
		tIncludeTextBuffs = {  },
		sExcludeBuffs = "",
		tExcludeBuffs = {  },
		tExcludeTextBuffs = {  },
		
		sIncludeDebuffs = "",
		tIncludeDebuffs = { },
		tIncludeTextDebuffs = {  },
		sExcludeDebuffs = "",
		tExcludeDebuffs = { },
		tExcludeTextDebuffs = {  },
		
		nBuffRows = 1,
		nBuffsPerRow = 6,
		nBuffIconSize = 35,
		bBuffShowNameTooltip = true,
		bBuffShowFullTooltip = true,
		
		nDebuffRows = 1,
		nDebuffsPerRow = 6,
		nDebuffIconSize = 35,
		bDebuffShowNameTooltip = true,
		bDebuffShowFullTooltip = true
	
	}
end


-- Set Default Colors
function OptiPlates:DefaultColors()
	for propKey,prop in pairs(tUserSettings) do 
		if string.sub(propKey,1,8) == "setColor" then
			self[propKey] = prop.default
		end
	end
	self:UpdateColorObjects()
end

-- checks that the default profile and current profile exist
function OptiPlates:CheckSettingsSanity()
	if self.setProfiles == nil then self.setProfiles = {} end
	
	if self.setCurrentProfileScope == nil then self.setCurrentProfileScope = GameLib.CodeEnumAddonSaveLevel.Character end
	if self.setCurrentProfileName == nil then self.setCurrentProfileName = "Default profile" end
	
	if self.setProfiles[GameLib.CodeEnumAddonSaveLevel.Character] == nil then 
		self.setProfiles[GameLib.CodeEnumAddonSaveLevel.Character] = {}
	end
	if self.setProfiles[GameLib.CodeEnumAddonSaveLevel.Realm] == nil then 
		self.setProfiles[GameLib.CodeEnumAddonSaveLevel.Realm] = {}
	end
	if self.setProfiles[GameLib.CodeEnumAddonSaveLevel.Account] == nil then 
		self.setProfiles[GameLib.CodeEnumAddonSaveLevel.Account] = {}
	end	
	if self.setProfiles[GameLib.CodeEnumAddonSaveLevel.General] == nil then 
		self.setProfiles[GameLib.CodeEnumAddonSaveLevel.General] = {}
	end
	
	if self.defaultBuffProfileSettings == nil then
		self:DefaultBuffSettings()
	end
	
	if self.setProfiles[self.setCurrentProfileScope][self.setCurrentProfileName] == nil then
		self.setCurrentProfileScope = GameLib.CodeEnumAddonSaveLevel.Character
		self.setCurrentProfileName = "Default profile"
		
		for propKey, prop in pairs(tUserSettings) do self[propKey] = prop.default end
		self:SaveProfile(self.setCurrentProfileScope,self.setCurrentProfileName)
	end
	
	if self.setProfiles["Buffprofiles"] == nil then
		self.setProfiles["Buffprofiles"] = { }
	end
	if self.setProfiles["Buffprofiles"]["Default"] == nil then
		self:CreateBuffProfile("Default", nil)
	end
	-- check that all settings have values to avoid nil errors
	local defaultBuffProfile = self.defaultBuffProfileSettings
	for profile, arr in pairs(self.setProfiles["Buffprofiles"]) do
		for key,val in pairs(defaultBuffProfile) do
			if arr[key] == nil then self.setProfiles["Buffprofiles"][profile][key] = val end
		end
	end
	
end

-- Updates the GUI controls to reflect the current settings
function OptiPlates:RefreshSettings()
	local versionLabel = self.wndMain:FindChild("Label_AddonVersion")
	if versionLabel ~= nil then
		versionLabel:SetText(string.format("Version: %s", _AddonVersion or "Undefined"))
	end
	-- GENERAL SETTINGS
	if self.wndMain:FindChild("SettingsGeneral") then
		if self.setEnableOcclusion ~= nil then 
			self.wndMain:FindChild("Button_ToggleOcclusion"):SetCheck(self.setEnableOcclusion)
			if Apollo.GetConsoleVariable("ui.occludeNameplatePositions") ~= self.setEnableOcclusion then
				Apollo.SetConsoleVariable("ui.occludeNameplatePositions", self.setEnableOcclusion) end end
			
		if self.setShowThreatAlert ~= nil then 
			self.wndMain:FindChild("Button_ShowThreatAlert"):SetCheck(self.setShowThreatAlert) 
			self.wndMain:FindChild("Button_ShowThreatLossAlert"):SetCheck(self.setShowThreatLossAlert) 
			
			end
		if self.setHideFullFriendlyBars ~= nil then 
			self.wndMain:FindChild("Button_HideFullFriendlyBars"):SetCheck(self.setHideFullFriendlyBars) end

		local btnCleansePosLeft = self.wndMain:FindChild("Button_CleanseIconPositionLeft")
		local btnCleansePosRight = self.wndMain:FindChild("Button_CleanseIconPositionRight")
		
		if self.setStyleCleanseIconPosition == "Left" then
			btnCleansePosLeft:SetCheck(true)
			btnCleansePosRight:SetCheck(false)
		else
			btnCleansePosLeft:SetCheck(false)
			btnCleansePosRight:SetCheck(true)
		end				
	end
	-- STYLE SETTINGS
	if self.wndMain:FindChild("SettingsStyle") then	
		if self.setStyle ~= nil then
			self.wndMain:FindChild("Button_StyleFlat"):SetCheck(self.setStyle == "flat")
			self.wndMain:FindChild("Button_StyleSAO"):SetCheck(self.setStyle == "sao")
		end	
	end
	
	if self.wndMain:FindChild("SettingsAdvanced") then
		local btnVertAnchorName = self.wndMain:FindChild("Button_NameplateAnchorName")
		local btnVertAnchorBars = self.wndMain:FindChild("Button_NameplateAnchorBars")
		local btnVertAnchorNone = self.wndMain:FindChild("Button_NameplateAnchorNone")
		
		if self.setStyleVerticalAnchor == "Name" then
			btnVertAnchorName:SetCheck(true)
			btnVertAnchorBars:SetCheck(false)
			btnVertAnchorNone:SetCheck(false)
		elseif self.setStyleVerticalAnchor == "Bars" then
			btnVertAnchorName:SetCheck(false)
			btnVertAnchorBars:SetCheck(true)
			btnVertAnchorNone:SetCheck(false)
		else
			btnVertAnchorName:SetCheck(false)
			btnVertAnchorBars:SetCheck(false)
			btnVertAnchorNone:SetCheck(true)
		end	
	end
	
	for idx, opt in pairs(self.optionsUI_DropdownListSettings) do
		if opt.wnd ~= nil then
			if opt.labelarray ~= nil then
				opt.btn:SetText("    "..self[opt.labelarray][self[opt.var]])
			else
				opt.btn:SetText("    "..self[opt.var])
			end
			local grid = opt.grid
			local arr = self[opt.array]
			grid:DeleteAll()
			for id, row in pairs(arr) do
				if opt.labelarray ~= nil then
					grid:AddRow(self[opt.labelarray][row]) 
				else
					grid:AddRow(row)
				end
			end
		end
	end
	
	for idx, opt in pairs(self.optionsUI_CheckboxSettings) do
		if opt.wnd ~= nil then 
			local val = self[opt.var] or tUserSettings[opt.var][default]
			opt.wnd:SetCheck(val)
			if opt.togglewnd then
				opt.togglewnd:Show(opt.wnd:IsChecked())
				local wndResize = opt.wnd:GetParent()
				local nLeft, nTop, nRight, nBottom = wndResize:GetAnchorOffsets()
				if opt.wnd:IsChecked() then
					wndResize:SetAnchorOffsets(nLeft, nTop, nRight, opt.parentheighton)
				else
					wndResize:SetAnchorOffsets(nLeft, nTop, nRight, opt.parentheightoff)
				end
			end
		end
	end	

	for idx, opt in pairs(self.optionsUI_EditboxSettings) do
		if opt.wnd ~= nil then
			local val = (self[opt.var] or tUserSettings[opt.var][default])
			--if tUserSettings[opt.var].setDefaultOnEmptyString and val == "" then
			--	self[opt.var] = tUserSettings[opt.var][default]
			--	val = self[opt.var]
			--end
			if (val ~= nil) then
				opt.wnd:SetText(val)
				if opt.colorlabel ~= nil then
					opt.colorlabel:SetTextColor(ApolloColor.new(self.setOpacity..val))
				end	
			end			
		end
	end	
	
	for idx, opt in pairs(self.optionsUI_SliderSettings) do
		if opt.wnd ~= nil then
			local val = (self[opt.var] or tUserSettings[opt.var][default])
			opt.wnd:SetValue(val)
			if opt.label ~= nil then
				if opt.divideby ~= nil then
					val = val / opt.divideby
				end
				if opt.stringformat ~= nil then
					opt.label:SetText(string.format(opt.stringformat,val))
				else
					opt.label:SetText(string.format("%s",val))
				end
			end
		end
	end	
	
	if self.setCurrentProfileScope ~= nil then
		local scopeStr = ""
		local scope = self.setCurrentProfileScope
		if scope == GameLib.CodeEnumAddonSaveLevel.General then
			scopeStr = "General"
		elseif scope == GameLib.CodeEnumAddonSaveLevel.Account then
			scopeStr = "Account"
		elseif scope == GameLib.CodeEnumAddonSaveLevel.Realm then
			scopeStr = "Realm"
		elseif scope == GameLib.CodeEnumAddonSaveLevel.Character then
			scopeStr = "Character"
		end
	
		self.wndMain:FindChild("Label_ProfileScope"):SetText(scopeStr)
	end
	
	if self.setCurrentProfileName ~= nil then
		self.wndMain:FindChild("Label_ProfileName"):SetText(self.setCurrentProfileName)
	end
	
	if self.wndMain:FindChild("SettingsProfiles") then	
		self.wndMain:FindChild("Button_NewProfileDefaultBase"):SetCheck(true)
		self.wndMain:FindChild("Button_NewProfileCopyBase"):SetCheck(false)
		self.wndMain:FindChild("Button_NewProfileListToggle"):Enable(false)
		local wndNewProfileScope = self.wndMain:FindChild("Grid_NewProfileScopeList")
		wndNewProfileScope:DeleteAll()
		wndNewProfileScope:AddRow("Character")
		wndNewProfileScope:AddRow("Realm")
		wndNewProfileScope:AddRow("Account")
		wndNewProfileScope:AddRow("General")
		wndNewProfileScope:SetCurrentRow(1)
	end
	self:UpdateOptionMenuPixies()
	self:CreateAllProfileLists()
	self:PopulateAllFontLists()
	self:CreateAllBuffProfileLists()
end

-----------------------------------------------------------------------------------------------
-- OptiPlates Functions: CREATE THE PLATES
-----------------------------------------------------------------------------------------------

-- New unit created, build main options here.
function OptiPlates:OnUnitCreated(unitNew)
	if not unitNew then return end
	local idUnit = unitNew:GetId()
	-- don't create unit if we already know about it
	local tNameplate = self.arUnit2Nameplate[idUnit]
	-- if player unit/gui hasn't been loaded, don't create any plates.
	-- OnFrame will recall OnUnitCreated with the contents of all tInitialNewUnits units once
	-- initialization is done
	if not self.bInitialLoadAllClear and (not self.playerUnit) and (not tNameplate) then
		self.tInitialNewUnits[idUnit] = unitNew
		return
	end
	if unitNew:GetDispositionTo(self.playerUnit) == tDisposition.Unknown then
		self.tInitialNewUnits[idUnit] = unitNew
		return
	end
	if tNameplate == nil then 
		tNameplate = self:CreateNameplateObject(unitNew)
	elseif not (unitNew == self.playerUnit) then
		tNameplate = self:ReloadNameplate(tNameplate)
	end
	if tNameplate.bIsPlayer then
		if tNameplate.unitOwner:IsMounted() then
			self.arMount2Unit[tNameplate.unitOwner:GetUnitMount():GetId()] = tNameplate.unitOwner
		end
	end
	
	tNameplate.bNeedsRedraw = true
end

function OptiPlates:OnUnitGibbed(unitUpdated)
	local tNameplate = self.arUnit2Nameplate[unitUpdated:GetId()]
	if tNameplate ~= nil then
		tNameplate.bGibbed = true
		self:UpdateNameplateVisibility(tNameplate)
	end
end


-- Create nameplate object from the unit.
function OptiPlates:CreateNameplateObject(unit)
	local unitId = unit:GetId()
	if unitId == nil then return end
	if self.arUnit2Nameplate[unitId] ~= nil then return self.arUnit2Nameplate[unitId] end
	-- main initialization will be done by ReloadNameplate, which can get called for both new or existing nameplate objects
	local tNameplate = {
		unitOwner	= unit,
		idUnit 		= unit:GetId(),
	}
	self.arUnit2Nameplate[tNameplate.idUnit] = tNameplate
	tNameplate = self:ReloadNameplate(tNameplate, false)

	return tNameplate
end

function OptiPlates:ReloadNameplate(tNameplate, bKeepNameplate)
	-- destroy nameplate just in case, sometimes stealthing stalkers can bug nameplates so they are
	-- left hovering until destroyed
	if tNameplate.wndNameplate ~= nil then
		if not bKeepNameplate then
			self.arDisplayedNameplates[tNameplate.wndNameplate:GetId()] = nil
			tNameplate.wndNameplate:SetUnit(nil)
			tNameplate.wndNameplate:Show(false)
			tNameplate.wndNameplate:Destroy()
			tNameplate.wndNameplate = nil
			tNameplate.wnd = nil
			tNameplate.tBuffWindowRefs 		= { tInUse = {}, tFreeWindows = {}, numBuffs = 0 }
			-- { tInUse = {idBuff = { wndRef, valid, numBuff }}, tFreeWindows = { wnd,wnd }, numBuffs }
			tNameplate.tDebuffWindowRefs 	= { tInUse = {}, tFreeWindows = {}, numBuffs = 0 }
		else
			tNameplate.wnd.ToprowContainer:DestroyAllPixies()
			tNameplate.wnd.NameContainer:DestroyAllPixies()
			tNameplate.wnd.Plates:DestroyAllPixies()		
			tNameplate.wndNameplate:DestroyAllPixies()
		end
	else
		tNameplate.tBuffWindowRefs 		= { tInUse = {}, tFreeWindows = {}, numBuffs = 0 }
		-- { tInUse = {idBuff = { wndRef, valid, numBuff }}, tFreeWindows = { wnd,wnd }, numBuffs }
		tNameplate.tDebuffWindowRefs 	= { tInUse = {}, tFreeWindows = {}, numBuffs = 0 }	
	end
	local unit = tNameplate.unitOwner
	local unitType = unit:GetType()
	
	tNameplate.bShow				= false -- whether nameplate should be shown or not
	tNameplate.bGibbed				= false
	tNameplate.bSpeechBubble 		= false -- does nameplate have a speech bubble visible
	--- unit type checks
	tNameplate.bHasActivationState	= false
	-- Some unit type checking
	tNameplate.bIsNPC				= (unitType == "NonPlayer")
	tNameplate.bIsCollectible		= (unitType == "Collectible")
	tNameplate.bIsPet				= (unitType == "Pet" or unitType == "Scanner")
	tNameplate.bIsPickup			= (unitType == "Pickup")
	tNameplate.bIsMount				= (unitType == "Mount")
	tNameplate.bIsPinataLoot		= (unitType == "PinataLoot")
	tNameplate.bIsPlayer			= (unitType == "Player")
	tNameplate.bIsHarvest			= (unitType == "Harvest")
	tNameplate.bIsOre				= (unitType == "Harvest" and unit:GetHarvestRequiredTradeskillName() == "Mining")
	tNameplate.bIsRelic				= (unitType == "Harvest" and unit:GetHarvestRequiredTradeskillName() == "Relic Hunter")
	tNameplate.bIsTree				= (unitType == "Harvest" and unit:GetHarvestRequiredTradeskillName() == "Survivalist")
	tNameplate.bIsPlant				= (unitType == "Harvest" and unit:GetHarvestRequiredTradeskillName() == "Farmer")	
	tNameplate.bIsImportantNPC		= false -- Is unit important, updated in UpdateNameplateVisibility
	tNameplate.bIsUnimportantNPC	= false -- updated in UpdateNameplateVisibility
	tNameplate.bIsVendorNPC			= false -- Is unit a vendor, updated in UpdateNameplateVisibility
	tNameplate.bIsDatacube			= false
	tNameplate.bIsQuestGiverNew		= false -- Quest-related variables are updated in UpdateRewardInfo()
	tNameplate.bIsQuestGiverNewDaily = false
	tNameplate.bIsQuestGiverNewGrey	= false
	tNameplate.bIsQuestReceiver		= false
	tNameplate.bIsQuestFinisher		= false
	tNameplate.bIsQuestUnit			= false
	tNameplate.bIsWeapon			= false
	tNameplate.bIsGuildMember		= false
	tNameplate.bIsObjective			= false -- Is unit an objective (path missions, quests etc)
	tNameplate.bIsQuestObjective	= false
	tNameplate.bIsEventObjective	= false
	tNameplate.bIsChallengeObjective = false
	tNameplate.bIsContractObjective	= false
	tNameplate.bIsPathObjective		= false
	tNameplate.nRewardCount			= 0
	tNameplate.bNeedsRedraw			= true -- whether bar needs to be redrawn
	tNameplate.nHealthPercent		= -1 -- Health percentage during last draw
	tNameplate.nTotalHealth			= -1 -- Health (health+shield+absorb) during last frame update
	tNameplate.nInterruptArmor		= -2 -- current interrupt armor
	tNameplate.nInterruptArmorMax	= -2 -- current max interrupt armor
	tNameplate.bShowShieldBar		= true -- updated in SetNormalNameplate
	tNameplate.bShowHealthBar		= true -- updated in SetNormalNameplate
	tNameplate.nVulnerableTime 		= 0 -- max duration of the current MoO
	tNameplate.nSubdueTime			= 0 -- max duration of the current subdue
	tNameplate.nCCState				= -1 -- current subdued state, from tSubdueCCStates
	tNameplate.sCCStateText			= "" -- current subdued state text
	tNameplate.tInCCStates			= {} -- list of affected cc states
	tNameplate.nPrevScale			= -1 -- scaling on previous frame, to avoid rescaling calculations
	tNameplate.nPrevOpacity			= -1 -- opacity on previous frame, as above
	tNameplate.nPrevBGOpacity		= -1 
	tNameplate.nSquaredDistance		= -1 -- updated in self:CheckDrawDistance, squared distance from player
	tNameplate.previousDamageTicks	= 0 -- to calculate shield reboot bar, value is Apollo:GetTickCount()
	tNameplate.bShowBars			= false -- Show Health Bars?
	tNameplate.bShowCastBar			= false -- Show Cast Bar?
	tNameplate.bShowDistance		= false -- Show distance text?
	tNameplate.bShowFocus			= false -- Show focus bar?
	tNameplate.bShowResource		= false -- Show resource bar?
	tNameplate.bShowDash			= false -- Show dash/sprint bar?
	tNameplate.bInCluster			= false -- Is Unit In targeted cluster, updated below
	tNameplate.bIsTarget			= false -- Is unit the current target
	tNameplate.bMouseOverSettingsEnabled = (unit == self.mouseOverUnit) or (unit:IsMounted() and unit:GetUnitMount() == self.mouseOverUnit)
	tNameplate.nNameWidth			= nil
	tNameplate.nHeadingPixieWidth	= nil
	tNameplate.nNameHeightOffset	= nil
	tNameplate.tPixieTables			= { 
										Target = nil,
										InterruptArmor = nil,
										Level = nil,
										HP = nil,
										SP = nil,
										CleanseIcon = nil,
										ClassIcon = nil,
										RoleIcon = nil,
										Heading = nil,
										HeadingText = nil,
										QuestIcon = nil
									}
	tNameplate.nLegacyTargetPixie	= nil
	tNameplate.nTargetPixie			= nil
	tNameplate.nInterruptArmorPixie = nil
	tNameplate.nInterruptArmorPixieInf  = nil
	tNameplate.nLevelPixie			= nil
	tNameplate.nLevelWidth			= nil
	tNameplate.nHPTextPixie			= nil
	tNameplate.nHPTextWidth			= nil
	tNameplate.nSPTextPixie			= nil
	tNameplate.nSPTextWidth			= nil
	tNameplate.nCleanseIconPixie	= nil
	tNameplate.nClassIconPixie		= nil
	tNameplate.nRoleIconPixie		= nil
	tNameplate.nHeadingPixie		= nil
	tNameplate.nHeadingTextPixie	= nil
	tNameplate.nQuestIconPixie		= nil
	tNameplate.cDispositionColor	= nil
	tNameplate.nUnitHeight			= 0 
	tNameplate.nOverheadAnchorY		= 0
	tNameplate.bDefaultHeight		= true
	tNameplate.nNameplateAnchor		= 1
	tNameplate.nHealthMax			= -1
	tNameplate.nHealth				= -1
	tNameplate.nShield				= -1
	tNameplate.nShieldMax			= -1
	tNameplate.nFocusMax			= 1000 -- GetMaximumMana returns 0 for units other than the player
	tNameplate.nAbsorb				= 0
	tNameplate.nPrevCastElapsed		= -1
	tNameplate.nPreviousLocation	= nil
	tNameplate.nPreviousLocationTime= 0	-- timestamp since the previous location changed
	tNameplate.eDisposition			= unit:GetDispositionTo(self.playerUnit)
	tNameplate.bIsDPS				= false
	tNameplate.bIsTank				= false
	tNameplate.bIsHealer			= false
	tNameplate.nGroupIndex			= -1
	tNameplate.nVerticalOffset		= 0
	tNameplate.nPrevVerticalOffset  = 0
	tNameplate.nScalingHorzOffset	= 0
	tNameplate.nScalingVertOffset	= 0
	tNameplate.nNameplateHeight		= 0
	tNameplate.bCastbarArranged		= false
	tNameplate.bLegacyHeightAdjusted = false
	tNameplate.bCleanseIconPositioned = false
	tNameplate.bBuffProfileDirty	= true
	tNameplate.tBuffProfile			= nil
	tNameplate.sUnitType			= nil
	tNameplate.sUnitTypeSettings	= nil
	tNameplate.bUpdateRewards		= false
	tNameplate.bFriendlyNameplateHealthVisibility = false
	
	-- Update CC states
	if tNameplate.bIsPlayer or tNameplate.bIsNPC then
		for name,idCCState in pairs(tSubdueCCStates) do
			if unit:IsInCCState(idCCState) then
				 if tNameplate.nCCState >= 0 then
					if unit:GetCCStateTimeRemaining(tNameplate.nCCState) < unit:GetCCStateTimeRemaining(idCCState) then
						tNameplate.nCCState = idCCState
						tNameplate.sCCStateText = name
					end
				 else
					tNameplate.nCCState = idCCState
					tNameplate.sCCStateText = name
				 end
			end
		end
	end
	
	-- update cluster/target information
	if self.playerUnit ~= nil then
		local unitCurrentTarget = self.playerUnit:GetTarget()
		if unitCurrentTarget == unit then
			tNameplate.bIsTarget = true
		end
		if (unitCurrentTarget ~= nil) then
			local bClusterInformation = unitCurrentTarget:GetClusterInformation()
			if bClusterInformation ~= nil and bClusterInformation.nCount ~= nil and bClusterInformation.nCount > 1 then
				local tClusterUnits = bClusterInformation.tUnits
				for idx = 1, #tClusterUnits do
					if tClusterUnits[idx] == unit then
						tNameplate.bInCluster = true
					end
				end
			end
		end
	end
	self:ParseRewards(tNameplate)
	self:UpdateNameplateVisibility(tNameplate)

	-- Update cached buff window entries
	if tNameplate.wndNameplate ~= nil then
		local debuffs = tNameplate.wnd.Debuffs:GetChildren()
		if debuffs then
			for idx,wnd in pairs(debuffs) do
				if wnd and wnd:IsValid() then
					wnd:Show(false)
					wnd:SetText("")
					if wnd:GetName() == "BuffWnd" then
						table.insert(tNameplate.tBuffWindowRefs.tFreeWindows,wnd)
					else
						table.insert(tNameplate.tDebuffWindowRefs.tFreeWindows,wnd)
					end
				end
			end
		end
	end
	self:StyleConfig(tNameplate)
	tNameplate.bNeedsRedraw = true
	self.arUnit2Nameplate[tNameplate.idUnit] = tNameplate
	if tNameplate.bIsPlayer and tNameplate.eDisposition == tDisposition.Friendly and unit:GetMaxHealth() ~= nil and unit:GetMaxHealth() > 0 then
		self.arFriendlyUnitNameplates[tNameplate.idUnit] = tNameplate
	else
		self.arFriendlyUnitNameplates[tNameplate.idUnit] = nil
	end
	return tNameplate
end

-- The following updates the struct that stores references to different windows, so
-- there's no need to call FindChild often, especially in OnFrame
function OptiPlates:RefreshWindowRefs(tNameplate, wnd)
	if tNameplate == nil or wnd == nil then return end
	local myStyle = self.setStyle
	local nWidth = 0
	local nHeight = 0
	local tLocation = wnd:GetLocation():ToTable()	
	if tNameplate.wnd ~= nil then 
		nWidth = tNameplate.wnd.Width
		nHeight = tNameplate.wnd.Height
	else

		local fPoints = tLocation.fPoints
		local nOffsets = tLocation.nOffsets
		if fPoints[1] ~= fPoints[3] then
			nWidth = Apollo:GetDisplaySize().nWidth - (math.abs(nOffsets[1]-nOffsets[3]))
		else
			nWidth = (math.abs(nOffsets[1]-nOffsets[3]))
		end
		if fPoints[2] ~= fPoints[4] then
			nHeight = Apollo:GetDisplaySize().nWidth - (math.abs(nOffsets[2]-nOffsets[4]))
		else
			nHeight = (math.abs(nOffsets[2]-nOffsets[4]))			
		end
	end
	
	local wndRefs = 
	{
		Name 				= wnd:FindChild("Name"),
		NameContainer		= wnd:FindChild("NameContainer"),
		GuildName			= wnd:FindChild("GuildName"),
		Plates				= wnd:FindChild("Plates"),
		HealthBar			= wnd:FindChild("HealthBar"),
		HealthBarBack		= wnd:FindChild("HealthBarBack"),
		AbsorbBar			= wnd:FindChild("AbsorbBar"),
		ShieldBar			= wnd:FindChild("ShieldBar"),
		ShieldTimerBar		= wnd:FindChild("ShieldTimerBar"),
		VulnerableBar		= wnd:FindChild("VulnerableBar"),
		Debuffs				= wnd:FindChild("Debuffs"),
		Cast				= wnd:FindChild("Cast"),
		CastBar				= wnd:FindChild("CastBar"),
		CastBarLabel		= wnd:FindChild("CastBarLabel"),
		ClassResource		= wnd:FindChild("ClassResource"),
		ClassResourceBar	= wnd:FindChild("ClassResourceBar"),
		FocusBar			= wnd:FindChild("FocusBar"),
		DashContainer		= wnd:FindChild("SprintDashContainer"),
		SprintBar			= wnd:FindChild("SprintBar"),
		DashBar				= wnd:FindChild("DashBar"),
		NamePlate			= wnd:FindChild("NamePlate_"..self.setStyle),
		NamePlateContainer	= wnd:FindChild("NamePlateContainer"),
		Distance			= wnd:FindChild("Distance"),
		NotTargeting		= wnd:FindChild("NotTargeting"),
		QuestRewards		= wnd:FindChild("QuestRewards"),
		ToprowContainer		= wnd:FindChild("ToprowContainer"),
		Width				= nWidth,
		Height				= nHeight,
		Location			= tLocation
	}
	tNameplate.wnd = wndRefs
	tNameplate.wndNameplate = wnd
end


-- Create nameplae window from the nameplate object.
function OptiPlates:CreateNameplateWindow(tNameplate)
	if tNameplate.wndNameplate ~= nil then return end -- Nameplate alrady has a window? Bail out.
	if tNameplate.unitOwner == nil then return end -- Nameplate has no owner? Bail out.
	local wnd = nil
	local wndRefs = nil
	-- Use unused in-memory window objects if available
	local cacheWindowArray = nil
	if tNameplate.bIsPlayer or tNameplate.bIsNPC then
		cacheWindowArray = next(self.arFreeWindows[1]) ~= nil and self.arFreeWindows[1] or self.arFreeWindows[2]
	else
		cacheWindowArray = next(self.arFreeWindows[2]) ~= nil and self.arFreeWindows[2] or self.arFreeWindows[1]
	end
	
	if next(cacheWindowArray) ~= nil then
		local poolEntry = table.remove(cacheWindowArray)
		wnd = poolEntry[1]
		wndRefs = poolEntry[2]
		self.nFreeWindows = self.nFreeWindows - 1
	end
	-- Create a new nameplate window if none is available
	if wnd == nil or not wnd:IsValid() then
		if self.bStyleSAO then
			wnd = Apollo.LoadForm(self.xmlDoc, "NamePlate_sao", self.setNameplateDefaultStratum, self)
		else
			wnd = Apollo.LoadForm(self.xmlDoc, "NamePlate_flat", self.setNameplateDefaultStratum, self)
		end
		
		local wndCast = nil
		local wndClassresource = nil
		local wndNameplateContainer = wnd:FindChild("NamePlateContainer")
	end
	local wndTopContainer = wnd:FindChild("ToprowContainer")
	wndTopContainer:DestroyChildren()
	wnd:Show(false, true)
	
	self:RefreshWindowRefs(tNameplate,wnd)
	wndRefs = tNameplate.wnd
	
	wndRefs.Name:SetData(tNameplate.unitOwner)

	wndRefs.NameContainer:Show(false)
	wndRefs.GuildName:Show(false)
	wndRefs.Plates:Show(false)
	wndRefs.Debuffs:Show(false)
	wndRefs.Cast:Show(false)
	wndRefs.ClassResource:Show(false)
	wndRefs.Distance:Show(false)
	wndRefs.NotTargeting:Show(false)
	wndRefs.ToprowContainer:Show(false)
	
	wnd:SetUnit(tNameplate.unitOwner, 1)  -- 1 for overhead, 0 for underfoot

	
	-- Set Up SAO Sprites
	---------------------------------
	local tempPlates = wnd:FindChild("Plates");	
	if self.bStyleSAO then
		tempPlates:FindChild("HealthBar"):SetFullSprite("SAO_Health")
		tempPlates:FindChild("HealthBarBack"):SetEmptySprite("SAO_Health")
		tempPlates:FindChild("HealthBarBack"):SetFullSprite("SAO_Health")
		tempPlates:FindChild("NotTargeting"):SetSprite("SAO_Health_Border_Thick")
		
		tempPlates:FindChild("ShieldBar"):SetEmptySprite("SAO_Shield")
		tempPlates:FindChild("ShieldBar"):SetFullSprite("SAO_Shield")
	else
	-- Set Up Flat Sprites
	---------------------------------
		tempPlates:FindChild("HealthBar"):SetFullSprite("Flat_Health")
		tempPlates:FindChild("HealthBarBack"):SetEmptySprite("Flat_Health")
		tempPlates:FindChild("HealthBarBack"):SetFullSprite("Flat_Health")
		tempPlates:FindChild("NotTargeting"):SetSprite("Flat_Border")
		
		tempPlates:FindChild("ShieldBar"):SetEmptySprite("Flat_Shield")
		tempPlates:FindChild("ShieldBar"):SetFullSprite("Flat_Shield")
	
	end
	
	tNameplate.wnd = wndRefs
	tNameplate.wndNameplate = wnd
	self.arDisplayedNameplates[wnd:GetId()] = tNameplate

	tNameplate.bNeedsRedraw = true	
	return tNameplate.wndNameplate
end

-- Destroy all nameplates and recreate them (called when changing style in options)
function OptiPlates:ReloadAllNameplates()
	local tempRecreate = {}
	for id,tNameplate in pairs(self.arUnit2Nameplate) do
		if tNameplate ~= nil and tNameplate.wndNameplate ~= nil then
			self:UnattachNameplateWindow(tNameplate)
			table.insert(tempRecreate, tNameplate)
		end
	end
	while next(self.arFreeWindows[1]) ~= nil do
		local poolEntry = table.remove(self.arFreeWindows[1])
		if poolEntry ~= nil then
			poolEntry[1]:Destroy()
		end
		self.nFreeWindows = self.nFreeWindows - 1		
	end
	while next(self.arFreeWindows[2]) ~= nil do
		local poolEntry = table.remove(self.arFreeWindows[2])
		if poolEntry ~= nil then
			poolEntry[1]:Destroy()
		end
		self.nFreeWindows = self.nFreeWindows - 1		
	end	
	for idx, tNameplate in pairs(tempRecreate) do
		self:ReloadNameplate(tNameplate)
		tNameplate.bNeedsRedraw = true
	end
end

-- Destroys or caches nameplate window, if cache space is available.
function OptiPlates:UnattachNameplateWindow(tNameplate)
	if tNameplate == nil then return end
	local wndNameplate = tNameplate.wndNameplate
	if tNameplate.wnd ~= nil then
		tNameplate.wnd.ToprowContainer:DestroyAllPixies()
		tNameplate.wnd.NameContainer:DestroyAllPixies()
		tNameplate.wnd.Plates:DestroyAllPixies()
	end
	tNameplate.tBuffWindowRefs = { tInUse = {}, tFreeWindows = {}, numBuffs = 0 }
	tNameplate.tDebuffWindowRefs = { tInUse = {}, tFreeWindows = {}, numBuffs = 0 }
	if wndNameplate ~= nil then
		self.arDisplayedNameplates[wndNameplate:GetId()] = nil
		local dbfCount = #wndNameplate:FindChild("Debuffs"):GetChildren()
		wndNameplate:DestroyAllPixies()
		wndNameplate:SetUnit(nil)
		wndNameplate:SetOpacity(self.setDefaultOpacityF or 1)
		wndNameplate:SetScale(self.setDefaultScaleF or 1)
		local nHalfSize = self.nStyleHalfWidth				
		wndNameplate:Show(false, true)
		
		if dbfCount > 1 then
			if self.nFreeWindows < self.nFreeWindowsPoolLimit or next(self.arFreeWindows[2]) ~= nil then

				table.insert(self.arFreeWindows[1], {wndNameplate, tNameplate.wnd})
				if self.nFreeWindows >= self.nFreeWindowsPoolLimit then
					local entry = table.remove(self.arFreeWindows[2])
					if entry ~= nil then
						entry[1]:Destroy()
					end
				else
					self.nFreeWindows = self.nFreeWindows + 1
				end
			else
				wndNameplate:Destroy()
			end
		elseif self.nFreeWindows < self.nFreeWindowsPoolLimit then
			table.insert(self.arFreeWindows[2], {wndNameplate, tNameplate.wnd})
			self.nFreeWindows = self.nFreeWindows + 1
		else
			wndNameplate:Destroy()
		end
	end
	tNameplate.wndNameplate = nil
	tNameplate.wndRefs = nil
end

function OptiPlates:OnUnitNameChanged(unitUpdated, strNewName)
	local tNameplate = self.arUnit2Nameplate[unitUpdated:GetId()]
	if tNameplate == nil or tNameplate.wndNameplate == nil then
		return
	end
	tNameplate.bNeedsRedraw = true
end

function OptiPlates:HelperGetUnitSetting(tNameplate, settingName, bInCombat)
	local unitOwner = tNameplate.unitOwner
	local unitPlayer = self.playerUnit
	if unitOwner == nil or unitPlayer == nil then return false end
	if self.tNameplateSettings == nil then return false end
	local arr = self.tNameplateSettings[tNameplate.sUnitType]

	local combat = nil
	if bInCombat then
		combat = "C"
	else
		combat = "OOC"
	end
	if (tNameplate.bIsTarget or tNameplate.bInCluster) and not (tNameplate.unitOwner == self.playerUnit and self.setNoTargetSettingsOnSelf) then
		local eDisposition = tNameplate.eDisposition 
		if eDisposition == Unit.CodeEnumDisposition.Hostile then
			return arr[combat][settingName] or self.tNameplateSettings.HostileTarget[combat][settingName]
		elseif eDisposition == Unit.CodeEnumDisposition.Friendly then
			return arr[combat][settingName] or self.tNameplateSettings.FriendlyTarget[combat][settingName]
		else
			return arr[combat][settingName] or self.tNameplateSettings.NeutralTarget[combat][settingName]
		end
	else
		return arr[combat][settingName]
	end
end

-- This function updates the visibility of parts of the nameplate that are at least somewhat dynamic
-- (in/out of combat visibility and other event-related changes)
function OptiPlates:SetNormalNameplate(tNameplate)
	if tNameplate.wndNameplate == nil then return end
	
	local namePlate = tNameplate.wndNameplate
	local unitOwner = tNameplate.unitOwner
	if not unitOwner:IsValid() then return end
	local unitPlayer = self.playerUnit
	local unitFocus = self.focusUnit
	local eDisposition = tNameplate.eDisposition
	local bIsPlayer = tNameplate.bIsPlayer
	local wndRefs = tNameplate.wnd
	--local sUnitSettingLetter = self:UnitSettingLookup(tNameplate)
	--local sUnitSettingLetterNoTarget = self:UnitSettingLookup(tNameplate, true)
	--if sUnitSettingLetter == nil then return end
	
	-- handle dynamic color change here
	local tempTextColor = self.arColorObjects.NeutralName
	local unitController = unitOwner:GetUnitOwner() or unitOwner
	if tNameplate.bIsHarvest or tNameplate.bIsDatacube then
		tempTextColor = self.arColorObjects.Harvestables
	elseif eDisposition == tDisposition.Neutral then
		tempTextColor = self.arColorObjects.NeutralName
	elseif eDisposition == tDisposition.Hostile then
		if bIsPlayer then
			if unitController:IsPvpFlagged() == true then
				tempTextColor = self.arColorObjects.HostilePlayerName
			else
				tempTextColor = self.arColorObjects.HostilePlayerNameUnflagged
			end
		else
			tempTextColor = self.arColorObjects.HostileName
		end
	elseif eDisposition == tDisposition.Friendly then
		if bIsPlayer then
			if tNameplate.unitOwner ~= self.playerUnit and self.setGP_Override and tNameplate.bIsGuildMember then
				tempTextColor = self.arColorObjects.GuildName
			elseif unitOwner:IsInYourGroup() then
				tempTextColor = self.arColorObjects.PartyName
			elseif tNameplate.unitOwner ~= self.playerUnit and tNameplate.bIsGuildMember then
				tempTextColor = self.arColorObjects.GuildName
			elseif unitController:IsPvpFlagged() == true then
				tempTextColor = self.arColorObjects.FriendlyPlayerName
			else
				tempTextColor = self.arColorObjects.FriendlyPlayerNameUnflagged
			end
		else
			tempTextColor = self.arColorObjects.FriendlyName
		end
	end	
	
	tNameplate.cDispositionColor = tempTextColor
	
	
	local bInCombat = unitOwner:IsInCombat()	
	local bShowName = false
	local bShowGuild = false
	local bShowBars = false -- also affected by DrawNameplateBars, called in OnFrame
	local bShowCastBar = false -- also affected by DrawNameplateCastbar, called in OnFrame
	local bShowBuffs = false -- also debuffs
	local bShowClassname = false
	local bShowDistance = false
	local bShowResource = false
	local bShowFocus = false
	local bShowDash = false
	
	-- Only enabled nodes have nameplates, so bShowName is always true
	if tNameplate.bIsHarvest or tNameplate.bIsDatacube then
		bShowName = true
		-- "guild text" for harvestable nameplate shows tier
		bShowGuild = self.setHarvestablesTierEnabled
		bShowBars = false
		bShowCastBar = false
		bShowBuffs = false
		bShowClassname = self.setHarvestablesIconEnabled
		bShowDistance = self.setHarvestablesDistanceEnabled
		bShowResource = false
		bShowFocus = false
	-- Targeted settings do not have combat-specific options		
	else
		bShowName = self:HelperGetUnitSetting(tNameplate,"Name", bInCombat) or ((tNameplate.bIsImportantNPC and self.setF_Important) or (tNameplate.bIsVendorNPC and self.setF_Vendor))
		bShowGuild = self:HelperGetUnitSetting(tNameplate,"Guild", bInCombat) or ((tNameplate.bIsImportantNPC and self.setF_Important) or (tNameplate.bIsVendorNPC and self.setF_Vendor))
		bShowBars = self:HelperGetUnitSetting(tNameplate,"Bars", bInCombat) and ((unitOwner:GetMaxHealth() or 0) > 0)
		bShowCastBar = self:HelperGetUnitSetting(tNameplate,"Castbar", bInCombat)
		bShowBuffs = self:HelperGetUnitSetting(tNameplate,"Buffs", bInCombat) -- Buffs visibility updated in DrawNameplate
		bShowClassname = self:HelperGetUnitSetting(tNameplate,"Class", bInCombat)
		bShowDistance = self:HelperGetUnitSetting(tNameplate,"Distance", bInCombat)
		bShowResource = tNameplate.bIsPlayer and self:HelperGetUnitSetting(tNameplate,"Resourcebar", bInCombat)
		bShowFocus = tNameplate.bIsPlayer and self:HelperGetUnitSetting(tNameplate,"Focusbar", bInCombat)
		bShowDash = tNameplate.bIsPlayer and self:HelperGetUnitSetting(tNameplate,"Sprintbar", bInCombat)
	end

	if tNameplate.bMouseOverSettingsEnabled then
		if eDisposition == tDisposition.Hostile then
			bShowName = bShowName or self.setMouseOverSettings_H_N
			bShowGuild = bShowGuild or self.setMouseOverSettings_H_G
			bShowBars = bShowBars or self.setMouseOverSettings_H_B
			bShowBuffs = bShowBuffs or self.setMouseOverSettings_H_D
			bShowDistance = bShowDistance or self.setMouseOverSettings_H_DI
			bShowFocus = bShowFocus or self.setMouseOverSettings_H_F
		else
			bShowName = bShowName or self.setMouseOverSettings_F_N
			bShowGuild = bShowGuild or self.setMouseOverSettings_F_G
			bShowBars = bShowBars or self.setMouseOverSettings_F_B
			bShowBuffs = bShowBuffs or self.setMouseOverSettings_F_D
			bShowDistance = bShowDistance or self.setMouseOverSettings_F_DI
			bShowFocus = bShowFocus or self.setMouseOverSettings_F_F		
		end
	end	
	if bIsPlayer then
		local nClassId = tNameplate.unitOwner:GetClassId()
		bShowFocus = bShowFocus and (nClassId == GameLib.CodeEnumClass.Esper 
						or nClassId == GameLib.CodeEnumClass.Medic 
						or nClassId == GameLib.CodeEnumClass.Spellslinger
						)
		if bShowClassname then 
			local enumClasses = GameLib.CodeEnumClass
			if nClassId == enumClasses.Warrior and not self.setStyleClassIconShowWarrior then
				bShowClassname = false
			elseif nClassId == enumClasses.Engineer and not self.setStyleClassIconShowEngineer then
				bShowClassname = false
			elseif nClassId == enumClasses.Esper and not self.setStyleClassIconShowEsper then
				bShowClassname = false
			elseif nClassId == enumClasses.Medic and not self.setStyleClassIconShowMedic then
				bShowClassname = false
			elseif nClassId == enumClasses.Stalker and not self.setStyleClassIconShowStalker then
				bShowClassname = false
			elseif nClassId == enumClasses.Spellslinger and not self.setStyleClassIconShowSpellslinger then
				bShowClassname = false
			end
		end
	else
		bShowFocus = false
	end
	-- Set Unit Name	
	----------------------			
	local strName = unitOwner:GetName()
	if not tNameplate.bIsHarvest then
		if self:HelperGetUnitSetting(tNameplate, "Title", bInCombat) then strName = unitOwner:GetTitleOrName() end
	end
	if not tNameplate.bIsHarvest then
		if bShowClassname and tNameplate.bIsPlayer and (self.setClassIconStyle == "Text" or self.setClassIconStyle == "IconAndText") then 
			local nClassId = (unitOwner:GetClassId() or "")
			local nClassName = (className[nClassId] or "[]")
			strName =  nClassName.. strName
		end
	end
	wndRefs.Name:SetText(strName)
	
	-- Default visibility
	if wndRefs.NameContainer:IsShown() ~= bShowName then
		wndRefs.NameContainer:Show(bShowName)
	end
	if wndRefs.Name:IsShown() ~= bShowName then
		wndRefs.Name:Show(bShowName)
	end
	if bShowName then
		wndRefs.Name:SetTextColor(tNameplate.cDispositionColor)
	end
	
	local nLevel = unitOwner:GetLevel()	
	if (nLevel ~= nil) and self:HelperGetUnitSetting(tNameplate, "LevelText", bInCombat) and ((self.setLevelTextPosition == "Bars") or bShowName) then
		if tNameplate.nLevelPixie == nil then tNameplate.nLevelPixie = -1 end
	else
		self:HelperUpdateLevelPixie(tNameplate, false)
	end	
	
	if self.setStyleCastbarTextInside or self.setStyleCastbarNoText then
		wndRefs.CastBarLabel:Show(false)
	else
		wndRefs.CastBar:SetText("")
	end
	if wndRefs.Cast:IsShown() ~= bShowCastBar then
		wndRefs.Cast:Show(bShowCastBar)
	end
	-- Some Static Color Settings
	wndRefs.AbsorbBar:SetBarColor(self.arColorObjects.AbsorbBar)
	wndRefs.HealthBarBack:SetBarColor(self.arColorObjects.BackdropBar)
	wndRefs.HealthBarBack:SetBGColor(self.arColorObjects.BackgroundBar)
	wndRefs.ShieldBar:SetBGColor(self.arColorObjects.BackgroundBar)
	if wndRefs.ClassResourceBar ~= nil then
		wndRefs.ClassResourceBar:SetBGColor(self.arColorObjects.BackgroundBar)
	end

	if wndRefs.FocusBar ~= nil then
		wndRefs.FocusBar:SetBGColor(self.arColorObjects.BackgroundBar)
	end
	if wndRefs.SprintBar ~= nil then
		wndRefs.SprintBar:SetBGColor(self.arColorObjects.BackgroundBar)
	end
	if wndRefs.DashBar ~= nil then
		wndRefs.DashBar:SetBGColor(self.arColorObjects.BackgroundBar)
	end
	
	local bResourceBarVisibilityChanged = false
	-- Class resource/focus/dash bar
	if wndRefs.FocusBar:IsShown() ~= bShowFocus then
		wndRefs.FocusBar:Show(bShowFocus)
		bResourceBarVisibilityChanged = true
	end
	if wndRefs.ClassResourceBar:IsShown() ~= bShowResource then
		wndRefs.ClassResourceBar:Show(bShowResource)
		bResourceBarVisibilityChanged = true
	end	
	if wndRefs.DashContainer:IsShown() ~= bShowDash then
		wndRefs.DashContainer:Show(bShowDash)
		bResourceBarVisibilityChanged = true
	end	
	
	local classResourceSetting = bShowResource or bShowFocus or bShowDash
	if wndRefs.ClassResource:IsShown() ~= classResourceSetting then
		wndRefs.ClassResource:Show(classResourceSetting)
		bResourceBarVisibilityChanged = true
	end
	if bResourceBarVisibilityChanged or tNameplate.bNeedsRedraw then
		local nHalfWidth = self.nStyleHalfWidth
		local nDashHeight = bShowDash and self.setStyleDashBarHeight or 0
		if bShowResource and bShowFocus and not self.setStyleCombineResourceBars then
			tNameplate.wnd.ClassResource:SetAnchorOffsets(-nHalfWidth, 0, nHalfWidth, self.setStyleFocusbarHeight+self.setStyleResourcebarHeight+nDashHeight)
		elseif bShowResource then
			tNameplate.wnd.ClassResource:SetAnchorOffsets(-nHalfWidth, 0, nHalfWidth, self.setStyleResourcebarHeight+nDashHeight)
		elseif bShowFocus then
			tNameplate.wnd.ClassResource:SetAnchorOffsets(-nHalfWidth, 0, nHalfWidth, self.setStyleFocusbarHeight+nDashHeight)
		end		
	end
	
	
	if (bResourceBarVisibilityChanged or tNameplate.bNeedsRedraw) then
		local nHalfWidth = self.nStyleHalfWidth
		if self.setStyleCombineResourceBars then
			if bShowFocus and bShowResource then
				tNameplate.wnd.ClassResourceBar:SetAnchorOffsets(0,0,nHalfWidth,self.setStyleResourcebarHeight)		
				tNameplate.wnd.FocusBar:SetAnchorOffsets(nHalfWidth,0,2*nHalfWidth,self.setStyleResourcebarHeight)
				tNameplate.wnd.DashContainer:SetAnchorOffsets(0,self.setStyleResourcebarHeight,0,self.setStyleResourcebarHeight+self.setStyleDashBarHeight)
			elseif bShowFocus or bShowResource then
				tNameplate.wnd.ClassResourceBar:SetAnchorOffsets(0,0,2*nHalfWidth,self.setStyleResourcebarHeight)		
				tNameplate.wnd.FocusBar:SetAnchorOffsets(0,0,2*nHalfWidth,self.setStyleFocusbarHeight)
				tNameplate.wnd.DashContainer:SetAnchorOffsets(0,self.setStyleResourcebarHeight,0,self.setStyleResourcebarHeight+self.setStyleDashBarHeight)
			elseif bShowDash then
				tNameplate.wnd.DashContainer:SetAnchorOffsets(0,0,0,self.setStyleDashBarHeight)			
			end
		else
			if bShowFocus and bShowResource then
				tNameplate.wnd.ClassResourceBar:SetAnchorOffsets(0,0,0,self.setStyleResourcebarHeight)		
				tNameplate.wnd.FocusBar:SetAnchorOffsets(0,self.setStyleResourcebarHeight,0,self.setStyleResourcebarHeight+self.setStyleFocusbarHeight)
				tNameplate.wnd.DashContainer:SetAnchorOffsets(0,self.setStyleResourcebarHeight+self.setStyleFocusbarHeight,0,self.setStyleResourcebarHeight+self.setStyleFocusbarHeight+self.setStyleDashBarHeight)
			elseif bShowFocus then
				tNameplate.wnd.FocusBar:SetAnchorOffsets(0,0,0,self.setStyleFocusbarHeight)
				tNameplate.wnd.DashContainer:SetAnchorOffsets(0,self.setStyleFocusbarHeight,0,self.setStyleFocusbarHeight+self.setStyleDashBarHeight)			
			elseif bShowResource then
				tNameplate.wnd.ClassResourceBar:SetAnchorOffsets(0,0,0,self.setStyleResourcebarHeight)		
				tNameplate.wnd.DashContainer:SetAnchorOffsets(0,self.setStyleResourcebarHeight,0,self.setStyleResourcebarHeight+self.setStyleDashBarHeight)
			elseif bShowDash then
				tNameplate.wnd.DashContainer:SetAnchorOffsets(0,0,0,self.setStyleDashBarHeight)			
			end
		end
	end
	-- Does the Unit have a guild?
	local strGuild = ""
	local hasGuild = false
	if bIsPlayer and unitOwner:GetGuildName() and string.len(unitOwner:GetGuildName()) > 0 then
		hasGuild = true
		strGuild = String_GetWeaselString(Apollo.GetString("Nameplates_GuildDisplay"), unitOwner:GetGuildName())
	elseif (not bIsPlayer) and unitOwner:GetAffiliationName() and string.len(unitOwner:GetAffiliationName()) > 0 then
		hasGuild = true
		strGuild = unitOwner:GetAffiliationName()
	elseif tNameplate.bIsHarvest then
		hasGuild = true
		local tradeskillName = (unitOwner:GetHarvestRequiredTradeskillName() or "")
		local tradeskillTier = (unitOwner:GetHarvestRequiredTradeskillTier() or "")
		strGuild = tradeskillName.." Tier "..tradeskillTier
	elseif tNameplate.bIsDatacube then
		hasGuild = true
		strGuild = "Lore item"
	end
	
	--Guild / Affiliation
	if wndRefs.GuildName ~= (hasGuild and bShowGuild) then
		wndRefs.GuildName:Show(hasGuild and bShowGuild)
	end

	if wndRefs.GuildName:IsShown() then
		wndRefs.GuildName:SetTextRaw(strGuild)
		wndRefs.GuildName:SetTextColor(tNameplate.cDispositionColor)
	end

	-- Class/Eliteness Icon
	local bShowEliteness = (tNameplate.bIsNPC and tNameplate.eDisposition ~= tDisposition.Friendly and (unitOwner:GetRank() > 4 or unitOwner:GetEliteness() > 0 or unitOwner:IsRare()) and self:HelperGetUnitSetting(tNameplate, "Eliteness", bInCombat))
	if bShowEliteness or (bShowClassname and ((tNameplate.bIsPlayer and (self.setClassIconStyle == "Icon" or self.setClassIconStyle == "IconAndText")) or (tNameplate.bIsHarvest or tNameplate.bIsDatacube)))
		and (self.setClassIconPosition == "Top" or bShowName) then
		if tNameplate.nClassIconPixie == nil then tNameplate.nClassIconPixie = -1 end
	else
		self:HelperUpdateClassIconPixie(tNameplate, false)
	end
	
	-- Targeted icon
	if (tNameplate.bIsTarget or (tNameplate.bInCluster and self.setShowTargetClusterIcons) or (unitOwner == unitFocus and self.setShowFocusTargetIndicator)) then 
		if tNameplate.bIsTarget then
			if self.setShowNormalTargetIndicator then
				if tNameplate.nTargetPixie == nil then tNameplate.nTargetPixie = -1 end
			else
				self:HelperUpdateTargetIconPixie(tNameplate, false)
			end
			if self.setShowLegacyTargetIndicator and tNameplate.nLegacyTargetPixie == nil then
				tNameplate.nLegacyTargetPixie = namePlate:AddPixie(tLegacyTargetIconPixie)
			end
		else
			if tNameplate.nTargetPixie == nil then tNameplate.nTargetPixie = -1 end
			if tNameplate.nLegacyTargetPixie ~= nil then
				namePlate:DestroyPixie(tNameplate.nLegacyTargetPixie)
				tNameplate.nLegacyTargetPixie = nil
			end				
		end
	else
		if tNameplate.nLegacyTargetPixie ~= nil then
			namePlate:DestroyPixie(tNameplate.nLegacyTargetPixie)
			tNameplate.nLegacyTargetPixie = nil
		end
		self:HelperUpdateTargetIconPixie(tNameplate, false)
	end	
	-- Quest indicator
	local nQuestIconPixieHeight = 0
	local bShowQuestIndicator = (
		tNameplate.nSquaredDistance >= self.setStyleQuestIconMinimumDistanceSquared and (
			(tNameplate.bIsQuestFinisher and self.setShowQuestIconReward) or
			(tNameplate.bIsQuestGiverNew and self.setShowQuestIconNew) or
			(tNameplate.bIsQuestGiverNewDaily and self.setShowQuestIconNewDaily) or
			(tNameplate.bIsQuestReceiver and self.setShowQuestIconReceiving) or
			(tNameplate.bIsQuestGiverNewGrey and self.setShowQuestIconNewGrey)
		)
	)
	if bShowQuestIndicator then
		if tNameplate.nQuestIconPixie == nil then tNameplate.nQuestIconPixie = -1 end
		nQuestIconPixieHeight = self.setStyleQuestIconSize
	else
		self:HelperUpdateQuestIconPixie(tNameplate, false)
	end

	-- Show distance
	if bShowQuestIndicator and self.setStyleQuestIconAlwaysShowDistance then
		bShowDistance = true
	end
	if bShowDistance then
		wndRefs.Distance:SetTextColor(tNameplate.cDispositionColor)
		wndRefs.Distance:Show(true)
	else
		wndRefs.Distance:Show(false)
	end	
	
	-- Heading arrow
	local bShowHeadingPixie = (unitOwner == self.playerUnit and ((bInCombat and self.setM_H_C) or (not bInCombat and self.setM_H)))
	local nHeadingPixieHeight = 0
	if bShowHeadingPixie then
		if tNameplate.nHeadingPixie == nil then tNameplate.nHeadingPixie = -1 end
		nHeadingPixieHeight = self.setStyleHeadingArrowSize
		local bShowHeadingText = self.setStyleHeadingShowText
		local nHeadingWidth = self.setStyleHeadingArrowSize		
		if bShowHeadingText then
			if tNameplate.nHeadingTextPixie == nil then tNameplate.nHeadingTextPixie = -1 end
			nHeadingPixieHeight = nHeadingPixieHeight + self.setStyleHeadingIndicatorTextFont.size+6
			tNameplate.nHeadingPixieWidth = math.max(Apollo.GetTextWidth(self.setStyleHeadingIndicatorTextFont.name,"NW"),self.setStyleHeadingArrowSize)
		else
			tNameplate.nHeadingPixieWidth = nHeadingWidth
		end
	end
	
	-- Group stuff
	local bShowRoleIcon = false
	if tNameplate.nGroupIndex > 0 then
		bShowRoleIcon = self:HelperGetUnitSetting(tNameplate, "Role", bInCombat)
	end
	if bShowRoleIcon and unitOwner:IsInYourGroup() 
		and (
			(tNameplate.bIsHealer and self.setStyleRoleIconShowHealer)
			or (tNameplate.bIsTank and self.setStyleRoleIconShowTank)
			or ((not (tNameplate.bIsHealer or tNameplate.bIsTank)) and self.setStyleRoleIconShowDPS)
		)
		then
		if tNameplate.nRoleIconPixie == nil then tNameplate.nRoleIconPixie = -1 end
	else
		self:HelperUpdateRoleIconPixie(tNameplate, false)
	end
	
	local nToprowContainerHeight = math.max(((not (tNameplate.bIsHarvest or tNameplate.bIsDatacube))and self.setClassIconPosition == "Top" and tNameplate.nClassIconPixie ~= nil) and self.setStyleClassIconSize or 0,
									tNameplate.nRoleIconPixie ~= nil and self.setStyleRoleIconSize or 0,
									tNameplate.nTargetPixie ~= nil and self.setStyleTargetedIconSize or 0,
									nHeadingPixieHeight,
									nQuestIconPixieHeight
								)
	if self.setStyleHeadingArrowBelowTargetMarker and tNameplate.nHeadingPixie ~= nil and tNameplate.nTargetPixie ~= nil then
		nToprowContainerHeight = nToprowContainerHeight + math.min(nHeadingPixieHeight,self.setStyleTargetedIconSize)
	end
	if tNameplate.nQuestIconPixie ~= nil and tNameplate.nTargetPixie ~= nil then
		nToprowContainerHeight = nToprowContainerHeight + math.min(nQuestIconPixieHeight,self.setStyleTargetedIconSize)
	end
	
	if nToprowContainerHeight > 0 then
		wndRefs.ToprowContainer:SetAnchorOffsets(0,0,0,nToprowContainerHeight)
		wndRefs.ToprowContainer:Show(true)
	else
		wndRefs.ToprowContainer:Show(false)
	end

	-- Show Quest/Path/Friend/Rival Icons
	self:DrawNameplateRewardIcons(tNameplate)

	-- Resize Name Window
	if bShowName then
		
		local nNameWidth = Apollo.GetTextWidth(self.setStyleUnitNameFont.name, strName .. " ")
		local nHalfNameWidth = nNameWidth*0.5
		local nNameHeight = self.setStyleUnitNameFont.size + 6
		wndRefs.QuestRewards:SetAnchorOffsets(nHalfNameWidth,-18,0,0)		
		local nHeight = math.max(wndRefs.QuestRewards:IsShown() and wndRefs.QuestRewards:GetHeight() or 0,
								nNameHeight,
								(tNameplate.nClassIconPixie ~= nil and (tNameplate.bIsNPC or tNameplate.bIsHarvest or tNameplate.bIsDatacube or self.setClassIconPosition ~= "Top")) and self.setStyleClassIconSize or 0
								)
		local nQuestRewardWidth = wndRefs.QuestRewards:IsShown() and wndRefs.QuestRewards:GetWidth() or 0
		wndRefs.Name:SetAnchorOffsets(-nHalfNameWidth, -nHeight, nHalfNameWidth, 0)
		wndRefs.NameContainer:SetAnchorOffsets(0, 0, 0, nHeight)	
		tNameplate.nNameWidth = nNameWidth
		tNameplate.nNameHeightOffset = (nHeight - nNameHeight) * 0.5
	else
		tNameplate.nNameWidth = nil
	end
	-- Resize Guild Window
	if bShowGuild then
		local nGuildWidth = Apollo.GetTextWidth(self.setStyleUnitGuildFont.name, strGuild .. " ")
		
		local nGuildHeight = self.setStyleUnitGuildFont.size + 6
		wndRefs.GuildName:SetAnchorOffsets(-nGuildWidth/2, 0, nGuildWidth/2, nGuildHeight)
	end

	if tNameplate.nLevelPixie ~= nil then
		self:HelperUpdateLevelPixie(tNameplate, true, nLevel, true)
	end	
	if tNameplate.nTargetPixie ~= nil then
		self:HelperUpdateTargetIconPixie(tNameplate, true, true)
	end
	if tNameplate.nQuestIconPixie ~= nil then
		self:HelperUpdateQuestIconPixie(tNameplate, true, true)
	end	
	if tNameplate.nClassIconPixie ~= nil then
		self:HelperUpdateClassIconPixie(tNameplate, true, true)
	end	
	if tNameplate.nRoleIconPixie ~= nil then
		self:HelperUpdateRoleIconPixie(tNameplate, true, true)
	end
	if tNameplate.nHPTextPixie ~= nil then
		-- destroy and recreate because changing color is kind of buggy
		self:HelperUpdateHPPixie(tNameplate, false)
		self:HelperUpdateHPPixie(tNameplate, true, nil, true)
	end
	if tNameplate.nSPTextPixie ~= nil then
		self:HelperUpdateSPPixie(tNameplate, true, nil, true)
	end
	if tNameplate.nCleanseIconPixie ~= nil then
		self:HelperUpdateCleanseIconPixie(tNameplate, true, nil, true)
	end		
	
	tNameplate.bShowBars = bShowBars;
	tNameplate.bShowCastBar = bShowCastBar;
	tNameplate.bShowDistance = bShowDistance	
	tNameplate.bShowFocus = bShowFocus
	tNameplate.bShowResource = bShowResource	
	tNameplate.bShowDash = bShowDash
	if self.bStyleSAO then
		if not self:HelperGetUnitSetting(tNameplate, "HPText", bInCombat) and (tNameplate.nLevelPixie == nil or self.setLevelTextPosition == "NameLeft") then
			tNameplate.wnd.Plates:SetAnchorOffsets(0,0,0,self.setHealthbarHeight+self.setShieldbarHeight+10)
		else
			tNameplate.wnd.Plates:SetAnchorOffsets(0,0,0,self.setHealthbarHeight+self.setShieldbarHeight+15)
		end
	end
	if self.bStyleFlat and self.setStyleFlatNumberInside then
		wndRefs.HealthBarBack:SetTextColor(tNameplate.cDispositionColor)
		wndRefs.ShieldBar:SetTextColor(self.arColorObjects.ShieldBar)
	end
end

function OptiPlates:DrawNameplateRewardIcons(tNameplate)
	local wndRefs = tNameplate.wnd
	local unitOwner = tNameplate.unitOwner
	if wndRefs.Name:IsShown() then
		if self.setShowIcons and not unitOwner:IsThePlayer() then
			local bHasRewards = (self:ShowRewardIcons(tNameplate) > 0)
			if wndRefs.QuestRewards:IsShown() ~= bHasRewards then
				wndRefs.QuestRewards:Show(bHasRewards)
			end
		else
			if wndRefs.QuestRewards:IsShown() then
				wndRefs.QuestRewards:Show(false)
			end
		end
	else
		if wndRefs.QuestRewards:IsShown() then
			wndRefs.QuestRewards:Show(false)
		end
	end	
	tNameplate.bUpdateRewards = false
end

function OptiPlates:HelperCalculateNameplateVerticalOffset(tNameplate)
	local wndRefs = tNameplate.wnd
	local nVerticalOffset = 0
	local nNameplateHeight = 0
	local nDebuffHeight = wndRefs.Debuffs:IsShown() and wndRefs.Debuffs:GetHeight() or 0
	local nTopHeight = wndRefs.ToprowContainer:IsShown() and wndRefs.ToprowContainer:GetHeight() or 0
	local nNameHeight = wndRefs.NameContainer:IsShown() and wndRefs.NameContainer:GetHeight() or 0
	local nGuildHeight = wndRefs.GuildName:IsShown() and wndRefs.GuildName:GetHeight() or 0
	local nPlatesHeight = wndRefs.Plates:IsShown() and wndRefs.Plates:GetHeight() or 0
	local nCastHeight = wndRefs.Cast:IsShown() and wndRefs.Cast:GetHeight() or 0
	local nResourceHeight = wndRefs.ClassResource:IsShown() and wndRefs.ClassResource:GetHeight() or 0
	local nDistanceHeight = wndRefs.Distance:IsShown() and wndRefs.Distance:GetHeight() or 0 
	
	if self.setStyleVerticalAnchor == "Name" then
		nVerticalOffset = nGuildHeight + nPlatesHeight + nResourceHeight + nCastHeight
	elseif self.setStyleVerticalAnchor == "Bars" then
		nVerticalOffset = nResourceHeight + nCastHeight
	end
	nNameplateHeight = nGuildHeight + nPlatesHeight + nResourceHeight + nCastHeight +
						nDebuffHeight + nTopHeight + nNameHeight + nDistanceHeight
	tNameplate.nVerticalOffset = nVerticalOffset
	tNameplate.nNameplateHeight = nNameplateHeight
	tNameplate.bDefaultHeight = false
end

-- Static initial config of nameplate elements, like width/height etc. Called once when nameplate is (re)created
function OptiPlates:StyleConfig(tNameplate)
	if tNameplate.wndNameplate == nil then return end
	local namePlate = tNameplate.wndNameplate
	self:RefreshWindowRefs(tNameplate,namePlate)
	local unitPlayer = self.playerUnit
	local unitOwner = tNameplate.unitOwner	
	local wndRefs = tNameplate.wnd
	tNameplate.bNeedsRedraw = true

	local nHalfWidth = self.nStyleHalfWidth
	if not self.setStyleCombineResourceBars then
		tNameplate.wnd.ClassResourceBar:SetAnchorPoints(0,0,1,0)
		tNameplate.wnd.ClassResourceBar:SetAnchorOffsets(0,0,0,self.setStyleResourcebarHeight)
		tNameplate.wnd.FocusBar:SetAnchorPoints(0,0,1,0)
		tNameplate.wnd.FocusBar:SetAnchorOffsets(0,self.setStyleResourcebarHeight,0,self.setStyleResourcebarHeight+self.setStyleFocusbarHeight)
		tNameplate.wnd.DashContainer:SetAnchorPoints(0,0,1,0)
		tNameplate.wnd.DashContainer:SetAnchorOffsets(0,self.setStyleResourcebarHeight+self.setStyleFocusbarHeight,0,self.setStyleResourcebarHeight+self.setStyleFocusbarHeight+self.setStyleDashBarHeight)
	else
		if tNameplate.bShowFocus then
			tNameplate.wnd.ClassResourceBar:SetAnchorPoints(0,0,0,0)
			tNameplate.wnd.ClassResourceBar:SetAnchorOffsets(0,0,nHalfWidth,self.setStyleResourcebarHeight)
			tNameplate.wnd.FocusBar:SetAnchorPoints(0,0,0,0)
			tNameplate.wnd.FocusBar:SetAnchorOffsets(nHalfWidth,0,2*nHalfWidth,self.setStyleResourcebarHeight)
		else
			tNameplate.wnd.ClassResourceBar:SetAnchorPoints(0,0,0,0)
			tNameplate.wnd.ClassResourceBar:SetAnchorOffsets(0,0,self.nStyleWidth,self.setStyleResourcebarHeight)
			tNameplate.wnd.FocusBar:SetAnchorPoints(0,0,0,0)
			tNameplate.wnd.FocusBar:SetAnchorOffsets(nHalfWidth,0,2*nHalfWidth,self.setStyleResourcebarHeight)
		end
		tNameplate.wnd.DashContainer:SetAnchorPoints(0,0,1,0)
		tNameplate.wnd.DashContainer:SetAnchorOffsets(0,self.setStyleResourcebarHeight,0,self.setStyleResourcebarHeight+self.setStyleDashBarHeight)		
	end
	local nHalfCastbarWidth = nHalfWidth + self.setStyleCastbarWidth
	if self.setCastbarTextStyle ~= "Inside" then
		wndRefs.CastBarLabel:SetFont(self.setStyleCastbarFont.name)
		wndRefs.Cast:SetAnchorOffsets(-nHalfCastbarWidth,
									0,
									nHalfCastbarWidth,
									self.setStyleCastbarHeight+self.setStyleCastbarPaddingTop + 6 + self.setStyleCastbarFont.size)		
		if self.setCastbarTextStyle == "Below" then
			wndRefs.CastBarLabel:SetAnchorOffsets(-nHalfCastbarWidth,
											self.setStyleCastbarPaddingTop + self.setStyleCastbarHeight,
											nHalfCastbarWidth,
											self.setStyleCastbarFont.size + 6 + self.setStyleCastbarPaddingTop + self.setStyleCastbarHeight
											)
			wndRefs.CastBar:SetAnchorOffsets(-nHalfCastbarWidth,
										self.setStyleCastbarPaddingTop,
										nHalfCastbarWidth,
										self.setStyleCastbarHeight+self.setStyleCastbarPaddingTop
										)
		else
			wndRefs.CastBarLabel:SetAnchorOffsets(-nHalfCastbarWidth,
											self.setStyleCastbarPaddingTop,
											nHalfCastbarWidth,
											self.setStyleCastbarFont.size + 6 + self.setStyleCastbarPaddingTop
											)
			wndRefs.CastBar:SetAnchorOffsets(-nHalfCastbarWidth,
										self.setStyleCastbarFont.size + 6 + self.setStyleCastbarPaddingTop,
										nHalfCastbarWidth,
										self.setStyleCastbarFont.size + 6 + self.setStyleCastbarHeight+self.setStyleCastbarPaddingTop
										)
		end
	else
		wndRefs.Cast:SetAnchorOffsets(-nHalfCastbarWidth,
									0,
									nHalfCastbarWidth,
									self.setStyleCastbarHeight+self.setStyleCastbarPaddingTop)
		wndRefs.CastBar:SetFont(self.setStyleCastbarFont.name)
		wndRefs.CastBar:SetAnchorOffsets(-nHalfCastbarWidth,self.setStyleCastbarPaddingTop,nHalfCastbarWidth,self.setStyleCastbarHeight+self.setStyleCastbarPaddingTop)	
		wndRefs.CastBarLabel:Show(false)
	end
	wndRefs.FocusBar:SetFont(self.setStyleFocusFont.name)
	wndRefs.ClassResourceBar:SetFont(self.setStyleClassResourceFont.name)
	wndRefs.Name:SetFont(self.setStyleUnitNameFont.name)
	wndRefs.GuildName:SetFont(self.setStyleUnitGuildFont.name)
	wndRefs.Distance:SetFont(self.setStyleDistanceFont.name)
	wndRefs.Distance:SetAnchorOffsets(0, 0, 0, self.setStyleDistanceFont.size + 6)
	
	
	
	tNameplate.wnd.HealthBarBack:SetAnchorOffsets(-nHalfWidth, 5+self.setShieldbarHeight, nHalfWidth, 5+self.setShieldbarHeight+self.setHealthbarHeight)
	tNameplate.wnd.ShieldBar:SetAnchorOffsets(-nHalfWidth, 5, nHalfWidth, 5+self.setShieldbarHeight)	
	if self.bStyleFlat then 
		local nPlatesHeight = self.setHealthbarHeight+self.setShieldbarHeight+5
		if nPlatesHeight < 25 then nPlatesHeight = 25 end
		tNameplate.wnd.Plates:SetAnchorOffsets(0,0,0,nPlatesHeight)
	
		tNameplate.wnd.HealthBar:SetFullSprite(healthTexture[self.setFlatHealthTexture])
		tNameplate.wnd.HealthBarBack:SetEmptySprite(healthTexture[self.setFlatHealthTexture])
		tNameplate.wnd.HealthBarBack:SetFullSprite(healthTexture[self.setFlatHealthTexture])
		tNameplate.wnd.NotTargeting:SetSprite("Flat_Border")
		
		tNameplate.wnd.ShieldBar:SetEmptySprite(shieldTexture[self.setFlatShieldTexture])
		tNameplate.wnd.ShieldBar:SetFullSprite(shieldTexture[self.setFlatShieldTexture])
	end

	--Resource Bar texture
	if namePlate:FindChild("ClassResource") ~= nil then
		namePlate:FindChild("ClassResourceBar"):SetEmptySprite(healthTexture[self.setFlatHealthTexture])
		namePlate:FindChild("ClassResourceBar"):SetFullSprite(healthTexture[self.setFlatHealthTexture])
		namePlate:FindChild("FocusBar"):SetEmptySprite(healthTexture[self.setFlatHealthTexture])
		namePlate:FindChild("FocusBar"):SetFullSprite(healthTexture[self.setFlatHealthTexture])
	end

	tNameplate.wnd.FocusBar:SetTextColor(self.arColorObjects.FocusBarText)
	tNameplate.wnd.FocusBar:SetBarColor(self.arColorObjects.FocusBar)

	tNameplate.wnd.ClassResourceBar:SetTextColor(self.arColorObjects.ResourceBarText)
	tNameplate.wnd.ClassResourceBar:SetBarColor(self.arColorObjects.ResourceBar)	
	
	tNameplate.wnd.SprintBar:SetBarColor(self.arColorObjects.SprintBar)
	tNameplate.wnd.SprintBar:SetTextColor(self.arColorObjects.SprintBarText)	
	
	tNameplate.wnd.DashBar:SetBarColor(self.arColorObjects.DashBar)
	tNameplate.wnd.DashBar:SetTextColor(self.arColorObjects.DashBarText)		
end

-- Config button to the side of OPTIONS
function OptiPlates:OnConfigure()
	self:OnOptiPlatesOn()
end


function OptiPlates:OnUnitDestroyed(unitDead)
	local idUnit = unitDead:GetId()
	local tNameplate = self.arUnit2Nameplate[idUnit]
	if tNameplate == nil then
		return
	end
	if tNameplate.bIsMount then
		if self.arMount2Unit[idUnit] ~= nil then
			local unitNameplate = self.arUnit2Nameplate[self.arMount2Unit[idUnit]:GetId()]
			if unitNameplate ~= nil and unitNameplate.wndNameplate ~= nil then unitNameplate.wndNameplate:SetUnit(self.arMount2Unit[idUnit]) end
			self.arMount2Unit[idUnit] = nil
		end
		
	end
	self:UnattachNameplateWindow(tNameplate)
	self.arUnit2Nameplate[idUnit] = nil
	self.arFriendlyUnitNameplates[idUnit] = nil
end

-- This function gets called when FrameCount is changed, does not fire when loading screen is loading
-- Useful for blocking main loop (OnFrame) until loading is done, not used outside of that.
function OptiPlates:OnFrameVarChange()
	if self.bInitialLoadAllClear then return end
	-- OptionsChanged sets the self.bInitialLoadAllClear to true, so the following is executed only once
	if self.playerUnit ~= nil and self.playerUnit:IsValid() and self.xmlLoaded then
		self:OptionsChanged()
		_AddonInfoObject = Apollo.GetAddonInfo("OptiPlates")
		if self.setOneVersionIntegrationEnabled then
			if _AddonVersionPatch ~= nil then
				Event_FireGenericEvent("OneVersion_ReportAddonInfo", "OptiPlates", _AddonVersionMajor, _AddonVersionMinor, _AddonVersionPatch)
			else
				Event_FireGenericEvent("OneVersion_ReportAddonInfo", "OptiPlates", _AddonVersionMajor, _AddonVersionMinor)
			end
		end
		Apollo.RemoveEventHandler("VarChange_FrameCount", self)
	end
end

-- This function gets called every frame
function OptiPlates:OnFrame()
	if self.xmlLoaded == false then return end
	local nCurrentTicks = Apollo.GetTickCount()	
	-- Limit full reward updates to once every nRewardUpdateMinInterval milliseconds (testing)
	if self.bRewardInfoDirty and (nCurrentTicks - nPrevRewardUpdateTicks) > nRewardUpdateMinInterval then 
		self:UpdateRewardInfo(); 
		nPrevRewardUpdateTicks = nCurrentTicks
	end
	
	
	self.frameCounter = self.frameCounter - 1
	local drawn = 0
	local drawnPlayers = 0
	local notDrawn = 0
	local total = 0
	if bOptionChangeQueued then
		self:OptionsChanged()
	end

	if self.frameCounter < 1 and self.playerUnit ~= nil and self.playerUnit:IsValid() and self.bInitialLoadAllClear then
		-- This conditional only gets called immediately after initial load is done,
		-- tInitialNewUnits have unit structs that had a UnitCreated event fire during loading.
		if self.tInitialNewUnits ~= nil then
			for idx,unit in pairs(self.tInitialNewUnits) do
				if unit and (unit:GetId() or -1) >= 0 then
					self:OnUnitCreated(unit)
				end
				self.tInitialNewUnits[idx] = nil
			end
		end
		subdued = false;
		local updateBuffsThisFrame = updateBuffs
		local bGroupInfoDirty = self.bGroupInfoDirty
		local playerWorldPosition = self.playerUnit:GetPosition()
		self.playerPosition = playerWorldPosition
		local currentTarget = self.playerUnit:GetTarget()
		
		if bGroupInfoDirty then
			self.arGroupRoles = {}
			self.arGroupRoles["Healer"] = {}
			self.arGroupRoles["Tank"] = {}
		end
		-- update group info if it's dirty
		if bGroupInfoDirty then 
			if self.tGroupUnits ~= nil then
				for groupId, nameplate in pairs(self.tGroupUnits) do
					nameplate.nGroupIndex = -1
					self.tGroupUnits[groupId] = nil
				end
			end
			local nGroupMaxSize = GroupLib.GetGroupMaxSize()
			for idx=1,nGroupMaxSize do
				local groupUnit = GroupLib.GetUnitForGroupMember(idx)
				if groupUnit ~= nil then
					local groupUnitId = groupUnit:GetId()
					local grpNameplate = self.arUnit2Nameplate[groupUnitId]
					local grpInfo = GroupLib.GetGroupMember(idx)
					if grpNameplate ~= nil then
						self.tGroupUnits[idx] = grpNameplate
						grpNameplate.nGroupIndex = idx
						if (grpNameplate.bIsTank ~= grpInfo.bTank) or
							(grpNameplate.bIsHealer ~= grpInfo.bHealer) or
							(grpNameplate.bIsDPS ~= grpInfo.bDPS) 
						  then
							grpNameplate.bIsDPS = grpInfo.bDPS
							grpNameplate.bIsHealer = grpInfo.bHealer
							grpNameplate.bIsTank = grpInfo.bTank
							grpNameplate.bNeedsRedraw = true
							self:UpdateNameplateVisibility(grpNameplate)
						end
						if grpNameplate.bIsHealer then table.insert(self.arGroupRoles["Healer"], grpNameplate) end
						if grpNameplate.bIsTank then table.insert(self.arGroupRoles["Tank"], grpNameplate) end
					end
				end
			end			
		end
		
		if self.clickTargetUnit ~= nil then
			GameLib.SetTargetUnit(self.clickTargetUnit)
			self.clickTargetUnit = nil
		end
		
		if self.setHideFullFriendlyBars then
			local nHealthThreshold = self.setHideFullFriendlyBarsHealthThreshold
			local nShieldThreshold = self.setHideFullFriendlyBarsShieldThreshold
			local bUseOneThreshold = self.setHideFullFriendlyBarsOneThreshold
			for idx, tNameplate in pairs(self.arFriendlyUnitNameplates) do
				local unitOwner = tNameplate.unitOwner
				local nHealthCurrent = unitOwner:GetHealth()
				local nHealthMax = unitOwner:GetMaxHealth()
				local nShieldCurrent = unitOwner:GetShieldCapacity()
				local nShieldMax = unitOwner:GetShieldCapacityMax()
				local nHealthPercent = nil
				local nShieldPercent = nil
				if unitOwner ~= self.playerUnit and nHealthCurrent ~= nil and nHealthCurrent > 0 and nHealthMax ~= nil and nHealthMax > 0 then
					nHealthPercent = nHealthCurrent/nHealthMax*100
					if nShieldCurrent ~= nil and nShieldMax ~= nil and nShieldMax > 0 then
						nShieldPercent = nShieldCurrent/nShieldMax*100
					end
					if bUseOneThreshold then
						if (nHealthPercent ~= nil and nHealthPercent < nHealthThreshold) or
						   (nShieldPercent ~= nil and nShieldPercent < nShieldThreshold)
						then
							if tNameplate.bFriendlyNameplateHealthVisibility ~= true and (tNameplate.wndNameplate == nil or not tNameplate.bShow) then
								tNameplate.bFriendlyNameplateHealthVisibility = true
								self:UpdateNameplateVisibility(tNameplate)
								tNameplate.bNeedsRedraw = true
							end
						elseif tNameplate.bFriendlyNameplateHealthVisibility == true and (tNameplate.wndNameplate ~= nil or tNameplate.bShow) then
								tNameplate.bFriendlyNameplateHealthVisibility = false
								self:UpdateNameplateVisibility(tNameplate)
								tNameplate.bNeedsRedraw = true		
						end	
					else
						if (nHealthPercent ~= nil and nHealthPercent < nHealthThreshold) and
						   (nShieldPercent ~= nil and nShieldPercent < nShieldThreshold)
						then
							if tNameplate.bFriendlyNameplateHealthVisibility ~= true and (tNameplate.wndNameplate == nil or not tNameplate.bShow) then
								tNameplate.bFriendlyNameplateHealthVisibility = true
								self:UpdateNameplateVisibility(tNameplate)
								tNameplate.bNeedsRedraw = true
							end
						elseif tNameplate.bFriendlyNameplateHealthVisibility == true and (tNameplate.wndNameplate ~= nil or tNameplate.bShow) then
								tNameplate.bFriendlyNameplateHealthVisibility = false
								self:UpdateNameplateVisibility(tNameplate)
								tNameplate.bNeedsRedraw = true
						end			
					end				
				end				
			end
		end
		
		-- Main loop that draws the nameplates	
		local unitPlayer = self.playerUnit
		local tPlayerNameplate = self.arUnit2Nameplate[unitPlayer:GetId()]
		local unitMouseOverUnit = self.mouseOverUnit
		-- last damage taken needs to update even if bars are hidden
		if tPlayerNameplate ~= nil then
			local nTotalHealth = unitPlayer:GetHealth()
				+ (unitPlayer:GetShieldCapacity() or 0) 
				+ (unitPlayer:GetAbsorptionValue() or 0) 
			if nTotalHealth < tPlayerNameplate.nTotalHealth then
				tPlayerNameplate.previousDamageTicks = Apollo:GetTickCount()
			end
			tPlayerNameplate.nTotalHealth = nTotalHealth
		end
		
		local nBlinded = (unitPlayer:GetCCStateTimeRemaining(Unit.CodeEnumCCState.Blind) or 0)
		if nBlinded ~= 0 then 
			self.bBlinded = true 
		else
			self.bBlinded = false
		end
		local bBlindConditional = (not self.setStyleHideBarsWhenBlinded) or (self.setStyleHideBarsWhenBlinded and not self.bBlinded)
		local bNeverShowConditional = not self.setNeverShow	and not (self.setNeverShowCity and self:IsInCity())
		local bBuffProfilesDirty = self.bBuffProfilesDirty
		
		local bMouseOverOn = self.setMouseOverEnabled and
				(not self.setMouseOverModifiersAlt or Apollo.IsAltKeyDown()) and
				(not self.setMouseOverModifiersCtrl or Apollo.IsControlKeyDown()) and
				(not self.setMouseOverModifiersShift or Apollo.IsShiftKeyDown())
		
		if unitMouseOverUnit ~= nil and unitMouseOverUnit:IsValid() then
			local tmpNameplate = self.arUnit2Nameplate[unitMouseOverUnit:GetId()]
			if tmpNameplate ~= nil then
				if bMouseOverOn and tmpNameplate.bMouseOverSettingsEnabled == false then
					tmpNameplate.bMouseOverSettingsEnabled = true
					self:ReloadNameplate(tmpNameplate)
					tmpNameplate.bNeedsRedraw = true
				elseif not bMouseOverOn and tmpNameplate.bMouseOverSettingsEnabled then
					tmpNameplate.bMouseOverSettingsEnabled = false
					self:ReloadNameplate(tmpNameplate)
					tmpNameplate.bNeedsRedraw = true
				end
			end
		end
		local bPvpWorkaroundEnabled = self.setHostilePvpOcclusionEnabled
		local bPvpEnabled = self.playerUnit:IsPvpFlagged()
		for idx, tNameplate in pairs(self.arDisplayedNameplates) do
			total = total + 1
			local unitOwner = tNameplate.unitOwner
			local wndNameplate = tNameplate.wndNameplate
			-- see if buff profile has changed
			if bBuffProfilesDirty then
				tNameplate.bBuffProfileDirty = true
			end
			local tmpDisposition = unitOwner:GetDispositionTo(unitPlayer)
			if tmpDisposition ~= nil and tmpDisposition ~= tNameplate.eDisposition then
				tNameplate.eDisposition = tmpDisposition
				self:UpdateNameplateVisibility(tNameplate)
				tNameplate.bNeedsRedraw = true
			end
			if tNameplate ~= nil and tNameplate.wndNameplate ~= nil and unitOwner ~= nil then
				if tNameplate.bIsPlayer and unitOwner:IsMounted() then
					local mountId = tNameplate.unitOwner:GetUnitMount():GetId()
					if self.arMount2Unit[mountId] == nil then
						tNameplate.wndNameplate:SetUnit(tNameplate.unitOwner:GetUnitMount())
					end
					self.arMount2Unit[mountId] = tNameplate.unitOwner
				end
				local bHidePvpNameplate = false
				if (not self.setEnableOcclusion) and bPvpWorkaroundEnabled and bPvpEnabled and tNameplate.bIsPlayer and tNameplate.eDisposition == tDisposition.Hostile and unitOwner:IsPvpFlagged() then
					local tUnitPos = unitOwner:GetPosition()
					if self.playerUnit:IsDead() then
						tNameplate.tLastDeathLocation = tUnitPos
					elseif unitOwner:IsDead() then
						tNameplate.tLastDeathLocation = tUnitPos
					elseif tNameplate.nSquaredDistance > 10000 then -- 100 meters
						tNameplate.tLastDeathLocation = tUnitPos
					else
						local tLastDeathLocation = tNameplate.tLastDeathLocation
						if tLastDeathLocation ~= nil and
							tLastDeathLocation.x == tUnitPos.x and
							tLastDeathLocation.y == tUnitPos.y and
							tLastDeathLocation.y == tUnitPos.y
						then
							bHidePvpNameplate = true
						else
							tNameplate.tLastDeathLocation = nil
						end
					end
					if not bHidePvpNameplate and tNameplate.nPreviousLocation ~= nil then
						local tOldPos = tNameplate.tPreviousLocation
						if tUnitPos.x == tOldPos.x and
							tUnitPos.y == tOldPos.y and 
							tUnitPos.z == tOldPos.z and
							tNameplate.nPreviousLocationTime > 0
						then
							if (nCurrentTicks - tNameplate.nPreviousLocationTime) > self.setHostilePvpOcclusionTimeout then
								if unitOwner:IsMounted() then
									bHidePvpNameplate = unitOwner:GetUnitMount():IsOccluded()
								else
									bHidePvpNameplate = unitOwner:IsOccluded()
								end
							end
						else
							tNameplate.nPreviousLocationTime = nCurrentTicks
							tNameplate.tPreviousLocation = tUnitPos
						end
					else
						tNameplate.nPreviousLocationTime = nCurrentTicks
						tNameplate.tPreviousLocation = tUnitPos
					end
				end
				-- visibility conditional
				if bBlindConditional
					and tNameplate.bShow -- check if nameplate wants to be drawn
					and (bNeverShowConditional
						or ((tNameplate.bIsVendorNPC and self.setF_Vendor) or (tNameplate.bIsImportantNPC and self.setF_Important))
					)
					and not unitOwner:IsDead() and unitOwner:IsValid() and not bHidePvpNameplate
					and not self:IsOccluded(tNameplate)-- if occlusion check is on, check it here
					and wndNameplate:IsOnScreen() and self:CheckDrawDistance(tNameplate) -- most expensive checks, whether nameplate is on screen and draw distance
				-- end visibility conditional
				then
					-- sanity check, if self.unitLastTarget is not current target, something screwy is going on
					if unitOwner == currentTarget
						and self.unitLastTarget ~= tNameplate
					then
						self:OnTargetUnitChanged(unitOwner)
					end
					-- update unit height
					if tNameplate.bIsNPC and self.setAutoAdjustNameplatesDown then
						local tScreenPos = GameLib.GetUnitScreenPosition(unitOwner)
						if tScreenPos.bOnScreen then
							local tAnchor = unitOwner:GetOverheadAnchor()
							if tAnchor ~= nil then
								tNameplate.nUnitHeight = tScreenPos.nY - tAnchor.y
								tNameplate.nOverheadAnchorY = tAnchor.y
								tNameplate.bDefaultHeight = false
							end
						else
							tNameplate.nUnitHeight = 0
							tNameplate.nOverheadAnchorY = 0
							tNameplate.bDefaultHeight = true
						end
					end
					
					if tNameplate.bNeedsRedraw then
						self:SetNormalNameplate(tNameplate)
					end
					if unitOwner == self.playerUnit then
						local bInCombat = unitOwner:IsInCombat()
						if self:HelperGetUnitSetting(tNameplate, "Heading", bInCombat) then
							self:HelperUpdateHeadingPixie(tPlayerNameplate, true, tPlayerNameplate.bNeedsRedraw)
						else
							self:HelperUpdateHeadingPixie(tPlayerNameplate, false)
						end
					end
					drawn = drawn + 1

					if tNameplate.bIsPlayer then drawnPlayers = drawnPlayers + 1 end
					
					if wndNameplate:IsShown() == false then
						wndNameplate:Show(true,true)
					end
					-- DrawNameplate calls all other DrawNameplate* functions as necessary, and is the meat of the thing
					self:DrawNameplate(tNameplate, updateBuffsThisFrame)	
					
					if tNameplate.bIsPlayer and tNameplate.eDisposition == tDisposition.Friendly and
						self.setNameplateBringDamagedToFront and
						(tNameplate.nShieldPercent or 1000) < self.setNameplateBringDamagedToFrontShieldThreshold and
						(tNameplate.nHealthPercent or 1000) < self.setNameplateBringDamagedToFrontHealthThreshold
					then
						table.insert(self.arDamagedUnits, tNameplate)
					end
				else
					notDrawn = notDrawn + 1
					if wndNameplate ~= nil and wndNameplate:IsShown() then
						tNameplate.bNeedsRedraw = true
						wndNameplate:Show(false,true)
					end
				end
			end
		end

		updateBuffs = false;
		self.refreshHealthShieldBar = false
		self.bGroupInfoDirty = false
		self.bBuffProfilesDirty = false
		self.frameCounter = self.setNameplateUpdateInterval
	else
		self.playerUnit = GameLib.GetPlayerUnit()
	end -- end frameCounter check
	
	-- Bringing nameplates to front needs to be done every frame regardless of settings
	if self.playerUnit ~= nil and self.playerUnit:IsValid() then
		if self.setNameplateBringHealerToFront and self.arGroupRoles ~= nil and self.arGroupRoles["Healer"] ~= nil then
			for idx, np in pairs(self.arGroupRoles["Healer"]) do
				if np.wndNameplate ~= nil then
					np.wndNameplate:ToFront()
				end
			end
		end
		
		if self.setNameplateBringTankToFront and self.arGroupRoles ~= nil and self.arGroupRoles["Tank"] ~= nil then
			for idx, np in pairs(self.arGroupRoles["Tank"]) do
				if np.wndNameplate ~= nil then
					np.wndNameplate:ToFront()
				end
			end
		end
		if self.setNameplateBringFocusToFront then
			local focusTarget = self.playerUnit:GetAlternateTarget()
			if focusTarget ~= nil then
				local np = self.arUnit2Nameplate[focusTarget:GetId()]
				if np ~= nil and np.wndNameplate ~= nil then
					np.wndNameplate:ToFront()
				end
			end
		end

		if self.setNameplateBringDamagedToFront then
			local bClearArray = self.frameCounter <= 1
			for idx, np in pairs(self.arDamagedUnits) do
				if np.wndNameplate ~= nil then
					np.wndNameplate:ToFront()
				end
				if bClearArray then
					self.arDamagedUnits[idx] = nil
				end
			end
		end
		if self.setNameplateBringTargetToFront then
			local targetUnit = self.playerUnit:GetTarget()
			if targetUnit ~= nil then
				local np = self.arUnit2Nameplate[targetUnit:GetId()]
				if np ~= nil and np.wndNameplate ~= nil then
					np.wndNameplate:ToFront()
				end
			end
		end
		if self.setNameplateBringSelfToFront then
			local np = self.arUnit2Nameplate[self.playerUnit:GetId()]
			if np ~= nil and np.wndNameplate ~= nil then
				self.arUnit2Nameplate[self.playerUnit:GetId()].wndNameplate:ToFront()
			end
		end
		
	end
	
	if updateStatistics then
		local currentTicks = Apollo.GetTickCount()
		local arr = self.optionsUI_AboutSettings
	    if updateStatisticsInterval < (currentTicks - statisticsLastUpdated) then
			_AddonInfoObject = Apollo.GetAddonInfo("OptiPlates")
			statisticsLastUpdated = currentTicks
			local tmpUnits = 0
			local totalBuffs = 0				
			local visibleBuffs = 0
			for idx,unit in pairs(self.arUnit2Nameplate) do
				if unit then 
					tmpUnits = tmpUnits + 1 
					if unit.wndNameplate ~= nil then
						local tmpBuffs = #unit.wnd.Debuffs:GetChildren()
						totalBuffs = totalBuffs + tmpBuffs
						if unit.wndNameplate:IsShown() then
							visibleBuffs = visibleBuffs + tmpBuffs
						end
					end
				end
			end
			local formattedTime = string.gsub(_AddonInfoObject.strLastModified, "(%d%d?)/(%d%d?)/(%d%d?)", "%2.%1.%3")

			arr.TotalUnits:SetText(tmpUnits)			
			arr.BuffWindows:SetText(totalBuffs)			
			arr.VisibleBuffWindows:SetText(visibleBuffs)
			arr.Version:SetText(string.format("%s   (%s)",_AddonVersion,formattedTime))
			arr.MemoryUsage:SetText(string.format("%d kB",math.floor(_AddonInfoObject.nMemoryUsage/1024)))
			arr.LastSecondTime:SetText(string.format("%d ms (%.2f ms per frame)",math.floor(_AddonInfoObject.fCallTimePerSecond*1000),_AddonInfoObject.fCallTimePerSecond*1000/GameLib.GetFrameRate()))			
		end
		
		arr.CachedNameplates:SetText(self.nFreeWindows)
		if self.frameCounter == self.setNameplateUpdateInterval then
			arr.TotalNameplates:SetText(total)
			arr.ShownNameplates:SetText(drawn)
			arr.ShownPlayerNameplates:SetText(drawnPlayers)
		end
		if self.wndMain:IsShown() == false then
			updateStatistics = false
		end
	end
end

function OptiPlates:HelperCountChildWindows(wnd, visible)
	if wnd == nil then return 0 end
	local children = wnd:GetChildren()
	local count = self:GetTableLength(children)
	if count == 0 then 
		return 0
	elseif (visible and not wnd:IsShown()) then
		return 0
	else
		local tmpcount = 0
		for idx, tmp in pairs(children) do
			tmpcount = tmpcount + self:HelperCountChildWindows(tmp, visible) + 1
		end
		return tmpcount
	end
end

-- returns the letter of the currently applicable settings (setT_* is target settings, would return T, setM_* is my settings, so M etc)
function OptiPlates:UnitSettingLookup(tNameplate,bNoTargetOrGuild)
	local unitOwner = tNameplate.unitOwner
	local unitPlayer = self.playerUnit
	if unitOwner == nil or unitPlayer == nil then return nil end
	
	if (unitOwner == unitPlayer:GetTarget() or tNameplate.bInCluster) and (not bNoTargetOrGuild)
		and (not (unitOwner == unitPlayer and self.setNoTargetSettingsOnSelf))
		then
		return "T"
	elseif unitOwner:IsThePlayer() then
		return "M"
	elseif unitOwner:IsInYourGroup() and ((not self.setGP_Override) or bNoTargetOrGuild or (self.setGP_Override and not tNameplate.bIsGuildMember)) then
		return "PP"
	elseif unitOwner:IsACharacter() then
		local eDisposition = tNameplate.eDisposition 
		if Unit.CodeEnumDisposition.Hostile == eDisposition then
			return "HP"
		else
			if tNameplate.bIsGuildMember and not bNoTargetOrGuild then
				return "GP"
			end
			return "FP"
		end
	else 
		local eDisposition = tNameplate.eDisposition 
		if Unit.CodeEnumDisposition.Hostile == eDisposition then
			return "H"
		elseif Unit.CodeEnumDisposition.Neutral == eDisposition then
			return "N"
		else
			return "F"
		end
	end
end


-- check whether a unit should have nameplate window and whether or not it should be visible
function OptiPlates:UpdateNameplateVisibility(tNameplate)
	if tNameplate == nil then return end
	if self.tNameplateSettings == nil then 
		tNameplate.eDisposition = Unit.CodeEnumDisposition.Unknown
		return 
	end
	local unitOwner = tNameplate.unitOwner
	
	tNameplate.bIsWeapon			= self:IsPlayerWeapon(unitOwner) -- disarmed weapon
	tNameplate.bIsGuildMember		= self:IsGuildMember(unitOwner)
	tNameplate.bIsImportantNPC		= self:IsImportantNPC(unitOwner) -- Is unit important (as defined by self:IsImportantNPC)
	tNameplate.bIsUnimportantNPC	= (not tNameplate.bIsImportantNPC and self:IsUnimportantNPC(unitOwner))
	tNameplate.bIsVendorNPC			= self:IsVendorNPC(unitOwner) -- Is unit a vendor
	tNameplate.bSpeechBubble 		= unitOwner:HasTextBubble()
	
	tNameplate.sUnitType = self:HelperGetUnitType(tNameplate)
	tNameplate.tUnitTypeSettings = self.tNameplateSettings[tNameplate.sUnitType]

	
	local bShouldHaveNameplate = self:ShouldHaveNameplate(tNameplate)
	if not bShouldHaveNameplate and tNameplate.wndNameplate ~= nil then
		self:UnattachNameplateWindow(tNameplate)
		tNameplate.bNeedsRedraw = true
	elseif bShouldHaveNameplate and tNameplate.wndNameplate == nil then
		self:CreateNameplateWindow(tNameplate)
		self:StyleConfig(tNameplate)
		tNameplate.bNeedsRedraw = true
	end
	local bOldVisible = tNameplate.bShow
	local bIsVisible = self:HelperVerifyVisibilityOptions(tNameplate)
	if bIsVisible ~= bOldVisible then tNameplate.bNeedsRedraw = true end
	tNameplate.bShow = bIsVisible
end

function OptiPlates:HelperGetUnitType(tNameplate)
	local unitOwner = tNameplate.unitOwner
	if unitOwner:IsThePlayer() then
		return "Self"
	elseif unitOwner:IsInYourGroup() then
		return "Party"
	elseif unitOwner:IsACharacter() then
		local eDisposition = tNameplate.eDisposition 
		if Unit.CodeEnumDisposition.Hostile == eDisposition then
			return "HostilePlayer"
		else
			--if tNameplate.bIsGuildMember then
			--	arr = self.tNameplateSettings.Guild
			--end
			return "FriendlyPlayer"
		end
	else 
		local eDisposition = tNameplate.eDisposition 
		if Unit.CodeEnumDisposition.Hostile == eDisposition then
			return "HostileNPC"
		elseif Unit.CodeEnumDisposition.Neutral == eDisposition then
			return "Neutral"
		else
			return "FriendlyNPC"
		end
	end
end
-- Called every frame if bars should be shown
function OptiPlates:DrawNameplateBars(tNameplate)
	local hasHealth = tNameplate.bShowHealthBar
	local hasShield = tNameplate.bShowShieldBar
	local vulnerableBar = tNameplate.wnd.VulnerableBar
	local unitOwner = tNameplate.unitOwner
	local unitPlayer = self.playerUnit
	local bInCombat = tNameplate.unitOwner:IsInCombat()
	local nVulnerable = unitOwner:GetCCStateTimeRemaining(Unit.CodeEnumCCState.Vulnerability)	
	local nHealthCurrent = unitOwner:GetHealth() or 0
	
	local nHealthMax = unitOwner:GetMaxHealth() or 0
	local nHealthPercent = 100
	local nShieldPercent = 100
	
	local nAbsorbCurrent = unitOwner:GetAbsorptionValue() or 0
	local nAbsorbMax = unitOwner:GetAbsorptionMax() or 0

	local nShieldCurrent = unitOwner:GetShieldCapacity() or 0
	local nShieldMax = unitOwner:GetShieldCapacityMax() or 0

	nHealthPercent = tNameplate.nHealthPercent

	local eDisposition = tNameplate.eDisposition
	local bIsPlayer = tNameplate.bIsPlayer
	
	local bBarsChanged = not (nHealthCurrent == tNameplate.nHealth and
						  nShieldCurrent == tNameplate.nShield and
						  nShieldMax == tNameplate.nShieldMax and
						  nHealthMax == tNameplate.nHealthMax)
	
	local bShowBars = nil
	-- Shield reboot needs to be drawn every (or near every) frame, 
	-- so it is here and not inside the conditional below
	-- 
	-- It does not care where the "damage" comes from, so equipping items with less hp/shield will trigger the bar also
	if unitOwner == unitPlayer and self.setM_SR and nShieldMax ~= nil and nShieldMax > 0 then
		local nShieldRebootTime = unitOwner:GetShieldRebootTime()
		local shieldTimerBar = tNameplate.wnd.ShieldTimerBar				
		if nShieldRebootTime ~= nil and nShieldMax ~= nil and nShieldCurrent ~= nil and nShieldMax > 0 and nShieldCurrent < nShieldMax then 
			local nCurrentTicks = Apollo:GetTickCount()
			local ticksSinceLastDamage = nCurrentTicks - tNameplate.previousDamageTicks
			local nShieldReboot = (nShieldRebootTime - ticksSinceLastDamage)
			shieldTimerBar:SetMax(8000)
			shieldTimerBar:SetProgress(nShieldReboot)
			if not shieldTimerBar:IsShown() then
				shieldTimerBar:Show(true)
			end
		else
			if shieldTimerBar:IsShown() then
				shieldTimerBar:Show(false)
			end			
		end		
	end

	-- health and shield amounts are stored in the nameplate object
	-- they are redrawn if an event triggers redraw for a single nameplate (bNeedsRedraw)
	-- or for all of them (refreshHealthShieldBar)
	-- This is because they don't need updating every frame, and the game tells us
	-- when they should be updated.
	if bBarsChanged or tNameplate.bNeedsRedraw or ((tNameplate.bShowCleanseIcon or tNameplate.bInCluster or tNameplate.bIsTarget) and self.refreshHealthShieldBar) then 
		bShowBars = true
		-- Hit Points Text (hide if 0)
		if nHealthCurrent == nil or nHealthCurrent <= 0 then 
			if tNameplate.wnd.HealthBarBack:IsShown() then
				tNameplate.wnd.HealthBarBack:Show(false)
				tNameplate.bNeedsRedraw = true				
			end
			self:HelperUpdateHPPixie(tNameplate, false)
			hasHealth = false
			tNameplate.bShowHealthBar = false
			nHealthCurrent = 0
		else
			if not tNameplate.wnd.HealthBarBack:IsShown() then
				tNameplate.wnd.HealthBarBack:Show(true)
				tNameplate.bNeedsRedraw = true
			end
			tNameplate.bShowHealthBar = true
			
			nHealthPercent = (nHealthCurrent / nHealthMax) * 100

			local tmpText = nil
			if self:HelperGetUnitSetting(tNameplate, "HPText", bInCombat) then
				if self.setHPTextStyle == "PercentageDeficit"  then
					tmpText = tostring(100 - math.floor(nHealthPercent)).."%"
					if nHealthPercent < 100 then
						tmpText = "-"..tmpText
					end
				elseif self.setHPTextStyle == "NormalDeficit" then
					tmpText = self:HelperFormatBigNumber(unitOwner:GetMaxHealth() - nHealthCurrent)
					if nHealthPercent < 100 then
						tmpText = "-"..tmpText
					end
				elseif self.setHPTextStyle == "Percentage" then
					tmpText = tostring(math.floor(nHealthPercent)).."%"
				else
					tmpText = self:HelperFormatBigNumber(nHealthCurrent)
				end

				if tmpText ~= nil then
					if self.setStyleFlatNumberInside and self.bStyleFlat then
						tNameplate.wnd.HealthBar:SetText(tmpText)
						self:HelperUpdateHPPixie(tNameplate, false)
					else
						self:HelperUpdateHPPixie(tNameplate, true, tmpText, true)
						tNameplate.wnd.HealthBar:SetText("")
		
					end
				end
			else
				self:HelperUpdateHPPixie(tNameplate, false)
				tNameplate.wnd.HealthBar:SetText("")
			end
		end
		tNameplate.nHealthPercent = nHealthPercent
		-- Shield Points Text (hide if 0)
		if nShieldCurrent == nil or nShieldMax == nil or nShieldMax <= 0 then 
			if tNameplate.wnd.ShieldBar:IsShown() then
				tNameplate.wnd.ShieldBar:Show(false)
				tNameplate.bNeedsRedraw = true
			end
			self:HelperUpdateSPPixie(tNameplate, false)
			hasShield = false
			tNameplate.bShowShieldBar = false

			nShieldCurrent = 0
		else
			if not tNameplate.wnd.ShieldBar:IsShown() then
				tNameplate.wnd.ShieldBar:Show(true)
				tNameplate.bNeedsRedraw = true
			end
			tNameplate.bShowShieldBar = true
			nShieldPercent = (nShieldCurrent / nShieldMax) * 100
			hasShield = true
			local tmpText = nil
			if self:HelperGetUnitSetting(tNameplate, "SPText", bInCombat) then
				if self.setSPTextStyle == "PercentageDeficit" then
					tmpText = tostring(100 - math.floor(nShieldPercent)).."%"
					if nShieldPercent < 100 then
						tmpText = "-"..tmpText
					end
				elseif self.setSPTextStyle == "NormalDeficit" then
					tmpText = self:HelperFormatBigNumber(unitOwner:GetShieldCapacityMax() - nShieldCurrent)
					if nShieldPercent < 100 then
						tmpText = "-"..tmpText
					end
				elseif self.setSPTextStyle == "Percentage" then
					tmpText = tostring(math.floor(nShieldPercent)).."%"
				else
					tmpText = self:HelperFormatBigNumber(nShieldCurrent)
				end

				if tmpText ~= nil then
					if self.setStyleFlatNumberInside and self.bStyleFlat then
						if self.setFlatShieldHeight >= 9 then
							tNameplate.wnd.ShieldBar:SetText(tmpText)
						else
							tNameplate.wnd.HealthBar:SetText(tNameplate.wnd.HealthBar:GetText().." ("..tmpText..")")
							tNameplate.wnd.ShieldBar:SetText("")
						end
						self:HelperUpdateSPPixie(tNameplate, false)
					else
						tNameplate.wnd.ShieldBar:SetText("")
						self:HelperUpdateSPPixie(tNameplate, true, tmpText, true)
					end	
				end
			else 
				self:HelperUpdateSPPixie(tNameplate, false)
				tNameplate.wnd.ShieldBar:SetText("")
			end			
		end
		
		-- Fill Bars
		if tNameplate.nHealthMax ~= nHealthMax or tNameplate.bNeedsRedraw then 
			tNameplate.wnd.HealthBar:SetMax(nHealthMax)
			tNameplate.wnd.HealthBarBack:SetMax(nHealthMax)
			tNameplate.wnd.AbsorbBar:SetMax(nHealthMax)
			tNameplate.wnd.HealthBarBack:SetProgress(nHealthCurrent, nHealthMax)			
		end
		if tNameplate.nShieldMax ~= nShieldMax or tNameplate.bNeedsRedraw then
			tNameplate.wnd.ShieldBar:SetMax(nShieldMax)
		end		
		if tNameplate.nHealth ~= nHealthCurrent or tNameplate.bNeedsRedraw then
			tNameplate.wnd.HealthBar:SetProgress(nHealthCurrent)		
			tNameplate.wnd.HealthBarBack:SetProgress(nHealthCurrent, nHealthMax)			
		end
		if tNameplate.nShield ~= nShieldCurrent or tNameplate.bNeedsRedraw then
			tNameplate.wnd.ShieldBar:SetProgress(nShieldCurrent or 0)
		end		

		tNameplate.nShield = nShieldCurrent
		tNameplate.nShieldMax = nShieldMax
		tNameplate.nShieldPercent = nShieldPercent
		tNameplate.nHealth = nHealthCurrent
		tNameplate.nHealthMax = nHealthMax
		tNameplate.nHealthPercent = nHealthPercent
	end
	
	-- Absorbs don't trigger refreshhealthshieldbar events, so it is here
	if tNameplate.nAbsorb ~= nAbsorbCurrent or tNameplate.bNeedsRedraw then
		tNameplate.wnd.AbsorbBar:SetProgress(nAbsorbCurrent or 0)
	end	
	tNameplate.nAbsorb = nAbsorbCurrent	
	
	-- Disposition Color Changes (Bars)
	local tempBarColor = self.arColorObjects.NeutralBars
	if eDisposition == tDisposition.Neutral then -- NEUTRAL
		tempBarColor = nHealthPercent > self.setHealthCutoff 
				and self.arColorObjects.NeutralBar
				or  self.arColorObjects.NeutralBarCutoff
	elseif eDisposition == tDisposition.Hostile then -- HOSTILE
		if bIsPlayer then
			tempBarColor = nHealthPercent > self.setHealthCutoff 
				and self.arColorObjects.HostilePlayerBar
				or  self.arColorObjects.HostilePlayerBarCutoff
		else
			tempBarColor = nHealthPercent > self.setHealthCutoff 
				and self.arColorObjects.HostileBar
				or  self.arColorObjects.HostileBarCutoff
		end
	elseif eDisposition == tDisposition.Friendly then -- FRIENDLY
		if bIsPlayer then
			tempBarColor = nHealthPercent > self.setHealthCutoff 
				and self.arColorObjects.FriendlyPlayerBar
				or  self.arColorObjects.FriendlyPlayerBarCutoff
		else
			tempBarColor = nHealthPercent > self.setHealthCutoff 
				and self.arColorObjects.FriendlyBar
				or  self.arColorObjects.FriendlyBarCutoff
		end
	end
	
	-- Class Colors
	if bIsPlayer and (
		(eDisposition == tDisposition.Hostile and self.setShowClassColors) or 
		(eDisposition == tDisposition.Friendly and self.setShowAllyClassColors)
		) and
		nHealthPercent > self.setHealthCutoff
	then
		local unitOwnerClassId = unitOwner:GetClassId()
		if 	   unitOwnerClassId == 1 then tempBarColor = self.arColorObjects.WarriorBar 
		elseif unitOwnerClassId == 5 then tempBarColor = self.arColorObjects.StalkerBar 
		elseif unitOwnerClassId == 2 then tempBarColor = self.arColorObjects.EngineerBar 
		
		elseif unitOwnerClassId == 3 then tempBarColor = self.arColorObjects.EsperBar 
		elseif unitOwnerClassId == 7 then tempBarColor = self.arColorObjects.SpellslingerBar 
		elseif unitOwnerClassId == 4 then tempBarColor = self.arColorObjects.MedicBar 
		end
	end

	if nVulnerable ~= 0 then tempBarColor = self.arColorObjects.VulnerableBar end
	if self.setStyleCleanseOutlineToHealthBar and (tNameplate.bShowCleanseIcon or tNameplate.bShowThreatIcon) then
		if tNameplate.bShowCleanseIcon then tempBarColor = self.arColorObjects.CleanseOutline
		elseif tNameplate.bShowThreatIcon then tempBarColor = self.arColorObjects.ThreatOutline end
	end
	tNameplate.wnd.HealthBar:SetBarColor(tempBarColor)
	if bBarsChanged or tNameplate.bNeedsRedraw then
		if nShieldPercent < self.setHealthCutoff and (eDisposition == tDisposition.Friendly or unitOwner == self.playerUnit)  then
			tNameplate.wnd.ShieldBar:SetBarColor(self.arColorObjects.FriendlyShieldBarCutoff)
		else
			tNameplate.wnd.ShieldBar:SetBarColor(self.arColorObjects.ShieldBar)
		end
	end
	if self.setStyleCastbarShowMoO == false then
		if nVulnerable == nil or nVulnerable == 0 then
			if vulnerableBar:IsVisible() then
				vulnerableBar:Show(false)
			end
		elseif nVulnerable == 0 and nVulnerable ~= tNameplate.nVulnerableTime then
			tNameplate.nVulnerableTime = 0 -- casting done, set back to 0
			if vulnerableBar:IsVisible() then
				vulnerableBar:Show(false)
			end
		elseif nVulnerable ~= 0 and nVulnerable > tNameplate.nVulnerableTime then
			tNameplate.nVulnerableTime = nVulnerable
			vulnerableBar:SetBarColor(self.arColorObjects.VulnerableTimerBar)
			if not vulnerableBar:IsVisible() then
				vulnerableBar:Show(true)
			end			
		elseif nVulnerable ~= 0 and nVulnerable < tNameplate.nVulnerableTime then
			vulnerableBar:SetMax(tNameplate.nVulnerableTime)
			vulnerableBar:SetProgress(nVulnerable)
		end
	end
	local bHideFriendlyFull = (self.setHideFullFriendlyBars == true and eDisposition == tDisposition.Friendly and tNameplate.bIsTarget == false and unitOwner:IsThePlayer() == false)
	if bShowBars and bHideFriendlyFull then
		bHideFriendlyFull = false -- will get set back to true if thresholds are met
		if self.setHideFullFriendlyBarsOneThreshold then
			if (tNameplate.nHealthPercent ~= nil and tNameplate.nHealthPercent > self.setHideFullFriendlyBarsHealthThreshold) or
			   (tNameplate.nShieldPercent ~= nil and tNameplate.nShieldPercent > self.setHideFullFriendlyBarsShieldThreshold)
			then
				bHideFriendlyFull = true
			end
		else
			if (tNameplate.nHealthPercent == nil or tNameplate.nHealthPercent > self.setHideFullFriendlyBarsHealthThreshold) and
			   (tNameplate.nShieldPercent == nil or tNameplate.nShieldPercent > self.setHideFullFriendlyBarsShieldThreshold)
			then
				bHideFriendlyFull = true
			end			
		end
	end	
	local bHideMineFull = (unitOwner:IsThePlayer() and self.setHideFullFriendlyBarsMine)
	if bHideMineFull then
		bHideMineFull = false
		if self.setHideFullFriendlyBarsMineOneThreshold then
			if (tNameplate.nHealthPercent ~= nil and tNameplate.nHealthPercent > self.setHideFullFriendlyBarsMineHealthThreshold) or
			   (tNameplate.nShieldPercent ~= nil and tNameplate.nShieldPercent > self.setHideFullFriendlyBarsMineShieldThreshold)
			then
				bHideMineFull = true
			end
		else
			if (tNameplate.nHealthPercent == nil or tNameplate.nHealthPercent > self.setHideFullFriendlyBarsMineHealthThreshold) and
			   (tNameplate.nShieldPercent == nil or tNameplate.nShieldPercent > self.setHideFullFriendlyBarsMineShieldThreshold)
			then
				bHideMineFull = true
			end			
		end				
	end
	
	local bHideFull = ((bHideFriendlyFull or 
					  bHideMineFull) and
					 not (tNameplate.bMouseOverSettingsEnabled and ((eDisposition == tDisposition.Hostile and self.setMouseOverSettings_H_B) or self.setMouseOverSettings_F_B))
					 )
	if bShowBars and bHideFull then
		bShowBars = false
	end

	local wndPlates = tNameplate.wnd.Plates
	if nHealthMax == 0 then
		bShowBars = false
	end
	if bShowBars ~= nil and wndPlates:IsShown() ~= bShowBars then
		wndPlates:Show(bShowBars)
		tNameplate.bNeedsRedraw = true
	end
end

function OptiPlates:OnBuffAdded(unitCaster, tBuff)
	self.arBuffCasters[tBuff.idBuff] = unitCaster
end

function OptiPlates:OnBuffRemoved(unitTarget, tBuff)
	self.arBuffCasters[tBuff.idBuff] = nil
end

-- called every frame when buffs should update
function OptiPlates:DrawNameplateBuffs(tNameplate)
	local sUnitSettingLetter =  tNameplateMatrixCategoryToLetter[tNameplate.sUnitType]
	local unitOwner = tNameplate.unitOwner
	local wndDebuffs = tNameplate.wnd.Debuffs
	local namePlate = tNameplate.wndNameplate
	local unitPlayer = self.playerUnit
	local eDisposition = tNameplate.eDisposition	

	-- Buff Info
	---------------------------------------------------
	local myBuffs = nil
	local debuffs = nil 
	local buffs   = nil
	
	-- Cleanse Icon: colored borders/icon for cleansing
	---------------------------------------------------
	local bShowCleanseIcon = false
	if self.setShowCleanseAlert and (eDisposition == tDisposition.Friendly or unitOwner == self.playerUnit) then
		if myBuffs == nil then
			myBuffs = unitOwner:GetBuffs()
			buffs = myBuffs["arBeneficial"]
			debuffs = myBuffs["arHarmful"]
		end
		for key, value in pairs(debuffs) do
			if value["splEffect"]:GetClass() == Spell.CodeEnumSpellClass.DebuffDispellable then --dispellable
				bShowCleanseIcon = true
				break
			end
		end
	end
	tNameplate.bShowCleanseIcon = bShowCleanseIcon

	local bBuffNumChanged = false	
	local bDebuffNumChanged = false
	
	-- reloads buff settings and forces a redrawing of all buff windows
	if tNameplate.bBuffProfileDirty then
		local profileName = self.setBuffSelectedProfiles[sUnitSettingLetter]
		tNameplate.tBuffProfile = self.setProfiles["Buffprofiles"][profileName]
		for idBuff,tBuff in pairs(tNameplate.tBuffWindowRefs.tInUse) do 
			if tBuff == false then tNameplate.tBuffWindowRefs.tInUse[idBuff] = nil 
			elseif tBuff then
				tBuff.wndRef:Show(false)
				tBuff.wndRef:SetText("")
				table.insert(tNameplate.tBuffWindowRefs.tFreeWindows,tBuff.wndRef)
				tBuff.wndRef = nil
				tNameplate.tBuffWindowRefs.tInUse[idBuff] = nil
			end
		end
		for idBuff,bool in pairs(tNameplate.tBuffProfile.tIncludeBuffs) do
			tNameplate.tBuffProfile.tIncludeBuffs[idBuff] = nil
		end
		for idBuff,bool in pairs(tNameplate.tBuffProfile.tExcludeBuffs) do
			tNameplate.tBuffProfile.tExcludeBuffs[idBuff] = nil
		end		
		for idBuff,tBuff in pairs(tNameplate.tDebuffWindowRefs.tInUse) do 
			if tBuff == false then tNameplate.tDebuffWindowRefs.tInUse[idBuff] = nil 
			elseif tBuff then
				tBuff.wndRef:Show(false)
				tBuff.wndRef:SetText("")
				table.insert(tNameplate.tDebuffWindowRefs.tFreeWindows,tBuff.wndRef)
				tBuff.wndRef = nil
				tNameplate.tDebuffWindowRefs.tInUse[idBuff] = nil
			end
		end
		for idBuff,bool in pairs(tNameplate.tBuffProfile.tIncludeDebuffs) do
			tNameplate.tBuffProfile.tIncludeDebuffs[idBuff] = nil
		end		
		for idBuff,bool in pairs(tNameplate.tBuffProfile.tExcludeDebuffs) do
			tNameplate.tBuffProfile.tExcludeDebuffs[idBuff] = nil
		end				
		tNameplate.tBuffWindowRefs.numBuffs = 0
		tNameplate.tDebuffWindowRefs.numBuffs = 0
		tNameplate.bBuffProfileDirty = false
		bBuffNumChanged = true
		bDebuffNumChanged = true
	end

	local tSpellTypeEnum = Spell.CodeEnumSpellClass
	local tBuffOpts = tNameplate.tBuffProfile
	local bMouseOverSettingsEnabled = (tNameplate.bMouseOverSettingsEnabled and 
						((eDisposition == tDisposition.Hostile and self.setMouseOverSettings_H_D) or self.setMouseOverSettings_F_D)
						)
	local buffsEnabled = self:HelperGetUnitSetting(tNameplate, "Buffs", unitOwner:IsInCombat()) or bMouseOverSettingsEnabled
						
	if tBuffOpts == nil then return end
	
	-- Update buff windows
	if (not tBuffOpts.bExcludeAllBuffs) and buffsEnabled then
		if myBuffs == nil then
			myBuffs = unitOwner:GetBuffs()
			buffs = myBuffs["arBeneficial"]
			debuffs = myBuffs["arHarmful"]
		end
		local buffWndRefs = tNameplate.tBuffWindowRefs.tInUse
		local newBuffs = 0
		local existingBuffs = 0
		local prevExistingBuffs = tNameplate.tBuffWindowRefs.numBuffs
		local maxBuffs = tBuffOpts.nBuffRows*tBuffOpts.nBuffsPerRow

		local tNewBuffs = {}
		-- Iterate through buffs and update time or mark them as new
		for k,v in pairs(buffs) do
			local buffId = v.idBuff
			local buffWndRef = buffWndRefs[buffId]
			if buffWndRef ~= false and buffWndRef ~= nil then
				-- existing buff 
				buffWndRef.valid = true
				buffWndRef.bNew = false
				if v.fTimeRemaining > 0 then
					buffWndRef.wndRef:SetText(self:HelperFormatTime(v.fTimeRemaining))
				end
				if v.nCount > 1 then
					buffWndRef.wndRef:FindChild("DebuffCount"):SetText(v.nCount)
				end
				existingBuffs = existingBuffs + 1
			elseif buffWndRef ~= false then
				-- new buff
				local buffClass = v.splEffect:GetClass()
				local buffName = v.splEffect:GetName()
				
				if tBuffOpts.tExcludeBuffs[buffId] == true then
					buffWndRefs[buffId] = false
				elseif tBuffOpts.bBuffOnlyShowMine and self.arBuffCasters[buffId] ~= self.playerUnit then
					buffWndRefs[buffId] = false					
				elseif tBuffOpts.tIncludeBuffs[buffId] == true then
					tNewBuffs[k] = v
				elseif tBuffOpts.tExcludeTextBuffs[buffName] == true then
					tBuffOpts.tExcludeBuffs[buffId] = true
					buffWndRefs[buffId] = false				
				elseif tBuffOpts.bIncludeAllBuffs
					or (tBuffOpts.bIncludeAllDispellableBuffs and buffClass == tSpellTypeEnum.BuffDispellable)
					or (tBuffOpts.bIncludeAllNonDispellableBuffs and (buffClass == tSpellTypeEnum.BuffNonDispellable or buffClass == Spell.CodeEnumSpellClass.BuffNonDispellableRightClickOk))
					or tBuffOpts.tIncludeTextBuffs[buffName] == true
				then
					tBuffOpts.tIncludeBuffs[buffId] = true
					tNewBuffs[k] = v
				else
					buffWndRefs[buffId] = false
				end
			end
		end
		local removedBuffs = prevExistingBuffs - existingBuffs
		-- remove expired buffs
		local tRemovedBuffs = {}
		if removedBuffs > 0 then
			for idBuff,tBuff in pairs(tNameplate.tBuffWindowRefs.tInUse) do 
				if tBuff and tBuff.valid == false then
					tRemovedBuffs[#tRemovedBuffs+1] = tBuff.numBuff
					tBuff.wndRef:Show(false)
					tBuff.wndRef:SetText("")
					table.insert(tNameplate.tBuffWindowRefs.tFreeWindows,tBuff.wndRef)
					tBuff.wndRef = nil
					tNameplate.tBuffWindowRefs.tInUse[idBuff] = nil
					tBuffOpts.tIncludeBuffs[idBuff] = nil
				end
			end
		end
		
		-- create windows for new buffs if there's room
		if existingBuffs < maxBuffs then
			local numBuffs = #buffs
			local bFullTooltipValid = false
			if tBuffOpts.bBuffShowFullTooltip and numBuffs <= 15 then
				local tmpWndRef = self.wndBuffTooltipRefs[numBuffs+1]
				tmpWndRef:SetUnit(unitOwner)
				local tooltip = tmpWndRef:GetBuffTooltip()
				if tooltip == nil or tooltip == "" then
					bFullTooltipValid = true
				end
			end
			for k,v in pairs(tNewBuffs) do
				local buffId = v.idBuff
				-- check for a cached window
				local buffWnd = table.remove(tNameplate.tBuffWindowRefs.tFreeWindows)
				if buffWnd == nil then
					-- no cached window
					buffWnd = Apollo.LoadForm(self.xmlDoc, "DebuffFrameNew",  wndDebuffs, self)
					buffWnd:SetName("BuffWnd")
				end
				buffWnd:SetSprite(v.splEffect:GetIcon())
				if v.fTimeRemaining == 0 then
					buffWnd:SetText("")
				else
					buffWnd:SetText(self:HelperFormatTime(v.fTimeRemaining))
				end

				if bFullTooltipValid and tBuffOpts.bBuffShowFullTooltip and k >= 1 and k <= 16 then
					self.wndBuffTooltipRefs[k]:SetUnit(unitOwner)
					buffWnd:SetTooltip("<P>"..v.splEffect:GetName()..":</P><P>"..self.wndBuffTooltipRefs[k]:GetBuffTooltip().."</P>")
					self.wndBuffTooltipRefs[k]:SetUnit(nil)
				elseif tBuffOpts.bBuffShowNameTooltip or tBuffOpts.bBuffShowFullTooltip then
					buffWnd:SetTooltip(v.splEffect:GetName())
				else
					buffWnd:SetTooltip("")
				end
				local stacks = v.nCount
				if stacks > 1 then
					buffWnd:FindChild("DebuffCount"):SetText(stacks)
				else
					buffWnd:FindChild("DebuffCount"):SetText("")
				end
				buffWnd:Show(true)
				newBuffs = newBuffs + 1
				prevExistingBuffs = prevExistingBuffs + 1
				buffWndRefs[buffId] = { wndRef = buffWnd, valid = true, numBuff = prevExistingBuffs, bNew = true }			
				
				if newBuffs+existingBuffs >= maxBuffs then break end
			end
		end
		local totalBuffs = newBuffs + existingBuffs
		tNameplate.tBuffWindowRefs.numBuffs = totalBuffs
		if removedBuffs > 0 or newBuffs > 0 then bBuffNumChanged = true end

		-- the following loop processes seperate buff windows and repositions
		-- based on the results of the previous loops
		for idBuff,tBuff in pairs(tNameplate.tBuffWindowRefs.tInUse) do 
			if tBuff and tBuff.valid then
				tBuff.valid = false
				if removedBuffs > 0 or newBuffs > 0 then
					local row = 0
					local nBuffMove = 0
					for idx,num in pairs(tRemovedBuffs) do
						if num < tBuff.numBuff then nBuffMove = nBuffMove + 1 end
					end
					if nBuffMove > 0 or tBuff.bNew then
						tBuff.numBuff = tBuff.numBuff - nBuffMove
						local numBuff = tBuff.numBuff
						while numBuff > tBuffOpts.nBuffsPerRow do
							numBuff = numBuff - tBuffOpts.nBuffsPerRow
							row = row + 1
						end
						numBuff = numBuff - 1
						local nLeft = numBuff*tBuffOpts.nBuffIconSize
						local nTop = row*tBuffOpts.nBuffIconSize
						
						if row < tBuffOpts.nBuffRows then
							local nBottom = nTop+tBuffOpts.nBuffIconSize
							tBuff.wndRef:SetAnchorOffsets(nLeft,nTop,nLeft+tBuffOpts.nBuffIconSize,nBottom)
							tBuff.wndRef:Show(true)
						else
							tBuff.wndRef:Show(false)
						end
					end
				end
			end
		end
	end
	
	-- Basically a copypaste of the previous block for buffs, except this time for debuffs
	if (not tBuffOpts.bExcludeAllDebuffs) and buffsEnabled then
		if myBuffs == nil then
			myBuffs = unitOwner:GetBuffs()
			buffs = myBuffs["arBeneficial"]
			debuffs = myBuffs["arHarmful"]
		end
		local buffWndRefs = tNameplate.tDebuffWindowRefs.tInUse
		local newBuffs = 0
		local existingBuffs = 0
		local prevExistingBuffs = tNameplate.tDebuffWindowRefs.numBuffs
		local maxBuffs = tBuffOpts.nDebuffRows*tBuffOpts.nDebuffsPerRow
		-- the following loop iterates through buffs and marks them as expired
		-- or creates new objects for them
		local tNewDebuffs = {}
		for k,v in pairs(debuffs) do
			local buffId = v.idBuff
			local buffWndRef = buffWndRefs[buffId]
			if buffWndRef ~= false and buffWndRef ~= nil then
				-- existing buff
				buffWndRef.valid = true
				buffWndRef.bNew = false
				if v.fTimeRemaining > 0 then
					buffWndRef.wndRef:SetText(self:HelperFormatTime(v.fTimeRemaining))
				end
				if v.nCount > 1 then
					buffWndRef.wndRef:FindChild("DebuffCount"):SetText(v.nCount)
				end
				existingBuffs = existingBuffs + 1
			elseif buffWndRef ~= false then
				-- new buff
				local buffClass = v.splEffect:GetClass()
				local buffName = v.splEffect:GetName()
				
				if tBuffOpts.tExcludeDebuffs[buffId] == true then
					buffWndRefs[buffId] = false					
				elseif tBuffOpts.bDebuffOnlyShowMine and self.arBuffCasters[buffId] ~= self.playerUnit then
					buffWndRefs[buffId] = false					
				elseif tBuffOpts.tIncludeDebuffs[buffId] == true then
					tNewDebuffs[k] = v
				elseif tBuffOpts.tExcludeTextDebuffs[buffName] == true then
					tBuffOpts.tExcludeDebuffs[buffId] = true
					buffWndRefs[buffId] = false									
				elseif tBuffOpts.bIncludeAllDebuffs
					or (tBuffOpts.bIncludeAllDispellableDebuffs and buffClass == tSpellTypeEnum.DebuffDispellable)
					or (tBuffOpts.bIncludeAllNonDispellableDebuffs and buffClass == tSpellTypeEnum.DebuffNonDispellable)
					or tBuffOpts.tIncludeTextDebuffs[buffName] == true
				then
					tBuffOpts.tIncludeDebuffs[buffId] = true
					tNewDebuffs[k] = v
				else
					buffWndRefs[buffId] = false
				end
			end
			
		end
		
		local removedBuffs = prevExistingBuffs - existingBuffs
		local tRemovedBuffs = {}
		-- remove expired buffs
		if removedBuffs > 0 then
			for idBuff,tBuff in pairs(tNameplate.tDebuffWindowRefs.tInUse) do 
				if tBuff and tBuff.valid == false then
					tRemovedBuffs[#tRemovedBuffs+1] = tBuff.numBuff
					tBuff.wndRef:Show(false)
					table.insert(tNameplate.tDebuffWindowRefs.tFreeWindows,tBuff.wndRef)
					tBuff.wndRef = nil
					tNameplate.tDebuffWindowRefs.tInUse[idBuff] = nil
					tBuffOpts.tIncludeDebuffs[idBuff] = nil
				end
			end
		end		
		
		if existingBuffs < maxBuffs then
			local numBuffs = #debuffs
			local bFullTooltipValid = false
			if tBuffOpts.bDebuffShowFullTooltip and numBuffs <= 15 then
				local tmpWndRef = self.wndDebuffTooltipRefs[numBuffs+1]
				tmpWndRef:SetUnit(unitOwner)
				local tooltip = tmpWndRef:GetBuffTooltip()
				if tooltip == nil or tooltip == "" then
					bFullTooltipValid = true
				end
			end		
			for k,v in pairs(tNewDebuffs) do
				local buffId = v.idBuff
				-- check for a cached window
				local buffWnd = table.remove(tNameplate.tDebuffWindowRefs.tFreeWindows)
				if buffWnd == nil then
					-- no cached window
					buffWnd = Apollo.LoadForm(self.xmlDoc, "DebuffFrameNew",  wndDebuffs, self)
					buffWnd:SetName("DebuffWnd")
				end
				buffWnd:SetSprite(v.splEffect:GetIcon())
				if v.fTimeRemaining == 0 then
					buffWnd:SetText("")
				else
					buffWnd:SetText(self:HelperFormatTime(v.fTimeRemaining))
				end

				if bFullTooltipValid and tBuffOpts.bDebuffShowFullTooltip and k >= 1 and k <= 16 then
					self.wndDebuffTooltipRefs[k]:SetUnit(unitOwner)
					buffWnd:SetTooltip("<P>"..v.splEffect:GetName()..":</P><P>"..self.wndDebuffTooltipRefs[k]:GetBuffTooltip().."</P>")
					self.wndDebuffTooltipRefs[k]:SetUnit(nil)
				elseif tBuffOpts.bDebuffShowNameTooltip or tBuffOpts.bDebuffShowFullTooltip then
					buffWnd:SetTooltip(v.splEffect:GetName())
				else
					buffWnd:SetTooltip("")
				end
				local stacks = v.nCount
				if stacks > 1 then
					buffWnd:FindChild("DebuffCount"):SetText(stacks)
				else
					buffWnd:FindChild("DebuffCount"):SetText("")
				end
				buffWnd:Show(true)
				newBuffs = newBuffs + 1
				prevExistingBuffs = prevExistingBuffs + 1
				buffWndRefs[buffId] = { wndRef = buffWnd, valid = true, numBuff = prevExistingBuffs, bNew = true }		
				if newBuffs+existingBuffs >= maxBuffs then break end
			end
		end
		local totalBuffs = newBuffs + existingBuffs
		tNameplate.tDebuffWindowRefs.numBuffs = totalBuffs
		if removedBuffs > 0 or newBuffs > 0 then bDebuffNumChanged = true end

		-- the following loop processes seperate debuff windows and repositions
		-- based on the results of the previous loops
		local buffWndBuffs = tNameplate.tBuffWindowRefs.numBuffs
		local buffWndHeight = 0
		if buffWndBuffs > 0 then
			buffWndHeight = math.ceil(buffWndBuffs/tBuffOpts.nBuffsPerRow)*tBuffOpts.nBuffIconSize
		end
		
		for idBuff,tBuff in pairs(tNameplate.tDebuffWindowRefs.tInUse) do 
			if tBuff and tBuff.valid then
				tBuff.valid = false
				if removedBuffs > 0 or newBuffs > 0 then
					local row = 0
					local nBuffMove = 0
					for idx,num in pairs(tRemovedBuffs) do
						if num < tBuff.numBuff then nBuffMove = nBuffMove + 1 end
					end
					if nBuffMove > 0 or tBuff.bNew then
						tBuff.numBuff = tBuff.numBuff - nBuffMove
						local numBuff = tBuff.numBuff
						while numBuff > tBuffOpts.nDebuffsPerRow do
							numBuff = numBuff - tBuffOpts.nDebuffsPerRow
							row = row + 1
						end
						numBuff = numBuff - 1
						local nLeft = numBuff*tBuffOpts.nDebuffIconSize
						local nTop = buffWndHeight+row*tBuffOpts.nDebuffIconSize
						if row < tBuffOpts.nDebuffRows then
							tBuff.wndRef:SetAnchorOffsets(nLeft,nTop,nLeft+tBuffOpts.nDebuffIconSize,nTop+tBuffOpts.nDebuffIconSize)
							tBuff.wndRef:Show(true)
						else
							tBuff.wndRef:Show(false)
						end
					end
				end
			end
		end
	end	
	
	-- Reposition buff/debuff windows if the count of either changed
	if bBuffNumChanged or bDebuffNumChanged or bMouseOverSettingsEnabled then
		local numBuffs = tNameplate.tBuffWindowRefs.numBuffs
		local numDebuffs = tNameplate.tDebuffWindowRefs.numBuffs
		local nWidth = 0
		local nHeight = 0
		if numBuffs > 0 and numBuffs <= tBuffOpts.nBuffsPerRow then
			nWidth = numBuffs*tBuffOpts.nBuffIconSize
			nHeight = tBuffOpts.nBuffIconSize
		elseif numBuffs > 0 then
			nWidth = tBuffOpts.nBuffsPerRow*tBuffOpts.nBuffIconSize
			nRows = math.ceil(numBuffs/tBuffOpts.nBuffsPerRow)
			nHeight = (nRows <= tBuffOpts.nBuffRows and nRows or tBuffOpts.nBuffRows)*tBuffOpts.nBuffIconSize
		end
		local nTempWidth = 0
		local nTempHeight = 0
		if numDebuffs > 0 and numDebuffs <= tBuffOpts.nDebuffsPerRow then
			nTempWidth = numDebuffs*tBuffOpts.nDebuffIconSize
			nTempHeight = tBuffOpts.nDebuffIconSize
		elseif numDebuffs > 0 then
			nTempWidth = tBuffOpts.nDebuffsPerRow*tBuffOpts.nDebuffIconSize
			nRows = math.ceil(numDebuffs/tBuffOpts.nDebuffsPerRow)
			nTempHeight = (nRows <= tBuffOpts.nDebuffRows and nRows or tBuffOpts.nDebuffRows)*tBuffOpts.nDebuffIconSize
		end
		if nTempWidth > nWidth then nWidth = nTempWidth end
		nHeight = nHeight + nTempHeight
		if nWidth > 0 and nHeight > 0 then
			local nHalfWidth = math.ceil(nWidth*0.5)
			wndDebuffs:SetAnchorOffsets(-nHalfWidth,0,nHalfWidth,nHeight)
			wndDebuffs:Show(true)
			tNameplate.bNeedsRedraw = true
		else
			wndDebuffs:Show(false)
			tNameplate.bNeedsRedraw = true
		end
	end
	
	if not buffsEnabled then 
		if wndDebuffs:IsShown() then
			wndDebuffs:Show(false)
			tNameplate.bNeedsRedraw = true
		end
	end
end

function OptiPlates:DrawNameplateCastbar(tNameplate)
	local unitOwner = tNameplate.unitOwner
	local bInCombat = tNameplate.unitOwner:IsInCombat()
	local cast = tNameplate.wnd.Cast
	local castbar = tNameplate.wnd.CastBar
	local castbarLabel = tNameplate.wnd.CastBarLabel
	if self.setCastbarTextStyle == "Inside" then
		castbarLabel = castbar
	end
	local nVulnerable = (unitOwner:GetCCStateTimeRemaining(Unit.CodeEnumCCState.Vulnerability) or 0)	
	local nSubdue = 0
	if tNameplate.nCCState >= 0 and self.setStyleCastbarShowSubdue then
		nSubdue = (unitOwner:GetCCStateTimeRemaining(tNameplate.nCCState) or 0)	
	end
	if self.setStyleCastbarShowMoO and nVulnerable ~= 0 then
		local vulnerableBar = castbar
		local tempBarColor = self.arColorObjects.VulnerableBar
		if nVulnerable ~= 0 and nVulnerable > tNameplate.nVulnerableTime then
			tNameplate.nVulnerableTime = nVulnerable
			vulnerableBar:SetMax(tNameplate.nVulnerableTime)
			vulnerableBar:SetProgress(nVulnerable)
			vulnerableBar:SetBarColor(self.arColorObjects.VulnerableTimerBar)
			if not vulnerableBar:IsVisible() then
				vulnerableBar:Show(true)
			end			
		elseif nVulnerable ~= 0 and nVulnerable < tNameplate.nVulnerableTime then
			vulnerableBar:SetProgress(nVulnerable)
			if not self:HelperGetUnitSetting(tNameplate, "CastText", bInCombat) then
				castbarLabel:SetText("")
				if castbarLabel:IsShown() then
					castbarLabel:Show(false)
				end					
			else
				castbarLabel:SetText(string.format("MoO! (%.2fs)",nVulnerable))
				if not castbarLabel:IsShown() then
					castbarLabel:Show(true)
				end					
			end
		end
	elseif self.setStyleCastbarShowSubdue and nSubdue > 0 then
		local vulnerableBar = castbar
		local tempBarColor = self.arColorObjects.VulnerableBar
		if nSubdue ~= 0 and nSubdue > tNameplate.nSubdueTime then
			tNameplate.nSubdueTime = nSubdue
			vulnerableBar:SetMax(tNameplate.nSubdueTime)
			vulnerableBar:SetProgress(nSubdue)
			vulnerableBar:SetBarColor(self.arColorObjects.VulnerableTimerBar)
			if not vulnerableBar:IsVisible() then
				vulnerableBar:Show(true)
			end			
		elseif nSubdue ~= 0 and nSubdue < tNameplate.nSubdueTime then
			vulnerableBar:SetProgress(nSubdue)
			local bShowText = self:HelperGetUnitSetting(tNameplate, "CastText", bInCombat)
			if not bShowText then
				castbarLabel:SetText("")
				if castbarLabel:IsShown() then
					castbarLabel:Show(false)
				end	
			else
				if not castbarLabel:IsShown() then
					castbarLabel:Show(true)
				end	
				local castName = string.format("%s! (%.2fs)",tNameplate.sCCStateText,nSubdue)
				castbarLabel:SetText(castName)
				if self.setStyleCastbarTextAutoFit then
					local nHalfWidth = math.floor(Apollo.GetTextWidth(self.setStyleCastbarFont.name, castName .. "   ") * 0.5)
					local nHalfStyleWidth = self.nStyleHalfWidth
					if nHalfWidth < (nHalfStyleWidth+self.setStyleCastbarWidth) then nHalfWidth = nHalfStyleWidth+self.setStyleCastbarWidth end
					local tOffsets = castbarLabel:GetLocation():ToTable().nOffsets
					castbarLabel:SetAnchorOffsets(-nHalfWidth,tOffsets[2],nHalfWidth,tOffsets[4])
				end
			end			
		end
	elseif tNameplate.bShowCastBar and unitOwner:ShouldShowCastBar() then
		local castElapsed = unitOwner:GetCastElapsed()
		local castName = unitOwner:GetCastName()
		local castDuration = unitOwner:GetCastDuration()			
		local bShowText = self:HelperGetUnitSetting(tNameplate, "CastText", bInCombat)
		if not castbar:IsShown() then
			castbar:Show(true)
		end
		if not castbarLabel:IsShown() then
			castbarLabel:Show(true)
		end			

		castbar:SetProgress(castElapsed)
		local bShowTimeRemaining = self.setCastbarTimeStyle == "Always" or (tNameplate.eDisposition == tDisposition.Hostile and self.setCastbarTimeStyle == "Hostile")
		if not bShowText then
			castName = ""
		elseif bShowTimeRemaining then
			castName = string.format("%s %.1fs",castName,(castDuration-castElapsed)/1000)
		end
		local nPrevCastElapsed = tNameplate.nPrevCastElapsed
		local bNewCast = nPrevCastElapsed <= 0 or nPrevCastElapsed > castElapsed
		if bNewCast or bShowTimeRemaining then
			if bNewCast then
				if tNameplate.eDisposition == tDisposition.Hostile then
					castbar:SetBarColor(self.arColorObjects.HostileCastBar)
				else
					castbar:SetBarColor(self.arColorObjects.FriendlyCastBar)
				end
				castbar:SetMax(castDuration)
			end
			castbarLabel:SetText(castName)
			if self.setStyleCastbarTextAutoFit and bShowText then
				local nHalfWidth = math.floor(Apollo.GetTextWidth(self.setStyleCastbarFont.name, castName .. "   ") * 0.5)
				local nHalfStyleWidth = self.nStyleHalfWidth
				if nHalfWidth < (nHalfStyleWidth+self.setStyleCastbarWidth) then nHalfWidth = nHalfStyleWidth+self.setStyleCastbarWidth end
				local tOffsets = castbarLabel:GetLocation():ToTable().nOffsets
				castbarLabel:SetAnchorOffsets(-nHalfWidth,tOffsets[2],nHalfWidth,tOffsets[4])
			end
			if not cast:IsShown() then 
				cast:Show(true)
				tNameplate.bNeedsRedraw = true
			end
		end
		tNameplate.nPrevCastElapsed = castElapsed
		if not tNameplate.bCastbarArranged then
			tNameplate.bNeedsRedraw = true
			tNameplate.bCastbarArranged = true
		end		
	else
		if castbar:IsShown() then
			castbar:SetProgress(0)
			castbar:Show(false)
		end
		if castbarLabel:IsShown() then
			castbarLabel:Show(false)
		end	
		tNameplate.bNeedsRedraw = true
		tNameplate.bCastbarArranged = false
		tNameplate.nPrevCastElapsed = -1
		tNameplate.nSubdueTime = -1
		tNameplate.nVulnerableTime = -1
	end	
end

function OptiPlates:DrawNameplateThreatBar(tNameplate)
	local shouldShowIcon = false
	local unitOwner = tNameplate.unitOwner
	local unitPlayer = self.playerUnit
	local eDisposition = tNameplate.eDisposition	
	local notTargeting = tNameplate.wnd.NotTargeting

	if (eDisposition == tDisposition.Neutral or eDisposition == tDisposition.Hostile) and unitOwner:IsInCombat() then	
		if self.setShowThreatLossAlert 
				and unitOwner:GetTarget() ~= nil 
				and unitOwner:GetTarget():IsThePlayer() == false 
				and tNameplate.bIsNPC then
			shouldShowIcon = true
		elseif self.setShowThreatAlert
				and unitOwner:GetTarget() ~= nil 
				and unitOwner:GetTarget():IsThePlayer() == true
				and tNameplate.bIsNPC then		
			shouldShowIcon = true
		end
		local bIcon = (shouldShowIcon and self.setStyleShowCleanseIcon)
		local bOutline = (shouldShowIcon and self.setStyleCleanseOutlineShow)
		tNameplate.bShowThreatIcon = shouldShowIcon
		self:HelperUpdateCleanseIconPixie(tNameplate, bIcon, self.setStyleThreatIconSprite)
		if notTargeting:IsShown() ~= bOutline then
			notTargeting:Show(bOutline)
			if bOutline then
				notTargeting:SetBGColor(self.arColorObjects.ThreatOutline)
			end
		end
	elseif eDisposition == tDisposition.Friendly then
		local bShowCleanseIcon = tNameplate.bShowCleanseIcon
		local bIcon = (bShowCleanseIcon and self.setStyleShowCleanseIcon)
		local bOutline = (bShowCleanseIcon and self.setStyleCleanseOutlineShow)		
		tNameplate.bShowThreatIcon = false
		self:HelperUpdateCleanseIconPixie(tNameplate, bIcon, self.setStyleCleanseIconSprite)
		if notTargeting:IsShown() ~= bOutline then		
			if bOutline then
				notTargeting:SetBGColor(self.arColorObjects.CleanseOutline)		
			end
			notTargeting:Show(bShowCleanseIcon)	
		end
	else
		if notTargeting:IsShown() ~= false then
			notTargeting:Show(false)
		end		
		self:HelperUpdateCleanseIconPixie(tNameplate, false)
		tNameplate.bShowThreatIcon = false
	end
end

-- This function gets called every frame for every visible nameplate
function OptiPlates:DrawNameplate(tNameplate,updateBuffsThisFrame)
	local unitOwner = tNameplate.unitOwner
	local unitPlayer = self.playerUnit
	local bInCombat = tNameplate.unitOwner:IsInCombat()
	local namePlate = tNameplate.wndNameplate
	
	local healthPlate = tNameplate.wnd.HealthBarBack
	local shieldPlate = tNameplate.wnd.ShieldBar
	
	local eDisposition = tNameplate.eDisposition
	local wndRefs = tNameplate.wnd
	

	-- If nameplate wants its bars shown, check plates
	if tNameplate.bShowBars or tNameplate.bShowCleanseIcon then
		self:DrawNameplateBars(tNameplate)
	else
		if tNameplate.wnd.Plates:IsShown() then
			tNameplate.wnd.Plates:Show(false)
			tNameplate.bNeedsRedraw = true
		end
		if healthPlate:IsShown() then
			healthPlate:Show(false)
			tNameplate.bNeedsRedraw = true
		end
		if shieldPlate:IsShown() then
			shieldPlate:Show(false)
			tNameplate.bNeedsRedraw = true
		end
		
		if tNameplate.nHPTextPixie ~= nil then
			self:HelperUpdateHPPixie(tNameplate, false)
			tNameplate.bNeedsRedraw = true
		end		
		if tNameplate.nSPTextPixie ~= nil then
			self:HelperUpdateSPPixie(tNameplate, false)
			tNameplate.bNeedsRedraw = true
		end		
	end
	
	-- Interrupt Armor
	if self:HelperGetUnitSetting(tNameplate, "InterruptArmor", bInCombat) then
		local nArmorNumber = unitOwner:GetInterruptArmorValue() or 0
		local nArmorMax = unitOwner:GetInterruptArmorMax() or 0	
		if nArmorNumber ~= tNameplate.nInterruptArmor or nArmorMax ~= tNameplate.nInterruptArmorMax then
			tNameplate.nInterruptArmor = nArmorNumber
			tNameplate.nInterruptArmorMax = nArmorMax
			tNameplate.bCleanseIconPositioned = false
			if nArmorMax > 0 then
				local bDegraded = false
				if nArmorNumber < nArmorMax then
					bDegraded = true
				end
				self:HelperUpdateInterruptArmorPixie(tNameplate, true, nArmorNumber, bDegraded)
			elseif nArmorMax == 0 then
				self:HelperUpdateInterruptArmorPixie(tNameplate, false)
			elseif nArmorMax == -1 then
				self:HelperUpdateInterruptArmorPixie(tNameplate, true, nArmorMax, false, true)
			end
		end
	else
		self:HelperUpdateInterruptArmorPixie(tNameplate, false)
		tNameplate.bCleanseIconPositioned = false
	end

	-- Cast bar.
	---------------------------------------------------
	if tNameplate.bShowCastBar then
		self:DrawNameplateCastbar(tNameplate)
	else
		local cast = tNameplate.wnd.Cast
		local castbar = tNameplate.wnd.CastBar
		
		if cast ~= nil and cast:IsShown() then
			castbar:SetProgress(0)
			cast:Show(false)
		end
	end

	if tNameplate.bShowDistance then
		local nDistance = self.squaredDistanceLookupTable[tNameplate.nSquaredDistance]
		if nDistance == nil then
			nDistance = math.floor(math.sqrt(tNameplate.nSquaredDistance))
		end
		tNameplate.wnd.Distance:SetText(string.format("%dm", nDistance))
	end
	
	-- Resource Bars (Resource/Focus/Dash)
	local classResource = wndRefs.ClassResource
	if tNameplate.bIsPlayer and (unitOwner:GetFocus() or 0) > tNameplate.nFocusMax then
		tNameplate.nFocusMax = unitOwner:GetFocus()
	end

	if tNameplate.bShowFocus or tNameplate.bShowResource or tNameplate.bShowDash then
		if tNameplate.bShowResource then
			local classResourceBar = tNameplate.wnd.ClassResourceBar
			local resID = 1
			local classID = unitOwner:GetClassId()
			if classID == GameLib.CodeEnumClass.Stalker then resID = 3 end
			if classID == GameLib.CodeEnumClass.Spellslinger then resID = 4 end
			
			local nResourceMax = unitOwner:GetMaxResource(resID) or 0
			local nResourceCurrent = math.floor(unitOwner:GetResource(resID) or 0)
			classResourceBar:SetMax(nResourceMax )
			classResourceBar:SetProgress(nResourceCurrent )
			if self:HelperGetUnitSetting(tNameplate, "ResourceText", bInCombat) then
				classResourceBar:SetText(nResourceCurrent )
			else
				classResourceBar:SetText("")
			end
		end
		if tNameplate.bShowFocus then
			local focusBar = tNameplate.wnd.FocusBar
			local nFocusMax = (unitOwner == self.playerUnit and unitOwner:GetMaxFocus() or tNameplate.nFocusMax)
			local nFocusCurrent = math.floor(unitOwner:GetFocus() or 0)

			focusBar:SetMax(nFocusMax)
			focusBar:SetProgress(nFocusCurrent)
			if self:HelperGetUnitSetting(tNameplate, "FocusText", bInCombat) then
				if self.setFocusTextStyle == "Normal" then
					focusBar:SetText(nFocusCurrent)
				elseif self.setFocusTextStyle == "Percentage" then
					focusBar:SetText(string.format("%d%%",nFocusCurrent/nFocusMax*100))
				end
			else
				focusBar:SetText("")
			end
		end
		if tNameplate.bShowDash then
			local nDashResourceIndex = 7
			local nSprintResourceIndex = 0
			local sprintBar = tNameplate.wnd.SprintBar
			local dashBar = tNameplate.wnd.DashBar
			
			local nDashMax = unitOwner:GetMaxResource(nDashResourceIndex)
			local nDashCurrent = unitOwner:GetResource(nDashResourceIndex)
			dashBar:SetMax(nDashMax)
			dashBar:SetProgress(nDashCurrent)
			
			local nSprintMax = unitOwner:GetMaxResource(nSprintResourceIndex)
			local nSprintCurrent = unitOwner:GetResource(nSprintResourceIndex)
			sprintBar:SetMax(nSprintMax)
			sprintBar:SetProgress(nSprintCurrent)
			if self:HelperGetUnitSetting(tNameplate, "SprintText", bInCombat) then
				sprintBar:SetText(string.format("%d%%", nSprintCurrent/nSprintMax*100))
				dashBar:SetText(string.format("%.2f", nDashCurrent/100))			
			else
				sprintBar:SetText("")
				dashBar:SetText("")
			end
		end
	end
	
	

	-- Update Buff Related Info. (timer)
	if updateBuffsThisFrame or tNameplate.bMouseOverSettingsEnabled then
		self:DrawNameplateBuffs(tNameplate)
	end	
	
	self:DrawNameplateThreatBar(tNameplate)	
	
	-- tNameplate.bNeedsRedraw does not need checking here, it is done before DrawNameplate call in SetNormalNameplate
	if tNameplate.bUpdateRewards then
		self:DrawNameplateRewardIcons(tNameplate)
	end
	
	if tNameplate.bNeedsRedraw then
		self:HelperCalculateNameplateVerticalOffset(tNameplate)
	end
	
	if tNameplate.wndNameplate ~= nil then
		self:DoNameplateScaling(tNameplate)
		self:DoNameplateOpacity(tNameplate)
	end
	self:DoNameplateRepositioning(tNameplate)
	
	if tNameplate.bNeedsRedraw then
		tNameplate.wnd.NamePlateContainer:ArrangeChildrenVert(2)
		if not self.setStyleCastbarAboveResourcebar and tNameplate.wnd.Cast:IsShown() and tNameplate.wnd.ClassResource:IsShown() then
			local tCastLoc = tNameplate.wnd.Cast:GetLocation():ToTable().nOffsets
			local nCastOffsetL = tCastLoc[1]
			local nCastOffsetT = tCastLoc[2]
			local nCastOffsetR = tCastLoc[3]
			local nCastOffsetB = tCastLoc[4]
			
			local tClassResourceLoc = tNameplate.wnd.ClassResource:GetLocation():ToTable().nOffsets
			local nCROffsetL = tClassResourceLoc[1]
			local nCROffsetT = tClassResourceLoc[2]
			local nCROffsetR = tClassResourceLoc[3]
			local nCROffsetB = tClassResourceLoc[4]
			local nClassResourceHeight = math.abs(nCROffsetT-nCROffsetB)
			
			tNameplate.wnd.Cast:SetAnchorOffsets(nCastOffsetL,nCastOffsetT+nClassResourceHeight,nCastOffsetR,nCastOffsetB+nClassResourceHeight)
			tNameplate.wnd.ClassResource:SetAnchorOffsets(nCROffsetL,nCastOffsetT,nCROffsetR,nCastOffsetT+nClassResourceHeight)
		end
	end	
	tNameplate.bNeedsRedraw = false
end

-- Vertically repositions the nameplate as required
function OptiPlates:DoNameplateRepositioning(tNameplate)
	if tNameplate == nil 
		or tNameplate.nOverheadAnchorY == nil 
		or tNameplate.nUnitHeight == nil 
		or tNameplate.wndNameplate == nil
	then return end
	local wnd = tNameplate.wndNameplate

	local nHalfSize = self.nStyleHalfWidth
	local nVerticalOffsetSetting = self.setNameplateVerticalOffset
	if tNameplate.unitOwner == self.playerUnit then nVerticalOffsetSetting = self.setNameplateVerticalOffsetMine end
	local nVerticalOffset = (tNameplate.nVerticalOffset - nVerticalOffsetSetting)-tNameplate.nScalingVertOffset
	local nNameplateHeight = tNameplate.nNameplateHeight
	
	-- below 80 pixels the legacy target indicator distorts or cuts off
	-- in other cases add a padding of 20 pixels per side to make things look nicer (maybe?)
	if tNameplate.bIsTarget and self.setShowLegacyTargetIndicator then 
		local tmp = 0
		if nNameplateHeight < 80  then 
			tmp = math.floor((80 - nNameplateHeight)*0.5) 
		end
		if tmp < 20 then tmp = 20 end
		nNameplateHeight = nNameplateHeight + tmp*2
		tNameplate.wnd.NamePlateContainer:SetAnchorOffsets(0,tmp,0,-tmp)
		nVerticalOffset = nVerticalOffset + tmp
		tNameplate.bLegacyHeightAdjusted = true
	elseif tNameplate.bLegacyHeightAdjusted then
		tNameplate.wnd.NamePlateContainer:SetAnchorOffsets(0,0,0,0)
		tNameplate.bLegacyHeightAdjusted = false
	end
	
	local nOffset = 0
	-- tNameplate.bDefaultHeight is updated in self:OnFrame()
	if self.setAutoAdjustNameplatesDown and tNameplate.bIsNPC and not tNameplate.bDefaultHeight and 
		(
		  (tNameplate.nUnitHeight > self.setOffscreenNameplatesMinHeight) or
		  (tNameplate.nOverheadAnchorY < self.setOffscreenNameplatesTopOffset)
		)
    then
		-- too tall unit, change anchor if the option is set
		nOffset = self.setOffscreenNameplatesAdjustAnchorOffset
		if tNameplate.nNameplateAnchor == 1 then
			tNameplate.wndNameplate:SetUnit(tNameplate.unitOwner,0)
			tNameplate.nNameplateAnchor = 0
		end
	else 
		-- reset to default
		if tNameplate.nNameplateAnchor == 0 then
			tNameplate.wndNameplate:SetUnit(tNameplate.unitOwner,1)
			tNameplate.nNameplateAnchor = 1
		end
	end
	
	nVerticalOffset = nVerticalOffset + nOffset
	if nVerticalOffset == tNameplate.nPrevVerticalOffset and not tNameplate.bNeedsRedraw then return end 
	tNameplate.nPrevVerticalOffset = nVerticalOffset
	wnd:SetAnchorOffsets(-100-nHalfSize-tNameplate.nScalingHorzOffset,-nNameplateHeight+nVerticalOffset,100+nHalfSize-tNameplate.nScalingHorzOffset,nVerticalOffset)
end

-- Scale nameplate as required
function OptiPlates:DoNameplateScaling(tNameplate)
	if tNameplate == nil or tNameplate.unitOwner == nil then return end
	local nNameplateDistanceSquared = tNameplate.nSquaredDistance
	local nPrevScale = tNameplate.nPrevScale
	local nBaseScale = self.setDefaultScaleF
	local wndRefs = tNameplate.wnd
	local unitOwner = tNameplate.unitOwner
	local unitPlayer = self.playerUnit
	local eDisposition = tNameplate.eDisposition
	if unitOwner == unitPlayer:GetTarget() then
		nBaseScale = self.setTargetDefaultScaleF
	elseif unitOwner == unitPlayer:GetAlternateTarget() then
		nBaseScale = self.setAltTargetDefaultScaleF
	end
	
	local nScale = 1
	if self.setDistanceScalingEnabled and nNameplateDistanceSquared >= 0
		and not (
			(self.setDistanceScalingNormalTargetScale and (tNameplate.bIsTarget or tNameplate.bInCluster))
			or (self.setDistanceScalingNormalObjectiveScale and tNameplate.bIsObjective)
			or (self.setHarvestablesDistanceScalingDisabled and (tNameplate.bIsHarvest or tNameplate.bIsDatacube))
		)
		then
		if nNameplateDistanceSquared <= self.setDistanceScalingMinRangeSquared then
			nScale = 1
		elseif nNameplateDistanceSquared >= self.setDistanceScalingMaxRangeSquared then
			nScale = (1 - self.setDistanceScalingScaleFactorF)
		else
			nScale = 1 - (((nNameplateDistanceSquared - self.setDistanceScalingMinRangeSquared) / (self.setDistanceScalingMaxRangeSquared - self.setDistanceScalingMinRangeSquared)) * self.setDistanceScalingScaleFactorF)
		end
	end
	if self.setHealthDeficitScalingEnabled and unitOwner:IsInCombat()
		and (not self.setHealthDeficitScalingPlayerInCombat or (self.setHealthDeficitScalingPlayerInCombat and unitPlayer:IsInCombat())) 
		and (not self.setHealthDeficitScalingOnlyFriendlyPlayers or (self.setHealthDeficitScalingOnlyFriendlyPlayers and tNameplate.bIsPlayer)) 
	then
		local nHealth = 0
		local nMaxHealth = 0
		if eDisposition == tDisposition.Friendly then
			if self.setHealthDeficitScalingUseHealth then
				nHealth = (unitOwner:GetHealth() or 0)
				nMaxHealth = (unitOwner:GetMaxHealth() or 0)
			end
			if self.setHealthDeficitScalingUseShield then
				nHealth = nHealth + (unitOwner:GetShieldCapacity() or 0)
				nMaxHealth = nMaxHealth + (unitOwner:GetShieldCapacityMax() or 0)
			end	
			if nMaxHealth > 0 then
				local nHealthPercent = nHealth/nMaxHealth
				if nHealthPercent <= self.setHealthDeficitScalingMinCutoffF then
					nScale = nScale + self.setHealthDeficitScalingMaxChangeF
				elseif nHealthPercent <= self.setHealthDeficitScalingMaxCutoffF then
					nScale = nScale + self.setHealthDeficitScalingMaxChangeF / 
					(self.setHealthDeficitScalingMaxCutoffF - self.setHealthDeficitScalingMinCutoffF)* (self.setHealthDeficitScalingMaxCutoffF - nHealthPercent )
				end
			end
		end
	end
	if tNameplate.nQuestIconPixie ~= nil and self.setStyleQuestIconUseDefaultScaling then
		nScale = nBaseScale
	else
		nScale = nScale*nBaseScale
	end
	if nScale ~= 1 and math.abs(nScale - 1) < self.setScalingUpdateMinChangeF then nScale = 1; nPrevScale = -1 end
	if nPrevScale >= 0
		and not tNameplate.bNeedsRedraw
		and (math.abs(nScale - nPrevScale) < self.setScalingUpdateMinChangeF)
	then
		return
	end
	
	tNameplate.bNeedsRedraw = true
	tNameplate.nPrevScale = nScale
	if wndRefs.NamePlate ~= nil and wndRefs.NamePlateContainer ~= nil then
		local namePlate = wndRefs.NamePlate
		wndRefs.NamePlate:SetScale(nScale)
		local normalWidth = 140+self.nStyleWidth
		local normalHeight = tNameplate.nNameplateHeight
		local newWidth = normalWidth * (nScale - 1)
		local newHeight = normalHeight * (nScale - 1)
		
		tNameplate.nScalingHorzOffset = math.floor(newWidth*0.5)
		tNameplate.nScalingVertOffset = math.floor(newHeight*0.5)
	end	
end

-- Change nameplate opacity as necessary
function OptiPlates:DoNameplateOpacity(tNameplate)
	local nPrevOpacity = tNameplate.nPrevOpacity
	local nNameplateDistanceSquared = tNameplate.nSquaredDistance	
	local nBaseOpacity = self.setDefaultOpacityF
	local nOpacity = 1
	local wndRefs = tNameplate.wnd
	local unitOwner = tNameplate.unitOwner
	local unitPlayer = self.playerUnit
	if unitOwner == nil or unitPlayer == nil then return end

	if self.setDistanceOpacityEnabled and nNameplateDistanceSquared >= 0
		and not (
			(self.setDistanceOpacityNormalTargetOpacity and (tNameplate.bIsTarget or tNameplate.bInCluster))
			or (self.setDistanceOpacityNormalObjectiveOpacity and tNameplate.bIsObjective)
			or (self.setHarvestablesDistanceScalingDisabled and (tNameplate.bIsHarvest or tNameplate.bIsDatacube))
		)
		then
		if nNameplateDistanceSquared < self.setDistanceOpacityMaxRangeSquared then
			nOpacity = 1
		else
			nOpacity = (1 - self.setDistanceOpacityFactorF)
		end
	end	
	
	if (self.setCombatOpacityEnabled and unitOwner:IsInCombat()) 
		and (not self.setCombatOpacityPlayerInCombat or (self.setCombatOpacityPlayerInCombat and unitPlayer:IsInCombat()))
		and unitOwner ~= unitPlayer 
	then
		local eDisposition = tNameplate.eDisposition
		if eDisposition == nOpacity*tDisposition.Friendly then
			nOpacity = self.setCombatOpacityFriendlyF
		elseif eDisposition == tDisposition.Hostile or eDisposition == tDisposition.Neutral then
			if (not self.setCombatOpacityDefaultHostileCasting 
				or (self.setCombatOpacityDefaultHostileCasting and not unitOwner:ShouldShowCastBar())
				)
			then
				nOpacity = nOpacity*self.setCombatOpacityHostileF	
			end
		end
	end
	local eDisposition = tNameplate.eDisposition
	if tNameplate.bSpeechBubble and eDisposition == tDisposition.Friendly then
		nOpacity = nOpacity*self.setSpeechBubbleOpacityF
	end
	
	if tNameplate.bShowBars and tNameplate.nPrevBGOpacity ~= self.setBarBGOpacity then
		local sBGOpacity = string.format("%x",math.floor(self.setBarBGOpacityF*255))
		local newColor = ApolloColor.new(sBGOpacity..self.setColorBackgroundBar)
		wndRefs.HealthBarBack:SetBGColor(newColor)
		wndRefs.ShieldBar:SetBGColor(newColor)
		if wndRefs.FocusBar:IsShown() then
			wndRefs.FocusBar:SetBGColor(newColor)
		end
		if wndRefs.ClassResourceBar:IsShown() then
			wndRefs.ClassResourceBar:SetBGColor(newColor)
		end		
		tNameplate.nPrevBGOpacity = self.setBarBGOpacity
	end
	if self.setTargetOpacityEmphasis > 0 and unitPlayer:GetTarget() ~= nil then
		if not (tNameplate.bIsTarget or (tNameplate.bInCluster and self.setShowTargetClusterIcons)) then
			nOpacity = nOpacity*(1 - self.setTargetOpacityEmphasisF)
		end
	end
	nOpacity = nOpacity * nBaseOpacity
	-- Perhaps redundant sanity checks
	if nOpacity < 0 then nOpacity = 0 end
	if nOpacity > 1 then nOpacity = 1 end

	if tNameplate.nPrevOpacity < 0 then tNameplate.bNeedsRedraw = true end
	
	if nPrevOpacity == nOpacity and not tNameplate.bNeedsRedraw then
		return
	end

	tNameplate.nPrevOpacity = nOpacity
	if wndRefs.NamePlateContainer ~= nil then
		wndRefs.NamePlateContainer:SetOpacity(nOpacity,20000000)
	end
end

-----------------------------------------------------------------------------------------------
-- GAME STATE UPDATE FUNCTIONS
-----------------------------------------------------------------------------------------------

-- Called from the nameplate window. 
function OptiPlates:OnWorldLocationOnScreen(wndHandler, wndControl, bOnScreen)
	local tNameplate = self.arDisplayedNameplates[wndHandler:GetId()]
	if tNameplate ~= nil then
		tNameplate.bNeedsRedraw = true
		self:UpdateNameplateVisibility(tNameplate)
	end
end

-- Called from the nameplate window.
function OptiPlates:OnUnitOcclusionChanged(wndHandler, wndControl, bOccluded)
	local tNameplate = self.arDisplayedNameplates[wndHandler:GetId()]
	if tNameplate ~= nil then
		if self.setEnableOcclusion then
			tNameplate.bNeedsRedraw = true
		end			
	end
end


function OptiPlates:OnEnteredCombat(unitChecked, bInCombat)
	if unitChecked == self.playerUnit then
		if self.setQuickselectWindowHideInCombat and self.setQuickselectWindowShow then
			self.wndQuickChange:Show(not bInCombat)
		end
	end
	if unitChecked ~= nil then 
		local tNameplate = self.arUnit2Nameplate[unitChecked:GetId()]
		if tNameplate ~= nil then 
			self:UpdateNameplateVisibility(tNameplate)
			tNameplate.bNeedsRedraw = true
		end
	end
end

function OptiPlates:OnMount()
	if self.playerUnit == nil then return end
	local tNameplate = self.arUnit2Nameplate[self.playerUnit:GetId()]
	if tNameplate ~= nil then
		self:UpdateNameplateVisibility(tNameplate)
	end
end

function OptiPlates:OnUnitTextBubbleToggled(tUnitArg, strText)
	local nSpeechBubbleMaxWidth = 250 -- empirical, can be wrong
	local idUnit = tUnitArg:GetId()
	if not self.arUnit2Nameplate or not self.arUnit2Nameplate[idUnit] then
		return
	end
	local tNameplate = self.arUnit2Nameplate[idUnit]
	local bSpeechBubble = tNameplate.bSpeechBubble
	if strText and strText ~= "" then
		if bSpeechBubble ~= true then
			local nTextWidth = Apollo.GetTextWidth("CRB_InterfaceMedium", strText)
			tNameplate.bSpeechBubble = true
			tNameplate.nSpeechBubbleWidth = math.min(nSpeechBubbleMaxWidth,nTextWidth)
			tNameplate.nSpeechBubbleHeight = (math.floor(nTextWidth/nSpeechBubbleMaxWidth)+1)*12
			tNameplate.bNeedsRedraw = true
		end
	else
		if bSpeechBubble ~= false then	
			tNameplate.bSpeechBubble = false
			tNameplate.bNeedsRedraw = true			
		end
	end
end

function OptiPlates:OnAlternateTargetUnitChanged(unitOwner)
	
	if self.focusUnit ~= nil then 
		local tNameplate = self.arUnit2Nameplate[self.focusUnit:GetId()]
		if tNameplate ~= nil then
			tNameplate.bNeedsRedraw = true
		end
	end
	self.focusUnit = unitOwner
end

-- build targeted options here; we get this event when a creature attacks, too
function OptiPlates:OnTargetUnitChanged(unitOwner) 
	local tNameplate = nil
	-- see if new target has nameplate and create it if it doesn't
	if unitOwner ~= nil then 
		tNameplate = self.arUnit2Nameplate[unitOwner:GetId()]
		if tNameplate == nil then
			tNameplate = self:CreateNameplateObject(unitOwner)
		else
			tNameplate.bNeedsRedraw = true		
		end
		tNameplate.bIsTarget = true
	end 
	if self.unitLastTarget ~= nil then
		self.unitLastTarget.bNeedsRedraw = true
		self.unitLastTarget.bIsTarget = false
		if self.unitLastTarget.unitOwner ~= self.playerUnit then
			self:ReloadNameplate(self.unitLastTarget)
		else
			self:UpdateNameplateVisibility(self.unitLastTarget)
		end
	end
	
	-- Check for old target's cluster status, and make sure their nameplates get updated
	local prevTarget = self.unitLastTarget
	local bSameCluster = false	
	if (prevTarget ~= nil) then
		prevTarget.bIsTarget = false
		if not prevTarget.bIsPlayer then
			local bClusterInformation = prevTarget.unitOwner:GetClusterInformation()
			if bClusterInformation ~= nil and bClusterInformation.nCount ~= nil and bClusterInformation.nCount > 1 then
				local tClusterUnits = bClusterInformation.tUnits
				for idx = 1, #tClusterUnits do
					local clusterUnit = self.arUnit2Nameplate[tClusterUnits[idx]:GetId()]
					if clusterUnit ~= nil and tNameplate ~= nil and clusterUnit.unitOwner:GetId() == tNameplate.unitOwner:GetId() then
						bSameCluster = true
					end
				end
				for idx = 1, #tClusterUnits do
					local clusterUnit = self.arUnit2Nameplate[tClusterUnits[idx]:GetId()]
					if (clusterUnit ~= nil) then
						clusterUnit.bNeedsRedraw = true
						if not bSameCluster then
							clusterUnit.bInCluster = false
							clusterUnit.bIsTarget = false
						end
						self:UpdateNameplateVisibility(clusterUnit)
					end
				end
			end
		end
	end


	
	self.unitLastTarget = tNameplate
	self:UpdateNameplateVisibility(tNameplate)
	-- If current target is a member of cluster, redraw all cluster members, unless new target is in previous cluster
	if tNameplate ~= nil and tNameplate.unitOwner ~= nil and not tNameplate.bIsPlayer then
		if not bSameCluster then
			local unitPlayer = self.playerUnit
			if unitPlayer ~= nil then 
				local unitCurrentTarget = unitPlayer:GetTarget()
				if (unitCurrentTarget ~= nil) then
					local bClusterInformation = unitCurrentTarget:GetClusterInformation()
					if bClusterInformation ~= nil and bClusterInformation.nCount ~= nil and bClusterInformation.nCount > 1 then
						local tClusterUnits = bClusterInformation.tUnits
						for idx = 1, #tClusterUnits do
							tNameplate = self.arUnit2Nameplate[tClusterUnits[idx]:GetId()]
							if tNameplate ~= nil then
								tNameplate.bInCluster = true
								tNameplate.bNeedsRedraw = true
								self:UpdateNameplateVisibility(tNameplate)
							end
						end
					end
				end
			end	
		end
	end
end

function OptiPlates:OnUnitPvpFlagsChanged(unitUpdated, tNameplate)
	if tNameplate == nil then
		tNameplate = self.arUnit2Nameplate[unitUpdated:GetId()]
		if tNameplate == nil then
			return
		end
	end
	self:UpdateNameplateVisibility(tNameplate)
	tNameplate.bNeedsRedraw = true
end

function OptiPlates:OnUnitNameChanged(unitUpdated, strNewName)
	local tNameplate = self.arUnit2Nameplate[unitUpdated:GetId()]
	if tNameplate == nil or tNameplate.wndNameplate == nil then return end

	tNameplate.bNeedsRedraw = true
end

function OptiPlates:OnUnitTitleChanged(unitUpdated)
	local tNameplate = self.arUnit2Nameplate[unitUpdated:GetId()]
	if tNameplate == nil or tNameplate.wndNameplate == nil then return end

	tNameplate.bNeedsRedraw = true
end

function OptiPlates:OnPlayerTitleChanged()
	local unitUpdated = self.playerUnit
	local tNameplate = self.arUnit2Nameplate[unitUpdated:GetId()]
	if tNameplate == nil or tNameplate.wndNameplate == nil then return end
	
	tNameplate.bNeedsRedraw = true
end

function OptiPlates:OnUnitGuildNameplateChanged(unitUpdated)
	local tNameplate = self.arUnit2Nameplate[unitUpdated:GetId()]
	if tNameplate == nil or tNameplate.wndNameplate == nil then return end
	
	tNameplate.bNeedsRedraw = true
end

function OptiPlates:OnGuildDirty(unitOwner, tNameplate)
	if self.guildDisplayed then
		tNameplate.bIsGuildMember = self.guildDisplayed:IsUnitMember(unitOwner)
	else
		tNameplate.bIsGuildMember = false
	end

	if self.guildWarParty then
		tNameplate.bIsWarPartyMember = self.guildWarParty:IsUnitMember(unitOwner)
	else
		tNameplate.bIsWarPartyMember = false
	end
end

function OptiPlates:OnUnitMemberOfGuildChange(unitOwner)
	if unitOwner == nil then
		return
	end

	local tNameplate = self.arUnit2Nameplate[unitOwner:GetId()]
	if tNameplate == nil or tNameplate.wndNameplate == nil then
		return
	end

	self:OnGuildDirty(unitOwner, tNameplate)

	self:OnUnitPvpFlagsChanged(unitOwner, tNameplate)
end

function OptiPlates:OnGuildChange()
	self.guildDisplayed = nil
	self.guildWarParty = nil
	for key, guildCurr in pairs(GuildLib.GetGuilds()) do
		if guildCurr:GetType() == GuildLib.GuildType_Guild then
			self.guildDisplayed = guildCurr
		end
		if guildCurr:GetType() == GuildLib.GuildType_WarParty then
			self.guildWarParty = guildCurr
		end
	end

	for key, tNameplate in pairs(self.arUnit2Nameplate) do
		local unitOwner = tNameplate.unitOwner
		self:OnGuildDirty(unitOwner, tNameplate)
		self:OnUnitPvpFlagsChanged(unitOwner, tNameplate)
	end
end

function OptiPlates:OnKeyBindingUpdated(strKeybind)
	if strKeybind ~= "Path Action" and strKeybind ~= "Cast Objective Ability" then
		return
	end

	self.strPathActionKeybind = GameLib.GetKeyBinding("PathAction")
	self.bPathActionUsesIcon = false
	if self.strPathActionKeybind == Apollo.GetString("HUDAlert_Unbound") or #self.strPathActionKeybind > 1 then -- Don't show interact
		self.bPathActionUsesIcon = true
	end

	self.strQuestActionKeybind = GameLib.GetKeyBinding("CastObjectiveAbility")
	self.bQuestActionUsesIcon = false

	if self.strQuestActionKeybind == Apollo.GetString("HUDAlert_Unbound") or #self.strQuestActionKeybind > 1 then -- Don't show interact
		self.bQuestActionUsesIcon = true
	end

	self:OptionsChanged()
end




function OptiPlates:OnUnitActivationTypeChanged(unitUpdated)
	if unitUpdated == nil then return end
	
	local tNameplate = self.arUnit2Nameplate[unitUpdated:GetId()]
	if tNameplate == nil then return end
	
	tNameplate.bIsVendorNPC = self:IsVendorNPC(tNameplate.unitOwner)
	tNameplate.bIsUnimportantNPC	= (tNameplate.bIsImportantNPC and false or self:IsUnimportantNPC(unit))
	tNameplate.bIsImportantNPC = self:IsImportantNPC(tNameplate.unitOwner)
	
	self:ParseRewards(tNameplate)
	self:UpdateNameplateVisibility(tNameplate)
end

-----------------------------------------------------------------------------------------------
function OptiPlates:OnRewardInfoUpdated() -- we've recieved an external signal that reward stuff has been changed
	self.bRewardInfoDirty = true
end

function OptiPlates:UpdateRewardInfo()
	if self.bInitialLoadAllClear == false then
		return
	end

	self.bRewardInfoDirty = false
	for idx, tNameplate in pairs(self.arUnit2Nameplate) do
		if tNameplate ~= nil and tNameplate.unitOwner ~= nil and tNameplate.unitOwner:IsValid() then
			local nOldRewardCount = tNameplate.nRewardCount
			local nNewRewardCount = self:ParseRewards(tNameplate)
			-- Nameplate visibility can change if rewards increase from 0 or go to 0
			if (nOldRewardCount == 0 or nNewRewardCount == 0) and nOldRewardCount ~= nNewRewardCount then
				self:UpdateNameplateVisibility(tNameplate)
				-- Visibility changed, needs a full redraw
				tNameplate.bNeedsRedraw = true
			else
				-- Only rewards need redrawing
				tNameplate.bUpdateRewards = true
			end
		end
	end
end


function OptiPlates:ParseRewards(tNameplate)
	local unitOwner = tNameplate.unitOwner
	if unitOwner == nil or not unitOwner:IsValid() then return 0 end
	
	local tActivation = unitOwner:GetActivationState()
	local unitType = unitOwner:GetType()
	tNameplate.bIsQuestGiverNew		= (tActivation.QuestNew ~= nil or tActivation.QuestNewMain ~= nil)
	tNameplate.bIsQuestGiverNewDaily = (tActivation.QuestNewDaily ~= nil) 
	tNameplate.bIsQuestGiverNewGrey	= (tActivation.QuestNewSoon ~= nil)	
	tNameplate.bIsQuestReceiver		= (tActivation.QuestReceiving ~= nil)
	tNameplate.bIsQuestFinisher		= (tActivation.QuestReward ~= nil)
	tNameplate.bIsQuestUnit			= (
			(tNameplate.bIsQuestFinisher and self.setShowQuestIconReward) or
			(tNameplate.bIsQuestGiverNew and self.setShowQuestIconNew) or
			(tNameplate.bIsQuestGiverNewDaily and self.setShowQuestIconNewDaily) or
			(tNameplate.bIsQuestReceiver and self.setShowQuestIconReceiving) or
			(tNameplate.bIsQuestGiverNewGrey and self.setShowQuestIconNewGrey)
	)
	tNameplate.bIsDatacube			= (unitType == "Simple" and tActivation.Datacube ~= nil)
	tNameplate.bHasActivationState	= (tActivation ~= nil) and (self:GetTableLength(tActivation) > 0) and tActivation.Chair == nil
	
	
	if tNameplate.bIsPet or tNameplate.bIsMount then
		tNameplate.bIsObjective = false
		return
	end
	local tRewardInfo = unitOwner:GetRewardInfo()
	if tRewardInfo == nil or type(tRewardInfo) ~= "table" then
		tNameplate.bIsObjective = false
		return
	end
	local iRewards = 0
	local nRewardCount = #tRewardInfo
	local ePathId = PlayerPathLib.GetPlayerPathType()
	local bAlwaysShow = self.setShowObjectivesAlways
	if tNameplate.bIsQuestUnit then iRewards = iRewards + 1 end
	if nRewardCount > 0 then
		for idx = 1, nRewardCount do
			local tReward = tRewardInfo[idx]
			local strType = tReward.strType
			if (not bAlwaysShow or self.setIconQuest) and strType == "Quest" then
				if tReward.eType == 11 then -- Contract
					-- questid 9940 is Prime Executioner contract on Exile side
					if tReward.idQuest ~= 9940 or (tNameplate.bIsNPC and not unitOwner:GetArchetype() == Unit.CodeEnumArchetype.Townie and unitOwner:CanGrantXp() and not unitOwner:GetRank() > 3) then
						iRewards = iRewards + 1
					end
				else -- Quest
					iRewards = iRewards + 1
				end
				break
			elseif (not bAlwaysShow or self.setIconChallenge) and strType == "Challenge" and tReward.bActivated then
				iRewards = iRewards + 1
				break
			elseif (not bAlwaysShow or self.setIconEvent) and strType == "PublicEvent" then --and tReward.peoObjective ~= nil then
				iRewards = iRewards + 1				
				break
			elseif (not bAlwaysShow or self.setIconMission) and ePathId == PlayerPathLib.PlayerPathType_Explorer and strType == "Explorer" then
				iRewards = iRewards + 1
				break
			elseif (not bAlwaysShow or self.setIconMission) and ePathId == PlayerPathLib.PlayerPathType_Scientist and strType == "Scientist" then
				iRewards = iRewards + 1
				break
			elseif (not bAlwaysShow or self.setIconMission) and ePathId == PlayerPathLib.PlayerPathType_Soldier and strType == "Soldier" then
				iRewards = iRewards + 1
				break
			elseif (not bAlwaysShow or self.setIconMission) and ePathId == PlayerPathLib.PlayerPathType_Settler and strType == "Settler" then
				iRewards = iRewards + 1
				break
			end
		end
	end

	--now do questgivers, etc
	if tActivation.QuestTarget ~= nil then
		iRewards = iRewards + 1
	end
	if tActivation.TalkTo ~= nil then
		iRewards = iRewards + 1
	end
	
	if tNameplate.bIsCollectible and (tActivation.Interact == nil or not tActivation.Interact.bCanInteract) then
		tNameplate.bIsObjective = false
	--elseif tNameplate.bIsNPC and tNameplate.unitOwner:GetHealth() == nil 
	--	and (
	--		(tActivation.Interact == nil or not tActivation.Interact.bCanInteract) 
	--	  and (tActivation.QuestTarget == nil or not tActivation.QuestTarget.bIsActive)
	--	  and (tActivation["Public Event"] == nil or not tActivation["Public Event"].bIsActive)
	--	)
	--then
	--	tNameplate.bIsObjective = false
	else
		tNameplate.bIsObjective = (iRewards > 0)
	end

	tNameplate.nRewardCount = iRewards
	return iRewards
end




-----------------------------------------------------------------------------------------------
-- Reward Icon Helper Methods
-----------------------------------------------------------------------------------------------

function OptiPlates:HelperDrawRewardTooltip(tRewardInfo, wndRewardIcon, strBracketText, strUnitName, tRewardString)
	-- TODO: Delete this code.
	if not tRewardInfo or not wndRewardIcon then return end
	-- TODO: set icon paths too

	local strMessage = tRewardInfo.strTitle
	if tRewardInfo.pmMission and tRewardInfo.pmMission:GetName() then
		local pmMission = tRewardInfo.pmMission
		if tRewardInfo.bIsActivate and PlayerPathLib.GetPlayerPathType() ~= PlayerPathLib.PlayerPathType_Explorer then -- todo: see if we can remove this requirement
			strMessage = String_GetWeaselString(Apollo.GetString("Nameplates_ActivateForMission"), pmMission:GetName())
		else
			strMessage = string.format("%s %s/%s", pmMission:GetName(), pmMission:GetNumCompleted(), pmMission:GetNumNeeded()) -- TODO Need string weasel (code will be deleted)
		end
	end

	local strProgress = ""
	local strNeeded = tRewardInfo.nNeeded
	local strCompleted = tRewardInfo.nCompleted
	local bShowCount = tRewardInfo.bShowCount
	
	if tRewardInfo.strType == "PublicEvent" and tRewardInfo.peoObjective ~= nil then
		strCompleted = tRewardInfo.peoObjective:ShowPercent()
	end
	
	if strCompleted and strCompleted ~= "" and strNeeded and strNeeded ~= "" then
		if tRewardInfo.bShowCount then
			strProgress = string.format(": %s/%s", strCompleted, strNeeded)
		else
			strProgress = string.format(": %s%%", strCompleted)
		end
	end

	if wndRewardIcon:IsShown() then -- TODO Need string weasel (code will be deleted)
		tRewardString = string.format("%s<P Font=\"CRB_InterfaceMedium\" TextColor=\"ffffffff\">%s: %s%s</P>", tRewardString, strBracketText, strMessage, strProgress)
	else
		tRewardString = string.format("%s<P Font=\"CRB_InterfaceMedium\" TextColor=\"Yellow\">%s (%s)</P><P Font=\"CRB_InterfaceMedium\">%s%s</P>",
										tRewardString, strUnitName, strBracketText, strMessage, strProgress)
		--wndRewardIcon:SetTooltip(tRewardString)
	end

	return tRewardString
end

function OptiPlates:HelperLoadRewardIcon(wndRewardPanel, idx)
	if wndRewardPanel:FindChild(idx) then
		return wndRewardPanel:FindChild(idx)
	end

	local wndCurr = Apollo.LoadForm(self.xmlDoc, "RewardIcon", wndRewardPanel, self)
	wndCurr:SetName(idx)
	wndCurr:Show(false) -- Visibility is important

	local tRewardIcons = karRewardIcons

	wndCurr:FindChild("Single"):SetSprite(tRewardIcons[idx].strSingle)

	if wndCurr:FindChild("Multi") then
		wndCurr:FindChild("Multi"):SetSprite(tRewardIcons[idx].strMulti)
	end

	return wndCurr
end

function OptiPlates:HelperDrawRewardIcon(wndRewardIcon)
	if not wndRewardIcon then
		return 0
	end

	local nResult = 0
	if wndRewardIcon:FindChild("Multi") then -- Show multi if the Single icon if the window is already visible
		wndRewardIcon:FindChild("Multi"):Show(wndRewardIcon:IsShown())
		wndRewardIcon:FindChild("Multi"):ToFront()
	end

	if not wndRewardIcon:IsShown() then -- Plus one to the counter if this is the first instance
		nResult = 1
	end

	wndRewardIcon:Show(true) -- At the very end
	return nResult
end


function OptiPlates:HelperDrawSpellBind(wndIcon, idx)
	if idx ~= 1 then -- paths, not quest
		if self.bPathActionUsesIcon then
			wndIcon:FindChild("TargetMark"):Show(true)
			wndIcon:FindChild("Bind"):SetText("")
		else
			wndIcon:FindChild("TargetMark"):Show(false)
			wndIcon:FindChild("Bind"):SetText(self.strPathActionKeybind)
		end
	else -- quest
		if self.bQuestActionUsesIcon then
			wndIcon:FindChild("TargetMark"):Show(true)
			wndIcon:FindChild("Bind"):SetText("")
		else
			wndIcon:FindChild("TargetMark"):Show(false)
			wndIcon:FindChild("Bind"):SetText(self.strQuestActionKeybind)
		end
	end
end


function OptiPlates:ShowRewardIcons(tNameplate)
	local unitOwner = tNameplate.unitOwner

	if unitOwner == nil then
		return 0
	end

	if tNameplate.wndNameplate == nil then
		return
	end

	local bIsFriend = unitOwner:IsFriend()
	local bIsRival = unitOwner:IsRival()
	local bIsAccountFriend = unitOwner:IsAccountFriend()
	local nFriendshipCount = (bIsFriend and 1 or 0) + (bIsRival and 1 or 0) + (bIsAccountFriend and 1 or 0)

	local tRewardInfo = {}
	tRewardInfo = unitOwner:GetRewardInfo()
	if (tRewardInfo == nil or type(tRewardInfo) ~= "table") and nFriendshipCount == 0 then
		return 0
	end

	local wndRewardPanel = tNameplate.wndNameplate:FindChild("QuestRewards")
	--wndRewardPanel:DestroyChildren()
	--wndRewardPanel:SetAnchorOffsets(0, 0, 0, 0)

	local tRewardString = {} -- temp table to store quest descriptions (builds multi-objective tooltips)
	for idx,tRewardIcons in pairs(karRewardIcons) do
		local wndTemp = wndRewardPanel:FindChild(idx)
		if wndTemp then 
			wndTemp:Show(false)
			wndTemp:SetTooltip("")
		end
		tRewardString[idx] = ""
	end

	local nRewardInfoListCount = tRewardInfo ~= nil and #tRewardInfo or 0
	if nRewardInfoListCount <= 0 and nFriendshipCount == 0 then
		return 0
	end

	local iRewardCount = 0
	if tNameplate.bIsObjective then
		for idx = 1, nRewardInfoListCount do
			local strType = tRewardInfo[idx].strType

			if self.setIconQuest and strType == "Quest" then
				local nRewardId = Unit.CodeEnumRewardInfoType.Quest
				local sTypeString = Apollo.GetString("CRB_Quest")
				if tRewardInfo[idx].eType == 11 then
					nRewardId = Unit.CodeEnumRewardInfoType.Contract
					sTypeString = Apollo.GetString("CRB_Contract")
				end
				local wndCurr = self:HelperLoadRewardIcon(wndRewardPanel, nRewardId)
				iRewardCount = iRewardCount + self:HelperDrawRewardIcon(wndCurr)
				tRewardString[nRewardId] = self:HelperDrawRewardTooltip(tRewardInfo[idx], wndCurr, sTypeString, unitOwner:GetName(), tRewardString[nRewardId])
				if self.setIconTooltips then wndCurr:SetTooltip(tRewardString[nRewardId]) end
				wndCurr:ToFront()

				if tRewardInfo[idx].splAbility then
					self:HelperDrawSpellBind(wndCurr, nRewardId)
				end
			elseif self.setIconChallenge and strType == "Challenge" and tRewardInfo[idx].bActivated then
				local nRewardId = Unit.CodeEnumRewardInfoType.Challenge
				local wndCurr = self:HelperLoadRewardIcon(wndRewardPanel, nRewardId)
				iRewardCount = iRewardCount + self:HelperDrawRewardIcon(wndCurr)
				tRewardString[nRewardId] = self:HelperDrawRewardTooltip(tRewardInfo[idx], wndCurr, Apollo.GetString("CBCrafting_Challenge"), unitOwner:GetName(), tRewardString[nRewardId])
				if self.setIconTooltips then wndCurr:SetTooltip(tRewardString[nRewardId]) end
				wndCurr:ToFront()

				if tRewardInfo[idx].splAbility then
					self:HelperDrawSpellBind(wndCurr, nRewardId)
				end
			elseif strType == "Soldier" and PlayerPathLib.GetPlayerPathType() == PlayerPathLib.PlayerPathType_Soldier and self.setIconMission then
				local nRewardId = Unit.CodeEnumRewardInfoType.Soldier
				local wndCurr = self:HelperLoadRewardIcon(wndRewardPanel, nRewardId)
				iRewardCount = iRewardCount + self:HelperDrawRewardIcon(wndCurr)
				tRewardString[nRewardId] = self:HelperDrawRewardTooltip(tRewardInfo[idx], wndCurr, Apollo.GetString("Nameplates_Mission"), unitOwner:GetName(), tRewardString[nRewardId])
				if self.setIconTooltips then wndCurr:SetTooltip(tRewardString[nRewardId]) end

				if tRewardInfo[idx].splReward then
					self:HelperDrawSpellBind(wndCurr, nRewardId)
				end
			elseif strType == "Settler" and PlayerPathLib.GetPlayerPathType() == PlayerPathLib.PlayerPathType_Settler and self.setIconMission then
				local nRewardId = Unit.CodeEnumRewardInfoType.Settler
				local wndCurr = self:HelperLoadRewardIcon(wndRewardPanel, nRewardId)
				iRewardCount = iRewardCount + self:HelperDrawRewardIcon(wndCurr)
				tRewardString[nRewardId] = self:HelperDrawRewardTooltip(tRewardInfo[idx], wndCurr, Apollo.GetString("Nameplates_Mission"), unitOwner:GetName(), tRewardString[nRewardId])
				if self.setIconTooltips then wndCurr:SetTooltip(tRewardString[nRewardId]) end

				if tRewardInfo[idx].splReward then
					self:HelperDrawSpellBind(wndCurr, nRewardId)
				end
			elseif strType == "Scientist" and PlayerPathLib.GetPlayerPathType() == PlayerPathLib.PlayerPathType_Scientist and self.setIconMission then
				local nRewardId = Unit.CodeEnumRewardInfoType.Scientist
				if tRewardInfo[idx].splReward == nil then
					local wndCurr = self:HelperLoadRewardIcon(wndRewardPanel, nRewardId)

					local pmMission = tRewardInfo[idx].pmMission
					local strMission = ""
					if pmMission then
						if pmMission:GetMissionState() >= PathMission.PathMissionState_Unlocked then
							if pmMission:GetType() == PathMission.PathMissionType_Scientist_FieldStudy then
							strMission = string.format("%s %s", pmMission:GetName(), String_GetWeaselString(Apollo.GetString("CRB_ProgressSimple"), pmMission:GetNumCompleted(), pmMission:GetNumNeeded()))
								local tActions = pmMission:GetScientistFieldStudy()
								if tActions then
									for idx, tEntry in ipairs(tActions) do
										if not tEntry.bIsCompleted then
											strMission = string.format("%s ... %s", strMission , tEntry.strName)
										end
									end
								end
							else
								strMission = string.format("%s %s", pmMission:GetName(), String_GetWeaselString(Apollo.GetString("CRB_ProgressSimple"), pmMission:GetNumCompleted(), pmMission:GetNumNeeded()))
							end
						else
							strMission = "???"
						end
					end

					iRewardCount = iRewardCount + self:HelperDrawRewardIcon(wndCurr)

					local strProgress = "" -- specific to #7
					local strUnitName = unitOwner:GetName() -- specific to #7
					local strBracketText = Apollo.GetString("Nameplates_Missions") -- specific to #7
					if wndCurr:IsShown() then -- already have a tooltip
						tRewardString[nRewardId] = string.format("%s<P Font=\"CRB_InterfaceMedium\" TextColor=\"ffffffff\">%s%s</P>", tRewardString[7], strMission, strProgress)
					else
						tRewardString[nRewardId] = string.format("%s<P Font=\"CRB_InterfaceMedium\" TextColor=\"Yellow\">%s (%s)</P>"..
														 "<P Font=\"CRB_InterfaceMedium\">%s%s</P>", tRewardString[7], strUnitName, strBracketText, strMessage, strProgress)
					end

					if self.setIconTooltips then wndCurr:SetTooltip(tRewardString[nRewardId]) end
				end

				local splSpell = tRewardInfo[idx].splReward
				if splSpell then
					local wndCurr = self:HelperLoadRewardIcon(wndRewardPanel, Unit.CodeEnumRewardInfoType.ScientistSpell)
					iRewardCount = iRewardCount + self:HelperDrawRewardIcon(wndCurr)
					--Tooltip.GetSpellTooltipForm(self, wndCurr, splSpell)
				end

			elseif strType == "Explorer" and PlayerPathLib.GetPlayerPathType() == PlayerPathLib.PlayerPathType_Explorer and self.setIconMission then
				local nRewardId = Unit.CodeEnumRewardInfoType.Explorer
				local wndCurr = self:HelperLoadRewardIcon(wndRewardPanel, nRewardId)
				iRewardCount = iRewardCount + self:HelperDrawRewardIcon(wndCurr)
				tRewardString[nRewardId] = self:HelperDrawRewardTooltip(tRewardInfo[idx], wndCurr, "Mission", unitOwner:GetName(), tRewardString[nRewardId])
				if self.setIconTooltips then wndCurr:SetTooltip(tRewardString[nRewardId]) end

				if tRewardInfo[idx].splReward then
					self:HelperDrawSpellBind(wndCurr, nRewardId)
				end
			elseif strType == "PublicEvent" and self.setIconEvent then
				local nRewardId = Unit.CodeEnumRewardInfoType.PublicEvent
				local wndCurr = self:HelperLoadRewardIcon(wndRewardPanel, nRewardId)

				local peEvent = tRewardInfo[idx].peoObjective
				local strTitle = peEvent:GetEvent():GetName()
				local nCompleted = peEvent:GetCount()
				local nNeeded = peEvent:GetRequiredCount()

				iRewardCount = iRewardCount + self:HelperDrawRewardIcon(wndCurr)

				if wndCurr:IsShown() then -- already have a tooltip
					-- Do nothing. It has been cut below
				else
					local strPublicEventMarker = String_GetWeaselString(Apollo.GetString("Nameplates_PublicEvents"), unitOwner:GetName())
					tRewardString[nRewardId] = string.format("<P Font=\"CRB_InterfaceMedium\" TextColor=\"Yellow\">%s</P>", strPublicEventMarker)
				end

				if peEvent:GetObjectiveType() == PublicEventObjective.PublicEventObjectiveType_Exterminate then
					strNumRemaining = String_GetWeaselString(Apollo.GetString("Nameplates_NumRemaining"), strTitle, nCompleted)
					tRewardString[nRewardId] = string.format("%s<P Font=\"CRB_InterfaceMedium\">%s</P>", tRewardString[nRewardId], strNumRemaining)
				elseif peEvent:ShowPercent() then
					strPercentCompleted = String_GetWeaselString(Apollo.GetString("Nameplates_PercentCompleted"), strTitle, nCompleted / nNeeded * 100)
					tRewardString[nRewardId] = string.format("%s<P Font=\"CRB_InterfaceMedium\">%s</P>", tRewardString[nRewardId], strPercentCompleted)
				else
					tRewardString[nRewardId] = string.format("%s<P Font=\"CRB_InterfaceMedium\">%s: %s/%s</P>", tRewardString[nRewardId], strTitle, nCompleted, nNeeded)
				end

				wndCurr:ToFront()
				if self.setIconTooltips then wndCurr:SetTooltip(tRewardString[nRewardId]) end
			end
		end
	end

	if bIsRival and self.setIconRival then
		local nRewardId = Unit.CodeEnumRewardInfoType.Rival
		local wndCurr = self:HelperLoadRewardIcon(wndRewardPanel, nRewardId)
		iRewardCount = iRewardCount + self:HelperDrawRewardIcon(wndCurr)
		
		if self.setIconTooltips then wndCurr:SetTooltip("<P Font=\"CRB_InterfaceMedium\" TextColor=\"Yellow\">" .. Apollo.GetString("TargetFrame_Rival") .. "</P>") end
		wndCurr:ToFront()
		iRewardCount = iRewardCount + 1
		wndCurr:Show(true)
	end
	
	
	local accountFriend = false
	if bIsAccountFriend or bIsFriend and self.setIconFriend then
		local nRewardId = Unit.CodeEnumRewardInfoType.Friend
		local wndCurr = self:HelperLoadRewardIcon(wndRewardPanel, nRewardId)
		iRewardCount = iRewardCount + self:HelperDrawRewardIcon(wndCurr)
		if self.setIconTooltips then 
			if bIsAccountFriend then
				wndCurr:SetTooltip("<P Font=\"CRB_InterfaceMedium\" TextColor=\"Yellow\">" .. Apollo.GetString("TargetFrame_AccountFriend") .. "</P>")
			else
				wndCurr:SetTooltip("<P Font=\"CRB_InterfaceMedium\" TextColor=\"Yellow\">" .. Apollo.GetString("TargetFrame_Friend") .. "</P>")
			end
		end
		wndCurr:ToFront()
		
		wndCurr:Show(true)
		accountFriend = true
	end
	
	if iRewardCount > 0 then
		--wndRewardPanel:SetAnchorOffsets(0, 0, knRewardWidth * iRewardCount, knNameRewardHeight) -- set it (first number is a negative offset)
		wndRewardPanel:ArrangeChildrenHorz(0)
	end

	return iRewardCount
end



--------------------------------------------------------------------------------------
-- Make sure game is loaded. Call when options change.
--------------------------------------------------------------------------------------

function OptiPlates:CreateNameplateSettingsArrays() 
	local arr = self.tNameplateSettings or { }
	local missing = {}
	for idx,category in pairs(tNameplateMatrixCategories) do
		if arr[category] == nil then arr[category] = {} end
		if arr[category]["C"] == nil then arr[category]["C"] = {} end
		if arr[category]["OOC"] == nil then arr[category]["OOC"] = {} end
		for idy,setting in pairs(tNameplateMatrixSettings) do
			if tNameplateMatrixEnabledSettings[category][setting] then
				--local incombatsetting = self[self:HelperGetMatrixSettingName(category, setting, true)]
				--local outofcombatsetting = self[self:HelperGetMatrixSettingName(category, setting, false)]
				local incombatsetting = tUserSettings[self:HelperGetMatrixSettingName(category, setting, true)]
				local outofcombatsetting = tUserSettings[self:HelperGetMatrixSettingName(category, setting, false)]
				if incombatsetting == nil then missing[self:HelperGetMatrixSettingName(category, setting, true)] = true end
				if outofcombatsetting == nil then missing[self:HelperGetMatrixSettingName(category, setting, false)] = true end
				arr[category]["C"][setting] = self[self:HelperGetMatrixSettingName(category, setting, true)]
				arr[category]["OOC"][setting] = self[self:HelperGetMatrixSettingName(category, setting, false)]
			else
				arr[category]["C"][setting] = false
				arr[category]["OOC"][setting] = false
			end
		end
	end
	self.tNameplateSettings = arr
	self.tmpMissing = missing
end

function OptiPlates:ProcessDeprecatedSettings()
	if self.setShowHealth or self.setShowHealthPercent then
		local setting = "HPText"
		for category,settings in pairs(tNameplateMatrixEnabledSettings) do
			if settings[setting] then
				self[self:HelperGetMatrixSettingName(category, setting, false)] = true
				self[self:HelperGetMatrixSettingName(category, setting, true)] = true
			end
		end
		self.setShowHealth = nil
		self.setShowHealthPercent = nil
	end
	if self.setShowShield or self.setShowShieldPercent then
		local setting = "SPText"
		for category,settings in pairs(tNameplateMatrixEnabledSettings) do
			if settings[setting] then
				self[self:HelperGetMatrixSettingName(category, setting, false)] = true
				self[self:HelperGetMatrixSettingName(category, setting, true)] = true
			end
		end
		self.setShowShield = nil
		self.setShowShieldPercent = nil
	end
	
	if self.setFlatWidth ~= nil or self.setSAOSize ~= nil then
		if self.setStyle == "sao" then
			self.setStyleWidth = self.setSAOSize
			self.setHealthbarHeight = math.floor(self.setSAOSize*0.14)
			self.setShieldbarHeight = math.floor(self.setSAOSize*0.05)
		else
			self.setStyleWidth = self.setFlatWidth
			self.setHealthbarHeight = self.setFlatHealthHeight
			self.setShieldbarHeight = self.setFlatShieldHeight
		end
		self.setFlatWidth = nil
		self.setSAOSize = nil
	end
	if self.setStyleCastbarTextInside or self.setStyleCastbarAboveText then
		if self.setStyleCastbarTextInside ~= nil and self.setStyleCastbarTextInside then
			self.setStyleCastbarTextPosition = "Inside"
		elseif self.setStyleCastbarAboveText then
			self.setStyleCastbarTextPosition = "Below"
		end
		self.setStyleCastbarTextInside = nil
		self.setStyleCastbarAboveText = nil
	end
	if self.setStyleCastbarShowTimeRemaining or self.setStyleCastbarShowTimeRemainingHostile then
		if self.setStyleCastbarShowTimeRemaining ~= nil and self.setStyleCastbarShowTimeRemaining then
			self.setStyleCastbarTextTimeRemaining = "Always"
		elseif self.setStyleCastbarShowTimeRemainingHostile then
			self.setStyleCastbarTextTimeRemaining = "Hostile"
		end
		self.setStyleCastbarShowTimeRemaining = nil
		self.setStyleCastbarShowTimeRemainingHostile = nil
	end
	if self.setStyleCastbarNoText then
		local setting = "CastText"
		for category,settings in pairs(tNameplateMatrixEnabledSettings) do
			if settings[setting] then
				self[self:HelperGetMatrixSettingName(category, setting, false)] = true
				self[self:HelperGetMatrixSettingName(category, setting, true)] = true
			end
		end
		self.setStyleCastbarNoText = nil
	end
	if self.setStyleLevelIconPositionToNameLeft ~= nil then
		if self.setStyleLevelIconPositionToNameLeft then self.setLevelTextPosition = "NameLeft"
		else self.setLevelTextPosition = "Bars"
		end
		self.setStyleLevelIconPositionToNameLeft = nil
	end
	if self.setStyleClassIconPositionTop ~= nil then
		if self.setStyleClassIconPositionTop then self.setClassIconPosition = "Top"
		else self.setClassIconPosition = "NameLeft"
		end
		self.setStyleClassIconPositionTop = nil
	end	
	if self.setIconClass ~= nil then
		if self.setIconClass then 
			self.setClassIconStyle = "Icon"
		else
			self.setClassIconStyle = "Text"
		end
		self.setIconClass = nil
	end
end

function OptiPlates:OptionsChanged()
	local unitPlayer = self.playerUnit
	if unitPlayer == nil then
		return
	end
	local currentTicks = Apollo.GetTickCount()
	if currentTicks - nPrevOptionChangeTicks < nOptionChangeMinInterval then
		bOptionChangeQueued = true
		return
	end
	nPrevOptionChangeTicks = currentTicks
	bOptionChangeQueued = false
	
	self:CheckSettingsSanity()
	self:UpdateColorObjects()
	self:CreateNameplateSettingsArrays()	
	if self.setColorPickerIntegrationEnabled then
		colorpicker = Apollo.GetAddon("ColorPicker")
	else
		colorpicker = nil
	end	
	self.bBlinded = unitPlayer:IsInCCState(Unit.CodeEnumCCState.Blind)
	self.focusUnit = self.playerUnit:GetAlternateTarget()
	self.bInitialLoadAllClear = true
	self.bGroupInfoDirty = true
	self.bBuffProfilesDirty = true
	
	self.bStyleFlat = (self.setStyle == "flat")
	self.bStyleSAO = (self.setStyle == "sao")
	if self.bStyleFlat then
	elseif self.bStyleSAO then
		if self.setShieldbarHeight > 10 then self.setShieldbarHeight = 10 end
		if self.setHealthbarHeight > 19 then self.setHealthbarHeight = 19 end	
		if self.setHealthbarHeight < 9 then self.setHealthbarHeight = 9 end	
	end
	self.setFlatHealthTexture = math.floor(self.setFlatHealthTexture)
	self.setFlatShieldTexture = math.floor(self.setFlatShieldTexture)
	
	self.nStyleWidth = self.setStyleWidth
	self.nStyleHalfWidth = math.floor(self.nStyleWidth * 0.5)
	
	Apollo.SetConsoleVariable("ui.occludeNameplatePositions", self.setEnableOcclusion)	
	if self.xmlLoaded and self.playerUnit ~= nil then
		for idx, tNameplate in pairs(self.arUnit2Nameplate) do
			if tNameplate ~= nil then
				self:ReloadNameplate(tNameplate, true)
			end
		end
	end
	
	if math.abs(100 - self.setDefaultScale) <= 3 then
		self.setDefaultScale = 100
	end		

	if math.abs(100 - self.setTargetDefaultScale) <= 3 then
		self.setTargetDefaultScale = 100
	end		

	if math.abs(100 - self.setAltTargetDefaultScale) <= 3 then
		self.setAltTargetDefaultScale = 100
	end		
	
	self.setDefaultScaleF = self.setDefaultScale/100
	self.setTargetDefaultScaleF = self.setTargetDefaultScale/100
	self.setAltTargetDefaultScaleF = self.setAltTargetDefaultScale/100
	self.setScalingUpdateMinChangeF = self.setScalingUpdateMinChange/100
	
	self.setStyleQuestIconMinimumDistanceSquared = self.setStyleQuestIconMinimumDistance * self.setStyleQuestIconMinimumDistance
	self.setDrawDistanceSquared = self.setDrawDistance * self.setDrawDistance
	self.setDistanceScalingMinRangeSquared = self.setDistanceScalingMinRange * self.setDistanceScalingMinRange
	self.setDistanceScalingMaxRangeSquared = self.setDistanceScalingMaxRange * self.setDistanceScalingMaxRange
	self.setDistanceScalingScaleFactorF = self.setDistanceScalingScaleFactor/100

	self.setDistanceOpacityMaxRangeSquared = self.setDistanceOpacityMaxRange * self.setDistanceOpacityMaxRange
	self.setDistanceOpacityFactorF = self.setDistanceOpacityFactor/100
	
	self.setHealthDeficitScalingMinCutoffF = self.setHealthDeficitScalingMinCutoff/100
	self.setHealthDeficitScalingMaxCutoffF = self.setHealthDeficitScalingMaxCutoff/100
	self.setHealthDeficitScalingMaxChangeF = self.setHealthDeficitScalingMaxChange/100

	self.setDefaultOpacityF = self.setDefaultOpacity/100
	self.setBarBGOpacityF = self.setBarBGOpacity/100
	self.setSpeechBubbleOpacityF = self.setSpeechBubbleOpacity/100
	self.setTargetOpacityEmphasisF = self.setTargetOpacityEmphasis/100
	
	self.setCombatOpacityFriendlyF = self.setCombatOpacityFriendly/100
	self.setCombatOpacityHostileF = self.setCombatOpacityHostile/100
	
	self:SaveProfile(self.setCurrentProfileScope, self.setCurrentProfileName)
	
	--self.wndQuickChange:MoveToLocation(WindowLocation.new({ fPoints = {0,0,0,0}, nOffsets = self.setQuickselectWindowPosition}))
	self:UpdateQuickChangeWindowLocation()
	if self.setQuickselectWindowLocked then
		self.wndQuickChange:SetStyle("Moveable", false)
		self.wndQuickChange:FindChild("Button_ToggleLock"):SetBGColor("yellow")
	else
		self.wndQuickChange:SetStyle("Moveable", true)
		self.wndQuickChange:FindChild("Button_ToggleLock"):SetBGColor("UI_BtnBGDefault")
	end
	if not self.setQuickselectWindowHideInCombat then
		self.wndQuickChange:Show(self.setQuickselectWindowShow)
	else
		self.wndQuickChange:Show((not self.playerUnit:IsInCombat()) and self.setQuickselectWindowShow)
	end
	
	self:UpdateOptionMenuPixies()
	self:CreateAllProfileLists()
	self:PopulateAllFontLists()
	self:CreateAllBuffProfileLists()
	self:ChangeBuffProfile("Default")
end

-- Better to have all color objects in memory instead of concatenating strings and recreating them every frame
function OptiPlates:UpdateColorObjects()
	local sBGOpacity = (self.setBarBGOpacity and string.format("%x",math.floor(self.setBarBGOpacity*255/100)) or "ff")
	self.arColorObjects = {
		BackgroundBar = ApolloColor.new(sBGOpacity..self.setColorBackgroundBar),
		BackdropBar = ApolloColor.new(self.setOpacity..self.setColorBackdropBar),
		ShieldBar = ApolloColor.new(self.setOpacity..self.setColorShieldBar),
		FriendlyShieldBarCutoff = ApolloColor.new(self.setOpacity..self.setColorFriendlyShieldBarCutoff),
		AbsorbBar = ApolloColor.new(self.setOpacity..self.setColorAbsorbBar),
		VulnerableBar = ApolloColor.new(self.setOpacity..self.setColorVulnerableBar),
		VulnerableTimerBar = ApolloColor.new(self.setOpacity..self.setColorVulnerableTimerBar),
		FriendlyCastBar = ApolloColor.new(self.setOpacity..self.setColorFriendlyCastBar),
		HostileCastBar = ApolloColor.new(self.setOpacity..self.setColorHostileCastBar),
		
		NeutralName = ApolloColor.new(self.setOpacity..self.setColorNeutralName),
		NeutralBar = ApolloColor.new(self.setOpacity..self.setColorNeutralBar),
		NeutralBarCutoff = ApolloColor.new(self.setOpacity..self.setColorNeutralBarCutoff),
		
		FriendlyName = ApolloColor.new(self.setOpacity..self.setColorFriendlyName),
		FriendlyBar = ApolloColor.new(self.setOpacity..self.setColorFriendlyBar),
		FriendlyBarCutoff = ApolloColor.new(self.setOpacity..self.setColorFriendlyBarCutoff),
	
		FriendlyPlayerName = ApolloColor.new(self.setOpacity..self.setColorFriendlyPlayerName),
		FriendlyPlayerNameUnflagged = ApolloColor.new(self.setOpacity..self.setColorFriendlyPlayerNameUnflagged),		
		FriendlyPlayerBar = ApolloColor.new(self.setOpacity..self.setColorFriendlyPlayerBar),
		FriendlyPlayerBarCutoff = ApolloColor.new(self.setOpacity..self.setColorFriendlyPlayerBarCutoff),
	
		PartyName = ApolloColor.new(self.setOpacity..self.setColorPartyName),
		GuildName = ApolloColor.new(self.setOpacity..self.setColorGuildName),
	
		HostileName = ApolloColor.new(self.setOpacity..self.setColorHostileName),
		HostileBar = ApolloColor.new(self.setOpacity..self.setColorHostileBar),
		HostileBarCutoff = ApolloColor.new(self.setOpacity..self.setColorHostileBarCutoff),
	
		HostilePlayerName = ApolloColor.new(self.setOpacity..self.setColorHostilePlayerName),
		HostilePlayerNameUnflagged = ApolloColor.new(self.setOpacity..self.setColorHostilePlayerNameUnflagged),		
		HostilePlayerBar = ApolloColor.new(self.setOpacity..self.setColorHostilePlayerBar),
		HostilePlayerBarCutoff = ApolloColor.new(self.setOpacity..self.setColorHostilePlayerBarCutoff),	
		
		WarriorBar = ApolloColor.new(self.setOpacity..self.setColorWarriorBar),
		StalkerBar = ApolloColor.new(self.setOpacity..self.setColorStalkerBar),
		EngineerBar = ApolloColor.new(self.setOpacity..self.setColorEngineerBar),
		EsperBar = ApolloColor.new(self.setOpacity..self.setColorEsperBar),
		SpellslingerBar = ApolloColor.new(self.setOpacity..self.setColorSpellslingerBar),
		MedicBar = ApolloColor.new(self.setOpacity..self.setColorMedicBar),
		
		ThreatOutline = ApolloColor.new(self.setOpacity..self.setColorThreatOutline),
		CleanseOutline = ApolloColor.new(self.setOpacity..self.setColorCleanseOutline),
		Harvestables = ApolloColor.new(self.setOpacity..self.setColorHarvestables),
		
		FocusBar = ApolloColor.new(self.setOpacity..self.setColorFocusBar),
		FocusBarText = ApolloColor.new(self.setOpacity..self.setColorFocusBarText),
		ResourceBar = ApolloColor.new(self.setOpacity..self.setColorResourceBar),
		ResourceBarText = ApolloColor.new(self.setOpacity..self.setColorResourceBarText),
		SprintBar = ApolloColor.new(self.setOpacity..self.setColorSprintBar),
		SprintBarText = ApolloColor.new(self.setOpacity..self.setColorSprintBarText),
		DashBar = ApolloColor.new(self.setOpacity..self.setColorDashBar),
		DashBarText = ApolloColor.new(self.setOpacity..self.setColorDashBarText),
	}
end

-- returns whether a unit should have nameplate at all
function OptiPlates:ShouldHaveNameplate(tNameplate)
	local unitOwner = tNameplate.unitOwner
	
	if tNameplate.bIsTarget or tNameplate.bInCluster or unitOwner == self.focusUnit then return true end
	if self.setMouseOverEnabled and unitOwner == self.mouseOverUnit and tNameplate.bMouseOverSettingsEnabled then return true end
	if tNameplate.bIsImportantNPC or tNameplate.bIsVendorNPC then
		return true
	end
	if tNameplate.bFriendlyNameplateHealthVisibility then return true end
	if tNameplate.bIsMount or tNameplate.bIsPinataLoot then
		return false
	end	
	
	if tNameplate.bIsObjective and self.setShowObjectivesAlways then
		return true
	end
	
	if tNameplate.bIsHarvest then
		if not self.setHarvestablesEnabled then return false end
		if not (tNameplate.bIsOre or tNameplate.bIsRelic or tNameplate.bIsTree or tNameplate.bIsPlant) then
			tNameplate.bIsOre = unitOwner:GetHarvestRequiredTradeskillName() == "Mining"
			tNameplate.bIsRelic	= unitOwner:GetHarvestRequiredTradeskillName() == "Relic Hunter"
			tNameplate.bIsTree = unitOwner:GetHarvestRequiredTradeskillName() == "Survivalist"
			tNameplate.bIsPlant = unitOwner:GetHarvestRequiredTradeskillName() == "Farmer"
		end
		if tNameplate.bIsOre and self.setHarvestablesMiningEnabled then return true end
		if tNameplate.bIsRelic and self.setHarvestablesRelicEnabled then return true end
		if tNameplate.bIsTree and self.setHarvestablesTreeEnabled then return true end
		if tNameplate.bIsPlant and self.setHarvestablesPlantEnabled then return true end
		return false
	end
	if tNameplate.bIsDatacube and self.setHarvestablesDatacubeEnabled then return true end
	
	if tNameplate.bIsWeapon then return true end
	local bQuestGiver = (
		(tNameplate.bIsQuestFinisher and self.setShowQuestIconReward) or
		(tNameplate.bIsQuestGiverNew and self.setShowQuestIconNew) or
		(tNameplate.bIsQuestGiverNewDaily and self.setShowQuestIconNewDaily) or
		(tNameplate.bIsQuestReceiver and self.setShowQuestIconReceiving) or
		(tNameplate.bIsQuestGiverNewGrey and self.setShowQuestIconNewGrey)
	)
	if bQuestGiver then return true end
	if self.setShowInteractableAlways and tNameplate.bHasActivationState then return true end
	if tNameplate.bIsPlayer and 
		(
			self.setRoleIconAlwaysShowNameplate and (
			(tNameplate.bIsTank and self.setStyleRoleIconShowTank) or
			(tNameplate.bIsHealer and self.setStyleRoleIconShowHealer) or 
			(tNameplate.bIsDPS and self.setStyleRoleIconShowDPS)
			)
		)
	then
		return true
	end
	if not tNameplate.tUnitTypeSettings[unitOwner:IsInCombat() and "C" or "OOC"]["Nameplate"] then return false end
	if tNameplate.bIsPlayer then return true end	
	if self.setHideUnimportantNPCNameplates and tNameplate.bIsUnimportantNPC then return false end
	if unitOwner:ShouldShowNamePlate() or unitOwner:GetHealth() ~= nil then
		return true
	end
	
	--if tNameplate.bIsCollectible and tNameplate.bIsObjective then
	--	return true
	--end

	
	if unitOwner:GetHealth() == nil then
		return false
	end
	return not (
		unitOwner:GetType() == "Simple"
		)

end

function OptiPlates:IsOccluded(tNameplate)
	if not self.setEnableOcclusion then return false end
	if tNameplate == nil or tNameplate.wndNameplate == nil then return false end
	-- some mounts occlude the player, so check if the mount is occluded in that case
	if tNameplate.unitOwner:IsMounted() then
		local bMountOccluded = tNameplate.unitOwner:GetUnitMount():IsOccluded() 
		if bMountOccluded then
			return true
		else
			local tScreenPos = GameLib.GetUnitScreenPosition(tNameplate.unitOwner:GetUnitMount())
			if tScreenPos.bOnScreen then
				return false
			else
				return true
			end
		end
	else
		return tNameplate.wndNameplate:IsOccluded() or tNameplate.unitOwner:IsOccluded()
	end
end

function OptiPlates:HelperUpdateSPPixie(tNameplate, bShow, sText, bUpdatePosition)
	if tNameplate == nil or tNameplate.wnd == nil then return end
	
	local wndPixieContainer = tNameplate.wnd.Plates
	if not bShow then
		if tNameplate.nSPTextPixie ~= nil then
			wndPixieContainer:DestroyPixie(tNameplate.nSPTextPixie)
			tNameplate.nSPTextPixie = nil
		end
		return
	end
	local tTempPixie = tNameplate.tPixieTables.SP
	local nTextWidth = nil
	local bTextWidthChanged = false
	if sText ~= nil then 
		nTextWidth = Apollo.GetTextWidth(tSPTextPixie.strFont, tostring(sText))+1
		tNameplate.nSPTextWidth = nTextWidth
		if tNameplate.nSPTextWidth == nil or tNameplate.nSPTextWidth ~= nTextWidth then
			bTextWidthChanged = true
		end
	end
	
	-- positioning logic below, must work in all cases
	if bUpdatePosition or tNameplate.nSPTextPixie == nil or tTempPixie == nil or bTextWidthChanged then
		if nTextWidth == nil then nTextWidth = tNameplate.nSPTextWidth end
		if self.bStyleFlat then -- Flat
			local nRightEdge = -self.nStyleHalfWidth-3
			if tTempPixie == nil then			
				tTempPixie = {}
				tTempPixie.loc = {
					fPoints = { 0.5, 0, 0.5, 0},
					nOffsets = { nRightEdge-nTextWidth, 0, nRightEdge, 15 }
				}
			else
				tTempPixie.loc.nOffsets[1] = nRightEdge-nTextWidth
				tTempPixie.loc.nOffsets[3] = nRightEdge
			end
		else -- SAO
			local nRightEdge = -self.nStyleHalfWidth-5
			if tTempPixie == nil then
				tTempPixie = {}
				tTempPixie.loc = {
					fPoints = { 0.5, 0, 0.5, 0},
					nOffsets = { nRightEdge-nTextWidth, 0, nRightEdge, 20 }
				}
			else
				tTempPixie.loc.nOffsets[1] = nRightEdge-nTextWidth
				tTempPixie.loc.nOffsets[3] = nRightEdge
			end
			
		end
		tTempPixie.crText = self.arColorObjects.ShieldBar
		tTempPixie.strFont = tSPTextPixie.strFont
		tTempPixie.flagsText = tSPTextPixie.flagsText
	end
	if sText ~= nil then
		tTempPixie.strText = tostring(sText)
	end
	tNameplate.tPixieTables.SP = tTempPixie
	if tNameplate.nSPTextPixie == nil or tNameplate.nSPTextPixie == -1 then
		tNameplate.nSPTextPixie = wndPixieContainer:AddPixie(tTempPixie)
	else
		wndPixieContainer:UpdatePixie(tNameplate.nSPTextPixie,tTempPixie)
	end
end

function OptiPlates:HelperUpdateHPPixie(tNameplate, bShow, sText, bUpdatePosition)
	if tNameplate == nil or tNameplate.wnd == nil then return end
	
	local wndPixieContainer = tNameplate.wnd.Plates
	if not bShow then
		if tNameplate.nHPTextPixie ~= nil then
			wndPixieContainer:DestroyPixie(tNameplate.nHPTextPixie)
			tNameplate.nHPTextPixie = nil
		end
		return
	end
	local tTempPixie = tNameplate.tPixieTables.HP
	-- positioning logic below, must work in all cases
	local nTextWidth = nil
	local bTextWidthChanged = false
	if sText ~= nil then 
		nTextWidth = Apollo.GetTextWidth(tHPTextPixie.strFont, tostring(sText))+1
		tNameplate.nHPTextWidth = nTextWidth
		if tNameplate.nHPTextWidth == nil or tNameplate.nHPTextWidth ~= nTextWidth then
			bTextWidthChanged = true
		end
	end
	if bUpdatePosition or tNameplate.nHPTextPixie == nil or tTempPixie == nil or bTextWidthChanged then
		if nTextWidth == nil then nTextWidth = tNameplate.nHPTextWidth end
		if self.bStyleFlat then -- Flat
			local nRightEdge = -self.nStyleHalfWidth-3
			if tTempPixie == nil then
				tTempPixie = {}
				tTempPixie.loc = {
					fPoints = { 0.5, 0, 0.5, 0},
					nOffsets = { nRightEdge-nTextWidth, 10, nRightEdge, 25 }
				}
			else
				tTempPixie.loc.nOffsets[1] = nRightEdge-nTextWidth
				tTempPixie.loc.nOffsets[3] = nRightEdge
			end			
		else -- SAO
			--local nScalingOffsetX = math.floor((self.setSAOSize - 100)*0.05)
			local nScalingOffsetX = 0
			--local nScalingOffsetY = math.floor((self.setSAOSize - 100)*0.14)				
			local nScalingOffsetY = 7+self.setShieldbarHeight+math.floor(self.setHealthbarHeight*0.5)
			local nLeftEdge = nScalingOffsetX
			if tNameplate.nLevelPixie ~= nil and self.setLevelTextPosition == "Bars" then nLeftEdge = nLeftEdge + tNameplate.nLevelWidth+5 end
			local nTopEdge = nScalingOffsetY
			if tTempPixie == nil then
				tTempPixie = {}			
				tTempPixie.loc = {
					fPoints = { 0.5, 0, 0.5, 0},
					nOffsets = { nLeftEdge, nTopEdge, nLeftEdge+nTextWidth, nTopEdge+14 }
				}			
			else
				tTempPixie.loc.nOffsets[1] = nLeftEdge
				tTempPixie.loc.nOffsets[2] = nTopEdge
				tTempPixie.loc.nOffsets[3] = nLeftEdge+nTextWidth
				tTempPixie.loc.nOffsets[4] = nTopEdge+14			
			end
		end
		tTempPixie.strFont = tHPTextPixie.strFont
		tTempPixie.flagsText = tHPTextPixie.flagsText
	end
	tTempPixie.crText = tNameplate.cDispositionColor
	if sText ~= nil then
		tTempPixie.strText = tostring(sText)
	end
	tNameplate.tPixieTables.HP = tTempPixie
	if tNameplate.nHPTextPixie == nil or tNameplate.nHPTextPixie == -1 then
		tNameplate.nHPTextPixie = wndPixieContainer:AddPixie(tTempPixie)
	else
		wndPixieContainer:UpdatePixie(tNameplate.nHPTextPixie,tTempPixie)
	end
end

function OptiPlates:HelperUpdateLevelPixie(tNameplate, bShow, sText, bUpdatePosition)
	if tNameplate == nil or tNameplate.wnd == nil then return end
	
	local wndPixieContainer = tNameplate.wnd.Plates
	if self.setLevelTextPosition == "NameLeft" then wndPixieContainer = tNameplate.wnd.NameContainer end
	if not bShow then
		if tNameplate.nLevelPixie ~= nil then
			wndPixieContainer:DestroyPixie(tNameplate.nLevelPixie)
			tNameplate.nLevelPixie = nil
		end
		return
	end
	local tTempPixie = tNameplate.tPixieTables.Level
	local nTextWidth = nil
	local bTextWidthChanged = false
	if sText ~= nil then 
		if self.setLevelTextPosition == "NameLeft" then
			nTextWidth = Apollo.GetTextWidth(self.setStyleUnitNameFont.name, tostring(sText))
		else
			nTextWidth = Apollo.GetTextWidth(tLevelPixie.strFont, tostring(sText))
		end
		nTextWidth = nTextWidth + 1
		tNameplate.nLevelWidth = nTextWidth
		if tNameplate.nLevelWidth == nil or tNameplate.nLevelWidth ~= nTextWidth then
			bTextWidthChanged = true
		end
	end
	-- positioning logic below, must work in all cases
	if bUpdatePosition or tNameplate.nLevelPixie == nil or tTempPixie == nil or bTextWidthChanged then
		if nTextWidth == nil then nTextWidth = tNameplate.nLevelWidth end	
		if self.setLevelTextPosition == "NameLeft" then
			local nNameWidth = tNameplate.nNameWidth or 0
			local nNameHeightOffset = tNameplate.nNameHeightOffset or 0
			local nLeftEdge = -nNameWidth * 0.5 - nTextWidth - 3
			if tTempPixie == nil then
				tTempPixie = { }
				tTempPixie.loc = {
					fPoints = { 0.5, 1, 0.5, 1},
					nOffsets = { nLeftEdge, -(self.setStyleUnitNameFont.size+6+nNameHeightOffset), nLeftEdge+nTextWidth, -nNameHeightOffset }
				}
			else
				tTempPixie.loc.nOffsets[1] = nLeftEdge
				tTempPixie.loc.nOffsets[2] = -(self.setStyleUnitNameFont.size+6+nNameHeightOffset)
				tTempPixie.loc.nOffsets[3] = nLeftEdge+nTextWidth
				tTempPixie.loc.nOffsets[4] = -nNameHeightOffset
			end	
			tTempPixie.strFont = self.setStyleUnitNameFont.name			
		else
			if self.bStyleFlat then -- Flat
				local nLeftEdge = self.nStyleHalfWidth+3
				if tTempPixie == nil then
					tTempPixie = { }
					tTempPixie.loc = {
						fPoints = { 0.5, 0, 0.5, 0},
						nOffsets = { nLeftEdge, 9, nLeftEdge+nTextWidth, 25 }
					}
				else
					tTempPixie.loc.nOffsets[1] = nLeftEdge
					tTempPixie.loc.nOffsets[3] = nLeftEdge+nTextWidth
				end
			else -- SAO
				--local nScalingOffsetX = math.floor((self.setSAOSize - 100)*0.05)
				--local nScalingOffsetY = math.floor((self.setSAOSize - 100)*0.14)				
				local nTopEdge = 5+self.setShieldbarHeight+math.floor(self.setHealthbarHeight*0.5)
				--local nTopEdge = 16+nScalingOffsetY
				if tTempPixie == nil then
					tTempPixie = { }			
					tTempPixie.loc = {
						fPoints = { 0.5, 0, 0.5, 0},
						nOffsets = { 0, nTopEdge, nTextWidth, nTopEdge+14 }
					}			
				else
					tTempPixie.loc.nOffsets[2] = nTopEdge
					tTempPixie.loc.nOffsets[3] = nTextWidth
					tTempPixie.loc.nOffsets[4] = nTopEdge+14
				end			
			end
			tTempPixie.strFont = tLevelPixie.strFont
		end
		tTempPixie.crText = ApolloColor.new("ff"..self.setColorLevelText)					
	end
	if sText ~= nil then
		tTempPixie.strText = tostring(sText)
	end
	tNameplate.tPixieTables.Level = tTempPixie
	if tNameplate.nLevelPixie == nil or tNameplate.nLevelPixie == -1 then
		tNameplate.nLevelPixie = wndPixieContainer:AddPixie(tTempPixie)
	else
		wndPixieContainer:UpdatePixie(tNameplate.nLevelPixie,tTempPixie)
	end
end

function OptiPlates:HelperUpdateInterruptArmorPixie(tNameplate, bShow, nArmor, bDegraded, bUpdatePosition)
	if tNameplate == nil or tNameplate.wnd == nil then return end

	local wndPixieContainer = tNameplate.wnd.Plates
	if not bShow then
		if tNameplate.nInterruptArmorPixie ~= nil then
			wndPixieContainer:DestroyPixie(tNameplate.nInterruptArmorPixie)
			if tNameplate.nInterruptArmorPixieInf ~= nil then
				wndPixieContainer:DestroyPixie(tNameplate.nInterruptArmorPixieInf)
			end
			tNameplate.nInterruptArmorPixieInf = nil
			tNameplate.nInterruptArmorPixie = nil
		end
		return
	end
	local tTempPixie = tNameplate.tPixieTables.InterruptArmor
	-- positioning logic below, must work in all cases
	if bUpdatePosition or tNameplate.nInterruptArmorPixie == nil or tTempPixie == nil then
		if self.bStyleFlat then -- Flat
			local nLeftEdge = self.nStyleHalfWidth
			if tNameplate.nLevelPixie ~= nil then nLeftEdge = nLeftEdge + tNameplate.nLevelWidth end
			if tTempPixie == nil then
				tTempPixie = { }
				tTempPixie.loc = {
					fPoints = { 0.5, 0, 0.5, 0},
					nOffsets = { nLeftEdge, 0, nLeftEdge+30, 30 }
				}
			else
				tTempPixie.loc.nOffsets[1] = nLeftEdge
				tTempPixie.loc.nOffsets[3] = nLeftEdge+30
			end
		else -- SAO
			local nLeftEdge = self.nStyleHalfWidth
			if tTempPixie == nil then
				tTempPixie = { }			
				tTempPixie.loc = {
					fPoints = { 0.5, 0, 0.5, 0},
					nOffsets = { nLeftEdge, 0, nLeftEdge + 30, 30 }
				}
			else
				tTempPixie.loc.nOffsets[1] = nLeftEdge
				tTempPixie.loc.nOffsets[3] = nLeftEdge+30			
			end
		end
		--tTempPixie.crText = tInterruptArmorPixie.crText
		tTempPixie.flagsText = tInterruptArmorPixie.flagsText
		tTempPixie.strFont = tInterruptArmorPixie.strFont
	end
	if nArmor >= 0 then
		tTempPixie.strText = tostring(nArmor)
	else
		tTempPixie.strText = ""
	end
	if bDegraded then
		tTempPixie.strSprite = "CRB_ActionBarFrameSprites:sprResourceBar_InterruptCircleRed"
	else
		tTempPixie.strSprite = "CRB_ActionBarFrameSprites:sprResourceBar_InterruptCircleOrange"
	end
	if tNameplate.nInterruptArmorPixie == nil or tNameplate.nInterruptArmorPixie == -1 then
		tNameplate.nInterruptArmorPixie = wndPixieContainer:AddPixie(tTempPixie)
	else
		wndPixieContainer:UpdatePixie(tNameplate.nInterruptArmorPixie,tTempPixie)
	end
	tNameplate.tPixieTables.InterruptArmor = tTempPixie
	if nArmor == -1 then
		local nLeft = tTempPixie.loc.nOffsets[1]
		tInterruptArmorInfPixie.loc.nOffsets[1] = nLeft +7
		tInterruptArmorInfPixie.loc.nOffsets[2] = 7
		tInterruptArmorInfPixie.loc.nOffsets[3] = nLeft +23
		tInterruptArmorInfPixie.loc.nOffsets[4] = 23
		if tNameplate.nInterruptArmorPixieInf == nil then
			tNameplate.nInterruptArmorPixieInf = wndPixieContainer:AddPixie(tInterruptArmorInfPixie)
		else
			wndPixieContainer:UpdatePixie(tNameplate.nInterruptArmorPixieInf,tInterruptArmorInfPixie)
		end
	elseif tNameplate.nInterruptArmorPixieInf ~= nil then
		wndPixieContainer:DestroyPixie(tNameplate.nInterruptArmorPixieInf)
		tNameplate.nInterruptArmorPixieInf = nil		
	end
end

function OptiPlates:HelperUpdateCleanseIconPixie(tNameplate, bShow, strSprite, bUpdatePosition)
	if tNameplate == nil or tNameplate.wnd == nil then return end
	
	local wndPixieContainer = tNameplate.wnd.Plates
	if not bShow then
		if tNameplate.nCleanseIconPixie ~= nil then
			wndPixieContainer:DestroyPixie(tNameplate.nCleanseIconPixie)
			tNameplate.nCleanseIconPixie = nil
		end
		return
	end
	local tTempPixie = tNameplate.tPixieTables.CleanseIcon
	-- positioning logic below, must work in all cases
	if bUpdatePosition or tNameplate.nCleanseIconPixie == nil or tTempPixie == nil then
		local nIconSize = self.setStyleCleanseIconSize
		local nLeftEdge = self.nStyleHalfWidth
		if self.bStyleFlat then -- Flat
			if self.setStyleCleanseIconPosition == "Left" then
				if tNameplate.nSPTextPixie ~= nil or tNameplate.nHPTextPixie ~= nil then
					nLeftEdge = nLeftEdge + math.max(tNameplate.nSPTextPixie and tNameplate.nSPTextWidth or 0,
													 tNameplate.nHPTextPixie and tNameplate.nHPTextWidth or 0
													)		
				end
				if tTempPixie == nil then
					tTempPixie = { }
					tTempPixie.loc = {
						fPoints = { 0.5, 0, 0.5, 0},
						nOffsets = { -nLeftEdge-nIconSize, 0,-nLeftEdge,nIconSize }
					}	
				else
					tTempPixie.loc.nOffsets[1] = -nLeftEdge-nIconSize
					tTempPixie.loc.nOffsets[3] = -nLeftEdge
					tTempPixie.loc.nOffsets[4] = nIconSize
				end
			else
				if tNameplate.nInterruptArmorPixie ~= nil then nLeftEdge = nLeftEdge + 25 end
				if tNameplate.nLevelPixie ~= nil then nLeftEdge = nLeftEdge + tNameplate.nLevelWidth end
				if tTempPixie == nil then
					tTempPixie = { }
					tTempPixie.loc = {
						fPoints = { 0.5, 0, 0.5, 0},
						nOffsets = { nLeftEdge, 0,nLeftEdge+nIconSize,nIconSize }
					}				
				else
					tTempPixie.loc.nOffsets[1] = nLeftEdge
					tTempPixie.loc.nOffsets[3] = nLeftEdge+nIconSize
					tTempPixie.loc.nOffsets[4] = nIconSize
				end				
			end
		else -- SAO
			if self.setStyleCleanseIconPosition == "Left" then
				if tNameplate.nSPTextPixie ~= nil then
					nLeftEdge = nLeftEdge + tNameplate.nSPTextWidth
				end
				if tTempPixie == nil then
					tTempPixie = { }
					tTempPixie.loc = {
						fPoints = { 0.5, 0, 0.5, 0},
						nOffsets = { -nLeftEdge-nIconSize, 0,-nLeftEdge,nIconSize }
					}
				else
					tTempPixie.loc.nOffsets[1] = -nLeftEdge-nIconSize
					tTempPixie.loc.nOffsets[3] = -nLeftEdge
					tTempPixie.loc.nOffsets[4] = nIconSize		
				end
			else
				if tNameplate.nInterruptArmorPixie ~= nil then nLeftEdge = nLeftEdge + 20 end
				if tTempPixie == nil then
					tTempPixie = { }				
					tTempPixie.loc = {
						fPoints = { 0.5, 0, 0.5, 0},
						nOffsets = { nLeftEdge, 0,nLeftEdge+nIconSize,nIconSize }
					}	
				else
					tTempPixie.loc.nOffsets[1] = nLeftEdge
					tTempPixie.loc.nOffsets[3] = nLeftEdge+nIconSize
					tTempPixie.loc.nOffsets[4] = nIconSize
				end								
			end
		end
		tTempPixie.strSprite = tCleanseIconPixie.strSprite
		tTempPixie.strFont = tCleanseIconPixie.strFont
	end
	if strSprite ~= nil then
		tTempPixie.strSprite = strSprite
	end
	tNameplate.tPixieTables.CleanseIcon = tTempPixie
	if tNameplate.nCleanseIconPixie == nil or tNameplate.nCleanseIconPixie == -1 then
		tNameplate.nCleanseIconPixie = wndPixieContainer:AddPixie(tTempPixie)
	else
		wndPixieContainer:UpdatePixie(tNameplate.nCleanseIconPixie,tTempPixie)
	end
end

function OptiPlates:HelperUpdateTargetIconPixie(tNameplate, bShow, bUpdatePosition)
	if tNameplate == nil or tNameplate.wnd == nil then return end
	
	local wndPixieContainer = tNameplate.wnd.ToprowContainer
	if not bShow then
		if tNameplate.nTargetPixie ~= nil and tNameplate.nTargetPixie ~= -1 then
			wndPixieContainer:DestroyPixie(tNameplate.nTargetPixie)
		end
		tNameplate.nTargetPixie = nil		
		return
	end
	local eDisposition = tNameplate.eDisposition
	local strSprite = nil
	local unitOwner = tNameplate.unitOwner
	local tTempPixie = tNameplate.tPixieTables.Target
	if tNameplate.nTargetPixie == -1 then tNameplate.nTargetPixie = nil end
	if tNameplate.bIsTarget then
		if self.setShowNormalTargetIndicator then
			if eDisposition == tDisposition.Hostile then 
				strSprite = self.setHostileTargetIndicatorIcon
			elseif unitOwner:IsInYourGroup() then
				strSprite = self.setPartyTargetIndicatorIcon
			elseif eDisposition == tDisposition.Neutral then
				strSprite = self.setNeutralTargetIndicatorIcon
			else
				strSprite = self.setFriendlyTargetIndicatorIcon
			end		
		end
	elseif unitOwner == self.focusUnit and self.setShowFocusTargetIndicator then
		if eDisposition == tDisposition.Hostile then 
			strSprite = self.setHostileAltTargetIndicatorIcon
		elseif eDisposition == tDisposition.Neutral then
			strSprite = self.setNeutralAltTargetIndicatorIcon
		else
			strSprite = self.setFriendlyAltTargetIndicatorIcon
		end		
	elseif tNameplate.bInCluster and self.setShowTargetClusterIcons then
		if eDisposition == tDisposition.Hostile then 
			strSprite = self.setHostileClusterIndicatorIcon
		elseif eDisposition == tDisposition.Neutral then
			strSprite = self.setNeutralClusterIndicatorIcon
		else
			strSprite = self.setFriendlyClusterIndicatorIcon
		end		
	end
	-- positioning logic below, must work in all cases
	if bUpdatePosition or tNameplate.nTargetPixie == nil or tTempPixie == nil then
		local nIconSize = self.setStyleTargetedIconSize
		local nLeftEdge = -nIconSize*0.5
		local nVertOffset = 0
		if self.setStyleHeadingArrowBelowTargetMarker and tNameplate.nHeadingPixie ~= nil then
			nVertOffset = self.setStyleHeadingArrowSize
			if tNameplate.nHeadingTextPixie ~= nil then
				nVertOffset = nVertOffset + self.setStyleHeadingIndicatorTextFont.size+6
			end
		end
		if tNameplate.nQuestIconPixie ~= nil then
			nVertOffset = self.setStyleQuestIconSize
		end
		if tTempPixie == nil then
			tTempPixie = { }
			tTempPixie.loc = {
				fPoints = { 0.5, 1, 0.5, 1},
				nOffsets = { nLeftEdge, -nIconSize-nVertOffset,nLeftEdge+nIconSize,-nVertOffset }
			}				
		else
			tTempPixie.loc.nOffsets[1] = nLeftEdge
			tTempPixie.loc.nOffsets[2] = -nIconSize-nVertOffset
			tTempPixie.loc.nOffsets[3] = nLeftEdge+nIconSize
			tTempPixie.loc.nOffsets[4] = -nVertOffset
		end
	end
	if strSprite ~= nil then
		tTempPixie.strSprite = strSprite
	end
	tNameplate.tPixieTables.Target = tTempPixie
	if tNameplate.nTargetPixie == nil or tNameplate.nTargetPixie == -1 then
		tNameplate.nTargetPixie = wndPixieContainer:AddPixie(tTempPixie)
	else
		wndPixieContainer:UpdatePixie(tNameplate.nTargetPixie,tTempPixie)
	end
end

function OptiPlates:HelperUpdateClassIconPixie(tNameplate, bShow, bUpdatePosition)
	if tNameplate == nil or tNameplate.wnd == nil then return end
	
	local wndPixieContainer = tNameplate.wnd.ToprowContainer
	if self.setClassIconPosition ~= "Top" or tNameplate.bIsHarvest or tNameplate.bIsDatacube or tNameplate.bIsNPC then
		wndPixieContainer = tNameplate.wnd.NameContainer
	end
	if not bShow then
		if tNameplate.nClassIconPixie ~= nil and tNameplate.nClassIconPixie ~= -1 then
			wndPixieContainer:DestroyPixie(tNameplate.nClassIconPixie)
		end
		tNameplate.nClassIconPixie = nil
		return
	end
	local unitOwner = tNameplate.unitOwner
	if tNameplate.nClassIconPixie == -1 then tNameplate.nClassIconPixie = nil end
	-- positioning logic below, must work in all cases
	local tTempPixie = tNameplate.tPixieTables.ClassIcon
	if bUpdatePosition or tNameplate.nClassIconPixie == nil or tTempPixie == nil then
		local nIconSize = self.setStyleClassIconSize
		if wndPixieContainer == tNameplate.wnd.ToprowContainer then
			local nLeftEdge = -nIconSize
			local nMiddleIconHalfWidth = 0
			if self.setStyleHeadingArrowBelowTargetMarker then
				if tNameplate.nHeadingPixie ~= nil then nMiddleIconHalfWidth = tNameplate.nHeadingPixieWidth * 0.5
				elseif tNameplate.nTargetPixie ~= nil then nMiddleIconHalfWidth = self.setStyleTargetedIconSize*0.5
				end
			else
				if tNameplate.nTargetPixie ~= nil then nMiddleIconHalfWidth = self.setStyleTargetedIconSize*0.5 end
			end
			if nMiddleIconHalfWidth > 0 then nLeftEdge = nLeftEdge - nMiddleIconHalfWidth
			elseif tNameplate.nRoleIconPixie == nil then nLeftEdge = nLeftEdge*0.5 end

			if tTempPixie == nil then
				tTempPixie = {}
				tTempPixie.loc = {
					fPoints = { 0.5, 1, 0.5, 1},
					nOffsets = { nLeftEdge, -nIconSize,nLeftEdge+nIconSize,0 }
				}
			else
				tTempPixie.loc.nOffsets[1] = nLeftEdge
				tTempPixie.loc.nOffsets[2] = -nIconSize
				tTempPixie.loc.nOffsets[3] = nLeftEdge+nIconSize			
			end
		else
			local nNameWidth = tNameplate.nNameWidth or 0
			local nLeftEdge = -nNameWidth*0.5-nIconSize
			if tNameplate.nLevelPixie ~= nil and self.setLevelTextPosition == "NameLeft" then nLeftEdge = nLeftEdge - tNameplate.nLevelWidth - 5 end
			if tTempPixie == nil then
				tTempPixie = {}
				tTempPixie.loc = {
					fPoints = { 0.5, 1, 0.5, 1},
					nOffsets = { nLeftEdge, -nIconSize,nLeftEdge+nIconSize,0 }
				}
			else
				tTempPixie.loc.nOffsets[1] = nLeftEdge
				tTempPixie.loc.nOffsets[2] = -nIconSize
				tTempPixie.loc.nOffsets[3] = nLeftEdge+nIconSize			
			end			
		end
		local colorBGColor = nil
		local enumClasses = GameLib.CodeEnumClass
		if tNameplate.bIsPlayer then
			local nClass = (unitOwner:GetClassId() or -1)		
			if nClass == enumClasses.Warrior and self.setStyleClassIconShowWarrior then
				colorBGColor = self.arColorObjects.WarriorBar
			elseif nClass == enumClasses.Engineer and self.setStyleClassIconShowEngineer then
				colorBGColor = self.arColorObjects.EngineerBar
			elseif nClass == enumClasses.Esper and self.setStyleClassIconShowEsper then
				colorBGColor = self.arColorObjects.EsperBar
			elseif nClass == enumClasses.Medic and self.setStyleClassIconShowMedic then
				colorBGColor = self.arColorObjects.MedicBar
			elseif nClass == enumClasses.Stalker and self.setStyleClassIconShowStalker then
				colorBGColor = self.arColorObjects.StalkerBar
			elseif nClass == enumClasses.Spellslinger and self.setStyleClassIconShowSpellslinger then
				colorBGColor = self.arColorObjects.SpellslingerBar
			end
			strSprite = arClassToClassIcon[nClass]
		elseif tNameplate.bIsHarvest then
			colorBGColor = "ffffffff"
			local sNodeTradeskill = unitOwner:GetHarvestRequiredTradeskillName()
			if sNodeTradeskill == "Mining" then
				strSprite = "IconSprites:Icon_MapNode_Map_Node_Mining"
			elseif sNodeTradeskill == "Farmer" then
				strSprite = "IconSprites:Icon_MapNode_Map_Node_Plant"
			elseif sNodeTradeskill == "Relic Hunter" then
				strSprite = "IconSprites:Icon_MapNode_Map_Node_Relic"
			elseif sNodeTradeskill == "Survivalist" then
				strSprite = "IconSprites:Icon_MapNode_Map_Node_Tree"
			end
		elseif tNameplate.bIsDatacube then
			strSprite = "Icon_Windows32_UI_CRB_InterfaceMenu_Lore"
			colorBGColor = "ffffffff"
		elseif tNameplate.bIsNPC then
			--strSprite = "HUD_TargetFrame:spr_TargetFrame_ClassIcon_Elite"
			strSprite = "BK3:UI_BK3_StoryPanelAlert_Icon"
			if unitOwner:IsRare() then
				colorBGColor = "ffa5e9f7"
			elseif unitOwner:GetRank() > 4 or  unitOwner:GetEliteness() > 0 then -- 4 is superior, which is kind of solo content still
				colorBGColor = "ffff4444"
			end
		end
		tTempPixie.cr = colorBGColor
		tTempPixie.strSprite = strSprite
	end
	tNameplate.tPixieTables.ClassIcon = tTempPixie
	if tNameplate.nClassIconPixie == nil or tNameplate.nClassIconPixie == -1 then
		tNameplate.nClassIconPixie = wndPixieContainer:AddPixie(tTempPixie)
	else
		wndPixieContainer:UpdatePixie(tNameplate.nClassIconPixie,tTempPixie)
	end
end

function OptiPlates:HelperUpdateRoleIconPixie(tNameplate, bShow, bUpdatePosition)
	if tNameplate == nil or tNameplate.wnd == nil then return end
	
	local wndPixieContainer = tNameplate.wnd.ToprowContainer
	if not bShow then
		if tNameplate.nRoleIconPixie ~= nil and tNameplate.nRoleIconPixie ~= -1 then
			wndPixieContainer:DestroyPixie(tNameplate.nRoleIconPixie)
		end
		tNameplate.nRoleIconPixie = nil		
		return
	end
	local unitOwner = tNameplate.unitOwner
	if tNameplate.nRoleIconPixie == -1 then tNameplate.nRoleIconPixie = nil end
	-- positioning logic below, must work in all cases
	local tTempPixie = tNameplate.tPixieTables.RoleIcon
	if bUpdatePosition or tNameplate.nRoleIconPixie == nil or tTempPixie == nil then
		local nIconSize = self.setStyleRoleIconSize
		local nLeftEdge = 0
		local nMiddleIconHalfWidth = 0
		if self.setStyleHeadingArrowBelowTargetMarker then
			if tNameplate.nHeadingPixie ~= nil then nMiddleIconHalfWidth = tNameplate.nHeadingPixieWidth * 0.5
			elseif tNameplate.nTargetPixie ~= nil then nMiddleIconHalfWidth = self.setStyleTargetedIconSize*0.5
			end
		else
			if tNameplate.nTargetPixie ~= nil then nMiddleIconHalfWidth = self.setStyleTargetedIconSize*0.5 end
		end		
		if nMiddleIconHalfWidth > 0 then nLeftEdge = nLeftEdge + nMiddleIconHalfWidth end
		if tTempPixie == nil then 
		tTempPixie = {}
			tTempPixie.loc = {
				fPoints = { 0.5, 1, 0.5, 1},
				nOffsets = { nLeftEdge, -nIconSize,nLeftEdge+nIconSize,0 }
			}
		else 
			tTempPixie.loc.nOffsets[1] = nLeftEdge
			tTempPixie.loc.nOffsets[2] = -nIconSize
			tTempPixie.loc.nOffsets[3] = nLeftEdge+nIconSize
		end
		local colorBGColor = nil
		local strSprite = nil
		if tNameplate.bIsHealer then
			strSprite = arRoleIcons["healer"]
			colorBGColor = ApolloColor.new("xkcdLightGreen")
		elseif tNameplate.bIsTank then
			strSprite = arRoleIcons["tank"]
			colorBGColor = ApolloColor.new("xkcdSoftPurple")
		else
			strSprite = arRoleIcons["dps"]
			colorBGColor = ApolloColor.new("xkcdLightBlue")
		end		
		tTempPixie.cr = colorBGColor
		tTempPixie.strSprite = strSprite
	end
	tNameplate.tPixieTables.RoleIcon = tTempPixie
	if tNameplate.nRoleIconPixie == nil or tNameplate.nRoleIconPixie == -1 then
		tNameplate.nRoleIconPixie = wndPixieContainer:AddPixie(tTempPixie)
	else
		wndPixieContainer:UpdatePixie(tNameplate.nRoleIconPixie,tTempPixie)
	end
end

function OptiPlates:HelperUpdateHeadingPixie(tNameplate, bShow, bUpdatePosition)
	if tNameplate == nil or tNameplate.wnd == nil then return end
	
	local wndPixieContainer = tNameplate.wnd.ToprowContainer
	if not bShow then
		if tNameplate.nHeadingPixie ~= nil and tNameplate.nHeadingPixie ~= -1 then
			wndPixieContainer:DestroyPixie(tNameplate.nHeadingPixie)
		end
		tNameplate.nHeadingPixie = nil
		if tNameplate.nHeadingTextPixie ~= nil then
			wndPixieContainer:DestroyPixie(tNameplate.nHeadingTextPixie)
		end
		tNameplate.nHeadingTextPixie = nil
		return
	end
	local bShowTextPixie = self.setStyleHeadingShowText
	local nTextPixieHeight = 0
	if bShowTextPixie then nTextPixieHeight = self.setStyleHeadingIndicatorTextFont.size+6 end
	local unitOwner = tNameplate.unitOwner
	if tNameplate.nHeadingPixie == -1 then tNameplate.nHeadingPixie = nil end
	-- positioning logic below, must work in all cases
	local tTempPixie = tNameplate.tPixieTables.Heading
	if bUpdatePosition or tNameplate.nHeadingPixie == nil or tTempPixie == nil then
		local nIconSize = self.setStyleHeadingArrowSize
		local nLeftEdge = -nIconSize
		if (self.setStyleHeadingArrowBelowTargetMarker) then nLeftEdge = nLeftEdge*0.5 end
		if (not self.setStyleHeadingArrowBelowTargetMarker) then
			if tNameplate.nTargetPixie ~= nil then nLeftEdge = nLeftEdge - self.setStyleTargetedIconSize*0.5 end
			if tNameplate.nClassIconPixie ~= nil and self.setClassIconPosition == "Top" then nLeftEdge = nLeftEdge - self.setStyleClassIconSize end
		end
		if tTempPixie == nil then
			tTempPixie = {}
			tTempPixie.loc = {
				fPoints = { 0.5, 1, 0.5, 1},
				nOffsets = { nLeftEdge, -nIconSize-nTextPixieHeight,nLeftEdge+nIconSize, -nTextPixieHeight}
			}
		else
			tTempPixie.loc.nOffsets[1] = nLeftEdge
			tTempPixie.loc.nOffsets[2] = -nIconSize-nTextPixieHeight
			tTempPixie.loc.nOffsets[3] = nLeftEdge+nIconSize			
			tTempPixie.loc.nOffsets[4] = -nTextPixieHeight
		end
		tTempPixie.strSprite = tHeadingPixie.strSprite
		tTempPixie.cr = ApolloColor.new("ff"..self.setColorHeadingIndicatorArrow)
	end
	local tCircleTable = self.circleUnitVectorLookupTable
	local tUnitPos = self.playerUnit:GetPosition()
	local unitVector = Vector3.New(tUnitPos.x,tUnitPos.y,tUnitPos.z)
	local top = 100000
	local degrees = nil
	for idx, vec in pairs(tCircleTable) do
		local loc = GameLib.WorldLocToScreenPoint(vec+unitVector)
		if loc.y < top then 
			top = loc.y 
			degrees = idx
		end
	end
	tTempPixie.fRotation = (degrees + 180)/180*math.pi
	tNameplate.tPixieTables.Heading = tTempPixie
	if tNameplate.nHeadingPixie == nil or tNameplate.nHeadingPixie == -1 then
		tNameplate.nHeadingPixie = wndPixieContainer:AddPixie(tTempPixie)
	else
		wndPixieContainer:UpdatePixie(tNameplate.nHeadingPixie,tTempPixie)
	end
	
	if bShowTextPixie then
		if tNameplate.nHeadingTextPixie == -1 then tNameplate.nHeadingTextPixie = nil end
		local tTextPixie = tNameplate.tPixieTables.HeadingText
		local fSegment = degrees/360 *16
		local nSegment = math.floor(fSegment)
		local strText = tHeadingSegmentText[nSegment+1]
		--local nWidth = Apollo.GetTextWidth(tHeadingTextPixie.strFont,strText)
		local nIconSize = tNameplate.nHeadingPixieWidth
		local nLeftEdge = -nIconSize
		if (self.setStyleHeadingArrowBelowTargetMarker) then nLeftEdge = nLeftEdge*0.5 end
		if (not self.setStyleHeadingArrowBelowTargetMarker) then
			if tNameplate.nTargetPixie ~= nil then nLeftEdge = nLeftEdge - self.setStyleTargetedIconSize*0.5 end
			if tNameplate.nClassIconPixie ~= nil and self.setClassIconPosition == "Top" then nLeftEdge = nLeftEdge - self.setStyleClassIconSize end
		end
		if tTextPixie == nil then
			tTextPixie = { }
			tTextPixie.loc = {
				fPoints = { 0.5, 1, 0.5, 1},
				nOffsets = { nLeftEdge, -nTextPixieHeight,nLeftEdge+nIconSize,0 }
			}
			tTextPixie.strFont = tHeadingTextPixie.strFont
			tTextPixie.crText = ApolloColor.new("ff"..self.setColorHeadingIndicatorText)
			tTextPixie.strFont = self.setStyleHeadingIndicatorTextFont.name
			tTextPixie.flagsText = tHeadingTextPixie.flagsText
		else
			tTextPixie.loc.nOffsets[1] = nLeftEdge
			tTextPixie.loc.nOffsets[2] = -nTextPixieHeight
			tTextPixie.loc.nOffsets[3] = nLeftEdge+nIconSize
		end		
		tTextPixie.strText = strText
		tNameplate.tPixieTables.HeadingText = tTextPixie
		if tNameplate.nHeadingTextPixie == nil or tNameplate.nHeadingTextPixie == -1 then
			tNameplate.nHeadingTextPixie = wndPixieContainer:AddPixie(tTextPixie)
		else
			wndPixieContainer:UpdatePixie(tNameplate.nHeadingTextPixie,tTextPixie)
		end		
	elseif tNameplate.nHeadingTextPixie ~= nil then
		wndPixieContainer:DestroyPixie(tNameplate.nHeadingTextPixie)
		tNameplate.nHeadingTextPixie = nil	
	end
	
end

function OptiPlates:HelperUpdateQuestIconPixie(tNameplate, bShow, bUpdatePosition)
	if tNameplate == nil or tNameplate.wnd == nil then return end
	
	local wndPixieContainer = tNameplate.wnd.ToprowContainer
	if not bShow then
		if tNameplate.nQuestIconPixie ~= nil and tNameplate.nQuestIconPixie ~= -1 then
			wndPixieContainer:DestroyPixie(tNameplate.nQuestIconPixie)
		end
		tNameplate.nQuestIconPixie = nil		
		return
	end
	local unitOwner = tNameplate.unitOwner
	if tNameplate.nQuestIconPixie == -1 then tNameplate.nQuestIconPixie = nil end
	-- positioning logic below, must work in all cases
	local tTempPixie = tNameplate.tPixieTables.QuestIcon
	if bUpdatePosition or tNameplate.nQuestIconPixie == nil or tTempPixie == nil then
		local nIconSize = self.setStyleQuestIconSize
		local nLeftEdge = -nIconSize * 0.5
		if tTempPixie == nil then
			tTempPixie = {}
			tTempPixie.loc = {
				fPoints = { 0.5, 1, 0.5, 1},
				nOffsets = { nLeftEdge, -nIconSize,nLeftEdge+nIconSize, 0}
			}
		else
			tTempPixie.loc.nOffsets[1] = nLeftEdge
			tTempPixie.loc.nOffsets[2] = -nIconSize
			tTempPixie.loc.nOffsets[3] = nLeftEdge+nIconSize			
		end
		if unitOwner:IsValid() then
			local tActivation = unitOwner:GetActivationState()
			local strSprite = nil
			if tActivation.QuestReward then
				strSprite = arQuestIcons.QuestReward
			elseif tActivation.QuestNew or tActivation.QuestNewMain then
				strSprite = arQuestIcons.QuestNew
			elseif tActivation.QuestNewDaily then
				strSprite = arQuestIcons.QuestNewDaily
			elseif tActivation.QuestReceiving then
				strSprite = arQuestIcons.QuestReceiving
			elseif tActivation.QuestNewSoon then
				strSprite = arQuestIcons.QuestNewGrey			
			end		
			tTempPixie.strSprite = strSprite
		end
	end
	tNameplate.tPixieTables.QuestIcon = tTempPixie
	if tNameplate.nQuestIconPixie == nil or tNameplate.nQuestIconPixie == -1 then
		tNameplate.nQuestIconPixie = wndPixieContainer:AddPixie(tTempPixie)
	else
		wndPixieContainer:UpdatePixie(tNameplate.nQuestIconPixie,tTempPixie)
	end
end

-- See if a nameplate should be displayed
function OptiPlates:HelperVerifyVisibilityOptions(tNameplate)
	if tNameplate == nil then return false end
	local unitPlayer = self.playerUnit
	local unitOwner = tNameplate.unitOwner
	local isImportant = tNameplate.bIsImportantNPC
	local isVendor = tNameplate.bIsVendorNPC
	if unitOwner == nil or not unitOwner:IsValid() or unitPlayer == nil or not unitPlayer:IsValid() then
		return false
	end
	local eDisposition = unitOwner:GetDispositionTo(unitPlayer)	
	
	-- Always show target nameplate, regardless of what it is
	if unitPlayer ~= nil then 
		local unitPlayerTarget = unitPlayer:GetTarget()
		if unitPlayerTarget ~= nil and unitPlayerTarget == unitOwner then
			return true
		end
	end
	
	if tNameplate.bHasActivationState and self.setShowInteractableAlways then return true end
	if self.setMouseOverEnabled and unitOwner == self.mouseOverUnit and tNameplate.bMouseOverSettingsEnabled then return true end
	-- PET PLATES

	if tNameplate.bGibbed then
		return false
	end

	local bShowNameplate = false
	tNameplate.eDisposition = eDisposition	
	if tNameplate.bIsPet then 
		if not unitOwner:GetUnitOwner() or unitOwner:GetUnitOwner() == nil then
			if eDisposition == tDisposition.Friendly then
				bShowNameplate = self.setShowFriendlyPets
			else
				bShowNameplate = self.setShowHostilePets
			end
		elseif unitOwner:GetUnitOwner():IsThePlayer() == true then
			bShowNameplate = self.setShowMyPets
		else
			if eDisposition == tDisposition.Friendly then
				bShowNameplate = self.setShowFriendlyPets
			else
				bShowNameplate = self.setShowHostilePets
			end
		end
	elseif eDisposition == tDisposition.Friendly and unitOwner:GetHealth() ~= nil then
		bShowNameplate = true
	elseif eDisposition == tDisposition.Hostile then
		bShowNameplate = true
	elseif eDisposition == tDisposition.Neutral and unitOwner:GetHealth() ~= nil then
		bShowNameplate = true
	end
	
	if tNameplate.bIsObjective and self.setIconQuest then
		bShowNameplate = true
	end
	
	-- this is unimplemented, always evaluates to false
	if self.bShowMainGroupOnly and unitOwner:IsInYourGroup() then
		bShowNameplate = true
	end

	if tNameplate.bIsWeapon then
		bShowNameplate = true
	end
	if tNameplate.bIsHarvest then
		bShowNameplate = true
	end
	local tActivation = unitOwner:GetActivationState()
	if tActivation.FlightPathSettler ~= nil or tActivation.FlightPath ~= nil or tActivation.FlightPathNew then
		bShowNameplate = true
	
	--Vendors
	elseif self.setF_Vendor and isVendor then
		bShowNameplate = true
	elseif self.setF_Important and isImportant then
		bShowNameplate = true				
		--QuestGivers too
	elseif tActivation.QuestReward ~= nil then
		bShowNameplate = true
	elseif tActivation.QuestNew ~= nil or tActivation.QuestNewMain ~= nil or tActivation.QuestNewDaily ~= nil then
		bShowNameplate = true
	elseif tActivation.QuestReceiving ~= nil then
		bShowNameplate = true
	elseif tActivation.TalkTo ~= nil then
		bShowNameplate = true
	end

	if tActivation.Datacube ~= nil and tActivation.Datacube.bIsActive then
		bShowNameplate = true
	end
	
	if bShowNameplate == true then
		bShowNameplate = not (self.playerUnit:IsInCombat() and self.bHideInCombat)
	end

	if unitOwner:IsThePlayer() then
		if not unitOwner:IsDead() then
			bShowNameplate = true
		else
			bShowNameplate = false
		end
	end
	
	--if unitPlayer ~= nil and unitPlayer:IsMounted() and unitPlayer:GetUnitMount() == unitOwner then
	--	bShowNameplate = false
	--end
	
	
	return bShowNameplate
end


function OptiPlates:IsImportantNPC(unitOwner)	
	if unitOwner == nil or not unitOwner:IsValid() then return false end
	local tActivation = unitOwner:GetActivationState()
	
	--Units without health
	if tActivation.Bank ~= nil then
		return true
	elseif tActivation.CREDDExchange ~= nil then
		return true
	elseif tActivation.Mail ~= nil then
		return true
	end
	
	if tActivation.InstancePortal ~= nil then
		return true
	end
	
	--Should we show?
	if not unitOwner:ShouldShowNamePlate() or unitOwner:GetHealth() == nil then
		return false
	end
	
	--Flight paths
	if tActivation.FlightPathSettler ~= nil or tActivation.FlightPath ~= nil or tActivation.FlightPathNew then
		return true
	end
	local sAffiliation = (unitOwner:GetAffiliationName() or "")
	if tActivation.Spell ~= nil and (sAffiliation == "<Exile Transit Authority>" or sAffiliation == "<Dominion Mass Transit>") then
		return true
	end
	
	--Quests
	if tActivation.QuestReward ~= nil then
		return true
	elseif tActivation.QuestNew ~= nil or tActivation.QuestNewMain ~= nil or tActivation.QuestNewDaily ~= nil then
		return true
	elseif tActivation.QuestReceiving ~= nil then
		return true
	elseif tActivation.QuestGiving ~= nil or tActivation.QuestGivingDaily ~= nil then
		return true
	elseif tActivation.TalkTo ~= nil then
		return true
	end
	
	--Vendors
	if tActivation.CommodityMarketplace ~= nil then
		return true
	elseif tActivation.ItemAuctionhouse then
		return true
	end
	
	--Trainers
	if tActivation.TradeskillTrainer then
		return true
	end
	return false
end

function OptiPlates:IsUnimportantNPC(unitOwner)
	if unitOwner == nil or not unitOwner:IsValid() then return true end
	local tNameplate = self.arUnit2Nameplate[unitOwner:GetId()]
	local tActivation = unitOwner:GetActivationState()
	if (tNameplate ~= nil and tNameplate.bIsPet and unitOwner:GetHealth() ~= nil) then return false end
	if unitOwner:GetDispositionTo(self.playerUnit) == tDisposition.Friendly and
		(tActivation == nil or self:GetTableLength(tActivation) == 0 or tActivation.Chair ~= nil)
	then
		return true
	end
	return false
end

function OptiPlates:IsVendorNPC(unitOwner)
	if unitOwner == nil or not unitOwner:IsValid() then return false end
	local tActivation = unitOwner:GetActivationState()
	
	if tActivation.Vendor then
		return true
	end
	return false
end

function OptiPlates:IsPlayerWeapon(unitOwner)
	local unitPlayer = GameLib.GetPlayerUnit()
	if unitPlayer == nil or not unitPlayer:IsValid() then return false end
	if unitOwner:GetType() == "Pickup" then
		local compString = unitPlayer:GetName().."'s Weapon"
		if compString == unitOwner:GetName() then
			return true
		end
		compString = "Waffe von "..unitPlayer:GetName()
		if compString == unitOwner:GetName() then
			return true
		end		
	end
	return false
end

function OptiPlates:IsGuildMember(unitOwner)
	if unitOwner ~= nil and not unitOwner:IsThePlayer() then
		local arGuilds = GuildLib.GetGuilds()
		
		for idx, guild in pairs(arGuilds) do
			local eGuildType = guild:GetType()
			if eGuildType == GuildLib.GuildType_Guild then
				return guild:IsUnitMember(unitOwner)
			end
		end
	end
	return false
end

function OptiPlates:IsInCity()
	local unitPlayer = self.playerUnit
	local currentZone = GameLib.GetCurrentZoneMap()
	
	if unitPlayer ~= nil and currentZone ~= nil then
		if currentZone.strName == "Thayd" or currentZone.strName == "Illium" then
			return true
		end
	end
	return false
end

-- Check Draw Distance of player, return if in range.
function OptiPlates:CheckDrawDistance(tNameplate)
	local unitPlayer = self.playerUnit
	local unitOwner = tNameplate.unitOwner

	if not unitOwner or not unitPlayer or self.setDrawDistance == nil then
	    return false
	end

	tPosTarget = unitOwner:GetPosition()
	tPosPlayer = self.playerPosition

	if tPosTarget == nil then
		return false
	end

	local nDeltaX = tPosTarget.x - tPosPlayer.x
	local nDeltaY = tPosTarget.y - tPosPlayer.y
	local nDeltaZ = tPosTarget.z - tPosPlayer.z

	local nDistance = (nDeltaX * nDeltaX) + (nDeltaY * nDeltaY) + (nDeltaZ * nDeltaZ)
	
	if tNameplate.nSquaredDistance < 0 then tNameplate.bNeedsRedraw = true end
	if tNameplate.bIsQuestUnit then
		local nMinDistanceSquared = self.setStyleQuestIconMinimumDistanceSquared
		if ((tNameplate.nSquaredDistance <= nMinDistanceSquared and 
			nDistance > nMinDistanceSquared) or 
			(tNameplate.nSquaredDistance > nMinDistanceSquared and 
			nDistance <= nMinDistanceSquared))
		then
			tNameplate.bNeedsRedraw = true
		end
	end	
	tNameplate.nSquaredDistance = math.floor(nDistance)
	if tNameplate.bIsQuestUnit and self.setStyleQuestIconIgnoreMaximumDistance then return true end	
	if (tNameplate.bIsHarvest or tNameplate.bIsDatacube) and self.setHarvestablesNoDistanceHiding then return true end

	
	local bInRange = (nDistance < knTargetRange)
	if bInRange then
		bInRange = (nDistance < (self.setDrawDistanceSquared))
		return bInRange
	else
		return false
	end
end



-----------------------------------------------------------------------------------------------
-- HELPER FUNCTIONS
-----------------------------------------------------------------------------------------------

-- Format numbers into shorter readable strings
-- Turns 99999 into 99.9k and 90000 into 90k
function OptiPlates:HelperFormatBigNumber(nArg)
	local strResult
	if nArg >= 1000000 then
		if math.floor(nArg%1000000/100000) == 0 then
			strResult = string.format("%sm", math.floor(nArg / 1000000))
		else
			strResult = string.format("%s.%sm", math.floor(nArg / 1000000), math.floor(nArg % 1000000 / 100000))
		end
	elseif nArg >= 1000 then
		if math.floor(nArg%1000/100) == 0 then
			strResult = string.format("%sk", math.floor(nArg / 1000))
		else
			strResult = string.format("%s.%sk", math.floor(nArg / 1000), math.floor(nArg % 1000 / 100))
		end
	else
		strResult = nArg
	end
	
	return strResult
end

function OptiPlates:HelperFormatTime(nArg)
	local strResult
	if nArg >= 86400 then
		strResult = string.format("%sd", math.floor(nArg / 86400))
	elseif nArg >= 3600 then
		strResult = string.format("%sh", math.floor(nArg / 3600))
	elseif nArg >= 60 then
		strResult = string.format("%sm", math.floor(nArg / 60))
	else
		strResult = string.format("%ss", math.floor(nArg))
	end
	
	return strResult
end


function OptiPlates:OnRefreshHealthShieldBar()
	self.refreshHealthShieldBar = true
end

-----------------------------------------------------------------------------------------------
-- UI Calls: Settings Page
-----------------------------------------------------------------------------------------------

-- Click on a name
-------------------------------------
function OptiPlates:OnNameClick(wndHandler, wndCtrl, nClick, nX, nY, bDoubleClick, bStopPropagation)
	--local unitOwner = wndCtrl:GetData()
	if not self.setMouseTargetingEnabled then return false end
	
	local bButtonMatches = 	
			(self.setMouseTargetingButtonLeft and nClick == GameLib.CodeEnumInputMouse.Left) or
			(self.setMouseTargetingButtonMiddle and nClick == GameLib.CodeEnumInputMouse.Middle) or
			(self.setMouseTargetingButtonRight and nClick == GameLib.CodeEnumInputMouse.Right)
			
	if not bButtonMatches then return false end

	local bModifiersMatch = 
			(not self.setMouseTargetingModifiersAlt or Apollo.IsAltKeyDown()) and
			(not self.setMouseTargetingModifiersCtrl or Apollo.IsControlKeyDown()) and
			(not self.setMouseTargetingModifiersShift or Apollo.IsShiftKeyDown())
	if not bModifiersMatch then return false end
	if self.clickTargetUnit ~= nil then return false end
	
	local unitOwner = nil
	-- this is not actually too slow, since clicking on nameplates is done relatively rarely (once or twice a second at most)
	for idx, tNameplate in pairs(self.arDisplayedNameplates) do
		if not (tNameplate == nil or tNameplate.unitOwner == nil) then
			if tNameplate.wnd.HealthBarBack == wndCtrl
				or tNameplate.wnd.ShieldBar == wndCtrl
				or tNameplate.wnd.GuildName == wndCtrl
				or tNameplate.wnd.Name == wndCtrl
				then
					unitOwner = tNameplate.unitOwner
				break
			end
		end
	end
	if unitOwner ~= nil and GameLib.GetTargetUnit() ~= unitOwner then
		self.clickTargetUnit = unitOwner
		return not self.setMouseTargetingClickthrough
	end
	return false
end

function OptiPlates:OnMouseOverUnitChanged(unit)
	if self.mouseOverUnit ~= nil and self.mouseOverUnit ~= unit then
		local tNameplate = self.arUnit2Nameplate[self.mouseOverUnit:GetId()]
		if tNameplate ~= nil and tNameplate.bMouseOverSettingsEnabled then
			tNameplate.bMouseOverSettingsEnabled = false
			tNameplate.bNeedsRedraw = true
		end
		self.mouseOverUnit = nil
		self:UpdateNameplateVisibility(tNameplate)					
	end
	if unit == nil then 
		self.mouseOverUnit = nil
		return
	end
	if unit:GetType() == "Mount" then
		self.mouseOverUnit = self.arMount2Unit[unit:GetId()]
		if self.mouseOverUnit == nil then
			for idx,tNameplate in pairs(self.arUnit2Nameplate) do
				local unitOwner = tNameplate.unitOwner
				if unitOwner:IsValid() and unitOwner:IsMounted() then
					local unitMount = unitOwner:GetUnitMount()
					self.arMount2Unit[unitOwner:GetId()] = unitMount
					if unit == unitMount then
						self.mouseOverUnit = unitOwner
						break
					end
				end
			end
		end
	else
		self.mouseOverUnit = unit
	end
end

-- General Settings
-------------------------------------
function OptiPlates:Button_ToggleOcclusion(wndHandler, wndCtrl)
	self.setEnableOcclusion = wndCtrl:IsChecked()
	Apollo.SetConsoleVariable("ui.occludeNameplatePositions", self.setEnableOcclusion)
	self:OptionsChanged()
end

function OptiPlates:Button_ShowThreatAlert(wndHandler, wndCtrl)
	local btnThreatLossAlert = self.wndMain:FindChild("Button_ShowThreatLossAlert")
	local btnThreatAlert = self.wndMain:FindChild("Button_ShowThreatAlert")
	local bThreatAlertChanged = false
	if wndCtrl == btnThreatLossAlert then
		self.setShowThreatLossAlert = btnThreatLossAlert:IsChecked()
		if self.setShowThreatAlert and self.setShowThreatAlert then
			self.setShowThreatAlert = not self.setShowThreatLossAlert
			btnThreatAlert:SetCheck(not self.setShowThreatLossAlert)
		end
	elseif wndCtrl == btnThreatAlert then
		self.setShowThreatAlert = btnThreatAlert:IsChecked()
		if self.setShowThreatAlert and self.setShowThreatAlert then
			self.setShowThreatLossAlert = not self.setShowThreatAlert
			btnThreatLossAlert:SetCheck(not self.setShowThreatAlert)
		end		
	end
	
	self:OptionsChanged()
end

-- COLORS
-------------------------------------
function OptiPlates:Button_DefaultColors(wndHandler, wndCtrl)
	self:DefaultColors()	
	self:RefreshSettings()
	self:OptionsChanged()
end

-- STYLE | sao, flat
-------------------------------------
function OptiPlates:Button_StyleSAO( wndHandler, wndControl, eMouseButton )
	self.setStyle = "sao";
	self.wndMain:FindChild("Flat Settings"):Show(false)
	local wndResize = self.wndMain:FindChild("Generic Settings")
	local nLeft, nTop, nRight, nBottom = wndResize:GetAnchorOffsets()
	wndResize:SetAnchorOffsets(nLeft, nTop, nRight, 255)
	self:OptionsChanged();
	self:ReloadAllNameplates()
end

function OptiPlates:Button_StyleFlat( wndHandler, wndControl, eMouseButton )
	self.setStyle = "flat";
	self.wndMain:FindChild("Flat Settings"):Show(true)
	local wndResize = self.wndMain:FindChild("Generic Settings")
	local nLeft, nTop, nRight, nBottom = wndResize:GetAnchorOffsets()
	wndResize:SetAnchorOffsets(nLeft, nTop, nRight, 345)
	self:OptionsChanged();
	self:ReloadAllNameplates()
end

function OptiPlates:CreateProfileList(grid, btn)
	if grid == nil then return end
	local profiles = self.setProfiles
	local scopeStr = ""
	local profileArray = {}
	local tempCounter = 1
	
	grid:DeleteAll()
	
	for scope,profile in pairs(profiles) do
		for profileName, profileTable in pairs(profile) do
			if scope == GameLib.CodeEnumAddonSaveLevel.General then
				scopeStr = "General"
			elseif scope == GameLib.CodeEnumAddonSaveLevel.Account then
				scopeStr = "Account"
			elseif scope == GameLib.CodeEnumAddonSaveLevel.Realm then
				scopeStr = "Realm"
			elseif scope == GameLib.CodeEnumAddonSaveLevel.Character then
				scopeStr = "Character"
			else
				break
			end
			local profileString = scopeStr.." - "..profileName
			grid:AddRow(profileString)
			profileArray[tempCounter] = {scope,profileName}
			tempCounter = tempCounter + 1
			if self.setCurrentProfileName == profileName and self.setCurrentProfileScope == scope then
				grid:SetCurrentRow(grid:GetRowCount())
				btn:SetText("   "..profileString)
			end
		end
	end
	self.tempProfileArray = profileArray
end

function OptiPlates:CreateAllProfileLists()
	self:CreateProfileList(self.wndMain:FindChild("Grid_NewProfileList"), self.wndMain:FindChild("Button_NewProfileListToggle"))
	self:CreateProfileList(self.wndMain:FindChild("Grid_RemoveProfileList"), self.wndMain:FindChild("Button_RemoveProfileListToggle"))
	self:CreateProfileList(self.wndMain:FindChild("Grid_ChangeProfileList"), self.wndMain:FindChild("Button_ChangeProfileListToggle"))	
	self:CreateProfileList(self.wndQuickChange:FindChild("Grid_QuickChangeProfileList"), self.wndQuickChange:FindChild("Button_QuickProfileListToggle"))	
end

function OptiPlates:CreateProfile(scope, name, baseArray)
	self.setProfiles[scope][name] = {}
	local tempProfile = {}
	for propKey, prop in pairs(tUserSettings) do 
		if baseArray == nil or baseArray[propKey] == nil then
			tempProfile[propKey] = prop.default
		else
			tempProfile[propKey] = baseArray[propKey] 
		end
	end	
	self.setProfiles[scope][name] = tempProfile	
	self:OptionsChanged()
	Event_FireGenericEvent("OptiPlates_ProfileCreated", scope, name)
end

function OptiPlates:RemoveProfile(scope, name)
	self.setProfiles[scope][name] = nil
	Event_FireGenericEvent("OptiPlates_ProfileRemoved", scope, name)
	if self.setCurrentProfileScope == scope and self.setCurrentProfileName == name then
		self:LoadProfile(GameLib.CodeEnumAddonSaveLevel.Character, "Default")
	end
	self:CheckSettingsSanity()
	self:OptionsChanged()
end

function OptiPlates:ChangeProfile(scope, name)
	local oldStyle = self.setStyle
	self:SaveProfile(self.setCurrentProfileScope,self.setCurrentProfileName)
	if self.setProfiles[scope] == nil or self.setProfiles[scope][name] == nil then return end
	self.setCurrentProfileScope = scope
	self.setCurrentProfileName = name
	self:LoadProfile(scope,name)
	if oldStyle ~= self.setStyle then
		self:ReloadAllNameplates()	
	end
	self:RefreshSettings()
	self:OptionsChanged()
	Event_FireGenericEvent("OptiPlates_ProfileChanged", scope, name)
end

-- saves the current settings to profile called name with scope
function OptiPlates:SaveProfile(scope,name)
	local tempProfile = {}
	for propKey,prop in pairs(tUserSettings) do 
		if not prop.deprecated then
			tempProfile[propKey] = self[propKey] 
		end
	end		
	if self.setProfiles[scope] == nil then self.setProfiles[scope] = {} end
	self.setProfiles[scope][name] = tempProfile
end

-- loads settings from profile name with scope 
function OptiPlates:LoadProfile(scope,name)
	local profile = self.setProfiles[scope][name]

	if profile == nil then return end
	local tmpTable = { }
	for propKey,prop in pairs(tUserSettings) do 
		if profile[propKey] == nil then
			if prop.defaultifexists ~= nil then
				tmpTable[propKey] = prop.defaultifexists
			else
				self[propKey] = prop.default
			end
		elseif not prop.percharacter then
			self[propKey] = profile[propKey] 
			if prop.setDefaultOnEmptyString and self[propKey] == "" then
				self[propKey] = prop.default
			end
		end
	end
	
	for newProp, oldProp in pairs(tmpTable) do 
		if self[oldProp] ~= nil then 
			self[newProp] = self[oldProp]
		else
			self[newProp] = tUserSettings.newProp.default 
		end
	end
end


function OptiPlates:Button_ChangeProfileListToggle( wndHandler, wndControl, eMouseButton )
	local wnd = self.wndMain:FindChild("Window_ChangeProfileListContainer")
	wnd:Show(not wnd:IsShown())
	if wnd:IsShown() then wnd:ToFront() end
end

function OptiPlates:Grid_OnChangeProfileList()
	local sel = self.wndMain:FindChild("Grid_ChangeProfileList"):GetCurrentRow()
	local btn = self.wndMain:FindChild("Button_ChangeProfileListToggle")
	local profile = self.tempProfileArray[sel]
	local profileString = self.scopeTextArray[profile[1]].." - "..profile[2]
	btn:SetText("   "..profileString)
	self.wndMain:FindChild("Window_ChangeProfileListContainer"):Show(false)
end

function OptiPlates:Grid_OnNewProfileScopeList()
	local sel = self.wndMain:FindChild("Grid_NewProfileScopeList"):GetCurrentRow()
	local btn = self.wndMain:FindChild("Button_NewProfileScopeListToggle")
	if sel == 1 then
		btn:SetText("  Character")
	elseif sel == 2 then
		btn:SetText("  Realm")
	elseif sel == 3 then
		btn:SetText("  Account")		
	elseif sel == 4 then
		btn:SetText("  General")
	end
	self.wndMain:FindChild("Window_NewProfileScopeListContainer"):Show(false)
end

function OptiPlates:Button_NewProfileListDisableToggle( wndHandler, wndControl, eMouseButton )
	local check = self.wndMain:FindChild("Button_NewProfileCopyBase"):IsChecked()
	local button = self.wndMain:FindChild("Button_NewProfileListToggle")
	if check then
		button:Enable(true)
	else
		button:Enable(false)
	end
	
end

function OptiPlates:Button_OnNewProfileScopeListToggle( wndHandler, wndControl, eMouseButton )
	local wnd = self.wndMain:FindChild("Window_NewProfileScopeListContainer")
	wnd:Show(not wnd:IsShown())
	if wnd:IsShown() then wnd:ToFront() end
end

function OptiPlates:Button_OnNewProfileListToggle( wndHandler, wndControl, eMouseButton )
	local wnd = self.wndMain:FindChild("Window_NewProfileListContainer")
	wnd:Show(not wnd:IsShown())
	if wnd:IsShown() then wnd:ToFront() end
end

function OptiPlates:Grid_OnNewProfileListChange()
	local sel = self.wndMain:FindChild("Grid_NewProfileList"):GetCurrentRow()
	local btn = self.wndMain:FindChild("Button_NewProfileListToggle")
	local profile = self.tempProfileArray[sel]
	local profileString = self.scopeTextArray[profile[1]].." - "..profile[2]
	btn:SetText("   "..profileString)
	self.wndMain:FindChild("Window_NewProfileListContainer"):Show(false)
end

function OptiPlates:Grid_OnRemoveProfileList()
	local sel = self.wndMain:FindChild("Grid_RemoveProfileList"):GetCurrentRow()
	local btn = self.wndMain:FindChild("Button_RemoveProfileListToggle")
	local profile = self.tempProfileArray[sel]
	local profileString = self.scopeTextArray[profile[1]].." - "..profile[2]
	btn:SetText("   "..profileString)
	self.wndMain:FindChild("Window_RemoveProfileListContainer"):Show(false)
end

function OptiPlates:Button_OnRemoveProfileListToggle( wndHandler, wndControl, eMouseButton )
	local wnd = self.wndMain:FindChild("Window_RemoveProfileListContainer")
	wnd:Show(not wnd:IsShown())
	if wnd:IsShown() then wnd:ToFront() end
end

function OptiPlates:Button_OnCreateProfile( wndHandler, wndControl, eMouseButton )
	local scope = ""
	local scopeRow = self.wndMain:FindChild("Grid_NewProfileScopeList"):GetCurrentRow()
	if scopeRow == 1 then
		scope = GameLib.CodeEnumAddonSaveLevel.Character
	elseif scopeRow == 2 then
		scope = GameLib.CodeEnumAddonSaveLevel.Realm
	elseif scopeRow == 3 then
		scope = GameLib.CodeEnumAddonSaveLevel.Account
	elseif scopeRow == 4 then
		scope = GameLib.CodeEnumAddonSaveLevel.General
	end
	
	local name = self.wndMain:FindChild("Edit_NewProfileName"):GetText()
	local baseArray = ""
	local check = self.wndMain:FindChild("Button_NewProfileCopyBase"):IsChecked()
	
	if check then
		baseArray = self.setProfiles[self.setCurrentProfileScope][self.setCurrentProfileName]
	else
		baseArray = nil
	end
	self:CreateProfile(scope,name,baseArray)
	
	self:CreateAllProfileLists()
end

function OptiPlates:Button_OnChangeProfile( wndHandler, wndControl, eMouseButton )
	local sel = self.wndMain:FindChild("Grid_ChangeProfileList"):GetCurrentRow()
	local profile = self.tempProfileArray[sel]
	local scope = profile[1]
	local name = profile[2]
	self:ChangeProfile(scope,name)
	self.wndMain:FindChild("Label_ProfileScope"):SetText(self.scopeTextArray[self.setCurrentProfileScope])
	self.wndMain:FindChild("Label_ProfileName"):SetText(self.setCurrentProfileName)
end

function OptiPlates:Button_OnRemoveProfile( wndHandler, wndControl, eMouseButton )
	local sel = self.wndMain:FindChild("Grid_RemoveProfileList"):GetCurrentRow()
	local profile = self.tempProfileArray[sel]
	local scope = profile[1]
	local name = profile[2]
	self:RemoveProfile(scope,name)
	self:CreateAllProfileLists()
	self.wndMain:FindChild("Grid_RemoveProfileList"):SetCurrentRow(1)
	self.wndMain:FindChild("Label_ProfileScope"):SetText(self.scopeTextArray[self.setCurrentProfileScope])
	self.wndMain:FindChild("Label_ProfileName"):SetText(self.setCurrentProfileName)
end
function OptiPlates:Button_OnClose( wndHandler, wndControl, eMouseButton )
	self.wndMain:Show(false)
end

function OptiPlates:OptionCheckboxChanged( wndHandler, wndControl, eMouseButton)
	for idx, opt in pairs(self.optionsUI_CheckboxSettings) do
		if opt.wnd == wndControl then
			self[opt.var] = opt.wnd:IsChecked()
			if opt.togglewnd then
				opt.togglewnd:Show(opt.wnd:IsChecked())
				local wndResize = opt.wnd:GetParent()
				local nLeft, nTop, nRight, nBottom = wndResize:GetAnchorOffsets()
				if opt.wnd:IsChecked() then
					wndResize:SetAnchorOffsets(nLeft, nTop, nRight, opt.parentheighton)
				else
					wndResize:SetAnchorOffsets(nLeft, nTop, nRight, opt.parentheightoff)
				end
			end			
			if opt.bneedsreload ~= nil and opt.bneedsreload == true then
				self:ReloadAllNameplates()
			end
			self:OptionsChanged()			
			return
		end
	end
end

function OptiPlates:OptionEditboxChanged( wndHandler, wndControl, eMouseButton)
	for idx, opt in pairs(self.optionsUI_EditboxSettings) do
		if opt.wnd == wndControl then
			if opt.vartype ~= nil then
				if opt.vartype == "number" then
					self[opt.var] = tonumber(opt.wnd:GetText() or "")
				end
			else
				self[opt.var] = (opt.wnd:GetText() or "")
			end
			if self[opt.var] == nil or (tUserSettings[opt.var].setDefaultOnEmptyString and self[opt.var] == "") then
				self[opt.var] = tUserSettings[opt.var].default
				if self[opt.var] ~= nil then
					opt.wnd:SetText(self[opt.var])
				end
			end
			if opt.colorlabel ~= nil then
				opt.colorlabel:SetTextColor(ApolloColor.new(self.setOpacity..self[opt.var] ))
			end
			self:OptionsChanged()
			return
		end
	end
end

function OptiPlates:OptionSliderChanged( wndHandler, wndControl, fNewValue, fOldValue)
	if fNewValue == fOldValue then return end
	for idx, opt in pairs(self.optionsUI_SliderSettings) do
		if opt.wnd == wndControl then
			local value = (fNewValue or 0)
			self[opt.var] = value
			if opt.label ~= nil then
				if opt.divideby ~= nil then
					value = value / opt.divideby
				end			
				if opt.stringformat ~= nil then
					opt.label:SetText(string.format(opt.stringformat,value))
				else
					opt.label:SetText(string.format("%s",value))
				end
			end
			self:OptionsChanged()
			return
		end
	end
end

function OptiPlates:OptionDropdownButtonPressed ( wndHandler, wndControl, eMouseButton)
	for idx,opt in pairs(self.optionsUI_DropdownListSettings) do
		if opt.btn == wndControl then
			opt.wnd:Show(not opt.wnd:IsShown())
			return
		end
	end
end

function OptiPlates:OptionDropdownSelectionChanged ( wndHandler, wndControl, iRow, iCol)
	for idx,opt in pairs(self.optionsUI_DropdownListSettings) do
		if opt.grid == wndControl then
			local selectedVal = self[opt.array][iRow]
			if opt.labelarray ~= nil then
				opt.btn:SetText("    "..self[opt.labelarray][selectedVal])
			else
				opt.btn:SetText("    "..selectedVal)
			end
			self[opt.var] = selectedVal
			opt.wnd:Show(false)
			self:OptionsChanged()
			self:ReloadAllNameplates()
			return
		end
	end	
end

function OptiPlates:OptionNavChanged(  wndHandler, wndControl, eMouseButton )
	local btn = wndControl
	local shownWnd = nil
	updateStatistics = false
	local wndSettingsFrame = self.wndMain:FindChild("SettingsFrame")
	for idx, opt in pairs(self.optionsUI_NavSettings) do
		if opt.btn == btn then 
			local wndName = opt.wnd 
			-- Clicked on currently selected option window
			if wndSettingsFrame:FindChild(wndName) then 
				return 
			else
				wndSettingsFrame:DestroyChildren()
			end
			self:HelperChangeOptionMenu(wndName)
			if opt.toggleStats ~= nil then
				updateStatistics = true
			end
		end
	end
end

function OptiPlates:HelperChangeOptionMenu(sPanelName)
	local wndSettingsFrame = self.wndMain:FindChild("SettingsFrame")
	local shownWnd = Apollo.LoadForm(self.xmlSettings, sPanelName, wndSettingsFrame, self)
	self:UpdateSettingsArrays()
	self:UpdateOptionMenuPixies()
	if sPanelName == "SettingsStyle" then
		if self.bStyleFlat then
			shownWnd:FindChild("Flat Settings"):Show(true)
			local wndResize = shownWnd:FindChild("Generic Settings")
			local nLeft, nTop, nRight, nBottom = wndResize:GetAnchorOffsets()
			wndResize:SetAnchorOffsets(nLeft, nTop, nRight, 345)		
		else
			shownWnd:FindChild("Flat Settings"):Show(false)		
			local wndResize = shownWnd:FindChild("Generic Settings")
			local nLeft, nTop, nRight, nBottom = wndResize:GetAnchorOffsets()
			wndResize:SetAnchorOffsets(nLeft, nTop, nRight, 255)
		end	
		shownWnd:FindChild("PreviewHealth_SAO"):SetSprite("SAO_Health")
		shownWnd:FindChild("PreviewShield_SAO"):SetSprite("SAO_Shield")
		shownWnd:FindChild("PreviewHealth_Flat"):SetSprite("Flat_Health")
		shownWnd:FindChild("PreviewShield_Flat"):SetSprite("Flat_Shield")				
	elseif sPanelName == "SettingsColors" or sPanelName == "SettingsHarvesting" then
		if colorpicker ~= nil then
			for propKey,arr in pairs(self.optionsUI_EditboxSettings) do 
				if arr.wnd ~= nil and arr.colorlabel ~= nil then
					arr.wnd:AddEventHandler("WindowGainedFocus", "ColorEditboxGainedFocus")
					arr.wnd:AddEventHandler("WindowLostFocus", "ColorEditboxLostFocus")
				end
			end
		end
	elseif sPanelName == "SettingsNameplate" then
		self:HelperUpdateNameplateMatrixButtons(shownWnd)
	elseif sPanelName == "SettingsBuffs" then
		self:CreateAllBuffProfileLists()
		self:Grid_OnChangeBuffProfileList()
	end
	self:RefreshSettings()
	wndSettingsFrame:SetVScrollPos(0)
	shownWnd:Show(true) 
end

function OptiPlates:NameplateMatrixButtonDown(wndHandler, wndCtrl, eMouseButton)
	local data = wndCtrl:GetData()
	local settingName = self:HelperGetMatrixSettingName(data.category, data.setting, false)
	local bOptionChanged = false
	if eMouseButton == GameLib.CodeEnumInputMouse.Left then -- Out of combat
		data.outofcombat = not data.outofcombat
		settingName = self:HelperGetMatrixSettingName(data.category, data.setting, false)
		self[settingName] = data.outofcombat
		self.tNameplateSettings[data.category]["OOC"][data.setting] = data.outofcombat
		bOptionChanged = true
	elseif eMouseButton == GameLib.CodeEnumInputMouse.Right then -- In combat
		data.incombat = not data.incombat
		settingName = self:HelperGetMatrixSettingName(data.category, data.setting, true)
		self[settingName] = data.incombat
		self.tNameplateSettings[data.category]["C"][data.setting] = data.incombat
		bOptionChanged = true
	else
		return
	end
	wndCtrl:SetData(data)
	
	self:HelperDrawNameplateMatrixButton(wndCtrl)
	if bOptionChanged then self:OptionsChanged() end
end

function OptiPlates:HelperDrawNameplateMatrixButton(wndCtrl)
	local data = wndCtrl:GetData()
	local color = nil
	wndCtrl:DestroyAllPixies()
	if data.incombat then
		wndCtrl:AddPixie(tNameplateMatrixInCombatPixie)	
	end
	if data.outofcombat then
		wndCtrl:AddPixie(tNameplateMatrixOutOfCombatPixie)	
	end
	wndCtrl:SetSprite("WhiteFill")
	--wndCtrl:SetBGColor(color)
	wndCtrl:SetBGColor(tNameplateMatrixColors[1])
	wndCtrl:SetTooltip(string.format("%s, %s\nOut of combat: %s\nIn combat: %s",data.category, data.setting, (data.outofcombat and "Show" or "Hide"),(data.incombat and "Show" or "Hide")))
	wndCtrl:Show(true)
end

function OptiPlates:HelperUpdateNameplateMatrixButtons(wndParent)
	local wnd = wndParent:FindChild("NameplateTable")
	local wndHeaders = wndParent:FindChild("NameplateTableHeaders")
	wnd:DestroyChildren()
	wnd:DestroyAllPixies()
	local x = 0
	local y = 0
	local nSize = 26
	local nSizeFiller = math.sqrt(nSize*nSize+nSize*nSize)
	local nTemp = nSize - math.floor(nSizeFiller*0.5)
	nSizeFiller = math.floor(nSizeFiller)
	local nTemp2 = nTemp-nSize
	tNameplateMatrixInCombatPixie.loc.nOffsets = { nTemp, nTemp, nTemp+nSizeFiller, nTemp+nSizeFiller }
	tNameplateMatrixOutOfCombatPixie.loc.nOffsets = { nTemp2, nTemp2, nTemp2+nSizeFiller, nTemp2+nSizeFiller }
	local nPad = 3
	local crWhite = ApolloColor.new("xkcdWhite")
	for idx, category in pairs(tNameplateMatrixCategories) do
		y = 0
		local bFirst = true
		for idy, setting in pairs(tNameplateMatrixSettings) do
			-- Setting names
			if category == "Self" or category == "FriendlyTarget" then
				if bFirst then
					x = x + 110
					bFirst = false
				end
				if not (category == "FriendlyTarget" and (setting == "ShieldReboot" or setting == "Heading" or setting == "Nameplate")) then
					tNameplateMatrixLabelPixie.strText = setting
					tNameplateMatrixLabelPixie.crText = crWhite
					tNameplateMatrixLabelPixie.fRotation = 0
					tNameplateMatrixLabelPixie.flagsText.DT_RIGHT = true
					tNameplateMatrixLabelPixie.flagsText.DT_VCENTER = true
					tNameplateMatrixLabelPixie.flagsText.DT_CENTER = false
					tNameplateMatrixLabelPixie.loc.nOffsets[1] = x - 110
					tNameplateMatrixLabelPixie.loc.nOffsets[3] = x - 10
					if setting == "Nameplate" then 
						tNameplateMatrixLabelPixie.loc.nOffsets[2] = 40
						tNameplateMatrixLabelPixie.loc.nOffsets[4] = 40+nSize+nPad
						wndHeaders:AddPixie(tNameplateMatrixLabelPixie)
					else
						tNameplateMatrixLabelPixie.loc.nOffsets[2] = y
						tNameplateMatrixLabelPixie.loc.nOffsets[4] = y+nSize+nPad
						wnd:AddPixie(tNameplateMatrixLabelPixie)
					end
				end
			end
			if tNameplateMatrixEnabledSettings[category] ~= nil and tNameplateMatrixEnabledSettings[category][setting] then
				local newWnd = nil
				if setting == "Nameplate" then
					newWnd = Apollo.LoadForm(self.xmlSettings	, "NameplateMatrix", wndHeaders, self)
				else
					newWnd = Apollo.LoadForm(self.xmlSettings	, "NameplateMatrix", wnd, self)
				end
				newWnd:SetData({ outofcombat = self[self:HelperGetMatrixSettingName(category,setting,false)], incombat = self[self:HelperGetMatrixSettingName(category,setting,true)], category = category, setting = setting })
				newWnd:AddEventHandler("MouseButtonDown", "NameplateMatrixButtonDown")
				if setting == "Nameplate" then
					newWnd:SetAnchorOffsets(x, 40, x+nSize, 40+nSize)
					y = y - nSize - nPad
				else
					newWnd:SetAnchorOffsets(x, y, x+nSize, y+nSize)
				end
				self:HelperDrawNameplateMatrixButton(newWnd)
			end
			y = y + nSize + nPad
		end
		tNameplateMatrixLabelPixie.strText = tNameplateMatrixCategoryToLetter[category]
		tNameplateMatrixLabelPixie.crText = tNameplateMatrixCategoryToColor[category]
		--tNameplateMatrixLabelPixie.fRotation = 315/180*math.pi
		tNameplateMatrixLabelPixie.flagsText.DT_RIGHT = false
		tNameplateMatrixLabelPixie.flagsText.DT_VCENTER = false
		tNameplateMatrixLabelPixie.flagsText.DT_CENTER = true
		tNameplateMatrixLabelPixie.loc.nOffsets[1] = x
		tNameplateMatrixLabelPixie.loc.nOffsets[2] = 22
		tNameplateMatrixLabelPixie.loc.nOffsets[3] = x+nSize
		tNameplateMatrixLabelPixie.loc.nOffsets[4] = 38
		wndHeaders:AddPixie(tNameplateMatrixLabelPixie)
		x = x + nSize + nPad
	end
		--tNameplateMatrixLabelPixie.fRotation = 315/180*math.pi
	tNameplateMatrixLabelPixie.flagsText.DT_RIGHT = false
	tNameplateMatrixLabelPixie.flagsText.DT_VCENTER = false
	tNameplateMatrixLabelPixie.flagsText.DT_CENTER = true
	tNameplateMatrixLabelPixie.crText = crWhite	
	tNameplateMatrixLabelPixie.loc.nOffsets[1] = 110
	tNameplateMatrixLabelPixie.loc.nOffsets[2] = 5
	tNameplateMatrixLabelPixie.loc.nOffsets[3] = 226
	tNameplateMatrixLabelPixie.loc.nOffsets[4] = 21
	tNameplateMatrixLabelPixie.strText = "Player"
	wndHeaders:AddPixie(tNameplateMatrixLabelPixie)
	tNameplateMatrixLabelPixie.loc.nOffsets[1] = 226
	tNameplateMatrixLabelPixie.loc.nOffsets[2] = 5
	tNameplateMatrixLabelPixie.loc.nOffsets[3] = 313
	tNameplateMatrixLabelPixie.loc.nOffsets[4] = 21
	tNameplateMatrixLabelPixie.strText = "NPC"
	wndHeaders:AddPixie(tNameplateMatrixLabelPixie)
	tNameplateMatrixLabelPixie.loc.nOffsets[1] = 423
	tNameplateMatrixLabelPixie.loc.nOffsets[2] = 5
	tNameplateMatrixLabelPixie.loc.nOffsets[3] = 510
	tNameplateMatrixLabelPixie.loc.nOffsets[4] = 21
	tNameplateMatrixLabelPixie.strText = "Target"
	wndHeaders:AddPixie(tNameplateMatrixLabelPixie)
	
	wnd:SetAnchorOffsets(0,0,0,y)
end

function OptiPlates:HelperGetMatrixSettingName(sCategory,sSetting,bInCombat)
	return string.format("set%s_%s%s", tNameplateMatrixCategoryToLetter[sCategory],tNameplateMatrixSettingToLetter[sSetting],(bInCombat and "_C" or ""))
end

function OptiPlates:ColorPickerColorUpdated()
	local arr = colorpickerwndControlArray
	local self = Apollo.GetAddon("OptiPlates")
	if arr ~= nil then
		local colorString = self:CColorToHexString(colorpickerColorObject)
		self[arr.var] = colorString
		self:OptionsChanged()
		self:RefreshSettings()
	end
end

function OptiPlates:ColorEditboxGainedFocus(wndHandler, wndCtrl)
	for propKey,arr in pairs(self.optionsUI_EditboxSettings) do 
		if arr.wnd ~= nil and arr.wnd == wndCtrl then
			colorpickerwndControlArray = arr
			local r,g,b = self:StringToRGB(self[arr.var])
			colorpickerColorObject = CColor.new(r,g,b,1)
			local settingsLoc = self.wndMain:GetLocation():ToTable().nOffsets
			local mouseLoc = Apollo.GetMouse()
			local nLeft = mouseLoc.x+250
			local nTop = mouseLoc.y
			local nHeight = colorpicker.wndMain:GetHeight()
			local nWidth = colorpicker.wndMain:GetWidth()
			if settingsLoc[4] < nTop + nHeight then nTop = settingsLoc[4] - nHeight end -- if below options window
			if settingsLoc[2] > nTop + 70 then nTop = settingsLoc[2] + 70 end -- if above options window
			
			colorpicker.wndMain:Move(nLeft,nTop,nWidth,nHeight)
			colorpicker.wndMain:Show(true)
			colorpicker.AdjustCColor(colorpickerColorObject, false, OptiPlates.ColorPickerColorUpdated)
		end
	end
end

function OptiPlates:ColorEditboxLostFocus(wndHandler, wndCtrl)
	if colorpickerwndControlArray ~= nil and colorpickerwndControlArray.wnd == wndCtrl then
		colorpicker.wndMain:Show(false)
		colorpickerwndControlArray = nil
	end
end

function OptiPlates:UpdateOptionMenuPixies()
	local wnd = self.wndMain:FindChild("SettingsAdvanced")
	if wnd ~= nil then
		local wndAdvanced1 = wnd:FindChild("Window_AdvancedSettings1")
		local wndAdvanced2 = wnd:FindChild("Window_AdvancedSettings2")
		local tmpPixie = {
			strSprite = "",
			loc = {
				fPoints = { 0,0,0,0 },
				nOffsets = { 140,10,170,10 }
			}
		}
		if wndAdvanced1 ~= nil then
			wndAdvanced1:DestroyAllPixies()
			tmpPixie.strSprite = self.setStyleThreatIconSprite
			tmpPixie.loc.nOffsets[2] = 60;tmpPixie.loc.nOffsets[4] = 90
			wndAdvanced1:AddPixie(tmpPixie)

			tmpPixie.strSprite = self.setStyleCleanseIconSprite
			tmpPixie.loc.nOffsets[2] = 90;tmpPixie.loc.nOffsets[4] = 120
			wndAdvanced1:AddPixie(tmpPixie)
		end
		if wndAdvanced2 ~= nil then
			wndAdvanced2:DestroyAllPixies()
			tmpPixie.strSprite = self.setFriendlyTargetIndicatorIcon
			tmpPixie.loc.nOffsets[2] = 25;tmpPixie.loc.nOffsets[4] = 55
			wndAdvanced2:AddPixie(tmpPixie)
			tmpPixie.strSprite = self.setPartyTargetIndicatorIcon
			tmpPixie.loc.nOffsets[2] = 55;tmpPixie.loc.nOffsets[4] = 85	
			wndAdvanced2:AddPixie(tmpPixie)
			tmpPixie.strSprite = self.setNeutralTargetIndicatorIcon
			tmpPixie.loc.nOffsets[2] = 85;tmpPixie.loc.nOffsets[4] = 115
			wndAdvanced2:AddPixie(tmpPixie)
			tmpPixie.strSprite = self.setHostileTargetIndicatorIcon
			tmpPixie.loc.nOffsets[2] = 115;tmpPixie.loc.nOffsets[4] = 145
			wndAdvanced2:AddPixie(tmpPixie)
		
			tmpPixie.strSprite = self.setFriendlyClusterIndicatorIcon
			tmpPixie.loc.nOffsets[2] = 155;tmpPixie.loc.nOffsets[4] = 185
			wndAdvanced2:AddPixie(tmpPixie)	
			tmpPixie.strSprite = self.setNeutralClusterIndicatorIcon
			tmpPixie.loc.nOffsets[2] = 185;tmpPixie.loc.nOffsets[4] = 215
			wndAdvanced2:AddPixie(tmpPixie)
			tmpPixie.strSprite = self.setHostileClusterIndicatorIcon
			tmpPixie.loc.nOffsets[2] = 215;tmpPixie.loc.nOffsets[4] = 245
			wndAdvanced2:AddPixie(tmpPixie)		

			tmpPixie.strSprite = self.setFriendlyAltTargetIndicatorIcon
			tmpPixie.loc.nOffsets[2] = 255;tmpPixie.loc.nOffsets[4] = 285
			wndAdvanced2:AddPixie(tmpPixie)
			tmpPixie.strSprite = self.setNeutralAltTargetIndicatorIcon
			tmpPixie.loc.nOffsets[2] = 285;tmpPixie.loc.nOffsets[4] = 315
			wndAdvanced2:AddPixie(tmpPixie)
			tmpPixie.strSprite = self.setHostileAltTargetIndicatorIcon
			tmpPixie.loc.nOffsets[2] = 315;tmpPixie.loc.nOffsets[4] = 345
			wndAdvanced2:AddPixie(tmpPixie)
		end
	end
end


function OptiPlates:OnGroup_Updated()
	self.bGroupInfoDirty = true
end

function OptiPlates:OnGroup_Left()
	self.bGroupInfoDirty = true
end

function OptiPlates:OnGroup_Join()
	self.bGroupInfoDirty = true
end

function OptiPlates:OnGroup_Add()
	self.bGroupInfoDirty = true
end

function OptiPlates:OnGroup_Remove()
	self.bGroupInfoDirty = true
end

function OptiPlates:OnGroup_MemberFlagsChanged()
	self.bGroupInfoDirty = true
end

function OptiPlates:OnGroup_FlagsChanged()
	self.bGroupInfoDirty = true
end

function OptiPlates:OnShieldsOverloaded(unit, bOverloaded)
	
end

function OptiPlates:OnApplyCCState(nCCState, unit)
	if unit == nil then return end
	local tNameplate = self.arUnit2Nameplate[unit:GetId()]
	if tNameplate then
		tNameplate.tInCCStates[nCCState] = true
		for name,idCCState in pairs(tSubdueCCStates) do
			if tNameplate.tInCCStates[idCCState] then
				if tNameplate.nCCState >= 0 then
					if unit:GetCCStateTimeRemaining(tNameplate.nCCState) < unit:GetCCStateTimeRemaining(idCCState) then
						tNameplate.nCCState = idCCState
						tNameplate.sCCStateText = name
					end
				else
					tNameplate.nCCState = idCCState
					tNameplate.sCCStateText = name				
				end
			end
		end
	end
end

function OptiPlates:OnRemoveCCState(nCCState, unit)
	if unit == nil then return end
	local tNameplate = self.arUnit2Nameplate[unit:GetId()]
	if tNameplate then
		tNameplate.tInCCStates[nCCState] = false
		if tNameplate.nCCState == nCCState then
			tNameplate.nCCState = -1
		end
		for name,idCCState in pairs(tSubdueCCStates) do
			if tNameplate.tInCCStates[idCCState] then
				 if tNameplate.nCCState >= 0 then
					if unit:GetCCStateTimeRemaining(tNameplate.nCCState) < unit:GetCCStateTimeRemaining(idCCState) then
						tNameplate.nCCState = idCCState
						tNameplate.sCCStateText = name
					end
				 else
					tNameplate.nCCState = idCCState
					tNameplate.sCCStateText = name
				 end
			end
		end
	end
end

function OptiPlates:PopulateAllFontLists()
	self:HelperPopulateFontList(self.wndMain:FindChild("Grid_NameFontList"),self.wndMain:FindChild("Button_NameFontToggle"),self.setStyleUnitNameFont)
	self:HelperPopulateFontList(self.wndMain:FindChild("Grid_GuildFontList"),self.wndMain:FindChild("Button_GuildFontToggle"),self.setStyleUnitGuildFont)
	self:HelperPopulateFontList(self.wndMain:FindChild("Grid_CastFontList"),self.wndMain:FindChild("Button_CastFontToggle"),self.setStyleCastbarFont)
	self:HelperPopulateFontList(self.wndMain:FindChild("Grid_DistanceFontList"),self.wndMain:FindChild("Button_DistanceFontToggle"),self.setStyleDistanceFont)
	self:HelperPopulateFontList(self.wndMain:FindChild("Grid_ResourceBarFontList"),self.wndMain:FindChild("Button_ResourceBarFontToggle"),self.setStyleClassResourceFont)
	self:HelperPopulateFontList(self.wndMain:FindChild("Grid_FocusBarFontList"),self.wndMain:FindChild("Button_FocusBarFontToggle"),self.setStyleFocusFont)
	self:HelperPopulateFontList(self.wndMain:FindChild("Grid_HeadingTextFontList"),self.wndMain:FindChild("Button_HeadingTextFontToggle"),self.setStyleHeadingIndicatorTextFont)
end

function OptiPlates:HelperGetFontString(font)
	if font == nil or font.name == nil or font.face == nil or font.size == nil then return "" end
	local tmpFontName = font.name
	tmpFontName = string.gsub(tmpFontName,"CRB_Pixel", "Choad/CRB_Pixel")
	local nStyleStart, nStyleEnd = string.find(tmpFontName, "_[BOI]+$")
	local bHasStyles = (nStyleStart or false) 
	local fontname = tmpFontName.." "
	if bHasStyles then
		local styleString = string.sub(tmpFontName,nStyleStart+1,nStyleEnd)
		styleString = string.gsub(styleString,"BB", "ExtraBold ")
		styleString = string.gsub(styleString,"^B", "Bold ")
		styleString = string.gsub(styleString,"I", "Italic ")
		styleString = string.gsub(styleString,"O", "Outlined ")
		fontname = string.sub(tmpFontName,1,nStyleStart-1)..", "..styleString
	end
	return fontname.."("..font.face..", "..font.size..")"
end

function OptiPlates:HelperPopulateFontList(grid, button, selected)
	if grid == nil or button == nil then return end
	local fonts = self.gameFonts
	grid:DeleteAll()
	local nTemp = -1
	for idx,font in pairs(fonts) do
		local fontString = self:HelperGetFontString(font)
		grid:AddRow(fontString)
		if selected ~= nil and selected.name == font.name then
			grid:SetCurrentRow(grid:GetRowCount())
			button:SetText("   "..fontString)
			nTemp = 0
		end
		if nTemp >= 0 then nTemp = nTemp + 1 end
		if nTemp == 5 then grid:SetVScrollPos(grid:GetVScrollRange()); nTemp = -2 end
	end
	if nTemp ~= -2 then grid:SetVScrollPos(grid:GetVScrollRange()) end
end

function OptiPlates:Grid_OnNameFontChange()
	self:FontSelectionUpdated(
		self.wndMain:FindChild("Window_NameFontListContainer"),
		self.wndMain:FindChild("Button_NameFontToggle"),
		self.wndMain:FindChild("Grid_NameFontList"),
		"setStyleUnitNameFont"
	)
end

function OptiPlates:Grid_OnGuildFontChange()
	self:FontSelectionUpdated(
		self.wndMain:FindChild("Window_GuildFontListContainer"),
		self.wndMain:FindChild("Button_GuildFontToggle"),
		self.wndMain:FindChild("Grid_GuildFontList"),
		"setStyleUnitGuildFont"
	)
end

function OptiPlates:Grid_OnCastFontChange()
	self:FontSelectionUpdated(
		self.wndMain:FindChild("Window_CastFontListContainer"),
		self.wndMain:FindChild("Button_CastFontToggle"),
		self.wndMain:FindChild("Grid_CastFontList"),
		"setStyleCastbarFont"
	)
end

function OptiPlates:Grid_OnDistanceFontChange()
	self:FontSelectionUpdated(
		self.wndMain:FindChild("Window_DistanceFontListContainer"),
		self.wndMain:FindChild("Button_DistanceFontToggle"),
		self.wndMain:FindChild("Grid_DistanceFontList"),
		"setStyleDistanceFont"
	)
end

function OptiPlates:Grid_OnResourceBarFontChange()
	self:FontSelectionUpdated(
		self.wndMain:FindChild("Window_ResourceBarFontListContainer"),
		self.wndMain:FindChild("Button_ResourceBarFontToggle"),
		self.wndMain:FindChild("Grid_ResourceBarFontList"),
		"setStyleClassResourceFont"
	)
end

function OptiPlates:Grid_OnFocusBarFontChange()
	self:FontSelectionUpdated(
		self.wndMain:FindChild("Window_FocusBarFontListContainer"),
		self.wndMain:FindChild("Button_FocusBarFontToggle"),
		self.wndMain:FindChild("Grid_FocusBarFontList"),
		"setStyleFocusFont"
	)
end

function OptiPlates:Grid_OnHeadingTextFontChange()
	self:FontSelectionUpdated(
		self.wndMain:FindChild("Window_HeadingTextFontListContainer"),
		self.wndMain:FindChild("Button_HeadingTextFontToggle"),
		self.wndMain:FindChild("Grid_HeadingTextFontList"),
		"setStyleHeadingIndicatorTextFont"
	)
end

function OptiPlates:Button_NameFontToggle( wndHandler, wndControl, eMouseButton )
	self:ToggleFontSelectionWindow(self.wndMain:FindChild("Window_NameFontListContainer"))
end
function OptiPlates:Button_GuildFontToggle( wndHandler, wndControl, eMouseButton )
	self:ToggleFontSelectionWindow(self.wndMain:FindChild("Window_GuildFontListContainer"))
end
function OptiPlates:Button_CastFontToggle( wndHandler, wndControl, eMouseButton )
	self:ToggleFontSelectionWindow(self.wndMain:FindChild("Window_CastFontListContainer"))
end
function OptiPlates:Button_DistanceFontToggle( wndHandler, wndControl, eMouseButton )
	self:ToggleFontSelectionWindow(self.wndMain:FindChild("Window_DistanceFontListContainer"))
end
function OptiPlates:Button_ResourceBarFontToggle( wndHandler, wndControl, eMouseButton )
	self:ToggleFontSelectionWindow(self.wndMain:FindChild("Window_ResourceBarFontListContainer"))
end
function OptiPlates:Button_FocusBarFontToggle( wndHandler, wndControl, eMouseButton )
	self:ToggleFontSelectionWindow(self.wndMain:FindChild("Window_FocusBarFontListContainer"))
end
function OptiPlates:Button_HeadingTextFontToggle( wndHandler, wndControl, eMouseButton )
	self:ToggleFontSelectionWindow(self.wndMain:FindChild("Window_HeadingTextFontListContainer"))
end

function OptiPlates:FontSelectionUpdated(wnd,btn,grid,var)
	if wnd == nil or btn == nil or grid == nil or var == nil then return end
	local font = self.gameFonts[grid:GetCurrentRow()]
	btn:SetText("   "..self:HelperGetFontString(font))
	wnd:Show(false)
	self[var] = font
	self:OptionsChanged()
end
function OptiPlates:ToggleFontSelectionWindow(wnd)
	if wnd == nil then return end
	if wnd:IsShown() then
		wnd:Show(false)
	else
		wnd:Show(true)
		wnd:ToFront()
	end
end

function OptiPlates:Button_NameplateAnchorChanged(wndHandler, wndControl, eMouseButton)
	if self.wndMain:FindChild("Button_NameplateAnchorName"):IsChecked() then 
		self.setStyleVerticalAnchor = "Name"
	elseif self.wndMain:FindChild("Button_NameplateAnchorBars"):IsChecked() then 
		self.setStyleVerticalAnchor = "Bars"
	else
		self.setStyleVerticalAnchor = "None"
		local wnd = self.wndMain:FindChild("Button_NameplateAnchorNone")
		if not wnd:IsChecked() then
			wnd:SetCheck(true)
		end
	end
	self:OptionsChanged()
end

function OptiPlates:Button_CleanseIconAnchorChanged(wndHandler, wndControl, eMouseButton)
	if self.wndMain:FindChild("Button_CleanseIconPositionLeft"):IsChecked() then 
		self.setStyleCleanseIconPosition = "Left"
	else
		self.setStyleCleanseIconPosition = "Right"
		local wnd = self.wndMain:FindChild("Button_CleanseIconPositionRight")
		if not wnd:IsChecked() then
			wnd:SetCheck(true)
		end
	end
	self:OptionsChanged()

end

---------------------------------------------------------------------------------------------------
-- QuickselectWindow Functions
---------------------------------------------------------------------------------------------------
function OptiPlates:Grid_OnProfileQuickChange()
	local sel = self.wndQuickChange:FindChild("Grid_QuickChangeProfileList"):GetCurrentRow()
	local btn = self.wndQuickChange:FindChild("Button_QuickProfileListToggle")
	local profile = self.tempProfileArray[sel]
	local scope = profile[1]
	local name = profile[2]	
	local profileString = self.scopeTextArray[scope].." - "..name
	btn:SetText("   "..profileString)
	self.wndQuickChange:FindChild("Window_QuickProfileListContainer"):Show(false)
	self:ChangeProfile(scope,name)
	self.wndMain:FindChild("Label_ProfileScope"):SetText(self.scopeTextArray[self.setCurrentProfileScope])
	self.wndMain:FindChild("Label_ProfileName"):SetText(self.setCurrentProfileName)	
end

function OptiPlates:Button_ProfileQuickChangeToggle( wndHandler, wndControl, eMouseButton )
	if not self.setQuickselectDragDropStart then
		local wnd = self.wndQuickChange:FindChild("Window_QuickProfileListContainer")
		if wnd:IsShown() then
			wnd:Show(false)
		else
			wnd:Show(true)
			wnd:ToFront()
		end
	end
	self.setQuickselectDragDropStart = false
end

function OptiPlates:Button_ProfileQuickChangeOptions( wndHandler, wndControl, eMouseButton )
	self.wndMain:Show(true)
end

function OptiPlates:Button_ProfileQuickChangeToggleLock( wndHandler, wndControl, eMouseButton )
	self.setQuickselectWindowLocked = not self.setQuickselectWindowLocked
	self:OptionsChanged()
end

function OptiPlates:Window_QuickselectWindowMove( wndHandler, wndControl, nOldLeft, nOldTop, nOldRight, nOldBottom )
	self.setQuickselectWindowPosition = self.wndQuickChange:GetLocation():ToTable().nOffsets
end

function OptiPlates:Button_QuickchangeResetLocation( wndHandler, wndControl, eMouseButton )
	self.setQuickselectWindowPosition = tUserSettings.setQuickselectWindowPosition.default
	self.setQuickselectWindowScaling = tUserSettings.setQuickselectWindowScaling.default
	self:OptionsChanged()
end

function OptiPlates:Button_ProfileQuickChangeToggleMove( wndHandler, wndControl, nLastRelativeMouseX, nLastRelativeMouseY )
	if self.setQuickselectDragDropStart and not self.setQuickselectWindowLocked then
		local mouse = Apollo.GetMouse()
		local offsetX = mouse.x - self.setQuickselectDragDropStartX
		local offsetY = mouse.y - self.setQuickselectDragDropStartY
		self.setQuickselectWindowPosition = { self.setQuickselectDragDropStart[1] + offsetX,
											  self.setQuickselectDragDropStart[2] + offsetY,
											  self.setQuickselectDragDropStart[3] + offsetX,											  
											  self.setQuickselectDragDropStart[4] + offsetY		
											}
		self.wndQuickChange:MoveToLocation(WindowLocation.new({ fPoints = {0,0,0,0}, nOffsets = self.setQuickselectWindowPosition}))		
	end
end

function OptiPlates:Button_ProfileQuickChangeToggleButtonDown( wndHandler, wndControl, eMouseButton )
	self.setQuickselectDragDropStart = true
	self.setQuickselectDragDropStartX = Apollo.GetMouse().x
	self.setQuickselectDragDropStartY = Apollo.GetMouse().y
	self.setQuickselectDragDropStart = self.setQuickselectWindowPosition
end

function OptiPlates:Button_ProfileQuickChangeToggleButtonUp( wndHandler, wndControl, eMouseButton )
	self.setQuickselectDragDropStart = false
end

function OptiPlates:UpdateQuickChangeWindowLocation()
	local wndLeft = self.setQuickselectWindowPosition[1]
	local wndTop = self.setQuickselectWindowPosition[2]
	local wndWidth = 275
	local wndHeight = 31
	if self.setQuickselectWindowNoButtons then
		wndWidth = 223
	end
	local scaling = math.floor(self.setQuickselectWindowScaling)/100
	if scaling ~= 1 then
		self.wndQuickChange:SetScale(scaling)
	end
	self.setQuickselectWindowPosition = { wndLeft,
										  wndTop,
										  wndLeft+wndWidth,
										  wndTop+wndHeight	
										}
	self.wndQuickChange:MoveToLocation(WindowLocation.new({ fPoints = {0,0,0,0}, nOffsets = self.setQuickselectWindowPosition}))			
end

---------------------------------------------------------------------------------------------------
-- Buff profiles settings
---------------------------------------------------------------------------------------------------
function OptiPlates:CreateBuffProfileList(grid, btn, sel)
	if grid == nil then return end
	local profiles = self.setProfiles["Buffprofiles"]
	if profiles == nil then return end
	local profileArray = {}
	local tempCounter = 1
	
	grid:DeleteAll()
	
	for profileName, profileTable in pairs(profiles) do
		grid:AddRow(profileName)
		profileArray[tempCounter] = profileName
		if sel == profileName then
			grid:SetCurrentRow(grid:GetRowCount())
			btn:SetText("   "..profileName)
		elseif sel == nil and tempCounter == 1 then
			grid:SetCurrentRow(1)
			btn:SetText("   "..profileName)
		end
		tempCounter = tempCounter + 1
	end
	self.tempBuffProfileArray = profileArray
end

function OptiPlates:CreateAllBuffProfileLists()
	if self.wndMain:FindChild("SettingsBuffs") ~= nil then
		self:CreateBuffProfileList(self.wndMain:FindChild("Grid_NewBuffProfileList"), self.wndMain:FindChild("Button_NewBuffProfileListToggle"), nil)
		self:CreateBuffProfileList(self.wndMain:FindChild("Grid_RemoveBuffProfileList"), self.wndMain:FindChild("Button_RemoveBuffProfileListToggle"), nil)
		self:CreateBuffProfileList(self.wndMain:FindChild("Grid_ChangeBuffProfileList"), self.wndMain:FindChild("Button_ChangeBuffProfileListToggle"), nil)	
		self.wndMain:FindChild("Button_NewBuffProfileDefaultBase"):SetCheck(true)
		self.wndMain:FindChild("Button_NewBuffProfileCopyBase"):SetCheck(false)
		self:Button_NewBuffProfileListDisableToggle()
	end
	local wndNameplate = self.wndMain:FindChild("SettingsBuffs")
	if wndNameplate ~= nil then
		self:CreateBuffProfileList(wndNameplate:FindChild("Grid_ChangeBuffProfile_M"), wndNameplate:FindChild("Button_ChangeBuffProfileToggle_M"), self.setBuffSelectedProfiles["M"])	
		self:CreateBuffProfileList(wndNameplate:FindChild("Grid_ChangeBuffProfile_PP"), wndNameplate:FindChild("Button_ChangeBuffProfileToggle_PP"), self.setBuffSelectedProfiles["PP"])	
		self:CreateBuffProfileList(wndNameplate:FindChild("Grid_ChangeBuffProfile_FP"), wndNameplate:FindChild("Button_ChangeBuffProfileToggle_FP"), self.setBuffSelectedProfiles["FP"])	
		self:CreateBuffProfileList(wndNameplate:FindChild("Grid_ChangeBuffProfile_F"), wndNameplate:FindChild("Button_ChangeBuffProfileToggle_F"), self.setBuffSelectedProfiles["F"])	
		self:CreateBuffProfileList(wndNameplate:FindChild("Grid_ChangeBuffProfile_HP"), wndNameplate:FindChild("Button_ChangeBuffProfileToggle_HP"), self.setBuffSelectedProfiles["HP"])	
		self:CreateBuffProfileList(wndNameplate:FindChild("Grid_ChangeBuffProfile_H"), wndNameplate:FindChild("Button_ChangeBuffProfileToggle_H"), self.setBuffSelectedProfiles["H"])	
		self:CreateBuffProfileList(wndNameplate:FindChild("Grid_ChangeBuffProfile_N"), wndNameplate:FindChild("Button_ChangeBuffProfileToggle_N"), self.setBuffSelectedProfiles["N"])	
	end
end

function OptiPlates:CreateBuffProfile(name, baseArray)
	-- default settings
	if self.defaultBuffProfileSettings == nil then
		self:DefaultBuffSettings()
	end	
	if self.setProfiles["Buffprofiles"][name] == nil then
		self.setProfiles["Buffprofiles"][name] = {}
	else
		return
	end
	for key,val in pairs(self.defaultBuffProfileSettings) do
		self.setProfiles["Buffprofiles"][name][key] = val
	end

	if baseArray ~= nil then
		for k,v in pairs(self.setProfiles["Buffprofiles"][name]) do 
			if baseArray.k ~= nil then
				self.setProfiles["Buffprofiles"][name].k = baseArray.k
			end
		end
	end
	self:ChangeBuffProfile(name)
	self:CreateAllBuffProfileLists()
end

function OptiPlates:RemoveBuffProfile(name)
    -- attempting to remove currently selected profile
	if name == "Default" then return end
	self.setProfiles["Buffprofiles"][name] = nil
	self:CreateAllBuffProfileLists()
	self:BuffOptionsChanged()
end

function OptiPlates:ChangeBuffProfile(name)
	local profile = self.setProfiles["Buffprofiles"][name]
	if profile == nil then return end
	local wnd = self.wndMain:FindChild("Window_ChangeBuffProfile")
	if wnd ~= nil then
		wnd:FindChild("Button_AllBuffs"):SetCheck(profile.bIncludeAllBuffs)
		wnd:FindChild("Button_ExcludeAllBuffs"):SetCheck(profile.bExcludeAllBuffs)
		wnd:FindChild("Button_AllDispellableBuffs"):SetCheck(profile.bIncludeAllDispellableBuffs)
		wnd:FindChild("Button_AllNonDispellableBuffs"):SetCheck(profile.bIncludeAllNonDispellableBuffs)
		wnd:FindChild("Button_BuffTooltipSimple"):SetCheck(profile.bBuffShowNameTooltip)
		wnd:FindChild("Button_BuffTooltipFull"):SetCheck(profile.bBuffShowFullTooltip)
		wnd:FindChild("Button_BuffOnlyShowMine"):SetCheck(profile.bBuffOnlyShowMine)
		
		wnd:FindChild("EditBox_IncludeBuffs"):SetText(profile.sIncludeBuffs)
		wnd:FindChild("EditBox_ExcludeBuffs"):SetText(profile.sExcludeBuffs)
		
		wnd:FindChild("Slider_NumBuffRows"):SetValue(math.floor(profile.nBuffRows))
		wnd:FindChild("Label_NumBuffRowsDisplay"):SetText(math.floor(profile.nBuffRows))
		wnd:FindChild("Slider_NumBuffsPerRow"):SetValue(math.floor(profile.nBuffsPerRow))
		wnd:FindChild("Label_NumBuffsPerRowDisplay"):SetText(math.floor(profile.nBuffsPerRow))
		wnd:FindChild("Slider_BuffIconSize"):SetValue(math.floor(profile.nBuffIconSize))
		wnd:FindChild("Label_BuffIconSizeDisplay"):SetText(math.floor(profile.nBuffIconSize))

		wnd:FindChild("Button_AllDebuffs"):SetCheck(profile.bIncludeAllDebuffs)
		wnd:FindChild("Button_ExcludeAllDebuffs"):SetCheck(profile.bExcludeAllDebuffs)
		wnd:FindChild("Button_AllDispellableDebuffs"):SetCheck(profile.bIncludeAllDispellableDebuffs)
		wnd:FindChild("Button_AllNonDispellableDebuffs"):SetCheck(profile.bIncludeAllNonDispellableDebuffs)
		wnd:FindChild("Button_DebuffTooltipSimple"):SetCheck(profile.bDebuffShowNameTooltip)
		wnd:FindChild("Button_DebuffTooltipFull"):SetCheck(profile.bDebuffShowFullTooltip)
		wnd:FindChild("Button_DebuffOnlyShowMine"):SetCheck(profile.bDebuffOnlyShowMine)
		
		wnd:FindChild("EditBox_IncludeDebuffs"):SetText(profile.sIncludeDebuffs)
		wnd:FindChild("EditBox_ExcludeDebuffs"):SetText(profile.sExcludeDebuffs)
		
		wnd:FindChild("Slider_NumDebuffRows"):SetValue(math.floor(profile.nDebuffRows))
		wnd:FindChild("Label_NumDebuffRowsDisplay"):SetText(math.floor(profile.nDebuffRows))
		wnd:FindChild("Slider_NumDebuffsPerRow"):SetValue(math.floor(profile.nDebuffsPerRow))
		wnd:FindChild("Label_NumDebuffsPerRowDisplay"):SetText(math.floor(profile.nDebuffsPerRow))
		wnd:FindChild("Slider_DebuffIconSize"):SetValue(math.floor(profile.nDebuffIconSize))
		wnd:FindChild("Label_DebuffIconSizeDisplay"):SetText(math.floor(profile.nDebuffIconSize))
		self:CreateBuffProfileList(wnd:FindChild("Grid_ChangeBuffProfileList"),wnd:FindChild("Button_ChangeBuffProfileListToggle"),name)
	end
end

function OptiPlates:Grid_OnChangeBuffProfileList()
	local sel = self.wndMain:FindChild("Grid_ChangeBuffProfileList"):GetCurrentRow()
	local profileName = self.tempBuffProfileArray[sel]
	self.wndMain:FindChild("Window_ChangeBuffProfileListContainer"):Show(false)
	self:ChangeBuffProfile(profileName)
end

function OptiPlates:Button_NewBuffProfileListDisableToggle()
	local check = self.wndMain:FindChild("Button_NewBuffProfileCopyBase"):IsChecked()
	local button = self.wndMain:FindChild("Button_NewBuffProfileListToggle")
	if check then
		button:Enable(true)
	else
		button:Enable(false)
	end
	
end

function OptiPlates:Button_OnNewBuffProfileListToggle( wndHandler, wndControl, eMouseButton )
	local wnd = self.wndMain:FindChild("Window_NewBuffProfileListContainer")
	wnd:Show(not wnd:IsShown())
end

function OptiPlates:Grid_OnNewBuffProfileListChange()
	local sel = self.wndMain:FindChild("Grid_NewBuffProfileList"):GetCurrentRow()
	local btn = self.wndMain:FindChild("Button_NewBuffProfileListToggle")
	local profileName = self.tempBuffProfileArray[sel]
	btn:SetText("   "..profileName)
	self.wndMain:FindChild("Window_NewBuffProfileListContainer"):Show(false)
end

function OptiPlates:Grid_OnRemoveBuffProfileList()
	local sel = self.wndMain:FindChild("Grid_RemoveBuffProfileList"):GetCurrentRow()
	local btn = self.wndMain:FindChild("Button_RemoveBuffProfileListToggle")
	local profileName = self.tempBuffProfileArray[sel]
	btn:SetText("   "..profileName)
	self.wndMain:FindChild("Window_RemoveBuffProfileListContainer"):Show(false)
end

function OptiPlates:Button_OnRemoveBuffProfileListToggle( wndHandler, wndControl, eMouseButton )
	local wnd = self.wndMain:FindChild("Window_RemoveBuffProfileListContainer")
	wnd:Show(not wnd:IsShown())
end

function OptiPlates:Button_OnCreateBuffProfile( wndHandler, wndControl, eMouseButton )
	local name = self.wndMain:FindChild("Edit_NewBuffProfileName"):GetText()
	local baseArray = nil
	local check = self.wndMain:FindChild("Button_NewBuffProfileCopyBase"):IsChecked()
	
	if check then
		local baseName = self.tempBuffProfileArray[self.wndMain:FindChild("Grid_NewBuffProfileList"):GetCurrentRow()]
		baseArray = self.setProfiles["Buffprofiles"][baseName]
	else
		baseArray = nil
	end
	self:CreateBuffProfile(name,baseArray)
end

function OptiPlates:Button_OnRemoveBuffProfile( wndHandler, wndControl, eMouseButton )
	local sel = self.wndMain:FindChild("Grid_RemoveBuffProfileList"):GetCurrentRow()
	local profileName = self.tempBuffProfileArray[sel]
	self:RemoveBuffProfile(profileName)
	self:CreateAllBuffProfileLists()
	self.wndMain:FindChild("Grid_RemoveBuffProfileList"):SetCurrentRow(1)
end

function OptiPlates:BuffOptionsChanged()
	local sel = self.wndMain:FindChild("Grid_ChangeBuffProfileList"):GetCurrentRow()
	local profileName = self.tempBuffProfileArray[sel]
	local profile = self.setProfiles["Buffprofiles"][profileName]
	local wnd = self.wndMain:FindChild("Window_ChangeBuffProfile")
	
	profile.bIncludeAllBuffs = wnd:FindChild("Button_AllBuffs"):IsChecked()
	profile.bIncludeAllDispellableBuffs = wnd:FindChild("Button_AllDispellableBuffs"):IsChecked()
	profile.bIncludeAllNonDispellableBuffs = wnd:FindChild("Button_AllNonDispellableBuffs"):IsChecked()
	profile.bBuffShowNameTooltip = wnd:FindChild("Button_BuffTooltipSimple"):IsChecked()
	profile.bBuffShowFullTooltip = wnd:FindChild("Button_BuffTooltipFull"):IsChecked()
	profile.bExcludeAllBuffs = wnd:FindChild("Button_ExcludeAllBuffs"):IsChecked()
	profile.bBuffOnlyShowMine = wnd:FindChild("Button_BuffOnlyShowMine"):IsChecked()
	
	if profile.bExcludeAllBuffs or 
		(profile.bIncludeAllBuffs == false and
		profile.bIncludeAllDispellableBuffs == false and
		profile.bIncludeAllNonDispellableBuffs == false and
		(profile.sIncludeBuffs == "" or profile.sIncludeBuffs == nil)
		)
		then
			profile.bExcludeAllBuffs = true
	else
		profile.bExcludeAllBuffs = false
	end
	
	profile.nBuffRows = math.floor(wnd:FindChild("Slider_NumBuffRows"):GetValue())
	wnd:FindChild("Label_NumBuffRowsDisplay"):SetText(math.floor(profile.nBuffRows))
	profile.nBuffsPerRow = math.floor(wnd:FindChild("Slider_NumBuffsPerRow"):GetValue())
	wnd:FindChild("Label_NumBuffsPerRowDisplay"):SetText(math.floor(profile.nBuffsPerRow))
	profile.nBuffIconSize = math.floor(wnd:FindChild("Slider_BuffIconSize"):GetValue())
	wnd:FindChild("Label_BuffIconSizeDisplay"):SetText(math.floor(profile.nBuffIconSize))
	
	profile.bIncludeAllDebuffs = wnd:FindChild("Button_AllDebuffs"):IsChecked()
	profile.bIncludeAllDispellableDebuffs = wnd:FindChild("Button_AllDispellableDebuffs"):IsChecked()
	profile.bIncludeAllNonDispellableDebuffs = wnd:FindChild("Button_AllNonDispellableDebuffs"):IsChecked()
	profile.bDebuffShowNameTooltip = wnd:FindChild("Button_DebuffTooltipSimple"):IsChecked()
	profile.bDebuffShowFullTooltip = wnd:FindChild("Button_DebuffTooltipFull"):IsChecked()
	profile.bExcludeAllDebuffs = wnd:FindChild("Button_ExcludeAllDebuffs"):IsChecked()
	profile.bDebuffOnlyShowMine = wnd:FindChild("Button_DebuffOnlyShowMine"):IsChecked()
	
	if profile.bExcludeAllDebuffs or 
		(profile.bIncludeAllDebuffs == false and
		profile.bIncludeAllDispellableDebuffs == false and
		profile.bIncludeAllNonDispellableDebuffs == false and
		(profile.sIncludeDebuffs == "" or profile.sIncludeDebuffs == nil)
		)
		then
			profile.bExcludeAllDebuffs = true
	else
		profile.bExcludeAllDebuffs = false
	end
	
	profile.nDebuffRows = math.floor(wnd:FindChild("Slider_NumDebuffRows"):GetValue())
	wnd:FindChild("Label_NumDebuffRowsDisplay"):SetText(math.floor(profile.nDebuffRows))
	profile.nDebuffsPerRow = math.floor(wnd:FindChild("Slider_NumDebuffsPerRow"):GetValue())
	wnd:FindChild("Label_NumDebuffsPerRowDisplay"):SetText(math.floor(profile.nDebuffsPerRow))
	profile.nDebuffIconSize = math.floor(wnd:FindChild("Slider_DebuffIconSize"):GetValue())
	wnd:FindChild("Label_DebuffIconSizeDisplay"):SetText(math.floor(profile.nDebuffIconSize))
	self.bBuffProfilesDirty = true
end

function OptiPlates:Button_SaveBuffList()
	local sel = self.wndMain:FindChild("Grid_ChangeBuffProfileList"):GetCurrentRow()
	local profileName = self.tempBuffProfileArray[sel]
	local profile = self.setProfiles["Buffprofiles"][profileName]
	local wnd = self.wndMain:FindChild("Window_ChangeBuffProfile")
	local incText = wnd:FindChild("EditBox_IncludeBuffs"):GetText()
	local excText = wnd:FindChild("EditBox_ExcludeBuffs"):GetText()
	profile.sIncludeBuffs = incText
	profile.tIncludeTextBuffs = {}
	local array = string.gmatch(incText,"[^;\r\n]+")
	for line in array do
		local buff = string.match(line,"^%s*.-%s*$")
		profile.tIncludeTextBuffs[buff] = true
		profile.bExcludeAllBuffs = false
	end
	
	profile.sExcludeBuffs = excText
	profile.tExcludeTextBuffs = {}
	array = string.gmatch(excText,"[^;\r\n]+")
	for line in array do
		local buff = string.match(line,"^%s*.-%s*$")
		profile.tExcludeTextBuffs[buff] = true
	end	
	self.bBuffProfilesDirty = true
end

function OptiPlates:Button_SaveDebuffList()
	local sel = self.wndMain:FindChild("Grid_ChangeBuffProfileList"):GetCurrentRow()
	local profileName = self.tempBuffProfileArray[sel]
	local profile = self.setProfiles["Buffprofiles"][profileName]
	local wnd = self.wndMain:FindChild("Window_ChangeBuffProfile")
	local incText = wnd:FindChild("EditBox_IncludeDebuffs"):GetText()
	local excText = wnd:FindChild("EditBox_IncludeDebuffs"):GetText()
	profile.sIncludeDebuffs = incText
	profile.tIncludeTextDebuffs = {}
	local array = string.gmatch(incText,"[^;\r\n]+")
	for line in array do
		local buff = string.match(line,"^%s*.-%s*$")
		profile.tIncludeTextDebuffs[buff] = true
		profile.bExcludeAllBuffs = false
	end
	
	profile.sExcludeDebuffs = excText
	profile.tExcludeTextDebuffs = {}
	array = string.gmatch(excText,"[^;\r\n]+")
	for line in array do
		local buff = string.match(line,"^%s*.-%s*$")
		profile.tIncludeTextDebuffs[buff] = true
	end	
	self.bBuffProfilesDirty = true
	
end

function OptiPlates:Button_ChangeBuffProfileListToggle(wndHandler, wndControl, eMouseButton)
	if wndControl == self.wndMain:FindChild("Button_ChangeBuffProfileListToggle") then
		local wnd = self.wndMain:FindChild("Window_ChangeBuffProfileListContainer")
		wnd:Show(not wnd:IsShown())
	elseif wndControl == self.wndMain:FindChild("Button_ChangeBuffProfileToggle_M") then
		local wnd = self.wndMain:FindChild("Window_ChangeBuffProfileContainer_M")
		wnd:Show(not wnd:IsShown())
	elseif wndControl == self.wndMain:FindChild("Button_ChangeBuffProfileToggle_PP") then
		local wnd = self.wndMain:FindChild("Window_ChangeBuffProfileContainer_PP")
		wnd:Show(not wnd:IsShown())
	elseif wndControl == self.wndMain:FindChild("Button_ChangeBuffProfileToggle_FP") then
		local wnd = self.wndMain:FindChild("Window_ChangeBuffProfileContainer_FP")
		wnd:Show(not wnd:IsShown())
	elseif wndControl == self.wndMain:FindChild("Button_ChangeBuffProfileToggle_F") then
		local wnd = self.wndMain:FindChild("Window_ChangeBuffProfileContainer_F")
		wnd:Show(not wnd:IsShown())
	elseif wndControl == self.wndMain:FindChild("Button_ChangeBuffProfileToggle_HP") then
		local wnd = self.wndMain:FindChild("Window_ChangeBuffProfileContainer_HP")
		wnd:Show(not wnd:IsShown())
	elseif wndControl == self.wndMain:FindChild("Button_ChangeBuffProfileToggle_H") then
		local wnd = self.wndMain:FindChild("Window_ChangeBuffProfileContainer_H")
		wnd:Show(not wnd:IsShown())
	elseif wndControl == self.wndMain:FindChild("Button_ChangeBuffProfileToggle_N") then
		local wnd = self.wndMain:FindChild("Window_ChangeBuffProfileContainer_N")
		wnd:Show(not wnd:IsShown())
	end
end

function OptiPlates:Grid_OnChangeNameplateBuffProfile(wndHandler, wndControl)
	local sLetter = nil
	local wnd = nil
	if wndControl == self.wndMain:FindChild("Grid_ChangeBuffProfile_M") then
		sLetter = "M"
		wnd = self.wndMain:FindChild("Window_ChangeBuffProfileContainer_M")
	elseif wndControl == self.wndMain:FindChild("Grid_ChangeBuffProfile_PP") then
		sLetter = "PP"
		wnd = self.wndMain:FindChild("Window_ChangeBuffProfileContainer_PP")
	elseif wndControl == self.wndMain:FindChild("Grid_ChangeBuffProfile_FP") then
		sLetter = "FP"
		wnd = self.wndMain:FindChild("Window_ChangeBuffProfileContainer_FP")
	elseif wndControl == self.wndMain:FindChild("Grid_ChangeBuffProfile_F") then
		sLetter = "F"
		wnd = self.wndMain:FindChild("Window_ChangeBuffProfileContainer_F")
	elseif wndControl == self.wndMain:FindChild("Grid_ChangeBuffProfile_HP") then
		sLetter = "HP"
		wnd = self.wndMain:FindChild("Window_ChangeBuffProfileContainer_HP")
	elseif wndControl == self.wndMain:FindChild("Grid_ChangeBuffProfile_H") then
		sLetter = "H"
		wnd = self.wndMain:FindChild("Window_ChangeBuffProfileContainer_H")
	elseif wndControl == self.wndMain:FindChild("Grid_ChangeBuffProfile_N") then
		sLetter = "N"
		wnd = self.wndMain:FindChild("Window_ChangeBuffProfileContainer_N")
	else
		return
	end
	
	local sel = wndControl:GetCurrentRow()
	local profileName = self.tempBuffProfileArray[sel]
	local wndNameplate = self.wndMain:FindChild("SettingsBuffs")
	self.setBuffSelectedProfiles[sLetter] = profileName
	self:CreateBuffProfileList(wndNameplate:FindChild("Grid_ChangeBuffProfile_"..sLetter), wndNameplate:FindChild("Button_ChangeBuffProfileToggle_"..sLetter), profileName)	
	self.bBuffProfilesDirty = true
	wnd:Show(false)
end

function OptiPlates:GetTableLength(T)
	local count = 0
	for _ in pairs(T) do count = count + 1 end
	return count
end

function OptiPlates:StringToRGB(str)
	local r, g, b = 0, 0, 0
	local n = tonumber(str, 16)
	if n then r = math.floor(n / 65536); g = math.floor(n / 256) % 256; b = n % 256 end
	return r / 255, g / 255, b / 255
end

function OptiPlates:CColorToHexString(ccolor)
	if ccolor == nil then return end
	local r = ccolor.r
	local g = ccolor.g
	local b = ccolor.b
	return string.format("%02x%02x%02x", math.floor(r * 255 + 0.5), math.floor(g * 255 + 0.5), math.floor(b * 255 + 0.5))
end

-----------------------------------------------------------------------------------------------
OptiPlates:Init()

